Add LoadPath "./metatheory".

Require Import List.
Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import infrastructure.
Require Import isEqSpecification.
Require Import reductions.

Lemma weakening_lc_ectx : 
  forall D1 D2 D3, 
    lc_ectx D2 ->
    lc_ectx (D1 ++ D3)  ->
    lc_ectx (D1 ++ D2 ++ D3).
Proof.
  induction D1.
    induction D2.
      simpl;  tauto.
      simpl in *; destruct a as (e1',e2'); firstorder.

    intros; destruct a as (e1',e2').
    rewrite <- app_comm_cons in *.
    change (lc_ectx ((e1', e2') :: D1 ++ D3))
      with (lc_exp e1' /\ lc_exp e2' /\ lc_ectx (D1 ++ D3)) in H0.   
    firstorder.
Qed.


Lemma cut_lc_ectx : forall (D1 D2 : ectx) u1 u2,
  lc_ectx (D1 ++ ((u1,u2)::D2)) ->
  lc_ectx (D1 ++ D2).
Proof.
  induction D1.
    simpl. tauto.
    
    intros D2 u1 u2.
    rewrite <- app_comm_cons. 
    destruct a as (e1,e2).
    simpl.
    intros [H1 [H2 H3]].
    eauto.
Qed.

(* A module type abstracting the important properties from mreduce. *)
Module Type REDUCTION.
  Parameter areduce : exp -> exp -> Prop.

  Axiom lc_areduce : forall e1 e2,
    areduce e1 e2 ->
    lc_exp e1 /\ lc_exp e2.

  Inductive mareduce : exp -> exp -> Prop :=
  | mareduce_refl : forall e,
     lc_exp e ->
     mareduce e e
  | mareduce_step : forall e1 e2 e3,
     areduce e1 e2 ->
     mareduce e2 e3 ->
     mareduce e1 e3.

  Axiom areduce_beta : forall e1 e2,
    reduce e1 e2 ->
    areduce e1 e2.
  
  Axiom areduce_subst : forall w x e1 e2,
    areduce e1 e2 ->
    pure w ->
    areduce (esubst_exp w x e1) (esubst_exp w x e2).

  Axiom areduce_preserves_constructor : forall C e0 e',
    areduce (e_Con C e0) e' ->
    exists e0', 
      e' = (e_Con C e0') /\ areduce e0 e0'.

  Axiom mareduce_diamond : forall e e1 e2,
    mareduce e e1 ->
    mareduce e e2 ->
    exists u, mareduce e1 u /\ mareduce e2 u.
End REDUCTION.

(* Ditto, plus some basic lemmas about mareduce. *)
Module Type REDUCTION_PLUS.
  Include Type REDUCTION.

  Axiom lc_mareduce : forall e e'',
    mareduce e e'' ->  lc_exp e''.

  Axiom mareduce_trans : forall e1 e2 e3,
    mareduce e1 e2 ->
    mareduce e2 e3 ->
    mareduce e1 e3.

  Axiom mareduce_subst : forall w x e e',
    pure w -> 
    mareduce e e' ->
    mareduce (esubst_exp w x e) (esubst_exp w x e').
 
  Axiom mareduce_preserves_constructor : forall C e0 e',
    mareduce (e_Con C e0) e' ->
    exists e0', e' = (e_Con C e0') /\ mareduce e0 e0'.
End REDUCTION_PLUS.

Module ReductionPlus (R : REDUCTION) <: REDUCTION_PLUS.
  Include R.

  Lemma lc_mareduce : forall e e'',
    mareduce e e'' ->  lc_exp e''.
  Proof.
    induction 1; auto.
  Qed.

  Hint Constructors mareduce.
  Hint Resolve areduce_subst.

  Lemma mareduce_trans : forall e1 e2 e3,
    mareduce e1 e2 ->
    mareduce e2 e3 ->
    mareduce e1 e3.
  Proof.
    induction 1; eauto.
  Qed.

  Lemma mareduce_subst : forall w x e e',
    pure w -> 
    mareduce e e' ->
    mareduce (esubst_exp w x e) (esubst_exp w x e').
  Proof.
    induction 2; eauto.
  Qed.

  Lemma mareduce_preserves_constructor_helper : forall C e e',
    mareduce e e' -> forall e0, e = (e_Con C e0) -> exists e0', e' = (e_Con C e0') /\ mareduce e0 e0'.
  Proof.
    induction 1; intros; subst.
      Case "refl"; eauto. 
      Case "step";
        destruct (areduce_preserves_constructor _ _ _ H) as [e0' [Heq' HR]];
        decompose record (IHmareduce e0' Heq');
        eauto.
  Qed.

 
  Lemma mareduce_preserves_constructor : forall C e0 e',
    mareduce (e_Con C e0) e' ->
    exists e0', e' = (e_Con C e0') /\ mareduce e0 e0'.
  Proof.
    intros C e0 e' H.
    apply mareduce_preserves_constructor_helper with (e := e_Con C e0) (e0 := e0); auto.
  Qed.
End ReductionPlus.

Module Reduce_reduce <: REDUCTION.
  Definition areduce := reduce.

  Inductive mareduce : exp -> exp -> Prop :=
  | mareduce_refl : forall e,
     lc_exp e ->
     mareduce e e
  | mareduce_step : forall e1 e2 e3,
     areduce e1 e2 ->
     mareduce e2 e3 ->
     mareduce e1 e3.

  Definition lc_areduce := lc_reduce.
  Lemma areduce_beta : forall e1 e2,
    reduce e1 e2 ->  areduce e1 e2.
  Proof. tauto. Qed.

  Lemma areduce_subst : forall w x e1 e2,
    areduce e1 e2 ->
    pure w ->
    areduce (esubst_exp w x e1) (esubst_exp w x e2).
  Proof.
    intros; unfold areduce; auto using reduce_subst.
  Qed.

  Definition areduce_preserves_constructor := reduce_preserves_constructor.

  Lemma lc_mareduce : forall e e'',
    mareduce e e'' ->  lc_exp e''.
  Proof.
    induction 1; auto.
  Qed.

  Hint Resolve lc_mareduce lc_areduce.
  Hint Constructors mareduce.
  Lemma mareduce_diamond : forall e e' e'',
    mareduce e e' ->
    mareduce e e'' ->
    exists u,
      mareduce e' u /\ mareduce e'' u.
  Proof.
    induction 1.
      Case "refl"; eauto.
      Case "step".
        inversion 1; subst.
          exists e3; eauto.

          match goal with [ H1 : areduce ?e ?e',
                            H2 : areduce ?e ?e'' |- _ ] => rewrite (reduce_deterministic _ _ H1 _ H2) in * end; 
          eauto.
  Qed.

End Reduce_reduce.

Module isEqBeta (R : REDUCTION) <: ISEQ.
  Include  (ReductionPlus(R)).
  
  Definition isEq (D : ectx)  e e' :=
    (lc_ectx D /\ lc_exp e /\ lc_exp e') /\
    exists u, mareduce e u /\ mareduce e' u.
 
  Inductive ctxereq : ectx -> ectx -> Prop :=    (* defn ctxereq *)
  | CEQ_Empty : 
    ctxereq  nil   nil 
  | CEQ_Eq : forall (D:ectx) (e1 e2:exp) (D':ectx) (e1' e2':exp),
    ctxereq D D' ->
    isEq  D   e1   e1'  ->
    isEq  D   e2   e2'  ->
    ctxereq   ( (  e1  ,  e2  ) ::  D  )     ( (  e1'  ,  e2'  ) ::  D'  ).  

  Lemma lc_isEq  : forall D e1 e2, 
    isEq D e1 e2 -> 
    lc_ectx D /\ lc_exp e1 /\ lc_exp e2.
  Proof.
    unfold isEq; tauto.
  Qed.
  
  Lemma lc_ctxereq : forall D1 D2,
    ctxereq D1 D2 ->
    lc_ectx D1 /\ lc_ectx D2.
  Proof.
    induction 1.
      simpl ; auto.
  
      destruct IHctxereq;
      repeat match goal with [H: isEq _ _ _ |- _] =>
                          let H1 := fresh in let H2 := fresh in
                          destruct (lc_isEq _ _ _ H) as [_ [H1 H2]]; clear H
                 end;
      split; simpl; auto.
  Qed.

  Ltac propagate_lc := 
    unfold isEq; intros;
      split ; [repeat split; assumption | idtac].

  Hint Constructors mareduce.
  Hint Resolve lc_mareduce.

  Lemma ref_isEq : forall D e, 
   lc_ectx D -> lc_exp e ->
   isEq D e e.
  Proof.
    propagate_lc.
    exists e; split; auto.
  Qed.

  Lemma sym_isEq : forall D e1 e2, 
    isEq D e1 e2 -> isEq D e2 e1.
  Proof.
    unfold isEq; firstorder.
  Qed.
 
  Lemma trans_isEq : forall D e1 e2 e3, 
    isEq D e1 e2 -> isEq D e2 e3 -> isEq D e1 e3.
  Proof.
    intros D e1 e2 e3 [[lcD [lce1 lce2]] [u1 [Re1u1 Re2u1]]] [[_ [_ lce3]] [u2 [Re2u2 Re3u2]]].
    destruct (mareduce_diamond _ _ _ Re2u1 Re2u2) as [u3 [Ru1u3 Ru2u3]].
    propagate_lc.
    exists u3; eauto using mareduce_trans.
  Qed.

  Lemma abeta_isEq : forall e1 e2, 
    areduce e1 e2 -> isEq nil e1 e2.
  Proof.
    intros e1 e2 H.
    destruct (lc_areduce _ _ H).
    assert (lc_ectx nil) by (simpl; tauto).
    propagate_lc.
    exists e2; split; eauto.
  Qed.

  Lemma beta_isEq  : forall e1 e2, 
    reduce e1 e2 -> isEq nil e1 e2.
  Proof.
    auto using abeta_isEq, areduce_beta.
  Qed.

  Lemma empty_isEq : forall C1 C2 e1 e2, 
    pure e1 -> pure e2 -> 
    C1 <> C2 -> ~(isEq nil (e_Con C1 e1)(e_Con C2 e2)).
  Proof.
    unfold isEq. unfold not.
    intros.
    destruct H2 as [_ [u [R1 R2]]].
    destruct (mareduce_preserves_constructor _ _ _ R1) as [u1 [Hu1 _]].
    destruct (mareduce_preserves_constructor _ _ _ R2) as [u2 [Hu2 _]].
    assert (e_Con C1 u1 = e_Con C2 u2) as Heq by congruence.
    injection Heq. tauto.
  Qed.

  Lemma injection2_strengthened_isEq : forall D C w1 w2,
    isEq D (e_Con C w1) (e_Con C w2) ->
    isEq D w1 w2.
  Proof.
   intros D C w1 w2 [[lcD [lcCw1 lcCw2]] [u [HR1 HR2]]].
   destruct (mareduce_preserves_constructor _ _ _ HR1) as [w1' [Heq1 HR1']].
   destruct (mareduce_preserves_constructor _ _ _ HR2) as [w2' [Heq2 HR2']].
   assert (lc_exp w1) by eauto.
   assert (lc_exp w2) by eauto.
   propagate_lc.
   assert (e_Con C w1' = e_Con C w2') as Heq by congruence.
   injection Heq; intros; subst.
   exists w2'; split; assumption.
  Qed.

  Lemma injection2_isEq : forall D C w1 w2,
    pure w1 ->
    pure w2 ->
    isEq D (e_Con C w1) (e_Con C w2) ->
    isEq D w1 w2.
  Proof.
    intros; apply injection2_strengthened_isEq with (C:=C); assumption.
  Qed.
   
  Lemma ctx_conv_isEq : 
    forall D1 D2 e1 e2,
      ctxereq D1 D2 -> isEq D1 e1 e2 -> isEq D2 e1 e2.
  Proof.
    unfold isEq; intros D1 D2 e1 e2 H [[H1 [H2 H3]] H4].
    destruct (lc_ctxereq _  _ H). 
    auto.
  Qed.

  Lemma cut_isEq : 
    forall D1 D2 u1 u2 e1 e2, 
      isEq (D2 ++ (u1, u2) :: D1) e1 e2 -> 
      isEq D1 u1 u2 -> 
      isEq (D2 ++ D1) e1 e2.
  Proof.
    unfold isEq; intros D1 D2 u1 u2 e1 e2 [[H1 [H2 H3]] H4] _. 
    eauto using cut_lc_ectx.
  Qed.

  Lemma weakening_isEq : 
    forall D1 D2 D3 e1 e2, 
      lc_ectx D2 ->
      isEq (D1 ++ D3) e1 e2 ->
      isEq (D1 ++ D2 ++ D3) e1 e2.
  Proof.
    unfold isEq; intros D1 D2 D3 e1 e2 H [[HlcD1D2 [Hlc_e1 Hlc_e2]] H1]; 
      auto using weakening_lc_ectx.
  Qed.

  Lemma subst_isEq :
    forall D e1 e2 w x,
      pure w ->
      isEq D e1 e2 -> 
      isEq (esubst_ectx w x D)
           (esubst_exp w x e1)
           (esubst_exp w x e2).
  Proof.
    unfold isEq. intros. destruct H0 as [[Hlc_D [Hlc_e1 Hlc_e2]] [u [R1 R2]]].
    split.
      auto. 
      exists (esubst_exp w x u); auto using mareduce_subst. 
  Qed.

End isEqBeta.


Module isEqFiat (R : REDUCTION) : ISEQ.
  Include (ReductionPlus(R)).

  Inductive IEF (D : ectx) : exp -> exp -> Prop :=
  | IEF_refl : forall e,
    lc_exp e ->
    IEF D e e
  | IEF_beta : forall e1 e2,
    areduce e1 e2 ->
    IEF D e1 e2
  | IEF_delta : forall e1 e2,
    In (e1,e2) D ->
    IEF D e1 e2 
  | IEF_symm : forall e1 e2,
    IEF D e1 e2 ->
    IEF D e2 e1
  | IEF_trans : forall e1 e2 e3,
    IEF D e1 e2 ->
    IEF D e2 e3 ->
    IEF D e1 e3
  | IEF_inj : forall C e1 e2,
    IEF D (e_Con C e1) (e_Con C e2) ->
    IEF D e1 e2.
  Hint Constructors IEF.

  Module isEqBeta_areduce := isEqBeta(R).

  Lemma IEF_implies_isEqBeta : forall e1 e2,
    IEF nil e1 e2 -> isEqBeta_areduce.isEq nil e1 e2.
  Proof.
    induction 1.
    Case "IEF_refl";
      apply isEqBeta_areduce.ref_isEq; [apply I | assumption].
    Case "IEF_beta";
      apply isEqBeta_areduce.abeta_isEq. assumption. 
    Case "IEF_delta"; 
      contradict H.
    Case "IEF_symm";
      apply isEqBeta_areduce.sym_isEq; assumption.
    Case "IEF_trans";
      apply isEqBeta_areduce.trans_isEq with (e2:=e2); assumption.
    Case "IEF_inj".
      apply isEqBeta_areduce.injection2_strengthened_isEq with (C:=C); assumption.
  Qed.
  
  Definition isEq D e1 e2 := lc_ectx D /\ IEF D e1 e2.

  Inductive ctxereq : ectx -> ectx -> Prop :=    (* defn ctxereq *)
  | CEQ_Empty : 
    ctxereq  nil   nil 
  | CEQ_Eq : forall (D:ectx) (e1 e2:exp) (D':ectx) (e1' e2':exp),
    ctxereq D D' ->
    isEq  D   e1   e1'  ->
    isEq  D   e2   e2'  ->
    ctxereq   ( (  e1  ,  e2  ) ::  D  )     ( (  e1'  ,  e2'  ) ::  D'  ).  

  Lemma lc_In_ectx : forall (D : ectx) e1 e2,
    lc_ectx D ->
    In (e1,e2) D ->
    lc_exp e1 /\ lc_exp e2.
  Proof.
    induction D.
      intros; 
        match goal with [H: In ?x nil |- _] => contradict H; apply in_nil end.

      intros; destruct a;
        match goal with [H: lc_ectx ((?e1,?e2)::?D) |- _] =>
                          change (lc_ectx ((e1,e2)::D)) with (lc_exp e1 /\ lc_exp e2 /\ lc_ectx D) in H;
                          let H1 := fresh "H" in let H2 := fresh "H" in let H3 := fresh "H" in
                          destruct H as [H1 [H2 H3]]
                   end;
        match goal with [H: In (?e1', ?e2') ((?e1, ?e2)::?D) |- _] =>
                        change (In (e1',e2') ((e1,e2)::D)) with ((e1,e2) = (e1',e2') \/ In (e1',e2') D) in H;
                        destruct H as [H | H]; [injection H; intros; subst | idtac]
                   end;
        auto.
  Qed.

  Hint Resolve lc_areduce.

  Lemma lc_isEq  : forall D e1 e2, 
    isEq D e1 e2 -> 
    lc_ectx D /\ lc_exp e1 /\ lc_exp e2.
  Proof.
    unfold isEq.
    intros D e1 e2 [lcD H].
    split.
      assumption.
      induction H;
        repeat match goal with [H: ?H1 /\ ?H2 |- _] => destruct H
                             | [H: reduce ?e1 ?e2 |- _] => destruct (lc_reduce _ _ H) ; clear H
                             | [H:  lc_ectx ?D,
                                H1: In (?e1,?e2) D |- _] 
                                => assert (lc_exp e1 /\ lc_exp e2) by apply (lc_In_ectx _ _ _ H H1); clear H1 
                             | [H: lc_exp (e_Con ?C ?e) |- _] 
                                => (assert (lc_exp e) by apply (lc_exp_e_Con_body _ _ H); clear H)
                          end;
        auto.
  Qed.

  Lemma lc_ctxereq : forall D1 D2,
    ctxereq D1 D2 ->
    lc_ectx D1 /\ lc_ectx D2.
  Proof.
    induction 1.
      simpl ; auto.
  
      destruct IHctxereq.
      repeat match goal with [H: isEq _ _ _ |- _] =>
                               let H1 := fresh in let H2 := fresh in
                               destruct (lc_isEq _ _ _ H) as [_ [H1 H2]]; clear H
                 end;
      split; simpl;  auto.
  Qed.

  Lemma ref_isEq : forall D e, 
   lc_ectx D -> lc_exp e ->
   isEq D e e.
  Proof.
    unfold isEq; auto.
  Qed.

  Lemma sym_isEq : forall D e1 e2, 
    isEq D e1 e2 -> isEq D e2 e1.
  Proof.
    unfold isEq; intros D e1 e2 [H1 H2]; auto.
  Qed. 

  Lemma trans_isEq : forall D e1 e2 e3, 
    isEq D e1 e2 -> isEq D e2 e3 -> isEq D e1 e3.
  Proof.
    unfold isEq; intros D e1 e2 e3 [H1 H2] [H3 H4]; eauto.
  Qed.

  Hint Constructors True.
  Lemma beta_isEq : forall e1 e2, 
    reduce e1 e2 -> isEq nil e1 e2.
  Proof.
    unfold isEq; split; [apply I | auto using areduce_beta].
  Qed.

  Lemma empty_isEq : forall C1 C2 e1 e2, 
    pure e1 -> pure e2 -> 
    C1 <> C2 -> ~(isEq nil (e_Con C1 e1)(e_Con C2 e2)).
  Proof.
    unfold not.
    intros C1 C2 e1 e2 Hpure1 Hpure2 Hneq [_ H_IEF].
    assert (isEqBeta_areduce.isEq nil (e_Con C1 e1) (e_Con C2 e2)).
      apply IEF_implies_isEqBeta; auto.
      revert H; apply isEqBeta_areduce.empty_isEq; auto.
  Qed.

  Lemma injection2_isEq : forall D C w1 w2,
    pure w1 ->
    pure w2 ->
    isEq D (e_Con C w1) (e_Con C w2) ->
    isEq D w1 w2.
  Proof.
    unfold isEq. intros D C w1 w2 _ _ [H_lc H_IEF].
    split; eauto.
  Qed.

  Lemma weakening_In : forall (D1 D2 D3 : ectx) a,
    In a (D1 ++ D3) ->
    In a (D1 ++ D2 ++ D3).
  Proof.
    induction D1.
      induction D2; simpl in*; firstorder.
      simpl in *; firstorder.
  Qed.

  Lemma weakening_IEF : forall D1 D2 D3 e1 e2,
    IEF (D1++D3) e1 e2 ->
    IEF (D1++D2++D3) e1 e2.
  Proof.
    induction 1; auto using weakening_In; eauto.
  Qed.

  Lemma weakening_isEq : 
    forall D1 D2 D3 e1 e2, 
      lc_ectx D2 ->
      isEq (D1 ++ D3) e1 e2 ->
      isEq (D1 ++ D2 ++ D3) e1 e2.
  Proof.
    unfold isEq; intros D1 D2 D4 e1 e2 H [H1 H2].
    auto using weakening_lc_ectx, weakening_IEF.
  Qed.

  Lemma weakening1_isEq : 
    forall u1 u2 D e1 e2, 
      lc_exp u1 ->
      lc_exp u2 ->
      isEq D e1 e2 ->
      isEq ((u1,u2)::D) e1 e2.
  Proof.
    intros.  
    apply weakening_isEq with (D3:=D) (D2:=(u1,u2)::nil) (D1:=nil);
       simpl; auto.
  Qed.

  Lemma weakening1_IEF : 
    forall u1 u2 D e1 e2,
      IEF D e1 e2 ->
      IEF ((u1,u2)::D) e1 e2.
  Proof.
    induction 1; auto; eauto.
      Case "IEF_delta"; apply IEF_delta; simpl; auto.
  Qed.  
  Hint Resolve weakening1_isEq.
  Hint Resolve weakening1_IEF.
 

  Lemma ctx_conv_isEq : 
    forall D1 D2 e1 e2,
      ctxereq D1 D2 -> isEq D1 e1 e2 -> isEq D2 e1 e2.
  Proof.
    intros D1 D2 e1 e2 Hctxereq [Hlc_D H_IEF].
    destruct (lc_ctxereq _ _ Hctxereq).
    split.
      assumption.

      clear H H0 Hlc_D.
      revert e1 e2 H_IEF.
      induction Hctxereq.
        tauto.

        induction 1; auto.
          Case "IEF_delta".
          change (In (e0,e3) ((e1,e2)::D)) with ((e1,e2)=(e0,e3) \/ In (e0,e3) D) in H1.
          destruct H1.
            SCase "used equation".
            injection H1; intros; subst.
            assert (lc_exp e1') by apply (lc_isEq _ _ _ H).
            destruct H as [_ H].
            assert (lc_exp e2') by apply (lc_isEq _ _ _ H0).
            destruct H0 as [_ H0].
            assert (IEF ((e1',e2')::D') e1' e2') by (apply IEF_delta; simpl; auto).
            assert (IEF ((e1',e2')::D') e0 e1') by auto.
            assert (IEF ((e1',e2')::D') e3 e2') by auto.
            eauto.
            
            SCase "didn't use equation"; auto.
         Case "IEF_trans"; eauto.
         Case "IEF_inj"; eauto.
  Qed.
 
  Lemma In_middle : forall (D1 D2 : ectx) u1 u2 e1 e2,
    In (e1,e2) (D2 ++ (u1,u2) :: D1) ->
    (u1,u2) = (e1,e2) \/ In (e1,e2) (D2 ++ D1).
  Proof.
    intros. 
    remember (D2 ++ (u1,u2) :: D1) as D.
    revert H D1 D2 u1 u2 HeqD.
    induction D.
      intros. contradict H.

      intros.
      destruct D2.
        simpl in *. inversion HeqD; intros; subst. auto.
        

        rewrite <- app_comm_cons in HeqD. 
        rewrite <- app_comm_cons.
        inversion HeqD; intros.
        change (In (e1,e2) (a::D)) with (a=(e1,e2) \/  In (e1,e2) D) in H.        
        destruct H.
          subst; subst; simpl; auto.
          pose (IHD H D1 D2 u1 u2 H2); simpl; tauto.
  Qed.

  Lemma cut_isEq : 
    forall D1 D2 u1 u2 e1 e2, 
      isEq (D2 ++ (u1, u2) :: D1) e1 e2 -> 
      isEq D1 u1 u2 -> 
      isEq (D2 ++ D1) e1 e2.
  Proof.
    intros D1 D2 u1 u2 e1 e2 [Hlc_ectx H_IEF] [_ H_IEF']. 
    split.
      apply cut_lc_ectx with (u1:=u1) (u2:=u2); auto.

    induction H_IEF; auto.
      Case "IEF_delta". 
      destruct (In_middle _ _ _ _ _ _ H) as [H0 | H0].
        injection H0; intros; subst;
        apply weakening_IEF with (D1:=nil) (D2:=D2); assumption.
      
        auto.
      Case "IEF_trans"; eauto.
      Case "IEF_inj"; eauto.
  Qed.

  Lemma subst_In :
    forall u x D e1 e2,
      In (e1,e2) D ->
      In (esubst_exp u x e1, esubst_exp u x e2) (esubst_ectx u x D).
  Proof.
   induction D; intros. 
     auto.
     
     destruct a as (u1,u2).
     change (In (e1, e2) ((u1, u2) :: D)) with  ((u1,u2)=(e1,e2) \/ In (e1,e2) D) in H; destruct H.
       injection H; intros; subst; simpl; auto.
       simpl; auto.
  Qed.

  Lemma subst_isEq :
    forall D e1 e2 w x,
      pure w ->
      isEq D e1 e2 -> 
      isEq (esubst_ectx w x D)
           (esubst_exp w x e1)
           (esubst_exp w x e2).
  Proof.
    intros D e1 e2 w x Hpure [HlcD  H_IEF].
    split.
      auto.
      induction H_IEF; auto using reduce_subst, subst_In; eauto.
  Qed.
End isEqFiat.
  
