Add LoadPath "./metatheory".
Require Import Coq.Arith.Wf_nat.
Require Import Coq.Logic.FunctionalExtensionality.
Require Import Coq.Program.Equality.

Require Export Metatheory.
Require Export LibLNgen.

Require Export lang.

(** NOTE: Auxiliary theorems are hidden in generated documentation.
    In general, there is a [_rec] version of every lemma involving
    [open] and [close]. *)


(* *********************************************************************** *)
(** * Induction principles for nonterminals *)

Scheme Cs_ind' := Induction for Cs Sort Prop.

Definition Cs_mutind :=
  fun H1 H2 H3 =>
  Cs_ind' H1 H2 H3.

Scheme Cs_rec' := Induction for Cs Sort Set.

Definition Cs_mutrec :=
  fun H1 H2 H3 =>
  Cs_rec' H1 H2 H3.

Scheme exp_ind' := Induction for exp Sort Prop
  with peb_ind' := Induction for peb Sort Prop
  with pe_ind' := Induction for pe Sort Prop.

Definition exp_peb_pe_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17 =>
  (conj (exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)
  ((conj (peb_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)
  (pe_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)))).

Scheme exp_rec' := Induction for exp Sort Set
  with peb_rec' := Induction for peb Sort Set
  with pe_rec' := Induction for pe Sort Set.

Definition exp_peb_pe_mutrec :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17 =>
  (pair ((pair (exp_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)
  (peb_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)))
  (pe_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)).

Scheme type_ind' := Induction for type Sort Prop
  with ptt_ind' := Induction for ptt Sort Prop
  with pt_ind' := Induction for pt Sort Prop.

Definition type_ptt_pt_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (conj (type_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  ((conj (ptt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (pt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))).

Scheme type_rec' := Induction for type Sort Set
  with ptt_rec' := Induction for ptt Sort Set
  with pt_rec' := Induction for pt Sort Set.

Definition type_ptt_pt_mutrec :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (pair ((pair (type_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (ptt_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))
  (pt_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)).

Scheme kind_ind' := Induction for kind Sort Prop.

Definition kind_mutind :=
  fun H1 H2 H3 =>
  kind_ind' H1 H2 H3.

Scheme kind_rec' := Induction for kind Sort Set.

Definition kind_mutrec :=
  fun H1 H2 H3 =>
  kind_rec' H1 H2 H3.

Scheme bind_ind' := Induction for bind Sort Prop.

Definition bind_mutind :=
  fun H1 H2 H3 =>
  bind_ind' H1 H2 H3.

Scheme bind_rec' := Induction for bind Sort Set.

Definition bind_mutrec :=
  fun H1 H2 H3 =>
  bind_rec' H1 H2 H3.


(* *********************************************************************** *)
(** * Close *)

Fixpoint close_exp_wrt_exp_rec (n1 : nat) (x1 : expvar) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | e_Var_f x2 => if (x1 == x2) then (e_Var_b n1) else (e_Var_f x2)
    | e_Var_b n2 => if (lt_ge_dec n2 n1) then (e_Var_b n2) else (e_Var_b (S n2))
    | e_Con C1 e2 => e_Con C1 (close_exp_wrt_exp_rec n1 x1 e2)
    | e_Lam e2 => e_Lam (close_exp_wrt_exp_rec (S n1) x1 e2)
    | e_App e2 e3 => e_App (close_exp_wrt_exp_rec n1 x1 e2) (close_exp_wrt_exp_rec n1 x1 e3)
    | e_Case e2 pe1 => e_Case (close_exp_wrt_exp_rec n1 x1 e2) (close_pe_wrt_exp_rec n1 x1 pe1)
    | e_Fix e2 => e_Fix (close_exp_wrt_exp_rec (S n1) x1 e2)
    | e_Sie e2 e3 => e_Sie (close_exp_wrt_exp_rec n1 x1 e2) (close_exp_wrt_exp_rec n1 x1 e3)
    | e_Proj1 e2 => e_Proj1 (close_exp_wrt_exp_rec n1 x1 e2)
    | e_Proj2 e2 => e_Proj2 (close_exp_wrt_exp_rec n1 x1 e2)
    | e_unit => e_unit
  end

with close_peb_wrt_exp_rec (n1 : nat) (x1 : expvar) (peb1 : peb) {struct peb1} : peb :=
  match peb1 with
    | peb_Branch C1 e1 => peb_Branch C1 (close_exp_wrt_exp_rec (S n1) x1 e1)
  end

with close_pe_wrt_exp_rec (n1 : nat) (x1 : expvar) (pe1 : pe) {struct pe1} : pe :=
  match pe1 with
    | pe_Empty => pe_Empty
    | pe_Cons peb1 pe2 => pe_Cons (close_peb_wrt_exp_rec n1 x1 peb1) (close_pe_wrt_exp_rec n1 x1 pe2)
  end.

Definition close_exp_wrt_exp x1 e1 := close_exp_wrt_exp_rec 0 x1 e1.

Definition close_peb_wrt_exp x1 peb1 := close_peb_wrt_exp_rec 0 x1 peb1.

Definition close_pe_wrt_exp x1 pe1 := close_pe_wrt_exp_rec 0 x1 pe1.

Fixpoint close_type_wrt_exp_rec (n1 : nat) (x1 : expvar) (t1 : type) {struct t1} : type :=
  match t1 with
    | t_Con T1 => t_Con T1
    | t_Unit => t_Unit
    | t_Pi t2 t3 => t_Pi (close_type_wrt_exp_rec n1 x1 t2) (close_type_wrt_exp_rec (S n1) x1 t3)
    | t_Sigma t2 t3 => t_Sigma (close_type_wrt_exp_rec n1 x1 t2) (close_type_wrt_exp_rec (S n1) x1 t3)
    | t_App t2 e1 => t_App (close_type_wrt_exp_rec n1 x1 t2) (close_exp_wrt_exp_rec n1 x1 e1)
    | t_Case e1 T1 u1 pt1 => t_Case (close_exp_wrt_exp_rec n1 x1 e1) T1 (close_exp_wrt_exp_rec n1 x1 u1) (close_pt_wrt_exp_rec n1 x1 pt1)
  end

with close_ptt_wrt_exp_rec (n1 : nat) (x1 : expvar) (ptt1 : ptt) {struct ptt1} : ptt :=
  match ptt1 with
    | ptt_pattern C1 t1 => ptt_pattern C1 (close_type_wrt_exp_rec (S n1) x1 t1)
  end

with close_pt_wrt_exp_rec (n1 : nat) (x1 : expvar) (pt1 : pt) {struct pt1} : pt :=
  match pt1 with
    | pt_empty => pt_empty
    | pt_cons ptt1 pt2 => pt_cons (close_ptt_wrt_exp_rec n1 x1 ptt1) (close_pt_wrt_exp_rec n1 x1 pt2)
  end.

Definition close_type_wrt_exp x1 t1 := close_type_wrt_exp_rec 0 x1 t1.

Definition close_ptt_wrt_exp x1 ptt1 := close_ptt_wrt_exp_rec 0 x1 ptt1.

Definition close_pt_wrt_exp x1 pt1 := close_pt_wrt_exp_rec 0 x1 pt1.

Fixpoint close_kind_wrt_exp_rec (n1 : nat) (x1 : expvar) (k1 : kind) {struct k1} : kind :=
  match k1 with
    | k_type => k_type
    | k_pi t1 k2 => k_pi (close_type_wrt_exp_rec n1 x1 t1) (close_kind_wrt_exp_rec (S n1) x1 k2)
  end.

Definition close_kind_wrt_exp x1 k1 := close_kind_wrt_exp_rec 0 x1 k1.

Fixpoint close_bind_wrt_exp_rec (n1 : nat) (x1 : expvar) (bind1 : bind) {struct bind1} : bind :=
  match bind1 with
    | bind_t t1 => bind_t (close_type_wrt_exp_rec n1 x1 t1)
    | bind_k k1 => bind_k (close_kind_wrt_exp_rec n1 x1 k1)
  end.

Definition close_bind_wrt_exp x1 bind1 := close_bind_wrt_exp_rec 0 x1 bind1.


(* *********************************************************************** *)
(** * Size *)

Fixpoint size_Cs (Cs1 : Cs) {struct Cs1} : nat :=
  match Cs1 with
    | Cs_Empty => 1
    | Cs_Cons C1 Cs2 => 1 + (size_Cs Cs2)
  end.

Fixpoint size_exp (e1 : exp) {struct e1} : nat :=
  match e1 with
    | e_Var_f x1 => 1
    | e_Var_b n1 => 1
    | e_Con C1 e2 => 1 + (size_exp e2)
    | e_Lam e2 => 1 + (size_exp e2)
    | e_App e2 e3 => 1 + (size_exp e2) + (size_exp e3)
    | e_Case e2 pe1 => 1 + (size_exp e2) + (size_pe pe1)
    | e_Fix e2 => 1 + (size_exp e2)
    | e_Sie e2 e3 => 1 + (size_exp e2) + (size_exp e3)
    | e_Proj1 e2 => 1 + (size_exp e2)
    | e_Proj2 e2 => 1 + (size_exp e2)
    | e_unit => 1
  end

with size_peb (peb1 : peb) {struct peb1} : nat :=
  match peb1 with
    | peb_Branch C1 e1 => 1 + (size_exp e1)
  end

with size_pe (pe1 : pe) {struct pe1} : nat :=
  match pe1 with
    | pe_Empty => 1
    | pe_Cons peb1 pe2 => 1 + (size_peb peb1) + (size_pe pe2)
  end.

Fixpoint size_type (t1 : type) {struct t1} : nat :=
  match t1 with
    | t_Con T1 => 1
    | t_Unit => 1
    | t_Pi t2 t3 => 1 + (size_type t2) + (size_type t3)
    | t_Sigma t2 t3 => 1 + (size_type t2) + (size_type t3)
    | t_App t2 e1 => 1 + (size_type t2) + (size_exp e1)
    | t_Case e1 T1 u1 pt1 => 1 + (size_exp e1) + (size_exp u1) + (size_pt pt1)
  end

with size_ptt (ptt1 : ptt) {struct ptt1} : nat :=
  match ptt1 with
    | ptt_pattern C1 t1 => 1 + (size_type t1)
  end

with size_pt (pt1 : pt) {struct pt1} : nat :=
  match pt1 with
    | pt_empty => 1
    | pt_cons ptt1 pt2 => 1 + (size_ptt ptt1) + (size_pt pt2)
  end.

Fixpoint size_kind (k1 : kind) {struct k1} : nat :=
  match k1 with
    | k_type => 1
    | k_pi t1 k2 => 1 + (size_type t1) + (size_kind k2)
  end.

Fixpoint size_bind (bind1 : bind) {struct bind1} : nat :=
  match bind1 with
    | bind_t t1 => 1 + (size_type t1)
    | bind_k k1 => 1 + (size_kind k1)
  end.


(* *********************************************************************** *)
(** * Degree *)

(** These define only an upper bound, not a strict upper bound. *)

Inductive degree_exp_wrt_exp : nat -> exp -> Prop :=
  | degree_wrt_exp_e_Var_f : forall n1 x1,
    degree_exp_wrt_exp n1 (e_Var_f x1)
  | degree_wrt_exp_e_Var_b : forall n1 n2,
    lt n2 n1 ->
    degree_exp_wrt_exp n1 (e_Var_b n2)
  | degree_wrt_exp_e_Con : forall n1 C1 e1,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 (e_Con C1 e1)
  | degree_wrt_exp_e_Lam : forall n1 e1,
    degree_exp_wrt_exp (S n1) e1 ->
    degree_exp_wrt_exp n1 (e_Lam e1)
  | degree_wrt_exp_e_App : forall n1 e1 e2,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 e2 ->
    degree_exp_wrt_exp n1 (e_App e1 e2)
  | degree_wrt_exp_e_Case : forall n1 e1 pe1,
    degree_exp_wrt_exp n1 e1 ->
    degree_pe_wrt_exp n1 pe1 ->
    degree_exp_wrt_exp n1 (e_Case e1 pe1)
  | degree_wrt_exp_e_Fix : forall n1 e1,
    degree_exp_wrt_exp (S n1) e1 ->
    degree_exp_wrt_exp n1 (e_Fix e1)
  | degree_wrt_exp_e_Sie : forall n1 e1 e2,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 e2 ->
    degree_exp_wrt_exp n1 (e_Sie e1 e2)
  | degree_wrt_exp_e_Proj1 : forall n1 e1,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 (e_Proj1 e1)
  | degree_wrt_exp_e_Proj2 : forall n1 e1,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 (e_Proj2 e1)
  | degree_wrt_exp_e_unit : forall n1,
    degree_exp_wrt_exp n1 (e_unit)

with degree_peb_wrt_exp : nat -> peb -> Prop :=
  | degree_wrt_exp_peb_Branch : forall n1 C1 e1,
    degree_exp_wrt_exp (S n1) e1 ->
    degree_peb_wrt_exp n1 (peb_Branch C1 e1)

with degree_pe_wrt_exp : nat -> pe -> Prop :=
  | degree_wrt_exp_pe_Empty : forall n1,
    degree_pe_wrt_exp n1 (pe_Empty)
  | degree_wrt_exp_pe_Cons : forall n1 peb1 pe1,
    degree_peb_wrt_exp n1 peb1 ->
    degree_pe_wrt_exp n1 pe1 ->
    degree_pe_wrt_exp n1 (pe_Cons peb1 pe1).

Scheme degree_exp_wrt_exp_ind' := Induction for degree_exp_wrt_exp Sort Prop
  with degree_peb_wrt_exp_ind' := Induction for degree_peb_wrt_exp Sort Prop
  with degree_pe_wrt_exp_ind' := Induction for degree_pe_wrt_exp Sort Prop.

Definition degree_exp_wrt_exp_degree_peb_wrt_exp_degree_pe_wrt_exp_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17 =>
  (conj (degree_exp_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)
  ((conj (degree_peb_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)
  (degree_pe_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 H17)))).

Hint Constructors degree_exp_wrt_exp : core lngen.

Hint Constructors degree_peb_wrt_exp : core lngen.

Hint Constructors degree_pe_wrt_exp : core lngen.

Inductive degree_type_wrt_exp : nat -> type -> Prop :=
  | degree_wrt_exp_t_Con : forall n1 T1,
    degree_type_wrt_exp n1 (t_Con T1)
  | degree_wrt_exp_t_Unit : forall n1,
    degree_type_wrt_exp n1 (t_Unit)
  | degree_wrt_exp_t_Pi : forall n1 t1 t2,
    degree_type_wrt_exp n1 t1 ->
    degree_type_wrt_exp (S n1) t2 ->
    degree_type_wrt_exp n1 (t_Pi t1 t2)
  | degree_wrt_exp_t_Sigma : forall n1 t1 t2,
    degree_type_wrt_exp n1 t1 ->
    degree_type_wrt_exp (S n1) t2 ->
    degree_type_wrt_exp n1 (t_Sigma t1 t2)
  | degree_wrt_exp_t_App : forall n1 t1 e1,
    degree_type_wrt_exp n1 t1 ->
    degree_exp_wrt_exp n1 e1 ->
    degree_type_wrt_exp n1 (t_App t1 e1)
  | degree_wrt_exp_t_Case : forall n1 e1 T1 u1 pt1,
    degree_exp_wrt_exp n1 e1 ->
    degree_exp_wrt_exp n1 u1 ->
    degree_pt_wrt_exp n1 pt1 ->
    degree_type_wrt_exp n1 (t_Case e1 T1 u1 pt1)

with degree_ptt_wrt_exp : nat -> ptt -> Prop :=
  | degree_wrt_exp_ptt_pattern : forall n1 C1 t1,
    degree_type_wrt_exp (S n1) t1 ->
    degree_ptt_wrt_exp n1 (ptt_pattern C1 t1)

with degree_pt_wrt_exp : nat -> pt -> Prop :=
  | degree_wrt_exp_pt_empty : forall n1,
    degree_pt_wrt_exp n1 (pt_empty)
  | degree_wrt_exp_pt_cons : forall n1 ptt1 pt1,
    degree_ptt_wrt_exp n1 ptt1 ->
    degree_pt_wrt_exp n1 pt1 ->
    degree_pt_wrt_exp n1 (pt_cons ptt1 pt1).

Scheme degree_type_wrt_exp_ind' := Induction for degree_type_wrt_exp Sort Prop
  with degree_ptt_wrt_exp_ind' := Induction for degree_ptt_wrt_exp Sort Prop
  with degree_pt_wrt_exp_ind' := Induction for degree_pt_wrt_exp Sort Prop.

Definition degree_type_wrt_exp_degree_ptt_wrt_exp_degree_pt_wrt_exp_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (conj (degree_type_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  ((conj (degree_ptt_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (degree_pt_wrt_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))).

Hint Constructors degree_type_wrt_exp : core lngen.

Hint Constructors degree_ptt_wrt_exp : core lngen.

Hint Constructors degree_pt_wrt_exp : core lngen.

Inductive degree_kind_wrt_exp : nat -> kind -> Prop :=
  | degree_wrt_exp_k_type : forall n1,
    degree_kind_wrt_exp n1 (k_type)
  | degree_wrt_exp_k_pi : forall n1 t1 k1,
    degree_type_wrt_exp n1 t1 ->
    degree_kind_wrt_exp (S n1) k1 ->
    degree_kind_wrt_exp n1 (k_pi t1 k1).

Scheme degree_kind_wrt_exp_ind' := Induction for degree_kind_wrt_exp Sort Prop.

Definition degree_kind_wrt_exp_mutind :=
  fun H1 H2 H3 =>
  degree_kind_wrt_exp_ind' H1 H2 H3.

Hint Constructors degree_kind_wrt_exp : core lngen.

Inductive degree_bind_wrt_exp : nat -> bind -> Prop :=
  | degree_wrt_exp_bind_t : forall n1 t1,
    degree_type_wrt_exp n1 t1 ->
    degree_bind_wrt_exp n1 (bind_t t1)
  | degree_wrt_exp_bind_k : forall n1 k1,
    degree_kind_wrt_exp n1 k1 ->
    degree_bind_wrt_exp n1 (bind_k k1).

Scheme degree_bind_wrt_exp_ind' := Induction for degree_bind_wrt_exp Sort Prop.

Definition degree_bind_wrt_exp_mutind :=
  fun H1 H2 H3 =>
  degree_bind_wrt_exp_ind' H1 H2 H3.

Hint Constructors degree_bind_wrt_exp : core lngen.


(* *********************************************************************** *)
(** * Local closure (version in [Set], induction principles) *)

Inductive lc_set_exp : exp -> Set :=
  | lc_set_e_Var_f : forall x1,
    lc_set_exp (e_Var_f x1)
  | lc_set_e_Con : forall C1 e1,
    lc_set_exp e1 ->
    lc_set_exp (e_Con C1 e1)
  | lc_set_e_Lam : forall e1,
    (forall x1 : expvar, lc_set_exp (open_exp_wrt_exp e1 (e_Var_f x1))) ->
    lc_set_exp (e_Lam e1)
  | lc_set_e_App : forall e1 e2,
    lc_set_exp e1 ->
    lc_set_exp e2 ->
    lc_set_exp (e_App e1 e2)
  | lc_set_e_Case : forall e1 pe1,
    lc_set_exp e1 ->
    lc_set_pe pe1 ->
    lc_set_exp (e_Case e1 pe1)
  | lc_set_e_Fix : forall e1,
    (forall f1 : expvar, lc_set_exp (open_exp_wrt_exp e1 (e_Var_f f1))) ->
    lc_set_exp (e_Fix e1)
  | lc_set_e_Sie : forall e1 e2,
    lc_set_exp e1 ->
    lc_set_exp e2 ->
    lc_set_exp (e_Sie e1 e2)
  | lc_set_e_Proj1 : forall e1,
    lc_set_exp e1 ->
    lc_set_exp (e_Proj1 e1)
  | lc_set_e_Proj2 : forall e1,
    lc_set_exp e1 ->
    lc_set_exp (e_Proj2 e1)
  | lc_set_e_unit :
    lc_set_exp (e_unit)

with lc_set_peb : peb -> Set :=
  | lc_set_peb_Branch : forall C1 e1,
    (forall x1 : expvar, lc_set_exp (open_exp_wrt_exp e1 (e_Var_f x1))) ->
    lc_set_peb (peb_Branch C1 e1)

with lc_set_pe : pe -> Set :=
  | lc_set_pe_Empty :
    lc_set_pe (pe_Empty)
  | lc_set_pe_Cons : forall peb1 pe1,
    lc_set_peb peb1 ->
    lc_set_pe pe1 ->
    lc_set_pe (pe_Cons peb1 pe1).

Scheme lc_exp_ind' := Induction for lc_exp Sort Prop
  with lc_peb_ind' := Induction for lc_peb Sort Prop
  with lc_pe_ind' := Induction for lc_pe Sort Prop.

Definition lc_exp_lc_peb_lc_pe_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 =>
  (conj (lc_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)
  ((conj (lc_peb_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)
  (lc_pe_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)))).

Scheme lc_set_exp_ind' := Induction for lc_set_exp Sort Prop
  with lc_set_peb_ind' := Induction for lc_set_peb Sort Prop
  with lc_set_pe_ind' := Induction for lc_set_pe Sort Prop.

Definition lc_set_exp_lc_set_peb_lc_set_pe_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 =>
  (conj (lc_set_exp_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)
  ((conj (lc_set_peb_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)
  (lc_set_pe_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)))).

Scheme lc_set_exp_rec' := Induction for lc_set_exp Sort Set
  with lc_set_peb_rec' := Induction for lc_set_peb Sort Set
  with lc_set_pe_rec' := Induction for lc_set_pe Sort Set.

Definition lc_set_exp_lc_set_peb_lc_set_pe_mutrec :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16 =>
  (pair ((pair (lc_set_exp_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)
  (lc_set_peb_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)))
  (lc_set_pe_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 H13 H14 H15 H16)).

Hint Constructors lc_exp : core lngen.

Hint Constructors lc_peb : core lngen.

Hint Constructors lc_pe : core lngen.

Hint Constructors lc_set_exp : core lngen.

Hint Constructors lc_set_peb : core lngen.

Hint Constructors lc_set_pe : core lngen.

Inductive lc_set_type : type -> Set :=
  | lc_set_t_Con : forall T1,
    lc_set_type (t_Con T1)
  | lc_set_t_Unit :
    lc_set_type (t_Unit)
  | lc_set_t_Pi : forall t1 t2,
    lc_set_type t1 ->
    (forall x1 : expvar, lc_set_type (open_type_wrt_exp t2 (e_Var_f x1))) ->
    lc_set_type (t_Pi t1 t2)
  | lc_set_t_Sigma : forall t1 t2,
    lc_set_type t1 ->
    (forall x1 : expvar, lc_set_type (open_type_wrt_exp t2 (e_Var_f x1))) ->
    lc_set_type (t_Sigma t1 t2)
  | lc_set_t_App : forall t1 e1,
    lc_set_type t1 ->
    lc_set_exp e1 ->
    lc_set_type (t_App t1 e1)
  | lc_set_t_Case : forall e1 T1 u1 pt1,
    lc_set_exp e1 ->
    lc_set_exp u1 ->
    lc_set_pt pt1 ->
    lc_set_type (t_Case e1 T1 u1 pt1)

with lc_set_ptt : ptt -> Set :=
  | lc_set_ptt_pattern : forall C1 t1,
    (forall x1 : expvar, lc_set_type (open_type_wrt_exp t1 (e_Var_f x1))) ->
    lc_set_ptt (ptt_pattern C1 t1)

with lc_set_pt : pt -> Set :=
  | lc_set_pt_empty :
    lc_set_pt (pt_empty)
  | lc_set_pt_cons : forall ptt1 pt1,
    lc_set_ptt ptt1 ->
    lc_set_pt pt1 ->
    lc_set_pt (pt_cons ptt1 pt1).

Scheme lc_type_ind' := Induction for lc_type Sort Prop
  with lc_ptt_ind' := Induction for lc_ptt Sort Prop
  with lc_pt_ind' := Induction for lc_pt Sort Prop.

Definition lc_type_lc_ptt_lc_pt_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (conj (lc_type_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  ((conj (lc_ptt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (lc_pt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))).

Scheme lc_set_type_ind' := Induction for lc_set_type Sort Prop
  with lc_set_ptt_ind' := Induction for lc_set_ptt Sort Prop
  with lc_set_pt_ind' := Induction for lc_set_pt Sort Prop.

Definition lc_set_type_lc_set_ptt_lc_set_pt_mutind :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (conj (lc_set_type_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  ((conj (lc_set_ptt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (lc_set_pt_ind' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))).

Scheme lc_set_type_rec' := Induction for lc_set_type Sort Set
  with lc_set_ptt_rec' := Induction for lc_set_ptt Sort Set
  with lc_set_pt_rec' := Induction for lc_set_pt Sort Set.

Definition lc_set_type_lc_set_ptt_lc_set_pt_mutrec :=
  fun H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12 =>
  (pair ((pair (lc_set_type_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)
  (lc_set_ptt_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)))
  (lc_set_pt_rec' H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 H11 H12)).

Hint Constructors lc_type : core lngen.

Hint Constructors lc_ptt : core lngen.

Hint Constructors lc_pt : core lngen.

Hint Constructors lc_set_type : core lngen.

Hint Constructors lc_set_ptt : core lngen.

Hint Constructors lc_set_pt : core lngen.

Inductive lc_set_kind : kind -> Set :=
  | lc_set_k_type :
    lc_set_kind (k_type)
  | lc_set_k_pi : forall t1 k1,
    lc_set_type t1 ->
    (forall x1 : expvar, lc_set_kind (open_kind_wrt_exp k1 (e_Var_f x1))) ->
    lc_set_kind (k_pi t1 k1).

Scheme lc_kind_ind' := Induction for lc_kind Sort Prop.

Definition lc_kind_mutind :=
  fun H1 H2 H3 =>
  lc_kind_ind' H1 H2 H3.

Scheme lc_set_kind_ind' := Induction for lc_set_kind Sort Prop.

Definition lc_set_kind_mutind :=
  fun H1 H2 H3 =>
  lc_set_kind_ind' H1 H2 H3.

Scheme lc_set_kind_rec' := Induction for lc_set_kind Sort Set.

Definition lc_set_kind_mutrec :=
  fun H1 H2 H3 =>
  lc_set_kind_rec' H1 H2 H3.

Hint Constructors lc_kind : core lngen.

Hint Constructors lc_set_kind : core lngen.

Inductive lc_set_bind : bind -> Set :=
  | lc_set_bind_t : forall t1,
    lc_set_type t1 ->
    lc_set_bind (bind_t t1)
  | lc_set_bind_k : forall k1,
    lc_set_kind k1 ->
    lc_set_bind (bind_k k1).

Scheme lc_bind_ind' := Induction for lc_bind Sort Prop.

Definition lc_bind_mutind :=
  fun H1 H2 H3 =>
  lc_bind_ind' H1 H2 H3.

Scheme lc_set_bind_ind' := Induction for lc_set_bind Sort Prop.

Definition lc_set_bind_mutind :=
  fun H1 H2 H3 =>
  lc_set_bind_ind' H1 H2 H3.

Scheme lc_set_bind_rec' := Induction for lc_set_bind Sort Set.

Definition lc_set_bind_mutrec :=
  fun H1 H2 H3 =>
  lc_set_bind_rec' H1 H2 H3.

Hint Constructors lc_bind : core lngen.

Hint Constructors lc_set_bind : core lngen.


(* *********************************************************************** *)
(** * Body *)

Definition body_exp_wrt_exp e1 := forall x1, lc_exp (open_exp_wrt_exp e1 (e_Var_f x1)).

Definition body_peb_wrt_exp peb1 := forall x1, lc_peb (open_peb_wrt_exp peb1 (e_Var_f x1)).

Definition body_pe_wrt_exp pe1 := forall x1, lc_pe (open_pe_wrt_exp pe1 (e_Var_f x1)).

Hint Unfold body_exp_wrt_exp.

Hint Unfold body_peb_wrt_exp.

Hint Unfold body_pe_wrt_exp.

Definition body_type_wrt_exp t1 := forall x1, lc_type (open_type_wrt_exp t1 (e_Var_f x1)).

Definition body_ptt_wrt_exp ptt1 := forall x1, lc_ptt (open_ptt_wrt_exp ptt1 (e_Var_f x1)).

Definition body_pt_wrt_exp pt1 := forall x1, lc_pt (open_pt_wrt_exp pt1 (e_Var_f x1)).

Hint Unfold body_type_wrt_exp.

Hint Unfold body_ptt_wrt_exp.

Hint Unfold body_pt_wrt_exp.

Definition body_kind_wrt_exp k1 := forall x1, lc_kind (open_kind_wrt_exp k1 (e_Var_f x1)).

Hint Unfold body_kind_wrt_exp.

Definition body_bind_wrt_exp bind1 := forall x1, lc_bind (open_bind_wrt_exp bind1 (e_Var_f x1)).

Hint Unfold body_bind_wrt_exp.


(* *********************************************************************** *)
(** * Tactic support *)

(** Additional hint declarations. *)

Hint Resolve @plus_le_compat : lngen.

(** Redefine some tactics. *)

Ltac default_case_split ::=
  first
    [ progress destruct_notin
    | progress destruct_sum
    | progress safe_f_equal
    ].


(* *********************************************************************** *)
(** * Theorems about [size] *)

Ltac default_auto ::= auto with arith lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma size_Cs_min_mutual :
(forall Cs1, 1 <= size_Cs Cs1).
Proof.
apply_mutual_ind Cs_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_Cs_min :
forall Cs1, 1 <= size_Cs Cs1.
Proof.
pose proof size_Cs_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_Cs_min : lngen.

(* begin hide *)

Lemma size_exp_min_size_peb_min_size_pe_min_mutual :
(forall e1, 1 <= size_exp e1) /\
(forall peb1, 1 <= size_peb peb1) /\
(forall pe1, 1 <= size_pe pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_exp_min :
forall e1, 1 <= size_exp e1.
Proof.
pose proof size_exp_min_size_peb_min_size_pe_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_exp_min : lngen.

Lemma size_peb_min :
forall peb1, 1 <= size_peb peb1.
Proof.
pose proof size_exp_min_size_peb_min_size_pe_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_peb_min : lngen.

Lemma size_pe_min :
forall pe1, 1 <= size_pe pe1.
Proof.
pose proof size_exp_min_size_peb_min_size_pe_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pe_min : lngen.

(* begin hide *)

Lemma size_type_min_size_ptt_min_size_pt_min_mutual :
(forall t1, 1 <= size_type t1) /\
(forall ptt1, 1 <= size_ptt ptt1) /\
(forall pt1, 1 <= size_pt pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_type_min :
forall t1, 1 <= size_type t1.
Proof.
pose proof size_type_min_size_ptt_min_size_pt_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_type_min : lngen.

Lemma size_ptt_min :
forall ptt1, 1 <= size_ptt ptt1.
Proof.
pose proof size_type_min_size_ptt_min_size_pt_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_ptt_min : lngen.

Lemma size_pt_min :
forall pt1, 1 <= size_pt pt1.
Proof.
pose proof size_type_min_size_ptt_min_size_pt_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pt_min : lngen.

(* begin hide *)

Lemma size_kind_min_mutual :
(forall k1, 1 <= size_kind k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_kind_min :
forall k1, 1 <= size_kind k1.
Proof.
pose proof size_kind_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_kind_min : lngen.

(* begin hide *)

Lemma size_bind_min_mutual :
(forall bind1, 1 <= size_bind bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_bind_min :
forall bind1, 1 <= size_bind bind1.
Proof.
pose proof size_bind_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_bind_min : lngen.

(* begin hide *)

Lemma size_exp_close_exp_wrt_exp_rec_size_peb_close_peb_wrt_exp_rec_size_pe_close_pe_wrt_exp_rec_mutual :
(forall e1 x1 n1,
  size_exp (close_exp_wrt_exp_rec n1 x1 e1) = size_exp e1) /\
(forall peb1 x1 n1,
  size_peb (close_peb_wrt_exp_rec n1 x1 peb1) = size_peb peb1) /\
(forall pe1 x1 n1,
  size_pe (close_pe_wrt_exp_rec n1 x1 pe1) = size_pe pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_exp_close_exp_wrt_exp_rec :
forall e1 x1 n1,
  size_exp (close_exp_wrt_exp_rec n1 x1 e1) = size_exp e1.
Proof.
pose proof size_exp_close_exp_wrt_exp_rec_size_peb_close_peb_wrt_exp_rec_size_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_exp_close_exp_wrt_exp_rec : lngen.
Hint Rewrite size_exp_close_exp_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_peb_close_peb_wrt_exp_rec :
forall peb1 x1 n1,
  size_peb (close_peb_wrt_exp_rec n1 x1 peb1) = size_peb peb1.
Proof.
pose proof size_exp_close_exp_wrt_exp_rec_size_peb_close_peb_wrt_exp_rec_size_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_peb_close_peb_wrt_exp_rec : lngen.
Hint Rewrite size_peb_close_peb_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pe_close_pe_wrt_exp_rec :
forall pe1 x1 n1,
  size_pe (close_pe_wrt_exp_rec n1 x1 pe1) = size_pe pe1.
Proof.
pose proof size_exp_close_exp_wrt_exp_rec_size_peb_close_peb_wrt_exp_rec_size_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pe_close_pe_wrt_exp_rec : lngen.
Hint Rewrite size_pe_close_pe_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_type_close_type_wrt_exp_rec_size_ptt_close_ptt_wrt_exp_rec_size_pt_close_pt_wrt_exp_rec_mutual :
(forall t1 x1 n1,
  size_type (close_type_wrt_exp_rec n1 x1 t1) = size_type t1) /\
(forall ptt1 x1 n1,
  size_ptt (close_ptt_wrt_exp_rec n1 x1 ptt1) = size_ptt ptt1) /\
(forall pt1 x1 n1,
  size_pt (close_pt_wrt_exp_rec n1 x1 pt1) = size_pt pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_type_close_type_wrt_exp_rec :
forall t1 x1 n1,
  size_type (close_type_wrt_exp_rec n1 x1 t1) = size_type t1.
Proof.
pose proof size_type_close_type_wrt_exp_rec_size_ptt_close_ptt_wrt_exp_rec_size_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_type_close_type_wrt_exp_rec : lngen.
Hint Rewrite size_type_close_type_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_ptt_close_ptt_wrt_exp_rec :
forall ptt1 x1 n1,
  size_ptt (close_ptt_wrt_exp_rec n1 x1 ptt1) = size_ptt ptt1.
Proof.
pose proof size_type_close_type_wrt_exp_rec_size_ptt_close_ptt_wrt_exp_rec_size_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_ptt_close_ptt_wrt_exp_rec : lngen.
Hint Rewrite size_ptt_close_ptt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pt_close_pt_wrt_exp_rec :
forall pt1 x1 n1,
  size_pt (close_pt_wrt_exp_rec n1 x1 pt1) = size_pt pt1.
Proof.
pose proof size_type_close_type_wrt_exp_rec_size_ptt_close_ptt_wrt_exp_rec_size_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pt_close_pt_wrt_exp_rec : lngen.
Hint Rewrite size_pt_close_pt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_kind_close_kind_wrt_exp_rec_mutual :
(forall k1 x1 n1,
  size_kind (close_kind_wrt_exp_rec n1 x1 k1) = size_kind k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_kind_close_kind_wrt_exp_rec :
forall k1 x1 n1,
  size_kind (close_kind_wrt_exp_rec n1 x1 k1) = size_kind k1.
Proof.
pose proof size_kind_close_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_kind_close_kind_wrt_exp_rec : lngen.
Hint Rewrite size_kind_close_kind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_bind_close_bind_wrt_exp_rec_mutual :
(forall bind1 x1 n1,
  size_bind (close_bind_wrt_exp_rec n1 x1 bind1) = size_bind bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_bind_close_bind_wrt_exp_rec :
forall bind1 x1 n1,
  size_bind (close_bind_wrt_exp_rec n1 x1 bind1) = size_bind bind1.
Proof.
pose proof size_bind_close_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_bind_close_bind_wrt_exp_rec : lngen.
Hint Rewrite size_bind_close_bind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

Lemma size_exp_close_exp_wrt_exp :
forall e1 x1,
  size_exp (close_exp_wrt_exp x1 e1) = size_exp e1.
Proof.
unfold close_exp_wrt_exp; default_simp.
Qed.

Hint Resolve size_exp_close_exp_wrt_exp : lngen.
Hint Rewrite size_exp_close_exp_wrt_exp using solve [auto] : lngen.

Lemma size_peb_close_peb_wrt_exp :
forall peb1 x1,
  size_peb (close_peb_wrt_exp x1 peb1) = size_peb peb1.
Proof.
unfold close_peb_wrt_exp; default_simp.
Qed.

Hint Resolve size_peb_close_peb_wrt_exp : lngen.
Hint Rewrite size_peb_close_peb_wrt_exp using solve [auto] : lngen.

Lemma size_pe_close_pe_wrt_exp :
forall pe1 x1,
  size_pe (close_pe_wrt_exp x1 pe1) = size_pe pe1.
Proof.
unfold close_pe_wrt_exp; default_simp.
Qed.

Hint Resolve size_pe_close_pe_wrt_exp : lngen.
Hint Rewrite size_pe_close_pe_wrt_exp using solve [auto] : lngen.

Lemma size_type_close_type_wrt_exp :
forall t1 x1,
  size_type (close_type_wrt_exp x1 t1) = size_type t1.
Proof.
unfold close_type_wrt_exp; default_simp.
Qed.

Hint Resolve size_type_close_type_wrt_exp : lngen.
Hint Rewrite size_type_close_type_wrt_exp using solve [auto] : lngen.

Lemma size_ptt_close_ptt_wrt_exp :
forall ptt1 x1,
  size_ptt (close_ptt_wrt_exp x1 ptt1) = size_ptt ptt1.
Proof.
unfold close_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve size_ptt_close_ptt_wrt_exp : lngen.
Hint Rewrite size_ptt_close_ptt_wrt_exp using solve [auto] : lngen.

Lemma size_pt_close_pt_wrt_exp :
forall pt1 x1,
  size_pt (close_pt_wrt_exp x1 pt1) = size_pt pt1.
Proof.
unfold close_pt_wrt_exp; default_simp.
Qed.

Hint Resolve size_pt_close_pt_wrt_exp : lngen.
Hint Rewrite size_pt_close_pt_wrt_exp using solve [auto] : lngen.

Lemma size_kind_close_kind_wrt_exp :
forall k1 x1,
  size_kind (close_kind_wrt_exp x1 k1) = size_kind k1.
Proof.
unfold close_kind_wrt_exp; default_simp.
Qed.

Hint Resolve size_kind_close_kind_wrt_exp : lngen.
Hint Rewrite size_kind_close_kind_wrt_exp using solve [auto] : lngen.

Lemma size_bind_close_bind_wrt_exp :
forall bind1 x1,
  size_bind (close_bind_wrt_exp x1 bind1) = size_bind bind1.
Proof.
unfold close_bind_wrt_exp; default_simp.
Qed.

Hint Resolve size_bind_close_bind_wrt_exp : lngen.
Hint Rewrite size_bind_close_bind_wrt_exp using solve [auto] : lngen.

(* begin hide *)

Lemma size_exp_open_exp_wrt_exp_rec_size_peb_open_peb_wrt_exp_rec_size_pe_open_pe_wrt_exp_rec_mutual :
(forall e1 e2 n1,
  size_exp e1 <= size_exp (open_exp_wrt_exp_rec n1 e2 e1)) /\
(forall peb1 e1 n1,
  size_peb peb1 <= size_peb (open_peb_wrt_exp_rec n1 e1 peb1)) /\
(forall pe1 e1 n1,
  size_pe pe1 <= size_pe (open_pe_wrt_exp_rec n1 e1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_exp_open_exp_wrt_exp_rec :
forall e1 e2 n1,
  size_exp e1 <= size_exp (open_exp_wrt_exp_rec n1 e2 e1).
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_size_peb_open_peb_wrt_exp_rec_size_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_exp_open_exp_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_peb_open_peb_wrt_exp_rec :
forall peb1 e1 n1,
  size_peb peb1 <= size_peb (open_peb_wrt_exp_rec n1 e1 peb1).
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_size_peb_open_peb_wrt_exp_rec_size_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_peb_open_peb_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pe_open_pe_wrt_exp_rec :
forall pe1 e1 n1,
  size_pe pe1 <= size_pe (open_pe_wrt_exp_rec n1 e1 pe1).
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_size_peb_open_peb_wrt_exp_rec_size_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pe_open_pe_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_type_open_type_wrt_exp_rec_size_ptt_open_ptt_wrt_exp_rec_size_pt_open_pt_wrt_exp_rec_mutual :
(forall t1 e1 n1,
  size_type t1 <= size_type (open_type_wrt_exp_rec n1 e1 t1)) /\
(forall ptt1 e1 n1,
  size_ptt ptt1 <= size_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1)) /\
(forall pt1 e1 n1,
  size_pt pt1 <= size_pt (open_pt_wrt_exp_rec n1 e1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_type_open_type_wrt_exp_rec :
forall t1 e1 n1,
  size_type t1 <= size_type (open_type_wrt_exp_rec n1 e1 t1).
Proof.
pose proof size_type_open_type_wrt_exp_rec_size_ptt_open_ptt_wrt_exp_rec_size_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_type_open_type_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_ptt_open_ptt_wrt_exp_rec :
forall ptt1 e1 n1,
  size_ptt ptt1 <= size_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1).
Proof.
pose proof size_type_open_type_wrt_exp_rec_size_ptt_open_ptt_wrt_exp_rec_size_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_ptt_open_ptt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pt_open_pt_wrt_exp_rec :
forall pt1 e1 n1,
  size_pt pt1 <= size_pt (open_pt_wrt_exp_rec n1 e1 pt1).
Proof.
pose proof size_type_open_type_wrt_exp_rec_size_ptt_open_ptt_wrt_exp_rec_size_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pt_open_pt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_kind_open_kind_wrt_exp_rec_mutual :
(forall k1 e1 n1,
  size_kind k1 <= size_kind (open_kind_wrt_exp_rec n1 e1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_kind_open_kind_wrt_exp_rec :
forall k1 e1 n1,
  size_kind k1 <= size_kind (open_kind_wrt_exp_rec n1 e1 k1).
Proof.
pose proof size_kind_open_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_kind_open_kind_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_bind_open_bind_wrt_exp_rec_mutual :
(forall bind1 e1 n1,
  size_bind bind1 <= size_bind (open_bind_wrt_exp_rec n1 e1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_bind_open_bind_wrt_exp_rec :
forall bind1 e1 n1,
  size_bind bind1 <= size_bind (open_bind_wrt_exp_rec n1 e1 bind1).
Proof.
pose proof size_bind_open_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_bind_open_bind_wrt_exp_rec : lngen.

(* end hide *)

Lemma size_exp_open_exp_wrt_exp :
forall e1 e2,
  size_exp e1 <= size_exp (open_exp_wrt_exp e1 e2).
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve size_exp_open_exp_wrt_exp : lngen.

Lemma size_peb_open_peb_wrt_exp :
forall peb1 e1,
  size_peb peb1 <= size_peb (open_peb_wrt_exp peb1 e1).
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve size_peb_open_peb_wrt_exp : lngen.

Lemma size_pe_open_pe_wrt_exp :
forall pe1 e1,
  size_pe pe1 <= size_pe (open_pe_wrt_exp pe1 e1).
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve size_pe_open_pe_wrt_exp : lngen.

Lemma size_type_open_type_wrt_exp :
forall t1 e1,
  size_type t1 <= size_type (open_type_wrt_exp t1 e1).
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve size_type_open_type_wrt_exp : lngen.

Lemma size_ptt_open_ptt_wrt_exp :
forall ptt1 e1,
  size_ptt ptt1 <= size_ptt (open_ptt_wrt_exp ptt1 e1).
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve size_ptt_open_ptt_wrt_exp : lngen.

Lemma size_pt_open_pt_wrt_exp :
forall pt1 e1,
  size_pt pt1 <= size_pt (open_pt_wrt_exp pt1 e1).
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve size_pt_open_pt_wrt_exp : lngen.

Lemma size_kind_open_kind_wrt_exp :
forall k1 e1,
  size_kind k1 <= size_kind (open_kind_wrt_exp k1 e1).
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve size_kind_open_kind_wrt_exp : lngen.

Lemma size_bind_open_bind_wrt_exp :
forall bind1 e1,
  size_bind bind1 <= size_bind (open_bind_wrt_exp bind1 e1).
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve size_bind_open_bind_wrt_exp : lngen.

(* begin hide *)

Lemma size_exp_open_exp_wrt_exp_rec_var_size_peb_open_peb_wrt_exp_rec_var_size_pe_open_pe_wrt_exp_rec_var_mutual :
(forall e1 x1 n1,
  size_exp (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1) = size_exp e1) /\
(forall peb1 x1 n1,
  size_peb (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1) = size_peb peb1) /\
(forall pe1 x1 n1,
  size_pe (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1) = size_pe pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_exp_open_exp_wrt_exp_rec_var :
forall e1 x1 n1,
  size_exp (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1) = size_exp e1.
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_var_size_peb_open_peb_wrt_exp_rec_var_size_pe_open_pe_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_exp_open_exp_wrt_exp_rec_var : lngen.
Hint Rewrite size_exp_open_exp_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_peb_open_peb_wrt_exp_rec_var :
forall peb1 x1 n1,
  size_peb (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1) = size_peb peb1.
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_var_size_peb_open_peb_wrt_exp_rec_var_size_pe_open_pe_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_peb_open_peb_wrt_exp_rec_var : lngen.
Hint Rewrite size_peb_open_peb_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pe_open_pe_wrt_exp_rec_var :
forall pe1 x1 n1,
  size_pe (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1) = size_pe pe1.
Proof.
pose proof size_exp_open_exp_wrt_exp_rec_var_size_peb_open_peb_wrt_exp_rec_var_size_pe_open_pe_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pe_open_pe_wrt_exp_rec_var : lngen.
Hint Rewrite size_pe_open_pe_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_type_open_type_wrt_exp_rec_var_size_ptt_open_ptt_wrt_exp_rec_var_size_pt_open_pt_wrt_exp_rec_var_mutual :
(forall t1 x1 n1,
  size_type (open_type_wrt_exp_rec n1 (e_Var_f x1) t1) = size_type t1) /\
(forall ptt1 x1 n1,
  size_ptt (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1) = size_ptt ptt1) /\
(forall pt1 x1 n1,
  size_pt (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1) = size_pt pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_type_open_type_wrt_exp_rec_var :
forall t1 x1 n1,
  size_type (open_type_wrt_exp_rec n1 (e_Var_f x1) t1) = size_type t1.
Proof.
pose proof size_type_open_type_wrt_exp_rec_var_size_ptt_open_ptt_wrt_exp_rec_var_size_pt_open_pt_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_type_open_type_wrt_exp_rec_var : lngen.
Hint Rewrite size_type_open_type_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_ptt_open_ptt_wrt_exp_rec_var :
forall ptt1 x1 n1,
  size_ptt (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1) = size_ptt ptt1.
Proof.
pose proof size_type_open_type_wrt_exp_rec_var_size_ptt_open_ptt_wrt_exp_rec_var_size_pt_open_pt_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_ptt_open_ptt_wrt_exp_rec_var : lngen.
Hint Rewrite size_ptt_open_ptt_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_pt_open_pt_wrt_exp_rec_var :
forall pt1 x1 n1,
  size_pt (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1) = size_pt pt1.
Proof.
pose proof size_type_open_type_wrt_exp_rec_var_size_ptt_open_ptt_wrt_exp_rec_var_size_pt_open_pt_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_pt_open_pt_wrt_exp_rec_var : lngen.
Hint Rewrite size_pt_open_pt_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_kind_open_kind_wrt_exp_rec_var_mutual :
(forall k1 x1 n1,
  size_kind (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1) = size_kind k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_kind_open_kind_wrt_exp_rec_var :
forall k1 x1 n1,
  size_kind (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1) = size_kind k1.
Proof.
pose proof size_kind_open_kind_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_kind_open_kind_wrt_exp_rec_var : lngen.
Hint Rewrite size_kind_open_kind_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_bind_open_bind_wrt_exp_rec_var_mutual :
(forall bind1 x1 n1,
  size_bind (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1) = size_bind bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_bind_open_bind_wrt_exp_rec_var :
forall bind1 x1 n1,
  size_bind (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1) = size_bind bind1.
Proof.
pose proof size_bind_open_bind_wrt_exp_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_bind_open_bind_wrt_exp_rec_var : lngen.
Hint Rewrite size_bind_open_bind_wrt_exp_rec_var using solve [auto] : lngen.

(* end hide *)

Lemma size_exp_open_exp_wrt_exp_var :
forall e1 x1,
  size_exp (open_exp_wrt_exp e1 (e_Var_f x1)) = size_exp e1.
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve size_exp_open_exp_wrt_exp_var : lngen.
Hint Rewrite size_exp_open_exp_wrt_exp_var using solve [auto] : lngen.

Lemma size_peb_open_peb_wrt_exp_var :
forall peb1 x1,
  size_peb (open_peb_wrt_exp peb1 (e_Var_f x1)) = size_peb peb1.
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve size_peb_open_peb_wrt_exp_var : lngen.
Hint Rewrite size_peb_open_peb_wrt_exp_var using solve [auto] : lngen.

Lemma size_pe_open_pe_wrt_exp_var :
forall pe1 x1,
  size_pe (open_pe_wrt_exp pe1 (e_Var_f x1)) = size_pe pe1.
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve size_pe_open_pe_wrt_exp_var : lngen.
Hint Rewrite size_pe_open_pe_wrt_exp_var using solve [auto] : lngen.

Lemma size_type_open_type_wrt_exp_var :
forall t1 x1,
  size_type (open_type_wrt_exp t1 (e_Var_f x1)) = size_type t1.
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve size_type_open_type_wrt_exp_var : lngen.
Hint Rewrite size_type_open_type_wrt_exp_var using solve [auto] : lngen.

Lemma size_ptt_open_ptt_wrt_exp_var :
forall ptt1 x1,
  size_ptt (open_ptt_wrt_exp ptt1 (e_Var_f x1)) = size_ptt ptt1.
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve size_ptt_open_ptt_wrt_exp_var : lngen.
Hint Rewrite size_ptt_open_ptt_wrt_exp_var using solve [auto] : lngen.

Lemma size_pt_open_pt_wrt_exp_var :
forall pt1 x1,
  size_pt (open_pt_wrt_exp pt1 (e_Var_f x1)) = size_pt pt1.
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve size_pt_open_pt_wrt_exp_var : lngen.
Hint Rewrite size_pt_open_pt_wrt_exp_var using solve [auto] : lngen.

Lemma size_kind_open_kind_wrt_exp_var :
forall k1 x1,
  size_kind (open_kind_wrt_exp k1 (e_Var_f x1)) = size_kind k1.
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve size_kind_open_kind_wrt_exp_var : lngen.
Hint Rewrite size_kind_open_kind_wrt_exp_var using solve [auto] : lngen.

Lemma size_bind_open_bind_wrt_exp_var :
forall bind1 x1,
  size_bind (open_bind_wrt_exp bind1 (e_Var_f x1)) = size_bind bind1.
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve size_bind_open_bind_wrt_exp_var : lngen.
Hint Rewrite size_bind_open_bind_wrt_exp_var using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [degree] *)

Ltac default_auto ::= auto with lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma degree_exp_wrt_exp_S_degree_peb_wrt_exp_S_degree_pe_wrt_exp_S_mutual :
(forall n1 e1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp (S n1) e1) /\
(forall n1 peb1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_peb_wrt_exp (S n1) peb1) /\
(forall n1 pe1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_pe_wrt_exp (S n1) pe1).
Proof.
apply_mutual_ind degree_exp_wrt_exp_degree_peb_wrt_exp_degree_pe_wrt_exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma degree_exp_wrt_exp_S :
forall n1 e1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp (S n1) e1.
Proof.
pose proof degree_exp_wrt_exp_S_degree_peb_wrt_exp_S_degree_pe_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_exp_wrt_exp_S : lngen.

Lemma degree_peb_wrt_exp_S :
forall n1 peb1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_peb_wrt_exp (S n1) peb1.
Proof.
pose proof degree_exp_wrt_exp_S_degree_peb_wrt_exp_S_degree_pe_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_peb_wrt_exp_S : lngen.

Lemma degree_pe_wrt_exp_S :
forall n1 pe1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_pe_wrt_exp (S n1) pe1.
Proof.
pose proof degree_exp_wrt_exp_S_degree_peb_wrt_exp_S_degree_pe_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pe_wrt_exp_S : lngen.

(* begin hide *)

Lemma degree_type_wrt_exp_S_degree_ptt_wrt_exp_S_degree_pt_wrt_exp_S_mutual :
(forall n1 t1,
  degree_type_wrt_exp n1 t1 ->
  degree_type_wrt_exp (S n1) t1) /\
(forall n1 ptt1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_ptt_wrt_exp (S n1) ptt1) /\
(forall n1 pt1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_pt_wrt_exp (S n1) pt1).
Proof.
apply_mutual_ind degree_type_wrt_exp_degree_ptt_wrt_exp_degree_pt_wrt_exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma degree_type_wrt_exp_S :
forall n1 t1,
  degree_type_wrt_exp n1 t1 ->
  degree_type_wrt_exp (S n1) t1.
Proof.
pose proof degree_type_wrt_exp_S_degree_ptt_wrt_exp_S_degree_pt_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_type_wrt_exp_S : lngen.

Lemma degree_ptt_wrt_exp_S :
forall n1 ptt1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_ptt_wrt_exp (S n1) ptt1.
Proof.
pose proof degree_type_wrt_exp_S_degree_ptt_wrt_exp_S_degree_pt_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ptt_wrt_exp_S : lngen.

Lemma degree_pt_wrt_exp_S :
forall n1 pt1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_pt_wrt_exp (S n1) pt1.
Proof.
pose proof degree_type_wrt_exp_S_degree_ptt_wrt_exp_S_degree_pt_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pt_wrt_exp_S : lngen.

(* begin hide *)

Lemma degree_kind_wrt_exp_S_mutual :
(forall n1 k1,
  degree_kind_wrt_exp n1 k1 ->
  degree_kind_wrt_exp (S n1) k1).
Proof.
apply_mutual_ind degree_kind_wrt_exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma degree_kind_wrt_exp_S :
forall n1 k1,
  degree_kind_wrt_exp n1 k1 ->
  degree_kind_wrt_exp (S n1) k1.
Proof.
pose proof degree_kind_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_kind_wrt_exp_S : lngen.

(* begin hide *)

Lemma degree_bind_wrt_exp_S_mutual :
(forall n1 bind1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_bind_wrt_exp (S n1) bind1).
Proof.
apply_mutual_ind degree_bind_wrt_exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma degree_bind_wrt_exp_S :
forall n1 bind1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_bind_wrt_exp (S n1) bind1.
Proof.
pose proof degree_bind_wrt_exp_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_bind_wrt_exp_S : lngen.

Lemma degree_exp_wrt_exp_O :
forall n1 e1,
  degree_exp_wrt_exp O e1 ->
  degree_exp_wrt_exp n1 e1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_exp_wrt_exp_O : lngen.

Lemma degree_peb_wrt_exp_O :
forall n1 peb1,
  degree_peb_wrt_exp O peb1 ->
  degree_peb_wrt_exp n1 peb1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_peb_wrt_exp_O : lngen.

Lemma degree_pe_wrt_exp_O :
forall n1 pe1,
  degree_pe_wrt_exp O pe1 ->
  degree_pe_wrt_exp n1 pe1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_pe_wrt_exp_O : lngen.

Lemma degree_type_wrt_exp_O :
forall n1 t1,
  degree_type_wrt_exp O t1 ->
  degree_type_wrt_exp n1 t1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_type_wrt_exp_O : lngen.

Lemma degree_ptt_wrt_exp_O :
forall n1 ptt1,
  degree_ptt_wrt_exp O ptt1 ->
  degree_ptt_wrt_exp n1 ptt1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_ptt_wrt_exp_O : lngen.

Lemma degree_pt_wrt_exp_O :
forall n1 pt1,
  degree_pt_wrt_exp O pt1 ->
  degree_pt_wrt_exp n1 pt1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_pt_wrt_exp_O : lngen.

Lemma degree_kind_wrt_exp_O :
forall n1 k1,
  degree_kind_wrt_exp O k1 ->
  degree_kind_wrt_exp n1 k1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_kind_wrt_exp_O : lngen.

Lemma degree_bind_wrt_exp_O :
forall n1 bind1,
  degree_bind_wrt_exp O bind1 ->
  degree_bind_wrt_exp n1 bind1.
Proof.
induction n1; default_simp.
Qed.

Hint Resolve degree_bind_wrt_exp_O : lngen.

(* begin hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp_rec_degree_peb_wrt_exp_close_peb_wrt_exp_rec_degree_pe_wrt_exp_close_pe_wrt_exp_rec_mutual :
(forall e1 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp (S n1) (close_exp_wrt_exp_rec n1 x1 e1)) /\
(forall peb1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_peb_wrt_exp (S n1) (close_peb_wrt_exp_rec n1 x1 peb1)) /\
(forall pe1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_pe_wrt_exp (S n1) (close_pe_wrt_exp_rec n1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp_rec :
forall e1 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp (S n1) (close_exp_wrt_exp_rec n1 x1 e1).
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_degree_peb_wrt_exp_close_peb_wrt_exp_rec_degree_pe_wrt_exp_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_exp_wrt_exp_close_exp_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_peb_wrt_exp_close_peb_wrt_exp_rec :
forall peb1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_peb_wrt_exp (S n1) (close_peb_wrt_exp_rec n1 x1 peb1).
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_degree_peb_wrt_exp_close_peb_wrt_exp_rec_degree_pe_wrt_exp_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_peb_wrt_exp_close_peb_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pe_wrt_exp_close_pe_wrt_exp_rec :
forall pe1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_pe_wrt_exp (S n1) (close_pe_wrt_exp_rec n1 x1 pe1).
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_degree_peb_wrt_exp_close_peb_wrt_exp_rec_degree_pe_wrt_exp_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pe_wrt_exp_close_pe_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_close_type_wrt_exp_rec_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_degree_pt_wrt_exp_close_pt_wrt_exp_rec_mutual :
(forall t1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  degree_type_wrt_exp (S n1) (close_type_wrt_exp_rec n1 x1 t1)) /\
(forall ptt1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_ptt_wrt_exp (S n1) (close_ptt_wrt_exp_rec n1 x1 ptt1)) /\
(forall pt1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_pt_wrt_exp (S n1) (close_pt_wrt_exp_rec n1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_close_type_wrt_exp_rec :
forall t1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  degree_type_wrt_exp (S n1) (close_type_wrt_exp_rec n1 x1 t1).
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_degree_pt_wrt_exp_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_type_wrt_exp_close_type_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ptt_wrt_exp_close_ptt_wrt_exp_rec :
forall ptt1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_ptt_wrt_exp (S n1) (close_ptt_wrt_exp_rec n1 x1 ptt1).
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_degree_pt_wrt_exp_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ptt_wrt_exp_close_ptt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pt_wrt_exp_close_pt_wrt_exp_rec :
forall pt1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_pt_wrt_exp (S n1) (close_pt_wrt_exp_rec n1 x1 pt1).
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_degree_pt_wrt_exp_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pt_wrt_exp_close_pt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_close_kind_wrt_exp_rec_mutual :
(forall k1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  degree_kind_wrt_exp (S n1) (close_kind_wrt_exp_rec n1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_close_kind_wrt_exp_rec :
forall k1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  degree_kind_wrt_exp (S n1) (close_kind_wrt_exp_rec n1 x1 k1).
Proof.
pose proof degree_kind_wrt_exp_close_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_kind_wrt_exp_close_kind_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_close_bind_wrt_exp_rec_mutual :
(forall bind1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_bind_wrt_exp (S n1) (close_bind_wrt_exp_rec n1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_close_bind_wrt_exp_rec :
forall bind1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_bind_wrt_exp (S n1) (close_bind_wrt_exp_rec n1 x1 bind1).
Proof.
pose proof degree_bind_wrt_exp_close_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_bind_wrt_exp_close_bind_wrt_exp_rec : lngen.

(* end hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp :
forall e1 x1,
  degree_exp_wrt_exp 0 e1 ->
  degree_exp_wrt_exp 1 (close_exp_wrt_exp x1 e1).
Proof.
unfold close_exp_wrt_exp; default_simp.
Qed.

Hint Resolve degree_exp_wrt_exp_close_exp_wrt_exp : lngen.

Lemma degree_peb_wrt_exp_close_peb_wrt_exp :
forall peb1 x1,
  degree_peb_wrt_exp 0 peb1 ->
  degree_peb_wrt_exp 1 (close_peb_wrt_exp x1 peb1).
Proof.
unfold close_peb_wrt_exp; default_simp.
Qed.

Hint Resolve degree_peb_wrt_exp_close_peb_wrt_exp : lngen.

Lemma degree_pe_wrt_exp_close_pe_wrt_exp :
forall pe1 x1,
  degree_pe_wrt_exp 0 pe1 ->
  degree_pe_wrt_exp 1 (close_pe_wrt_exp x1 pe1).
Proof.
unfold close_pe_wrt_exp; default_simp.
Qed.

Hint Resolve degree_pe_wrt_exp_close_pe_wrt_exp : lngen.

Lemma degree_type_wrt_exp_close_type_wrt_exp :
forall t1 x1,
  degree_type_wrt_exp 0 t1 ->
  degree_type_wrt_exp 1 (close_type_wrt_exp x1 t1).
Proof.
unfold close_type_wrt_exp; default_simp.
Qed.

Hint Resolve degree_type_wrt_exp_close_type_wrt_exp : lngen.

Lemma degree_ptt_wrt_exp_close_ptt_wrt_exp :
forall ptt1 x1,
  degree_ptt_wrt_exp 0 ptt1 ->
  degree_ptt_wrt_exp 1 (close_ptt_wrt_exp x1 ptt1).
Proof.
unfold close_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve degree_ptt_wrt_exp_close_ptt_wrt_exp : lngen.

Lemma degree_pt_wrt_exp_close_pt_wrt_exp :
forall pt1 x1,
  degree_pt_wrt_exp 0 pt1 ->
  degree_pt_wrt_exp 1 (close_pt_wrt_exp x1 pt1).
Proof.
unfold close_pt_wrt_exp; default_simp.
Qed.

Hint Resolve degree_pt_wrt_exp_close_pt_wrt_exp : lngen.

Lemma degree_kind_wrt_exp_close_kind_wrt_exp :
forall k1 x1,
  degree_kind_wrt_exp 0 k1 ->
  degree_kind_wrt_exp 1 (close_kind_wrt_exp x1 k1).
Proof.
unfold close_kind_wrt_exp; default_simp.
Qed.

Hint Resolve degree_kind_wrt_exp_close_kind_wrt_exp : lngen.

Lemma degree_bind_wrt_exp_close_bind_wrt_exp :
forall bind1 x1,
  degree_bind_wrt_exp 0 bind1 ->
  degree_bind_wrt_exp 1 (close_bind_wrt_exp x1 bind1).
Proof.
unfold close_bind_wrt_exp; default_simp.
Qed.

Hint Resolve degree_bind_wrt_exp_close_bind_wrt_exp : lngen.

(* begin hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv_mutual :
(forall e1 x1 n1,
  degree_exp_wrt_exp (S n1) (close_exp_wrt_exp_rec n1 x1 e1) ->
  degree_exp_wrt_exp n1 e1) /\
(forall peb1 x1 n1,
  degree_peb_wrt_exp (S n1) (close_peb_wrt_exp_rec n1 x1 peb1) ->
  degree_peb_wrt_exp n1 peb1) /\
(forall pe1 x1 n1,
  degree_pe_wrt_exp (S n1) (close_pe_wrt_exp_rec n1 x1 pe1) ->
  degree_pe_wrt_exp n1 pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv :
forall e1 x1 n1,
  degree_exp_wrt_exp (S n1) (close_exp_wrt_exp_rec n1 x1 e1) ->
  degree_exp_wrt_exp n1 e1.
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv :
forall peb1 x1 n1,
  degree_peb_wrt_exp (S n1) (close_peb_wrt_exp_rec n1 x1 peb1) ->
  degree_peb_wrt_exp n1 peb1.
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv :
forall pe1 x1 n1,
  degree_pe_wrt_exp (S n1) (close_pe_wrt_exp_rec n1 x1 pe1) ->
  degree_pe_wrt_exp n1 pe1.
Proof.
pose proof degree_exp_wrt_exp_close_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_close_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_pe_wrt_exp_close_pe_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_close_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv_mutual :
(forall t1 x1 n1,
  degree_type_wrt_exp (S n1) (close_type_wrt_exp_rec n1 x1 t1) ->
  degree_type_wrt_exp n1 t1) /\
(forall ptt1 x1 n1,
  degree_ptt_wrt_exp (S n1) (close_ptt_wrt_exp_rec n1 x1 ptt1) ->
  degree_ptt_wrt_exp n1 ptt1) /\
(forall pt1 x1 n1,
  degree_pt_wrt_exp (S n1) (close_pt_wrt_exp_rec n1 x1 pt1) ->
  degree_pt_wrt_exp n1 pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_close_type_wrt_exp_rec_inv :
forall t1 x1 n1,
  degree_type_wrt_exp (S n1) (close_type_wrt_exp_rec n1 x1 t1) ->
  degree_type_wrt_exp n1 t1.
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_type_wrt_exp_close_type_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv :
forall ptt1 x1 n1,
  degree_ptt_wrt_exp (S n1) (close_ptt_wrt_exp_rec n1 x1 ptt1) ->
  degree_ptt_wrt_exp n1 ptt1.
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv :
forall pt1 x1 n1,
  degree_pt_wrt_exp (S n1) (close_pt_wrt_exp_rec n1 x1 pt1) ->
  degree_pt_wrt_exp n1 pt1.
Proof.
pose proof degree_type_wrt_exp_close_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_close_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_pt_wrt_exp_close_pt_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_close_kind_wrt_exp_rec_inv_mutual :
(forall k1 x1 n1,
  degree_kind_wrt_exp (S n1) (close_kind_wrt_exp_rec n1 x1 k1) ->
  degree_kind_wrt_exp n1 k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_close_kind_wrt_exp_rec_inv :
forall k1 x1 n1,
  degree_kind_wrt_exp (S n1) (close_kind_wrt_exp_rec n1 x1 k1) ->
  degree_kind_wrt_exp n1 k1.
Proof.
pose proof degree_kind_wrt_exp_close_kind_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_kind_wrt_exp_close_kind_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_close_bind_wrt_exp_rec_inv_mutual :
(forall bind1 x1 n1,
  degree_bind_wrt_exp (S n1) (close_bind_wrt_exp_rec n1 x1 bind1) ->
  degree_bind_wrt_exp n1 bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_close_bind_wrt_exp_rec_inv :
forall bind1 x1 n1,
  degree_bind_wrt_exp (S n1) (close_bind_wrt_exp_rec n1 x1 bind1) ->
  degree_bind_wrt_exp n1 bind1.
Proof.
pose proof degree_bind_wrt_exp_close_bind_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_bind_wrt_exp_close_bind_wrt_exp_rec_inv : lngen.

(* end hide *)

Lemma degree_exp_wrt_exp_close_exp_wrt_exp_inv :
forall e1 x1,
  degree_exp_wrt_exp 1 (close_exp_wrt_exp x1 e1) ->
  degree_exp_wrt_exp 0 e1.
Proof.
unfold close_exp_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_exp_wrt_exp_close_exp_wrt_exp_inv : lngen.

Lemma degree_peb_wrt_exp_close_peb_wrt_exp_inv :
forall peb1 x1,
  degree_peb_wrt_exp 1 (close_peb_wrt_exp x1 peb1) ->
  degree_peb_wrt_exp 0 peb1.
Proof.
unfold close_peb_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_peb_wrt_exp_close_peb_wrt_exp_inv : lngen.

Lemma degree_pe_wrt_exp_close_pe_wrt_exp_inv :
forall pe1 x1,
  degree_pe_wrt_exp 1 (close_pe_wrt_exp x1 pe1) ->
  degree_pe_wrt_exp 0 pe1.
Proof.
unfold close_pe_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_pe_wrt_exp_close_pe_wrt_exp_inv : lngen.

Lemma degree_type_wrt_exp_close_type_wrt_exp_inv :
forall t1 x1,
  degree_type_wrt_exp 1 (close_type_wrt_exp x1 t1) ->
  degree_type_wrt_exp 0 t1.
Proof.
unfold close_type_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_type_wrt_exp_close_type_wrt_exp_inv : lngen.

Lemma degree_ptt_wrt_exp_close_ptt_wrt_exp_inv :
forall ptt1 x1,
  degree_ptt_wrt_exp 1 (close_ptt_wrt_exp x1 ptt1) ->
  degree_ptt_wrt_exp 0 ptt1.
Proof.
unfold close_ptt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_ptt_wrt_exp_close_ptt_wrt_exp_inv : lngen.

Lemma degree_pt_wrt_exp_close_pt_wrt_exp_inv :
forall pt1 x1,
  degree_pt_wrt_exp 1 (close_pt_wrt_exp x1 pt1) ->
  degree_pt_wrt_exp 0 pt1.
Proof.
unfold close_pt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_pt_wrt_exp_close_pt_wrt_exp_inv : lngen.

Lemma degree_kind_wrt_exp_close_kind_wrt_exp_inv :
forall k1 x1,
  degree_kind_wrt_exp 1 (close_kind_wrt_exp x1 k1) ->
  degree_kind_wrt_exp 0 k1.
Proof.
unfold close_kind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_kind_wrt_exp_close_kind_wrt_exp_inv : lngen.

Lemma degree_bind_wrt_exp_close_bind_wrt_exp_inv :
forall bind1 x1,
  degree_bind_wrt_exp 1 (close_bind_wrt_exp x1 bind1) ->
  degree_bind_wrt_exp 0 bind1.
Proof.
unfold close_bind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_bind_wrt_exp_close_bind_wrt_exp_inv : lngen.

(* begin hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp_rec_degree_peb_wrt_exp_open_peb_wrt_exp_rec_degree_pe_wrt_exp_open_pe_wrt_exp_rec_mutual :
(forall e1 e2 n1,
  degree_exp_wrt_exp (S n1) e1 ->
  degree_exp_wrt_exp n1 e2 ->
  degree_exp_wrt_exp n1 (open_exp_wrt_exp_rec n1 e2 e1)) /\
(forall peb1 e1 n1,
  degree_peb_wrt_exp (S n1) peb1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_peb_wrt_exp n1 (open_peb_wrt_exp_rec n1 e1 peb1)) /\
(forall pe1 e1 n1,
  degree_pe_wrt_exp (S n1) pe1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pe_wrt_exp n1 (open_pe_wrt_exp_rec n1 e1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp_rec :
forall e1 e2 n1,
  degree_exp_wrt_exp (S n1) e1 ->
  degree_exp_wrt_exp n1 e2 ->
  degree_exp_wrt_exp n1 (open_exp_wrt_exp_rec n1 e2 e1).
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_degree_peb_wrt_exp_open_peb_wrt_exp_rec_degree_pe_wrt_exp_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_exp_wrt_exp_open_exp_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_peb_wrt_exp_open_peb_wrt_exp_rec :
forall peb1 e1 n1,
  degree_peb_wrt_exp (S n1) peb1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_peb_wrt_exp n1 (open_peb_wrt_exp_rec n1 e1 peb1).
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_degree_peb_wrt_exp_open_peb_wrt_exp_rec_degree_pe_wrt_exp_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_peb_wrt_exp_open_peb_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pe_wrt_exp_open_pe_wrt_exp_rec :
forall pe1 e1 n1,
  degree_pe_wrt_exp (S n1) pe1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pe_wrt_exp n1 (open_pe_wrt_exp_rec n1 e1 pe1).
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_degree_peb_wrt_exp_open_peb_wrt_exp_rec_degree_pe_wrt_exp_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pe_wrt_exp_open_pe_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_open_type_wrt_exp_rec_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_degree_pt_wrt_exp_open_pt_wrt_exp_rec_mutual :
(forall t1 e1 n1,
  degree_type_wrt_exp (S n1) t1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_type_wrt_exp n1 (open_type_wrt_exp_rec n1 e1 t1)) /\
(forall ptt1 e1 n1,
  degree_ptt_wrt_exp (S n1) ptt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_ptt_wrt_exp n1 (open_ptt_wrt_exp_rec n1 e1 ptt1)) /\
(forall pt1 e1 n1,
  degree_pt_wrt_exp (S n1) pt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pt_wrt_exp n1 (open_pt_wrt_exp_rec n1 e1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_open_type_wrt_exp_rec :
forall t1 e1 n1,
  degree_type_wrt_exp (S n1) t1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_type_wrt_exp n1 (open_type_wrt_exp_rec n1 e1 t1).
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_degree_pt_wrt_exp_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_type_wrt_exp_open_type_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ptt_wrt_exp_open_ptt_wrt_exp_rec :
forall ptt1 e1 n1,
  degree_ptt_wrt_exp (S n1) ptt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_ptt_wrt_exp n1 (open_ptt_wrt_exp_rec n1 e1 ptt1).
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_degree_pt_wrt_exp_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ptt_wrt_exp_open_ptt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pt_wrt_exp_open_pt_wrt_exp_rec :
forall pt1 e1 n1,
  degree_pt_wrt_exp (S n1) pt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pt_wrt_exp n1 (open_pt_wrt_exp_rec n1 e1 pt1).
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_degree_pt_wrt_exp_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pt_wrt_exp_open_pt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_open_kind_wrt_exp_rec_mutual :
(forall k1 e1 n1,
  degree_kind_wrt_exp (S n1) k1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_kind_wrt_exp n1 (open_kind_wrt_exp_rec n1 e1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_open_kind_wrt_exp_rec :
forall k1 e1 n1,
  degree_kind_wrt_exp (S n1) k1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_kind_wrt_exp n1 (open_kind_wrt_exp_rec n1 e1 k1).
Proof.
pose proof degree_kind_wrt_exp_open_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_kind_wrt_exp_open_kind_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_open_bind_wrt_exp_rec_mutual :
(forall bind1 e1 n1,
  degree_bind_wrt_exp (S n1) bind1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_bind_wrt_exp n1 (open_bind_wrt_exp_rec n1 e1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_open_bind_wrt_exp_rec :
forall bind1 e1 n1,
  degree_bind_wrt_exp (S n1) bind1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_bind_wrt_exp n1 (open_bind_wrt_exp_rec n1 e1 bind1).
Proof.
pose proof degree_bind_wrt_exp_open_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_bind_wrt_exp_open_bind_wrt_exp_rec : lngen.

(* end hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp :
forall e1 e2,
  degree_exp_wrt_exp 1 e1 ->
  degree_exp_wrt_exp 0 e2 ->
  degree_exp_wrt_exp 0 (open_exp_wrt_exp e1 e2).
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve degree_exp_wrt_exp_open_exp_wrt_exp : lngen.

Lemma degree_peb_wrt_exp_open_peb_wrt_exp :
forall peb1 e1,
  degree_peb_wrt_exp 1 peb1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_peb_wrt_exp 0 (open_peb_wrt_exp peb1 e1).
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve degree_peb_wrt_exp_open_peb_wrt_exp : lngen.

Lemma degree_pe_wrt_exp_open_pe_wrt_exp :
forall pe1 e1,
  degree_pe_wrt_exp 1 pe1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_pe_wrt_exp 0 (open_pe_wrt_exp pe1 e1).
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve degree_pe_wrt_exp_open_pe_wrt_exp : lngen.

Lemma degree_type_wrt_exp_open_type_wrt_exp :
forall t1 e1,
  degree_type_wrt_exp 1 t1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_type_wrt_exp 0 (open_type_wrt_exp t1 e1).
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve degree_type_wrt_exp_open_type_wrt_exp : lngen.

Lemma degree_ptt_wrt_exp_open_ptt_wrt_exp :
forall ptt1 e1,
  degree_ptt_wrt_exp 1 ptt1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_ptt_wrt_exp 0 (open_ptt_wrt_exp ptt1 e1).
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve degree_ptt_wrt_exp_open_ptt_wrt_exp : lngen.

Lemma degree_pt_wrt_exp_open_pt_wrt_exp :
forall pt1 e1,
  degree_pt_wrt_exp 1 pt1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_pt_wrt_exp 0 (open_pt_wrt_exp pt1 e1).
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve degree_pt_wrt_exp_open_pt_wrt_exp : lngen.

Lemma degree_kind_wrt_exp_open_kind_wrt_exp :
forall k1 e1,
  degree_kind_wrt_exp 1 k1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_kind_wrt_exp 0 (open_kind_wrt_exp k1 e1).
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve degree_kind_wrt_exp_open_kind_wrt_exp : lngen.

Lemma degree_bind_wrt_exp_open_bind_wrt_exp :
forall bind1 e1,
  degree_bind_wrt_exp 1 bind1 ->
  degree_exp_wrt_exp 0 e1 ->
  degree_bind_wrt_exp 0 (open_bind_wrt_exp bind1 e1).
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve degree_bind_wrt_exp_open_bind_wrt_exp : lngen.

(* begin hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv_mutual :
(forall e1 e2 n1,
  degree_exp_wrt_exp n1 (open_exp_wrt_exp_rec n1 e2 e1) ->
  degree_exp_wrt_exp (S n1) e1) /\
(forall peb1 e1 n1,
  degree_peb_wrt_exp n1 (open_peb_wrt_exp_rec n1 e1 peb1) ->
  degree_peb_wrt_exp (S n1) peb1) /\
(forall pe1 e1 n1,
  degree_pe_wrt_exp n1 (open_pe_wrt_exp_rec n1 e1 pe1) ->
  degree_pe_wrt_exp (S n1) pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv :
forall e1 e2 n1,
  degree_exp_wrt_exp n1 (open_exp_wrt_exp_rec n1 e2 e1) ->
  degree_exp_wrt_exp (S n1) e1.
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv :
forall peb1 e1 n1,
  degree_peb_wrt_exp n1 (open_peb_wrt_exp_rec n1 e1 peb1) ->
  degree_peb_wrt_exp (S n1) peb1.
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv :
forall pe1 e1 n1,
  degree_pe_wrt_exp n1 (open_pe_wrt_exp_rec n1 e1 pe1) ->
  degree_pe_wrt_exp (S n1) pe1.
Proof.
pose proof degree_exp_wrt_exp_open_exp_wrt_exp_rec_inv_degree_peb_wrt_exp_open_peb_wrt_exp_rec_inv_degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_pe_wrt_exp_open_pe_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_open_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv_mutual :
(forall t1 e1 n1,
  degree_type_wrt_exp n1 (open_type_wrt_exp_rec n1 e1 t1) ->
  degree_type_wrt_exp (S n1) t1) /\
(forall ptt1 e1 n1,
  degree_ptt_wrt_exp n1 (open_ptt_wrt_exp_rec n1 e1 ptt1) ->
  degree_ptt_wrt_exp (S n1) ptt1) /\
(forall pt1 e1 n1,
  degree_pt_wrt_exp n1 (open_pt_wrt_exp_rec n1 e1 pt1) ->
  degree_pt_wrt_exp (S n1) pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_type_wrt_exp_open_type_wrt_exp_rec_inv :
forall t1 e1 n1,
  degree_type_wrt_exp n1 (open_type_wrt_exp_rec n1 e1 t1) ->
  degree_type_wrt_exp (S n1) t1.
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_type_wrt_exp_open_type_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv :
forall ptt1 e1 n1,
  degree_ptt_wrt_exp n1 (open_ptt_wrt_exp_rec n1 e1 ptt1) ->
  degree_ptt_wrt_exp (S n1) ptt1.
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv :
forall pt1 e1 n1,
  degree_pt_wrt_exp n1 (open_pt_wrt_exp_rec n1 e1 pt1) ->
  degree_pt_wrt_exp (S n1) pt1.
Proof.
pose proof degree_type_wrt_exp_open_type_wrt_exp_rec_inv_degree_ptt_wrt_exp_open_ptt_wrt_exp_rec_inv_degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_pt_wrt_exp_open_pt_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_open_kind_wrt_exp_rec_inv_mutual :
(forall k1 e1 n1,
  degree_kind_wrt_exp n1 (open_kind_wrt_exp_rec n1 e1 k1) ->
  degree_kind_wrt_exp (S n1) k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_kind_wrt_exp_open_kind_wrt_exp_rec_inv :
forall k1 e1 n1,
  degree_kind_wrt_exp n1 (open_kind_wrt_exp_rec n1 e1 k1) ->
  degree_kind_wrt_exp (S n1) k1.
Proof.
pose proof degree_kind_wrt_exp_open_kind_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_kind_wrt_exp_open_kind_wrt_exp_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_open_bind_wrt_exp_rec_inv_mutual :
(forall bind1 e1 n1,
  degree_bind_wrt_exp n1 (open_bind_wrt_exp_rec n1 e1 bind1) ->
  degree_bind_wrt_exp (S n1) bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_bind_wrt_exp_open_bind_wrt_exp_rec_inv :
forall bind1 e1 n1,
  degree_bind_wrt_exp n1 (open_bind_wrt_exp_rec n1 e1 bind1) ->
  degree_bind_wrt_exp (S n1) bind1.
Proof.
pose proof degree_bind_wrt_exp_open_bind_wrt_exp_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_bind_wrt_exp_open_bind_wrt_exp_rec_inv : lngen.

(* end hide *)

Lemma degree_exp_wrt_exp_open_exp_wrt_exp_inv :
forall e1 e2,
  degree_exp_wrt_exp 0 (open_exp_wrt_exp e1 e2) ->
  degree_exp_wrt_exp 1 e1.
Proof.
unfold open_exp_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_exp_wrt_exp_open_exp_wrt_exp_inv : lngen.

Lemma degree_peb_wrt_exp_open_peb_wrt_exp_inv :
forall peb1 e1,
  degree_peb_wrt_exp 0 (open_peb_wrt_exp peb1 e1) ->
  degree_peb_wrt_exp 1 peb1.
Proof.
unfold open_peb_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_peb_wrt_exp_open_peb_wrt_exp_inv : lngen.

Lemma degree_pe_wrt_exp_open_pe_wrt_exp_inv :
forall pe1 e1,
  degree_pe_wrt_exp 0 (open_pe_wrt_exp pe1 e1) ->
  degree_pe_wrt_exp 1 pe1.
Proof.
unfold open_pe_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_pe_wrt_exp_open_pe_wrt_exp_inv : lngen.

Lemma degree_type_wrt_exp_open_type_wrt_exp_inv :
forall t1 e1,
  degree_type_wrt_exp 0 (open_type_wrt_exp t1 e1) ->
  degree_type_wrt_exp 1 t1.
Proof.
unfold open_type_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_type_wrt_exp_open_type_wrt_exp_inv : lngen.

Lemma degree_ptt_wrt_exp_open_ptt_wrt_exp_inv :
forall ptt1 e1,
  degree_ptt_wrt_exp 0 (open_ptt_wrt_exp ptt1 e1) ->
  degree_ptt_wrt_exp 1 ptt1.
Proof.
unfold open_ptt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_ptt_wrt_exp_open_ptt_wrt_exp_inv : lngen.

Lemma degree_pt_wrt_exp_open_pt_wrt_exp_inv :
forall pt1 e1,
  degree_pt_wrt_exp 0 (open_pt_wrt_exp pt1 e1) ->
  degree_pt_wrt_exp 1 pt1.
Proof.
unfold open_pt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_pt_wrt_exp_open_pt_wrt_exp_inv : lngen.

Lemma degree_kind_wrt_exp_open_kind_wrt_exp_inv :
forall k1 e1,
  degree_kind_wrt_exp 0 (open_kind_wrt_exp k1 e1) ->
  degree_kind_wrt_exp 1 k1.
Proof.
unfold open_kind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_kind_wrt_exp_open_kind_wrt_exp_inv : lngen.

Lemma degree_bind_wrt_exp_open_bind_wrt_exp_inv :
forall bind1 e1,
  degree_bind_wrt_exp 0 (open_bind_wrt_exp bind1 e1) ->
  degree_bind_wrt_exp 1 bind1.
Proof.
unfold open_bind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate degree_bind_wrt_exp_open_bind_wrt_exp_inv : lngen.


(* *********************************************************************** *)
(** * Theorems about [open] and [close] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma close_exp_wrt_exp_rec_inj_close_peb_wrt_exp_rec_inj_close_pe_wrt_exp_rec_inj_mutual :
(forall e1 e2 x1 n1,
  close_exp_wrt_exp_rec n1 x1 e1 = close_exp_wrt_exp_rec n1 x1 e2 ->
  e1 = e2) /\
(forall peb1 peb2 x1 n1,
  close_peb_wrt_exp_rec n1 x1 peb1 = close_peb_wrt_exp_rec n1 x1 peb2 ->
  peb1 = peb2) /\
(forall pe1 pe2 x1 n1,
  close_pe_wrt_exp_rec n1 x1 pe1 = close_pe_wrt_exp_rec n1 x1 pe2 ->
  pe1 = pe2).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_exp_wrt_exp_rec_inj :
forall e1 e2 x1 n1,
  close_exp_wrt_exp_rec n1 x1 e1 = close_exp_wrt_exp_rec n1 x1 e2 ->
  e1 = e2.
Proof.
pose proof close_exp_wrt_exp_rec_inj_close_peb_wrt_exp_rec_inj_close_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_exp_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_peb_wrt_exp_rec_inj :
forall peb1 peb2 x1 n1,
  close_peb_wrt_exp_rec n1 x1 peb1 = close_peb_wrt_exp_rec n1 x1 peb2 ->
  peb1 = peb2.
Proof.
pose proof close_exp_wrt_exp_rec_inj_close_peb_wrt_exp_rec_inj_close_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_peb_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pe_wrt_exp_rec_inj :
forall pe1 pe2 x1 n1,
  close_pe_wrt_exp_rec n1 x1 pe1 = close_pe_wrt_exp_rec n1 x1 pe2 ->
  pe1 = pe2.
Proof.
pose proof close_exp_wrt_exp_rec_inj_close_peb_wrt_exp_rec_inj_close_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_pe_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_inj_close_ptt_wrt_exp_rec_inj_close_pt_wrt_exp_rec_inj_mutual :
(forall t1 t2 x1 n1,
  close_type_wrt_exp_rec n1 x1 t1 = close_type_wrt_exp_rec n1 x1 t2 ->
  t1 = t2) /\
(forall ptt1 ptt2 x1 n1,
  close_ptt_wrt_exp_rec n1 x1 ptt1 = close_ptt_wrt_exp_rec n1 x1 ptt2 ->
  ptt1 = ptt2) /\
(forall pt1 pt2 x1 n1,
  close_pt_wrt_exp_rec n1 x1 pt1 = close_pt_wrt_exp_rec n1 x1 pt2 ->
  pt1 = pt2).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_inj :
forall t1 t2 x1 n1,
  close_type_wrt_exp_rec n1 x1 t1 = close_type_wrt_exp_rec n1 x1 t2 ->
  t1 = t2.
Proof.
pose proof close_type_wrt_exp_rec_inj_close_ptt_wrt_exp_rec_inj_close_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_type_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ptt_wrt_exp_rec_inj :
forall ptt1 ptt2 x1 n1,
  close_ptt_wrt_exp_rec n1 x1 ptt1 = close_ptt_wrt_exp_rec n1 x1 ptt2 ->
  ptt1 = ptt2.
Proof.
pose proof close_type_wrt_exp_rec_inj_close_ptt_wrt_exp_rec_inj_close_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_ptt_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pt_wrt_exp_rec_inj :
forall pt1 pt2 x1 n1,
  close_pt_wrt_exp_rec n1 x1 pt1 = close_pt_wrt_exp_rec n1 x1 pt2 ->
  pt1 = pt2.
Proof.
pose proof close_type_wrt_exp_rec_inj_close_ptt_wrt_exp_rec_inj_close_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_pt_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_inj_mutual :
(forall k1 k2 x1 n1,
  close_kind_wrt_exp_rec n1 x1 k1 = close_kind_wrt_exp_rec n1 x1 k2 ->
  k1 = k2).
Proof.
apply_mutual_ind kind_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_inj :
forall k1 k2 x1 n1,
  close_kind_wrt_exp_rec n1 x1 k1 = close_kind_wrt_exp_rec n1 x1 k2 ->
  k1 = k2.
Proof.
pose proof close_kind_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_kind_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_inj_mutual :
(forall bind1 bind2 x1 n1,
  close_bind_wrt_exp_rec n1 x1 bind1 = close_bind_wrt_exp_rec n1 x1 bind2 ->
  bind1 = bind2).
Proof.
apply_mutual_ind bind_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_inj :
forall bind1 bind2 x1 n1,
  close_bind_wrt_exp_rec n1 x1 bind1 = close_bind_wrt_exp_rec n1 x1 bind2 ->
  bind1 = bind2.
Proof.
pose proof close_bind_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_bind_wrt_exp_rec_inj : lngen.

(* end hide *)

Lemma close_exp_wrt_exp_inj :
forall e1 e2 x1,
  close_exp_wrt_exp x1 e1 = close_exp_wrt_exp x1 e2 ->
  e1 = e2.
Proof.
unfold close_exp_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_exp_wrt_exp_inj : lngen.

Lemma close_peb_wrt_exp_inj :
forall peb1 peb2 x1,
  close_peb_wrt_exp x1 peb1 = close_peb_wrt_exp x1 peb2 ->
  peb1 = peb2.
Proof.
unfold close_peb_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_peb_wrt_exp_inj : lngen.

Lemma close_pe_wrt_exp_inj :
forall pe1 pe2 x1,
  close_pe_wrt_exp x1 pe1 = close_pe_wrt_exp x1 pe2 ->
  pe1 = pe2.
Proof.
unfold close_pe_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_pe_wrt_exp_inj : lngen.

Lemma close_type_wrt_exp_inj :
forall t1 t2 x1,
  close_type_wrt_exp x1 t1 = close_type_wrt_exp x1 t2 ->
  t1 = t2.
Proof.
unfold close_type_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_type_wrt_exp_inj : lngen.

Lemma close_ptt_wrt_exp_inj :
forall ptt1 ptt2 x1,
  close_ptt_wrt_exp x1 ptt1 = close_ptt_wrt_exp x1 ptt2 ->
  ptt1 = ptt2.
Proof.
unfold close_ptt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_ptt_wrt_exp_inj : lngen.

Lemma close_pt_wrt_exp_inj :
forall pt1 pt2 x1,
  close_pt_wrt_exp x1 pt1 = close_pt_wrt_exp x1 pt2 ->
  pt1 = pt2.
Proof.
unfold close_pt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_pt_wrt_exp_inj : lngen.

Lemma close_kind_wrt_exp_inj :
forall k1 k2 x1,
  close_kind_wrt_exp x1 k1 = close_kind_wrt_exp x1 k2 ->
  k1 = k2.
Proof.
unfold close_kind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_kind_wrt_exp_inj : lngen.

Lemma close_bind_wrt_exp_inj :
forall bind1 bind2 x1,
  close_bind_wrt_exp x1 bind1 = close_bind_wrt_exp x1 bind2 ->
  bind1 = bind2.
Proof.
unfold close_bind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate close_bind_wrt_exp_inj : lngen.

(* begin hide *)

Lemma close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual :
(forall e1 x1 n1,
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp_rec n1 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1) = e1) /\
(forall peb1 x1 n1,
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp_rec n1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1) = peb1) /\
(forall pe1 x1 n1,
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp_rec n1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1) = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_exp_wrt_exp_rec_open_exp_wrt_exp_rec :
forall e1 x1 n1,
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp_rec n1 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1) = e1.
Proof.
pose proof close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_exp_wrt_exp_rec_open_exp_wrt_exp_rec : lngen.
Hint Rewrite close_exp_wrt_exp_rec_open_exp_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_peb_wrt_exp_rec_open_peb_wrt_exp_rec :
forall peb1 x1 n1,
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp_rec n1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1) = peb1.
Proof.
pose proof close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_peb_wrt_exp_rec_open_peb_wrt_exp_rec : lngen.
Hint Rewrite close_peb_wrt_exp_rec_open_peb_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pe_wrt_exp_rec_open_pe_wrt_exp_rec :
forall pe1 x1 n1,
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp_rec n1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1) = pe1.
Proof.
pose proof close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_pe_wrt_exp_rec_open_pe_wrt_exp_rec : lngen.
Hint Rewrite close_pe_wrt_exp_rec_open_pe_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_open_type_wrt_exp_rec_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual :
(forall t1 x1 n1,
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp_rec n1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x1) t1) = t1) /\
(forall ptt1 x1 n1,
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp_rec n1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1) = ptt1) /\
(forall pt1 x1 n1,
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp_rec n1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1) = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_open_type_wrt_exp_rec :
forall t1 x1 n1,
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp_rec n1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x1) t1) = t1.
Proof.
pose proof close_type_wrt_exp_rec_open_type_wrt_exp_rec_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_type_wrt_exp_rec_open_type_wrt_exp_rec : lngen.
Hint Rewrite close_type_wrt_exp_rec_open_type_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec :
forall ptt1 x1 n1,
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp_rec n1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1) = ptt1.
Proof.
pose proof close_type_wrt_exp_rec_open_type_wrt_exp_rec_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec : lngen.
Hint Rewrite close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pt_wrt_exp_rec_open_pt_wrt_exp_rec :
forall pt1 x1 n1,
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp_rec n1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1) = pt1.
Proof.
pose proof close_type_wrt_exp_rec_open_type_wrt_exp_rec_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_pt_wrt_exp_rec_open_pt_wrt_exp_rec : lngen.
Hint Rewrite close_pt_wrt_exp_rec_open_pt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_open_kind_wrt_exp_rec_mutual :
(forall k1 x1 n1,
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp_rec n1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1) = k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_open_kind_wrt_exp_rec :
forall k1 x1 n1,
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp_rec n1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1) = k1.
Proof.
pose proof close_kind_wrt_exp_rec_open_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_kind_wrt_exp_rec_open_kind_wrt_exp_rec : lngen.
Hint Rewrite close_kind_wrt_exp_rec_open_kind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_open_bind_wrt_exp_rec_mutual :
(forall bind1 x1 n1,
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp_rec n1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1) = bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_open_bind_wrt_exp_rec :
forall bind1 x1 n1,
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp_rec n1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1) = bind1.
Proof.
pose proof close_bind_wrt_exp_rec_open_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_bind_wrt_exp_rec_open_bind_wrt_exp_rec : lngen.
Hint Rewrite close_bind_wrt_exp_rec_open_bind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

Lemma close_exp_wrt_exp_open_exp_wrt_exp :
forall e1 x1,
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp x1 (open_exp_wrt_exp e1 (e_Var_f x1)) = e1.
Proof.
unfold close_exp_wrt_exp; unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve close_exp_wrt_exp_open_exp_wrt_exp : lngen.
Hint Rewrite close_exp_wrt_exp_open_exp_wrt_exp using solve [auto] : lngen.

Lemma close_peb_wrt_exp_open_peb_wrt_exp :
forall peb1 x1,
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp x1 (open_peb_wrt_exp peb1 (e_Var_f x1)) = peb1.
Proof.
unfold close_peb_wrt_exp; unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve close_peb_wrt_exp_open_peb_wrt_exp : lngen.
Hint Rewrite close_peb_wrt_exp_open_peb_wrt_exp using solve [auto] : lngen.

Lemma close_pe_wrt_exp_open_pe_wrt_exp :
forall pe1 x1,
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp x1 (open_pe_wrt_exp pe1 (e_Var_f x1)) = pe1.
Proof.
unfold close_pe_wrt_exp; unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve close_pe_wrt_exp_open_pe_wrt_exp : lngen.
Hint Rewrite close_pe_wrt_exp_open_pe_wrt_exp using solve [auto] : lngen.

Lemma close_type_wrt_exp_open_type_wrt_exp :
forall t1 x1,
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp x1 (open_type_wrt_exp t1 (e_Var_f x1)) = t1.
Proof.
unfold close_type_wrt_exp; unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve close_type_wrt_exp_open_type_wrt_exp : lngen.
Hint Rewrite close_type_wrt_exp_open_type_wrt_exp using solve [auto] : lngen.

Lemma close_ptt_wrt_exp_open_ptt_wrt_exp :
forall ptt1 x1,
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp x1 (open_ptt_wrt_exp ptt1 (e_Var_f x1)) = ptt1.
Proof.
unfold close_ptt_wrt_exp; unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve close_ptt_wrt_exp_open_ptt_wrt_exp : lngen.
Hint Rewrite close_ptt_wrt_exp_open_ptt_wrt_exp using solve [auto] : lngen.

Lemma close_pt_wrt_exp_open_pt_wrt_exp :
forall pt1 x1,
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp x1 (open_pt_wrt_exp pt1 (e_Var_f x1)) = pt1.
Proof.
unfold close_pt_wrt_exp; unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve close_pt_wrt_exp_open_pt_wrt_exp : lngen.
Hint Rewrite close_pt_wrt_exp_open_pt_wrt_exp using solve [auto] : lngen.

Lemma close_kind_wrt_exp_open_kind_wrt_exp :
forall k1 x1,
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp x1 (open_kind_wrt_exp k1 (e_Var_f x1)) = k1.
Proof.
unfold close_kind_wrt_exp; unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve close_kind_wrt_exp_open_kind_wrt_exp : lngen.
Hint Rewrite close_kind_wrt_exp_open_kind_wrt_exp using solve [auto] : lngen.

Lemma close_bind_wrt_exp_open_bind_wrt_exp :
forall bind1 x1,
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp x1 (open_bind_wrt_exp bind1 (e_Var_f x1)) = bind1.
Proof.
unfold close_bind_wrt_exp; unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve close_bind_wrt_exp_open_bind_wrt_exp : lngen.
Hint Rewrite close_bind_wrt_exp_open_bind_wrt_exp using solve [auto] : lngen.

(* begin hide *)

Lemma open_exp_wrt_exp_rec_close_exp_wrt_exp_rec_open_peb_wrt_exp_rec_close_peb_wrt_exp_rec_open_pe_wrt_exp_rec_close_pe_wrt_exp_rec_mutual :
(forall e1 x1 n1,
  open_exp_wrt_exp_rec n1 (e_Var_f x1) (close_exp_wrt_exp_rec n1 x1 e1) = e1) /\
(forall peb1 x1 n1,
  open_peb_wrt_exp_rec n1 (e_Var_f x1) (close_peb_wrt_exp_rec n1 x1 peb1) = peb1) /\
(forall pe1 x1 n1,
  open_pe_wrt_exp_rec n1 (e_Var_f x1) (close_pe_wrt_exp_rec n1 x1 pe1) = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_exp_wrt_exp_rec_close_exp_wrt_exp_rec :
forall e1 x1 n1,
  open_exp_wrt_exp_rec n1 (e_Var_f x1) (close_exp_wrt_exp_rec n1 x1 e1) = e1.
Proof.
pose proof open_exp_wrt_exp_rec_close_exp_wrt_exp_rec_open_peb_wrt_exp_rec_close_peb_wrt_exp_rec_open_pe_wrt_exp_rec_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_exp_wrt_exp_rec_close_exp_wrt_exp_rec : lngen.
Hint Rewrite open_exp_wrt_exp_rec_close_exp_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_peb_wrt_exp_rec_close_peb_wrt_exp_rec :
forall peb1 x1 n1,
  open_peb_wrt_exp_rec n1 (e_Var_f x1) (close_peb_wrt_exp_rec n1 x1 peb1) = peb1.
Proof.
pose proof open_exp_wrt_exp_rec_close_exp_wrt_exp_rec_open_peb_wrt_exp_rec_close_peb_wrt_exp_rec_open_pe_wrt_exp_rec_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_peb_wrt_exp_rec_close_peb_wrt_exp_rec : lngen.
Hint Rewrite open_peb_wrt_exp_rec_close_peb_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pe_wrt_exp_rec_close_pe_wrt_exp_rec :
forall pe1 x1 n1,
  open_pe_wrt_exp_rec n1 (e_Var_f x1) (close_pe_wrt_exp_rec n1 x1 pe1) = pe1.
Proof.
pose proof open_exp_wrt_exp_rec_close_exp_wrt_exp_rec_open_peb_wrt_exp_rec_close_peb_wrt_exp_rec_open_pe_wrt_exp_rec_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_pe_wrt_exp_rec_close_pe_wrt_exp_rec : lngen.
Hint Rewrite open_pe_wrt_exp_rec_close_pe_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_close_type_wrt_exp_rec_open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec_open_pt_wrt_exp_rec_close_pt_wrt_exp_rec_mutual :
(forall t1 x1 n1,
  open_type_wrt_exp_rec n1 (e_Var_f x1) (close_type_wrt_exp_rec n1 x1 t1) = t1) /\
(forall ptt1 x1 n1,
  open_ptt_wrt_exp_rec n1 (e_Var_f x1) (close_ptt_wrt_exp_rec n1 x1 ptt1) = ptt1) /\
(forall pt1 x1 n1,
  open_pt_wrt_exp_rec n1 (e_Var_f x1) (close_pt_wrt_exp_rec n1 x1 pt1) = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_close_type_wrt_exp_rec :
forall t1 x1 n1,
  open_type_wrt_exp_rec n1 (e_Var_f x1) (close_type_wrt_exp_rec n1 x1 t1) = t1.
Proof.
pose proof open_type_wrt_exp_rec_close_type_wrt_exp_rec_open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec_open_pt_wrt_exp_rec_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_type_wrt_exp_rec_close_type_wrt_exp_rec : lngen.
Hint Rewrite open_type_wrt_exp_rec_close_type_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec :
forall ptt1 x1 n1,
  open_ptt_wrt_exp_rec n1 (e_Var_f x1) (close_ptt_wrt_exp_rec n1 x1 ptt1) = ptt1.
Proof.
pose proof open_type_wrt_exp_rec_close_type_wrt_exp_rec_open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec_open_pt_wrt_exp_rec_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec : lngen.
Hint Rewrite open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pt_wrt_exp_rec_close_pt_wrt_exp_rec :
forall pt1 x1 n1,
  open_pt_wrt_exp_rec n1 (e_Var_f x1) (close_pt_wrt_exp_rec n1 x1 pt1) = pt1.
Proof.
pose proof open_type_wrt_exp_rec_close_type_wrt_exp_rec_open_ptt_wrt_exp_rec_close_ptt_wrt_exp_rec_open_pt_wrt_exp_rec_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_pt_wrt_exp_rec_close_pt_wrt_exp_rec : lngen.
Hint Rewrite open_pt_wrt_exp_rec_close_pt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_close_kind_wrt_exp_rec_mutual :
(forall k1 x1 n1,
  open_kind_wrt_exp_rec n1 (e_Var_f x1) (close_kind_wrt_exp_rec n1 x1 k1) = k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_close_kind_wrt_exp_rec :
forall k1 x1 n1,
  open_kind_wrt_exp_rec n1 (e_Var_f x1) (close_kind_wrt_exp_rec n1 x1 k1) = k1.
Proof.
pose proof open_kind_wrt_exp_rec_close_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_kind_wrt_exp_rec_close_kind_wrt_exp_rec : lngen.
Hint Rewrite open_kind_wrt_exp_rec_close_kind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_close_bind_wrt_exp_rec_mutual :
(forall bind1 x1 n1,
  open_bind_wrt_exp_rec n1 (e_Var_f x1) (close_bind_wrt_exp_rec n1 x1 bind1) = bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_close_bind_wrt_exp_rec :
forall bind1 x1 n1,
  open_bind_wrt_exp_rec n1 (e_Var_f x1) (close_bind_wrt_exp_rec n1 x1 bind1) = bind1.
Proof.
pose proof open_bind_wrt_exp_rec_close_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_bind_wrt_exp_rec_close_bind_wrt_exp_rec : lngen.
Hint Rewrite open_bind_wrt_exp_rec_close_bind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

Lemma open_exp_wrt_exp_close_exp_wrt_exp :
forall e1 x1,
  open_exp_wrt_exp (close_exp_wrt_exp x1 e1) (e_Var_f x1) = e1.
Proof.
unfold close_exp_wrt_exp; unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve open_exp_wrt_exp_close_exp_wrt_exp : lngen.
Hint Rewrite open_exp_wrt_exp_close_exp_wrt_exp using solve [auto] : lngen.

Lemma open_peb_wrt_exp_close_peb_wrt_exp :
forall peb1 x1,
  open_peb_wrt_exp (close_peb_wrt_exp x1 peb1) (e_Var_f x1) = peb1.
Proof.
unfold close_peb_wrt_exp; unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve open_peb_wrt_exp_close_peb_wrt_exp : lngen.
Hint Rewrite open_peb_wrt_exp_close_peb_wrt_exp using solve [auto] : lngen.

Lemma open_pe_wrt_exp_close_pe_wrt_exp :
forall pe1 x1,
  open_pe_wrt_exp (close_pe_wrt_exp x1 pe1) (e_Var_f x1) = pe1.
Proof.
unfold close_pe_wrt_exp; unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve open_pe_wrt_exp_close_pe_wrt_exp : lngen.
Hint Rewrite open_pe_wrt_exp_close_pe_wrt_exp using solve [auto] : lngen.

Lemma open_type_wrt_exp_close_type_wrt_exp :
forall t1 x1,
  open_type_wrt_exp (close_type_wrt_exp x1 t1) (e_Var_f x1) = t1.
Proof.
unfold close_type_wrt_exp; unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve open_type_wrt_exp_close_type_wrt_exp : lngen.
Hint Rewrite open_type_wrt_exp_close_type_wrt_exp using solve [auto] : lngen.

Lemma open_ptt_wrt_exp_close_ptt_wrt_exp :
forall ptt1 x1,
  open_ptt_wrt_exp (close_ptt_wrt_exp x1 ptt1) (e_Var_f x1) = ptt1.
Proof.
unfold close_ptt_wrt_exp; unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve open_ptt_wrt_exp_close_ptt_wrt_exp : lngen.
Hint Rewrite open_ptt_wrt_exp_close_ptt_wrt_exp using solve [auto] : lngen.

Lemma open_pt_wrt_exp_close_pt_wrt_exp :
forall pt1 x1,
  open_pt_wrt_exp (close_pt_wrt_exp x1 pt1) (e_Var_f x1) = pt1.
Proof.
unfold close_pt_wrt_exp; unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve open_pt_wrt_exp_close_pt_wrt_exp : lngen.
Hint Rewrite open_pt_wrt_exp_close_pt_wrt_exp using solve [auto] : lngen.

Lemma open_kind_wrt_exp_close_kind_wrt_exp :
forall k1 x1,
  open_kind_wrt_exp (close_kind_wrt_exp x1 k1) (e_Var_f x1) = k1.
Proof.
unfold close_kind_wrt_exp; unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve open_kind_wrt_exp_close_kind_wrt_exp : lngen.
Hint Rewrite open_kind_wrt_exp_close_kind_wrt_exp using solve [auto] : lngen.

Lemma open_bind_wrt_exp_close_bind_wrt_exp :
forall bind1 x1,
  open_bind_wrt_exp (close_bind_wrt_exp x1 bind1) (e_Var_f x1) = bind1.
Proof.
unfold close_bind_wrt_exp; unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve open_bind_wrt_exp_close_bind_wrt_exp : lngen.
Hint Rewrite open_bind_wrt_exp_close_bind_wrt_exp using solve [auto] : lngen.

(* begin hide *)

Lemma open_exp_wrt_exp_rec_inj_open_peb_wrt_exp_rec_inj_open_pe_wrt_exp_rec_inj_mutual :
(forall e2 e1 x1 n1,
  x1 `notin` efv_exp e2 ->
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp_rec n1 (e_Var_f x1) e2 = open_exp_wrt_exp_rec n1 (e_Var_f x1) e1 ->
  e2 = e1) /\
(forall peb2 peb1 x1 n1,
  x1 `notin` efv_peb peb2 ->
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp_rec n1 (e_Var_f x1) peb2 = open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1 ->
  peb2 = peb1) /\
(forall pe2 pe1 x1 n1,
  x1 `notin` efv_pe pe2 ->
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp_rec n1 (e_Var_f x1) pe2 = open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1 ->
  pe2 = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_exp_wrt_exp_rec_inj :
forall e2 e1 x1 n1,
  x1 `notin` efv_exp e2 ->
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp_rec n1 (e_Var_f x1) e2 = open_exp_wrt_exp_rec n1 (e_Var_f x1) e1 ->
  e2 = e1.
Proof.
pose proof open_exp_wrt_exp_rec_inj_open_peb_wrt_exp_rec_inj_open_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_exp_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_peb_wrt_exp_rec_inj :
forall peb2 peb1 x1 n1,
  x1 `notin` efv_peb peb2 ->
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp_rec n1 (e_Var_f x1) peb2 = open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1 ->
  peb2 = peb1.
Proof.
pose proof open_exp_wrt_exp_rec_inj_open_peb_wrt_exp_rec_inj_open_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_peb_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pe_wrt_exp_rec_inj :
forall pe2 pe1 x1 n1,
  x1 `notin` efv_pe pe2 ->
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp_rec n1 (e_Var_f x1) pe2 = open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1 ->
  pe2 = pe1.
Proof.
pose proof open_exp_wrt_exp_rec_inj_open_peb_wrt_exp_rec_inj_open_pe_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_pe_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_inj_open_ptt_wrt_exp_rec_inj_open_pt_wrt_exp_rec_inj_mutual :
(forall t2 t1 x1 n1,
  x1 `notin` efv_type t2 ->
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp_rec n1 (e_Var_f x1) t2 = open_type_wrt_exp_rec n1 (e_Var_f x1) t1 ->
  t2 = t1) /\
(forall ptt2 ptt1 x1 n1,
  x1 `notin` efv_ptt ptt2 ->
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt2 = open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1 ->
  ptt2 = ptt1) /\
(forall pt2 pt1 x1 n1,
  x1 `notin` efv_pt pt2 ->
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp_rec n1 (e_Var_f x1) pt2 = open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1 ->
  pt2 = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_inj :
forall t2 t1 x1 n1,
  x1 `notin` efv_type t2 ->
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp_rec n1 (e_Var_f x1) t2 = open_type_wrt_exp_rec n1 (e_Var_f x1) t1 ->
  t2 = t1.
Proof.
pose proof open_type_wrt_exp_rec_inj_open_ptt_wrt_exp_rec_inj_open_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_type_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ptt_wrt_exp_rec_inj :
forall ptt2 ptt1 x1 n1,
  x1 `notin` efv_ptt ptt2 ->
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt2 = open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1 ->
  ptt2 = ptt1.
Proof.
pose proof open_type_wrt_exp_rec_inj_open_ptt_wrt_exp_rec_inj_open_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_ptt_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pt_wrt_exp_rec_inj :
forall pt2 pt1 x1 n1,
  x1 `notin` efv_pt pt2 ->
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp_rec n1 (e_Var_f x1) pt2 = open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1 ->
  pt2 = pt1.
Proof.
pose proof open_type_wrt_exp_rec_inj_open_ptt_wrt_exp_rec_inj_open_pt_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_pt_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_inj_mutual :
(forall k2 k1 x1 n1,
  x1 `notin` efv_kind k2 ->
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp_rec n1 (e_Var_f x1) k2 = open_kind_wrt_exp_rec n1 (e_Var_f x1) k1 ->
  k2 = k1).
Proof.
apply_mutual_ind kind_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_inj :
forall k2 k1 x1 n1,
  x1 `notin` efv_kind k2 ->
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp_rec n1 (e_Var_f x1) k2 = open_kind_wrt_exp_rec n1 (e_Var_f x1) k1 ->
  k2 = k1.
Proof.
pose proof open_kind_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_kind_wrt_exp_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_inj_mutual :
(forall bind2 bind1 x1 n1,
  x1 `notin` efv_bind bind2 ->
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp_rec n1 (e_Var_f x1) bind2 = open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1 ->
  bind2 = bind1).
Proof.
apply_mutual_ind bind_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_inj :
forall bind2 bind1 x1 n1,
  x1 `notin` efv_bind bind2 ->
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp_rec n1 (e_Var_f x1) bind2 = open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1 ->
  bind2 = bind1.
Proof.
pose proof open_bind_wrt_exp_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_bind_wrt_exp_rec_inj : lngen.

(* end hide *)

Lemma open_exp_wrt_exp_inj :
forall e2 e1 x1,
  x1 `notin` efv_exp e2 ->
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp e2 (e_Var_f x1) = open_exp_wrt_exp e1 (e_Var_f x1) ->
  e2 = e1.
Proof.
unfold open_exp_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_exp_wrt_exp_inj : lngen.

Lemma open_peb_wrt_exp_inj :
forall peb2 peb1 x1,
  x1 `notin` efv_peb peb2 ->
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp peb2 (e_Var_f x1) = open_peb_wrt_exp peb1 (e_Var_f x1) ->
  peb2 = peb1.
Proof.
unfold open_peb_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_peb_wrt_exp_inj : lngen.

Lemma open_pe_wrt_exp_inj :
forall pe2 pe1 x1,
  x1 `notin` efv_pe pe2 ->
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp pe2 (e_Var_f x1) = open_pe_wrt_exp pe1 (e_Var_f x1) ->
  pe2 = pe1.
Proof.
unfold open_pe_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_pe_wrt_exp_inj : lngen.

Lemma open_type_wrt_exp_inj :
forall t2 t1 x1,
  x1 `notin` efv_type t2 ->
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp t2 (e_Var_f x1) = open_type_wrt_exp t1 (e_Var_f x1) ->
  t2 = t1.
Proof.
unfold open_type_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_type_wrt_exp_inj : lngen.

Lemma open_ptt_wrt_exp_inj :
forall ptt2 ptt1 x1,
  x1 `notin` efv_ptt ptt2 ->
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp ptt2 (e_Var_f x1) = open_ptt_wrt_exp ptt1 (e_Var_f x1) ->
  ptt2 = ptt1.
Proof.
unfold open_ptt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_ptt_wrt_exp_inj : lngen.

Lemma open_pt_wrt_exp_inj :
forall pt2 pt1 x1,
  x1 `notin` efv_pt pt2 ->
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp pt2 (e_Var_f x1) = open_pt_wrt_exp pt1 (e_Var_f x1) ->
  pt2 = pt1.
Proof.
unfold open_pt_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_pt_wrt_exp_inj : lngen.

Lemma open_kind_wrt_exp_inj :
forall k2 k1 x1,
  x1 `notin` efv_kind k2 ->
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp k2 (e_Var_f x1) = open_kind_wrt_exp k1 (e_Var_f x1) ->
  k2 = k1.
Proof.
unfold open_kind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_kind_wrt_exp_inj : lngen.

Lemma open_bind_wrt_exp_inj :
forall bind2 bind1 x1,
  x1 `notin` efv_bind bind2 ->
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp bind2 (e_Var_f x1) = open_bind_wrt_exp bind1 (e_Var_f x1) ->
  bind2 = bind1.
Proof.
unfold open_bind_wrt_exp; eauto with lngen.
Qed.

Hint Immediate open_bind_wrt_exp_inj : lngen.


(* *********************************************************************** *)
(** * Theorems about [lc] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma degree_exp_wrt_exp_of_lc_exp_degree_peb_wrt_exp_of_lc_peb_degree_pe_wrt_exp_of_lc_pe_mutual :
(forall e1,
  lc_exp e1 ->
  degree_exp_wrt_exp 0 e1) /\
(forall peb1,
  lc_peb peb1 ->
  degree_peb_wrt_exp 0 peb1) /\
(forall pe1,
  lc_pe pe1 ->
  degree_pe_wrt_exp 0 pe1).
Proof.
apply_mutual_ind lc_exp_lc_peb_lc_pe_mutind;
intros;
let x1 := fresh "x1" in pick_fresh x1;
repeat (match goal with
          | H1 : _, H2 : _ |- _ => specialize H1 with H2
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_exp_wrt_exp_of_lc_exp :
forall e1,
  lc_exp e1 ->
  degree_exp_wrt_exp 0 e1.
Proof.
pose proof degree_exp_wrt_exp_of_lc_exp_degree_peb_wrt_exp_of_lc_peb_degree_pe_wrt_exp_of_lc_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_exp_wrt_exp_of_lc_exp : lngen.

Lemma degree_peb_wrt_exp_of_lc_peb :
forall peb1,
  lc_peb peb1 ->
  degree_peb_wrt_exp 0 peb1.
Proof.
pose proof degree_exp_wrt_exp_of_lc_exp_degree_peb_wrt_exp_of_lc_peb_degree_pe_wrt_exp_of_lc_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_peb_wrt_exp_of_lc_peb : lngen.

Lemma degree_pe_wrt_exp_of_lc_pe :
forall pe1,
  lc_pe pe1 ->
  degree_pe_wrt_exp 0 pe1.
Proof.
pose proof degree_exp_wrt_exp_of_lc_exp_degree_peb_wrt_exp_of_lc_peb_degree_pe_wrt_exp_of_lc_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pe_wrt_exp_of_lc_pe : lngen.

(* begin hide *)

Lemma degree_type_wrt_exp_of_lc_type_degree_ptt_wrt_exp_of_lc_ptt_degree_pt_wrt_exp_of_lc_pt_mutual :
(forall t1,
  lc_type t1 ->
  degree_type_wrt_exp 0 t1) /\
(forall ptt1,
  lc_ptt ptt1 ->
  degree_ptt_wrt_exp 0 ptt1) /\
(forall pt1,
  lc_pt pt1 ->
  degree_pt_wrt_exp 0 pt1).
Proof.
apply_mutual_ind lc_type_lc_ptt_lc_pt_mutind;
intros;
let x1 := fresh "x1" in pick_fresh x1;
repeat (match goal with
          | H1 : _, H2 : _ |- _ => specialize H1 with H2
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_type_wrt_exp_of_lc_type :
forall t1,
  lc_type t1 ->
  degree_type_wrt_exp 0 t1.
Proof.
pose proof degree_type_wrt_exp_of_lc_type_degree_ptt_wrt_exp_of_lc_ptt_degree_pt_wrt_exp_of_lc_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_type_wrt_exp_of_lc_type : lngen.

Lemma degree_ptt_wrt_exp_of_lc_ptt :
forall ptt1,
  lc_ptt ptt1 ->
  degree_ptt_wrt_exp 0 ptt1.
Proof.
pose proof degree_type_wrt_exp_of_lc_type_degree_ptt_wrt_exp_of_lc_ptt_degree_pt_wrt_exp_of_lc_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ptt_wrt_exp_of_lc_ptt : lngen.

Lemma degree_pt_wrt_exp_of_lc_pt :
forall pt1,
  lc_pt pt1 ->
  degree_pt_wrt_exp 0 pt1.
Proof.
pose proof degree_type_wrt_exp_of_lc_type_degree_ptt_wrt_exp_of_lc_ptt_degree_pt_wrt_exp_of_lc_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_pt_wrt_exp_of_lc_pt : lngen.

(* begin hide *)

Lemma degree_kind_wrt_exp_of_lc_kind_mutual :
(forall k1,
  lc_kind k1 ->
  degree_kind_wrt_exp 0 k1).
Proof.
apply_mutual_ind lc_kind_mutind;
intros;
let x1 := fresh "x1" in pick_fresh x1;
repeat (match goal with
          | H1 : _, H2 : _ |- _ => specialize H1 with H2
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_kind_wrt_exp_of_lc_kind :
forall k1,
  lc_kind k1 ->
  degree_kind_wrt_exp 0 k1.
Proof.
pose proof degree_kind_wrt_exp_of_lc_kind_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_kind_wrt_exp_of_lc_kind : lngen.

(* begin hide *)

Lemma degree_bind_wrt_exp_of_lc_bind_mutual :
(forall bind1,
  lc_bind bind1 ->
  degree_bind_wrt_exp 0 bind1).
Proof.
apply_mutual_ind lc_bind_mutind;
intros;
let x1 := fresh "x1" in pick_fresh x1;
repeat (match goal with
          | H1 : _, H2 : _ |- _ => specialize H1 with H2
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_bind_wrt_exp_of_lc_bind :
forall bind1,
  lc_bind bind1 ->
  degree_bind_wrt_exp 0 bind1.
Proof.
pose proof degree_bind_wrt_exp_of_lc_bind_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_bind_wrt_exp_of_lc_bind : lngen.

(* begin hide *)

Lemma lc_exp_of_degree_lc_peb_of_degree_lc_pe_of_degree_size_mutual :
forall i1,
(forall e1,
  size_exp e1 = i1 ->
  degree_exp_wrt_exp 0 e1 ->
  lc_exp e1) /\
(forall peb1,
  size_peb peb1 = i1 ->
  degree_peb_wrt_exp 0 peb1 ->
  lc_peb peb1) /\
(forall pe1,
  size_pe pe1 = i1 ->
  degree_pe_wrt_exp 0 pe1 ->
  lc_pe pe1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind exp_peb_pe_mutind;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_exp_of_degree :
forall e1,
  degree_exp_wrt_exp 0 e1 ->
  lc_exp e1.
Proof.
intros e1; intros;
pose proof (lc_exp_of_degree_lc_peb_of_degree_lc_pe_of_degree_size_mutual (size_exp e1));
intuition eauto.
Qed.

Hint Resolve lc_exp_of_degree : lngen.

Lemma lc_peb_of_degree :
forall peb1,
  degree_peb_wrt_exp 0 peb1 ->
  lc_peb peb1.
Proof.
intros peb1; intros;
pose proof (lc_exp_of_degree_lc_peb_of_degree_lc_pe_of_degree_size_mutual (size_peb peb1));
intuition eauto.
Qed.

Hint Resolve lc_peb_of_degree : lngen.

Lemma lc_pe_of_degree :
forall pe1,
  degree_pe_wrt_exp 0 pe1 ->
  lc_pe pe1.
Proof.
intros pe1; intros;
pose proof (lc_exp_of_degree_lc_peb_of_degree_lc_pe_of_degree_size_mutual (size_pe pe1));
intuition eauto.
Qed.

Hint Resolve lc_pe_of_degree : lngen.

(* begin hide *)

Lemma lc_type_of_degree_lc_ptt_of_degree_lc_pt_of_degree_size_mutual :
forall i1,
(forall t1,
  size_type t1 = i1 ->
  degree_type_wrt_exp 0 t1 ->
  lc_type t1) /\
(forall ptt1,
  size_ptt ptt1 = i1 ->
  degree_ptt_wrt_exp 0 ptt1 ->
  lc_ptt ptt1) /\
(forall pt1,
  size_pt pt1 = i1 ->
  degree_pt_wrt_exp 0 pt1 ->
  lc_pt pt1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind type_ptt_pt_mutind;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_type_of_degree :
forall t1,
  degree_type_wrt_exp 0 t1 ->
  lc_type t1.
Proof.
intros t1; intros;
pose proof (lc_type_of_degree_lc_ptt_of_degree_lc_pt_of_degree_size_mutual (size_type t1));
intuition eauto.
Qed.

Hint Resolve lc_type_of_degree : lngen.

Lemma lc_ptt_of_degree :
forall ptt1,
  degree_ptt_wrt_exp 0 ptt1 ->
  lc_ptt ptt1.
Proof.
intros ptt1; intros;
pose proof (lc_type_of_degree_lc_ptt_of_degree_lc_pt_of_degree_size_mutual (size_ptt ptt1));
intuition eauto.
Qed.

Hint Resolve lc_ptt_of_degree : lngen.

Lemma lc_pt_of_degree :
forall pt1,
  degree_pt_wrt_exp 0 pt1 ->
  lc_pt pt1.
Proof.
intros pt1; intros;
pose proof (lc_type_of_degree_lc_ptt_of_degree_lc_pt_of_degree_size_mutual (size_pt pt1));
intuition eauto.
Qed.

Hint Resolve lc_pt_of_degree : lngen.

(* begin hide *)

Lemma lc_kind_of_degree_size_mutual :
forall i1,
(forall k1,
  size_kind k1 = i1 ->
  degree_kind_wrt_exp 0 k1 ->
  lc_kind k1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind kind_mutind;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_kind_of_degree :
forall k1,
  degree_kind_wrt_exp 0 k1 ->
  lc_kind k1.
Proof.
intros k1; intros;
pose proof (lc_kind_of_degree_size_mutual (size_kind k1));
intuition eauto.
Qed.

Hint Resolve lc_kind_of_degree : lngen.

(* begin hide *)

Lemma lc_bind_of_degree_size_mutual :
forall i1,
(forall bind1,
  size_bind bind1 = i1 ->
  degree_bind_wrt_exp 0 bind1 ->
  lc_bind bind1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind bind_mutind;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_bind_of_degree :
forall bind1,
  degree_bind_wrt_exp 0 bind1 ->
  lc_bind bind1.
Proof.
intros bind1; intros;
pose proof (lc_bind_of_degree_size_mutual (size_bind bind1));
intuition eauto.
Qed.

Hint Resolve lc_bind_of_degree : lngen.

Ltac Cs_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              fail 1
          end).

Ltac exp_peb_pe_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_exp_wrt_exp_of_lc_exp in J1; clear H
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_peb_wrt_exp_of_lc_peb in J1; clear H
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_pe_wrt_exp_of_lc_pe in J1; clear H
          end).

Ltac type_ptt_pt_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_type_wrt_exp_of_lc_type in J1; clear H
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_ptt_wrt_exp_of_lc_ptt in J1; clear H
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_pt_wrt_exp_of_lc_pt in J1; clear H
          end).

Ltac kind_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_kind_wrt_exp_of_lc_kind in J1; clear H
          end).

Ltac bind_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_bind_wrt_exp_of_lc_bind in J1; clear H
          end).

Lemma lc_e_Lam_exists :
forall x1 e1,
  lc_exp (open_exp_wrt_exp e1 (e_Var_f x1)) ->
  lc_exp (e_Lam e1).
Proof.
intros; exp_peb_pe_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_e_Fix_exists :
forall f1 e1,
  lc_exp (open_exp_wrt_exp e1 (e_Var_f f1)) ->
  lc_exp (e_Fix e1).
Proof.
intros; exp_peb_pe_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_peb_Branch_exists :
forall x1 C1 e1,
  lc_exp (open_exp_wrt_exp e1 (e_Var_f x1)) ->
  lc_peb (peb_Branch C1 e1).
Proof.
intros; exp_peb_pe_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_t_Pi_exists :
forall x1 t1 t2,
  lc_type t1 ->
  lc_type (open_type_wrt_exp t2 (e_Var_f x1)) ->
  lc_type (t_Pi t1 t2).
Proof.
intros; type_ptt_pt_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_t_Sigma_exists :
forall x1 t1 t2,
  lc_type t1 ->
  lc_type (open_type_wrt_exp t2 (e_Var_f x1)) ->
  lc_type (t_Sigma t1 t2).
Proof.
intros; type_ptt_pt_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_ptt_pattern_exists :
forall x1 C1 t1,
  lc_type (open_type_wrt_exp t1 (e_Var_f x1)) ->
  lc_ptt (ptt_pattern C1 t1).
Proof.
intros; type_ptt_pt_lc_exists_tac; eauto with lngen.
Qed.

Lemma lc_k_pi_exists :
forall x1 t1 k1,
  lc_type t1 ->
  lc_kind (open_kind_wrt_exp k1 (e_Var_f x1)) ->
  lc_kind (k_pi t1 k1).
Proof.
intros; kind_lc_exists_tac; eauto with lngen.
Qed.

Hint Extern 1 (lc_exp (e_Lam _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_e_Lam_exists x1).

Hint Extern 1 (lc_exp (e_Fix _)) =>
  let f1 := fresh in
  pick_fresh f1;
  apply (lc_e_Fix_exists f1).

Hint Extern 1 (lc_peb (peb_Branch _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_peb_Branch_exists x1).

Hint Extern 1 (lc_type (t_Pi _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_t_Pi_exists x1).

Hint Extern 1 (lc_type (t_Sigma _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_t_Sigma_exists x1).

Hint Extern 1 (lc_ptt (ptt_pattern _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_ptt_pattern_exists x1).

Hint Extern 1 (lc_kind (k_pi _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_k_pi_exists x1).

Lemma lc_body_exp_wrt_exp :
forall e1 e2,
  body_exp_wrt_exp e1 ->
  lc_exp e2 ->
  lc_exp (open_exp_wrt_exp e1 e2).
Proof.
unfold body_exp_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
exp_peb_pe_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_exp_wrt_exp : lngen.

Lemma lc_body_peb_wrt_exp :
forall peb1 e1,
  body_peb_wrt_exp peb1 ->
  lc_exp e1 ->
  lc_peb (open_peb_wrt_exp peb1 e1).
Proof.
unfold body_peb_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
exp_peb_pe_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_peb_wrt_exp : lngen.

Lemma lc_body_pe_wrt_exp :
forall pe1 e1,
  body_pe_wrt_exp pe1 ->
  lc_exp e1 ->
  lc_pe (open_pe_wrt_exp pe1 e1).
Proof.
unfold body_pe_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
exp_peb_pe_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_pe_wrt_exp : lngen.

Lemma lc_body_type_wrt_exp :
forall t1 e1,
  body_type_wrt_exp t1 ->
  lc_exp e1 ->
  lc_type (open_type_wrt_exp t1 e1).
Proof.
unfold body_type_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
type_ptt_pt_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_type_wrt_exp : lngen.

Lemma lc_body_ptt_wrt_exp :
forall ptt1 e1,
  body_ptt_wrt_exp ptt1 ->
  lc_exp e1 ->
  lc_ptt (open_ptt_wrt_exp ptt1 e1).
Proof.
unfold body_ptt_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
type_ptt_pt_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_ptt_wrt_exp : lngen.

Lemma lc_body_pt_wrt_exp :
forall pt1 e1,
  body_pt_wrt_exp pt1 ->
  lc_exp e1 ->
  lc_pt (open_pt_wrt_exp pt1 e1).
Proof.
unfold body_pt_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
type_ptt_pt_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_pt_wrt_exp : lngen.

Lemma lc_body_kind_wrt_exp :
forall k1 e1,
  body_kind_wrt_exp k1 ->
  lc_exp e1 ->
  lc_kind (open_kind_wrt_exp k1 e1).
Proof.
unfold body_kind_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
kind_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_kind_wrt_exp : lngen.

Lemma lc_body_bind_wrt_exp :
forall bind1 e1,
  body_bind_wrt_exp bind1 ->
  lc_exp e1 ->
  lc_bind (open_bind_wrt_exp bind1 e1).
Proof.
unfold body_bind_wrt_exp;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
bind_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body_bind_wrt_exp : lngen.

Lemma lc_body_e_Lam_1 :
forall e1,
  lc_exp (e_Lam e1) ->
  body_exp_wrt_exp e1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_e_Lam_1 : lngen.

Lemma lc_body_e_Fix_1 :
forall e1,
  lc_exp (e_Fix e1) ->
  body_exp_wrt_exp e1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_e_Fix_1 : lngen.

Lemma lc_body_peb_Branch_2 :
forall C1 e1,
  lc_peb (peb_Branch C1 e1) ->
  body_exp_wrt_exp e1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_peb_Branch_2 : lngen.

Lemma lc_body_t_Pi_2 :
forall t1 t2,
  lc_type (t_Pi t1 t2) ->
  body_type_wrt_exp t2.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_t_Pi_2 : lngen.

Lemma lc_body_t_Sigma_2 :
forall t1 t2,
  lc_type (t_Sigma t1 t2) ->
  body_type_wrt_exp t2.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_t_Sigma_2 : lngen.

Lemma lc_body_ptt_pattern_2 :
forall C1 t1,
  lc_ptt (ptt_pattern C1 t1) ->
  body_type_wrt_exp t1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_ptt_pattern_2 : lngen.

Lemma lc_body_k_pi_2 :
forall t1 k1,
  lc_kind (k_pi t1 k1) ->
  body_kind_wrt_exp k1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_k_pi_2 : lngen.

(* begin hide *)

Lemma lc_exp_unique_lc_peb_unique_lc_pe_unique_mutual :
(forall e1 (proof2 proof3 : lc_exp e1), proof2 = proof3) /\
(forall peb1 (proof2 proof3 : lc_peb peb1), proof2 = proof3) /\
(forall pe1 (proof2 proof3 : lc_pe pe1), proof2 = proof3).
Proof.
apply_mutual_ind lc_exp_lc_peb_lc_pe_mutind;
intros;
let proof1 := fresh "proof1" in
rename_last_into proof1; dependent destruction proof1;
f_equal; default_simp; auto using @functional_extensionality_dep with lngen.
Qed.

(* end hide *)

Lemma lc_exp_unique :
forall e1 (proof2 proof3 : lc_exp e1), proof2 = proof3.
Proof.
pose proof lc_exp_unique_lc_peb_unique_lc_pe_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_exp_unique : lngen.

Lemma lc_peb_unique :
forall peb1 (proof2 proof3 : lc_peb peb1), proof2 = proof3.
Proof.
pose proof lc_exp_unique_lc_peb_unique_lc_pe_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_peb_unique : lngen.

Lemma lc_pe_unique :
forall pe1 (proof2 proof3 : lc_pe pe1), proof2 = proof3.
Proof.
pose proof lc_exp_unique_lc_peb_unique_lc_pe_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_pe_unique : lngen.

(* begin hide *)

Lemma lc_type_unique_lc_ptt_unique_lc_pt_unique_mutual :
(forall t1 (proof2 proof3 : lc_type t1), proof2 = proof3) /\
(forall ptt1 (proof2 proof3 : lc_ptt ptt1), proof2 = proof3) /\
(forall pt1 (proof2 proof3 : lc_pt pt1), proof2 = proof3).
Proof.
apply_mutual_ind lc_type_lc_ptt_lc_pt_mutind;
intros;
let proof1 := fresh "proof1" in
rename_last_into proof1; dependent destruction proof1;
f_equal; default_simp; auto using @functional_extensionality_dep with lngen.
Qed.

(* end hide *)

Lemma lc_type_unique :
forall t1 (proof2 proof3 : lc_type t1), proof2 = proof3.
Proof.
pose proof lc_type_unique_lc_ptt_unique_lc_pt_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_type_unique : lngen.

Lemma lc_ptt_unique :
forall ptt1 (proof2 proof3 : lc_ptt ptt1), proof2 = proof3.
Proof.
pose proof lc_type_unique_lc_ptt_unique_lc_pt_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_ptt_unique : lngen.

Lemma lc_pt_unique :
forall pt1 (proof2 proof3 : lc_pt pt1), proof2 = proof3.
Proof.
pose proof lc_type_unique_lc_ptt_unique_lc_pt_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_pt_unique : lngen.

(* begin hide *)

Lemma lc_kind_unique_mutual :
(forall k1 (proof2 proof3 : lc_kind k1), proof2 = proof3).
Proof.
apply_mutual_ind lc_kind_mutind;
intros;
let proof1 := fresh "proof1" in
rename_last_into proof1; dependent destruction proof1;
f_equal; default_simp; auto using @functional_extensionality_dep with lngen.
Qed.

(* end hide *)

Lemma lc_kind_unique :
forall k1 (proof2 proof3 : lc_kind k1), proof2 = proof3.
Proof.
pose proof lc_kind_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_kind_unique : lngen.

(* begin hide *)

Lemma lc_bind_unique_mutual :
(forall bind1 (proof2 proof3 : lc_bind bind1), proof2 = proof3).
Proof.
apply_mutual_ind lc_bind_mutind;
intros;
let proof1 := fresh "proof1" in
rename_last_into proof1; dependent destruction proof1;
f_equal; default_simp; auto using @functional_extensionality_dep with lngen.
Qed.

(* end hide *)

Lemma lc_bind_unique :
forall bind1 (proof2 proof3 : lc_bind bind1), proof2 = proof3.
Proof.
pose proof lc_bind_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_bind_unique : lngen.

(* begin hide *)

Lemma lc_exp_of_lc_set_exp_lc_peb_of_lc_set_peb_lc_pe_of_lc_set_pe_mutual :
(forall e1, lc_set_exp e1 -> lc_exp e1) /\
(forall peb1, lc_set_peb peb1 -> lc_peb peb1) /\
(forall pe1, lc_set_pe pe1 -> lc_pe pe1).
Proof.
apply_mutual_ind lc_set_exp_lc_set_peb_lc_set_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma lc_exp_of_lc_set_exp :
forall e1, lc_set_exp e1 -> lc_exp e1.
Proof.
pose proof lc_exp_of_lc_set_exp_lc_peb_of_lc_set_peb_lc_pe_of_lc_set_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_exp_of_lc_set_exp : lngen.

Lemma lc_peb_of_lc_set_peb :
forall peb1, lc_set_peb peb1 -> lc_peb peb1.
Proof.
pose proof lc_exp_of_lc_set_exp_lc_peb_of_lc_set_peb_lc_pe_of_lc_set_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_peb_of_lc_set_peb : lngen.

Lemma lc_pe_of_lc_set_pe :
forall pe1, lc_set_pe pe1 -> lc_pe pe1.
Proof.
pose proof lc_exp_of_lc_set_exp_lc_peb_of_lc_set_peb_lc_pe_of_lc_set_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_pe_of_lc_set_pe : lngen.

(* begin hide *)

Lemma lc_type_of_lc_set_type_lc_ptt_of_lc_set_ptt_lc_pt_of_lc_set_pt_mutual :
(forall t1, lc_set_type t1 -> lc_type t1) /\
(forall ptt1, lc_set_ptt ptt1 -> lc_ptt ptt1) /\
(forall pt1, lc_set_pt pt1 -> lc_pt pt1).
Proof.
apply_mutual_ind lc_set_type_lc_set_ptt_lc_set_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma lc_type_of_lc_set_type :
forall t1, lc_set_type t1 -> lc_type t1.
Proof.
pose proof lc_type_of_lc_set_type_lc_ptt_of_lc_set_ptt_lc_pt_of_lc_set_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_type_of_lc_set_type : lngen.

Lemma lc_ptt_of_lc_set_ptt :
forall ptt1, lc_set_ptt ptt1 -> lc_ptt ptt1.
Proof.
pose proof lc_type_of_lc_set_type_lc_ptt_of_lc_set_ptt_lc_pt_of_lc_set_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_ptt_of_lc_set_ptt : lngen.

Lemma lc_pt_of_lc_set_pt :
forall pt1, lc_set_pt pt1 -> lc_pt pt1.
Proof.
pose proof lc_type_of_lc_set_type_lc_ptt_of_lc_set_ptt_lc_pt_of_lc_set_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_pt_of_lc_set_pt : lngen.

(* begin hide *)

Lemma lc_kind_of_lc_set_kind_mutual :
(forall k1, lc_set_kind k1 -> lc_kind k1).
Proof.
apply_mutual_ind lc_set_kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma lc_kind_of_lc_set_kind :
forall k1, lc_set_kind k1 -> lc_kind k1.
Proof.
pose proof lc_kind_of_lc_set_kind_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_kind_of_lc_set_kind : lngen.

(* begin hide *)

Lemma lc_bind_of_lc_set_bind_mutual :
(forall bind1, lc_set_bind bind1 -> lc_bind bind1).
Proof.
apply_mutual_ind lc_set_bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma lc_bind_of_lc_set_bind :
forall bind1, lc_set_bind bind1 -> lc_bind bind1.
Proof.
pose proof lc_bind_of_lc_set_bind_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_bind_of_lc_set_bind : lngen.

(* begin hide *)

Lemma lc_set_exp_of_lc_exp_lc_set_peb_of_lc_peb_lc_set_pe_of_lc_pe_size_mutual :
forall i1,
(forall e1,
  size_exp e1 = i1 ->
  lc_exp e1 ->
  lc_set_exp e1) *
(forall peb1,
  size_peb peb1 = i1 ->
  lc_peb peb1 ->
  lc_set_peb peb1) *
(forall pe1,
  size_pe pe1 = i1 ->
  lc_pe pe1 ->
  lc_set_pe pe1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind exp_peb_pe_mutrec;
default_simp;
try solve [assert False by default_simp; tauto];
(* non-trivial cases *)
constructor; default_simp;
try first [apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb];
default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_set_exp_of_lc_exp :
forall e1,
  lc_exp e1 ->
  lc_set_exp e1.
Proof.
intros e1; intros;
pose proof (lc_set_exp_of_lc_exp_lc_set_peb_of_lc_peb_lc_set_pe_of_lc_pe_size_mutual (size_exp e1));
intuition eauto.
Qed.

Hint Resolve lc_set_exp_of_lc_exp : lngen.

Lemma lc_set_peb_of_lc_peb :
forall peb1,
  lc_peb peb1 ->
  lc_set_peb peb1.
Proof.
intros peb1; intros;
pose proof (lc_set_exp_of_lc_exp_lc_set_peb_of_lc_peb_lc_set_pe_of_lc_pe_size_mutual (size_peb peb1));
intuition eauto.
Qed.

Hint Resolve lc_set_peb_of_lc_peb : lngen.

Lemma lc_set_pe_of_lc_pe :
forall pe1,
  lc_pe pe1 ->
  lc_set_pe pe1.
Proof.
intros pe1; intros;
pose proof (lc_set_exp_of_lc_exp_lc_set_peb_of_lc_peb_lc_set_pe_of_lc_pe_size_mutual (size_pe pe1));
intuition eauto.
Qed.

Hint Resolve lc_set_pe_of_lc_pe : lngen.

(* begin hide *)

Lemma lc_set_type_of_lc_type_lc_set_ptt_of_lc_ptt_lc_set_pt_of_lc_pt_size_mutual :
forall i1,
(forall t1,
  size_type t1 = i1 ->
  lc_type t1 ->
  lc_set_type t1) *
(forall ptt1,
  size_ptt ptt1 = i1 ->
  lc_ptt ptt1 ->
  lc_set_ptt ptt1) *
(forall pt1,
  size_pt pt1 = i1 ->
  lc_pt pt1 ->
  lc_set_pt pt1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind type_ptt_pt_mutrec;
default_simp;
try solve [assert False by default_simp; tauto];
(* non-trivial cases *)
constructor; default_simp;
try first [apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_pt_of_lc_pt
 | apply lc_set_ptt_of_lc_ptt
 | apply lc_set_type_of_lc_type
 | apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_pt_of_lc_pt
 | apply lc_set_ptt_of_lc_ptt
 | apply lc_set_type_of_lc_type
 | apply lc_set_exp_of_lc_exp
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_pt_of_lc_pt
 | apply lc_set_ptt_of_lc_ptt
 | apply lc_set_type_of_lc_type];
default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_set_type_of_lc_type :
forall t1,
  lc_type t1 ->
  lc_set_type t1.
Proof.
intros t1; intros;
pose proof (lc_set_type_of_lc_type_lc_set_ptt_of_lc_ptt_lc_set_pt_of_lc_pt_size_mutual (size_type t1));
intuition eauto.
Qed.

Hint Resolve lc_set_type_of_lc_type : lngen.

Lemma lc_set_ptt_of_lc_ptt :
forall ptt1,
  lc_ptt ptt1 ->
  lc_set_ptt ptt1.
Proof.
intros ptt1; intros;
pose proof (lc_set_type_of_lc_type_lc_set_ptt_of_lc_ptt_lc_set_pt_of_lc_pt_size_mutual (size_ptt ptt1));
intuition eauto.
Qed.

Hint Resolve lc_set_ptt_of_lc_ptt : lngen.

Lemma lc_set_pt_of_lc_pt :
forall pt1,
  lc_pt pt1 ->
  lc_set_pt pt1.
Proof.
intros pt1; intros;
pose proof (lc_set_type_of_lc_type_lc_set_ptt_of_lc_ptt_lc_set_pt_of_lc_pt_size_mutual (size_pt pt1));
intuition eauto.
Qed.

Hint Resolve lc_set_pt_of_lc_pt : lngen.

(* begin hide *)

Lemma lc_set_kind_of_lc_kind_size_mutual :
forall i1,
(forall k1,
  size_kind k1 = i1 ->
  lc_kind k1 ->
  lc_set_kind k1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind kind_mutrec;
default_simp;
try solve [assert False by default_simp; tauto];
(* non-trivial cases *)
constructor; default_simp;
try first [apply lc_set_exp_of_lc_exp
 | apply lc_set_kind_of_lc_kind
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_pt_of_lc_pt
 | apply lc_set_ptt_of_lc_ptt
 | apply lc_set_type_of_lc_type];
default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_set_kind_of_lc_kind :
forall k1,
  lc_kind k1 ->
  lc_set_kind k1.
Proof.
intros k1; intros;
pose proof (lc_set_kind_of_lc_kind_size_mutual (size_kind k1));
intuition eauto.
Qed.

Hint Resolve lc_set_kind_of_lc_kind : lngen.

(* begin hide *)

Lemma lc_set_bind_of_lc_bind_size_mutual :
forall i1,
(forall bind1,
  size_bind bind1 = i1 ->
  lc_bind bind1 ->
  lc_set_bind bind1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind bind_mutrec;
default_simp;
try solve [assert False by default_simp; tauto];
(* non-trivial cases *)
constructor; default_simp;
try first [apply lc_set_bind_of_lc_bind
 | apply lc_set_exp_of_lc_exp
 | apply lc_set_kind_of_lc_kind
 | apply lc_set_pe_of_lc_pe
 | apply lc_set_peb_of_lc_peb
 | apply lc_set_pt_of_lc_pt
 | apply lc_set_ptt_of_lc_ptt
 | apply lc_set_type_of_lc_type];
default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_set_bind_of_lc_bind :
forall bind1,
  lc_bind bind1 ->
  lc_set_bind bind1.
Proof.
intros bind1; intros;
pose proof (lc_set_bind_of_lc_bind_size_mutual (size_bind bind1));
intuition eauto.
Qed.

Hint Resolve lc_set_bind_of_lc_bind : lngen.


(* *********************************************************************** *)
(** * More theorems about [open] and [close] *)

Ltac default_auto ::= auto with lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma close_exp_wrt_exp_rec_degree_exp_wrt_exp_close_peb_wrt_exp_rec_degree_peb_wrt_exp_close_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual :
(forall e1 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp_rec n1 x1 e1 = e1) /\
(forall peb1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp_rec n1 x1 peb1 = peb1) /\
(forall pe1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp_rec n1 x1 pe1 = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_exp_wrt_exp_rec_degree_exp_wrt_exp :
forall e1 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp_rec n1 x1 e1 = e1.
Proof.
pose proof close_exp_wrt_exp_rec_degree_exp_wrt_exp_close_peb_wrt_exp_rec_degree_peb_wrt_exp_close_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_exp_wrt_exp_rec_degree_exp_wrt_exp : lngen.
Hint Rewrite close_exp_wrt_exp_rec_degree_exp_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_peb_wrt_exp_rec_degree_peb_wrt_exp :
forall peb1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp_rec n1 x1 peb1 = peb1.
Proof.
pose proof close_exp_wrt_exp_rec_degree_exp_wrt_exp_close_peb_wrt_exp_rec_degree_peb_wrt_exp_close_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_peb_wrt_exp_rec_degree_peb_wrt_exp : lngen.
Hint Rewrite close_peb_wrt_exp_rec_degree_peb_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pe_wrt_exp_rec_degree_pe_wrt_exp :
forall pe1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp_rec n1 x1 pe1 = pe1.
Proof.
pose proof close_exp_wrt_exp_rec_degree_exp_wrt_exp_close_peb_wrt_exp_rec_degree_peb_wrt_exp_close_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_pe_wrt_exp_rec_degree_pe_wrt_exp : lngen.
Hint Rewrite close_pe_wrt_exp_rec_degree_pe_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_degree_type_wrt_exp_close_ptt_wrt_exp_rec_degree_ptt_wrt_exp_close_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual :
(forall t1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp_rec n1 x1 t1 = t1) /\
(forall ptt1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp_rec n1 x1 ptt1 = ptt1) /\
(forall pt1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp_rec n1 x1 pt1 = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_type_wrt_exp_rec_degree_type_wrt_exp :
forall t1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp_rec n1 x1 t1 = t1.
Proof.
pose proof close_type_wrt_exp_rec_degree_type_wrt_exp_close_ptt_wrt_exp_rec_degree_ptt_wrt_exp_close_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_type_wrt_exp_rec_degree_type_wrt_exp : lngen.
Hint Rewrite close_type_wrt_exp_rec_degree_type_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ptt_wrt_exp_rec_degree_ptt_wrt_exp :
forall ptt1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp_rec n1 x1 ptt1 = ptt1.
Proof.
pose proof close_type_wrt_exp_rec_degree_type_wrt_exp_close_ptt_wrt_exp_rec_degree_ptt_wrt_exp_close_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ptt_wrt_exp_rec_degree_ptt_wrt_exp : lngen.
Hint Rewrite close_ptt_wrt_exp_rec_degree_ptt_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_pt_wrt_exp_rec_degree_pt_wrt_exp :
forall pt1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp_rec n1 x1 pt1 = pt1.
Proof.
pose proof close_type_wrt_exp_rec_degree_type_wrt_exp_close_ptt_wrt_exp_rec_degree_ptt_wrt_exp_close_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_pt_wrt_exp_rec_degree_pt_wrt_exp : lngen.
Hint Rewrite close_pt_wrt_exp_rec_degree_pt_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_degree_kind_wrt_exp_mutual :
(forall k1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp_rec n1 x1 k1 = k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_kind_wrt_exp_rec_degree_kind_wrt_exp :
forall k1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp_rec n1 x1 k1 = k1.
Proof.
pose proof close_kind_wrt_exp_rec_degree_kind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_kind_wrt_exp_rec_degree_kind_wrt_exp : lngen.
Hint Rewrite close_kind_wrt_exp_rec_degree_kind_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_degree_bind_wrt_exp_mutual :
(forall bind1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp_rec n1 x1 bind1 = bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_bind_wrt_exp_rec_degree_bind_wrt_exp :
forall bind1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp_rec n1 x1 bind1 = bind1.
Proof.
pose proof close_bind_wrt_exp_rec_degree_bind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve close_bind_wrt_exp_rec_degree_bind_wrt_exp : lngen.
Hint Rewrite close_bind_wrt_exp_rec_degree_bind_wrt_exp using solve [auto] : lngen.

(* end hide *)

Lemma close_exp_wrt_exp_lc_exp :
forall e1 x1,
  lc_exp e1 ->
  x1 `notin` efv_exp e1 ->
  close_exp_wrt_exp x1 e1 = e1.
Proof.
unfold close_exp_wrt_exp; default_simp.
Qed.

Hint Resolve close_exp_wrt_exp_lc_exp : lngen.
Hint Rewrite close_exp_wrt_exp_lc_exp using solve [auto] : lngen.

Lemma close_peb_wrt_exp_lc_peb :
forall peb1 x1,
  lc_peb peb1 ->
  x1 `notin` efv_peb peb1 ->
  close_peb_wrt_exp x1 peb1 = peb1.
Proof.
unfold close_peb_wrt_exp; default_simp.
Qed.

Hint Resolve close_peb_wrt_exp_lc_peb : lngen.
Hint Rewrite close_peb_wrt_exp_lc_peb using solve [auto] : lngen.

Lemma close_pe_wrt_exp_lc_pe :
forall pe1 x1,
  lc_pe pe1 ->
  x1 `notin` efv_pe pe1 ->
  close_pe_wrt_exp x1 pe1 = pe1.
Proof.
unfold close_pe_wrt_exp; default_simp.
Qed.

Hint Resolve close_pe_wrt_exp_lc_pe : lngen.
Hint Rewrite close_pe_wrt_exp_lc_pe using solve [auto] : lngen.

Lemma close_type_wrt_exp_lc_type :
forall t1 x1,
  lc_type t1 ->
  x1 `notin` efv_type t1 ->
  close_type_wrt_exp x1 t1 = t1.
Proof.
unfold close_type_wrt_exp; default_simp.
Qed.

Hint Resolve close_type_wrt_exp_lc_type : lngen.
Hint Rewrite close_type_wrt_exp_lc_type using solve [auto] : lngen.

Lemma close_ptt_wrt_exp_lc_ptt :
forall ptt1 x1,
  lc_ptt ptt1 ->
  x1 `notin` efv_ptt ptt1 ->
  close_ptt_wrt_exp x1 ptt1 = ptt1.
Proof.
unfold close_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve close_ptt_wrt_exp_lc_ptt : lngen.
Hint Rewrite close_ptt_wrt_exp_lc_ptt using solve [auto] : lngen.

Lemma close_pt_wrt_exp_lc_pt :
forall pt1 x1,
  lc_pt pt1 ->
  x1 `notin` efv_pt pt1 ->
  close_pt_wrt_exp x1 pt1 = pt1.
Proof.
unfold close_pt_wrt_exp; default_simp.
Qed.

Hint Resolve close_pt_wrt_exp_lc_pt : lngen.
Hint Rewrite close_pt_wrt_exp_lc_pt using solve [auto] : lngen.

Lemma close_kind_wrt_exp_lc_kind :
forall k1 x1,
  lc_kind k1 ->
  x1 `notin` efv_kind k1 ->
  close_kind_wrt_exp x1 k1 = k1.
Proof.
unfold close_kind_wrt_exp; default_simp.
Qed.

Hint Resolve close_kind_wrt_exp_lc_kind : lngen.
Hint Rewrite close_kind_wrt_exp_lc_kind using solve [auto] : lngen.

Lemma close_bind_wrt_exp_lc_bind :
forall bind1 x1,
  lc_bind bind1 ->
  x1 `notin` efv_bind bind1 ->
  close_bind_wrt_exp x1 bind1 = bind1.
Proof.
unfold close_bind_wrt_exp; default_simp.
Qed.

Hint Resolve close_bind_wrt_exp_lc_bind : lngen.
Hint Rewrite close_bind_wrt_exp_lc_bind using solve [auto] : lngen.

(* begin hide *)

Lemma open_exp_wrt_exp_rec_degree_exp_wrt_exp_open_peb_wrt_exp_rec_degree_peb_wrt_exp_open_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual :
(forall e2 e1 n1,
  degree_exp_wrt_exp n1 e2 ->
  open_exp_wrt_exp_rec n1 e1 e2 = e2) /\
(forall peb1 e1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  open_peb_wrt_exp_rec n1 e1 peb1 = peb1) /\
(forall pe1 e1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  open_pe_wrt_exp_rec n1 e1 pe1 = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_exp_wrt_exp_rec_degree_exp_wrt_exp :
forall e2 e1 n1,
  degree_exp_wrt_exp n1 e2 ->
  open_exp_wrt_exp_rec n1 e1 e2 = e2.
Proof.
pose proof open_exp_wrt_exp_rec_degree_exp_wrt_exp_open_peb_wrt_exp_rec_degree_peb_wrt_exp_open_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_exp_wrt_exp_rec_degree_exp_wrt_exp : lngen.
Hint Rewrite open_exp_wrt_exp_rec_degree_exp_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_peb_wrt_exp_rec_degree_peb_wrt_exp :
forall peb1 e1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  open_peb_wrt_exp_rec n1 e1 peb1 = peb1.
Proof.
pose proof open_exp_wrt_exp_rec_degree_exp_wrt_exp_open_peb_wrt_exp_rec_degree_peb_wrt_exp_open_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_peb_wrt_exp_rec_degree_peb_wrt_exp : lngen.
Hint Rewrite open_peb_wrt_exp_rec_degree_peb_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pe_wrt_exp_rec_degree_pe_wrt_exp :
forall pe1 e1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  open_pe_wrt_exp_rec n1 e1 pe1 = pe1.
Proof.
pose proof open_exp_wrt_exp_rec_degree_exp_wrt_exp_open_peb_wrt_exp_rec_degree_peb_wrt_exp_open_pe_wrt_exp_rec_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_pe_wrt_exp_rec_degree_pe_wrt_exp : lngen.
Hint Rewrite open_pe_wrt_exp_rec_degree_pe_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_degree_type_wrt_exp_open_ptt_wrt_exp_rec_degree_ptt_wrt_exp_open_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual :
(forall t1 e1 n1,
  degree_type_wrt_exp n1 t1 ->
  open_type_wrt_exp_rec n1 e1 t1 = t1) /\
(forall ptt1 e1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  open_ptt_wrt_exp_rec n1 e1 ptt1 = ptt1) /\
(forall pt1 e1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  open_pt_wrt_exp_rec n1 e1 pt1 = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_type_wrt_exp_rec_degree_type_wrt_exp :
forall t1 e1 n1,
  degree_type_wrt_exp n1 t1 ->
  open_type_wrt_exp_rec n1 e1 t1 = t1.
Proof.
pose proof open_type_wrt_exp_rec_degree_type_wrt_exp_open_ptt_wrt_exp_rec_degree_ptt_wrt_exp_open_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_type_wrt_exp_rec_degree_type_wrt_exp : lngen.
Hint Rewrite open_type_wrt_exp_rec_degree_type_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ptt_wrt_exp_rec_degree_ptt_wrt_exp :
forall ptt1 e1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  open_ptt_wrt_exp_rec n1 e1 ptt1 = ptt1.
Proof.
pose proof open_type_wrt_exp_rec_degree_type_wrt_exp_open_ptt_wrt_exp_rec_degree_ptt_wrt_exp_open_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ptt_wrt_exp_rec_degree_ptt_wrt_exp : lngen.
Hint Rewrite open_ptt_wrt_exp_rec_degree_ptt_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_pt_wrt_exp_rec_degree_pt_wrt_exp :
forall pt1 e1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  open_pt_wrt_exp_rec n1 e1 pt1 = pt1.
Proof.
pose proof open_type_wrt_exp_rec_degree_type_wrt_exp_open_ptt_wrt_exp_rec_degree_ptt_wrt_exp_open_pt_wrt_exp_rec_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_pt_wrt_exp_rec_degree_pt_wrt_exp : lngen.
Hint Rewrite open_pt_wrt_exp_rec_degree_pt_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_degree_kind_wrt_exp_mutual :
(forall k1 e1 n1,
  degree_kind_wrt_exp n1 k1 ->
  open_kind_wrt_exp_rec n1 e1 k1 = k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_kind_wrt_exp_rec_degree_kind_wrt_exp :
forall k1 e1 n1,
  degree_kind_wrt_exp n1 k1 ->
  open_kind_wrt_exp_rec n1 e1 k1 = k1.
Proof.
pose proof open_kind_wrt_exp_rec_degree_kind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_kind_wrt_exp_rec_degree_kind_wrt_exp : lngen.
Hint Rewrite open_kind_wrt_exp_rec_degree_kind_wrt_exp using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_degree_bind_wrt_exp_mutual :
(forall bind1 e1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  open_bind_wrt_exp_rec n1 e1 bind1 = bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_bind_wrt_exp_rec_degree_bind_wrt_exp :
forall bind1 e1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  open_bind_wrt_exp_rec n1 e1 bind1 = bind1.
Proof.
pose proof open_bind_wrt_exp_rec_degree_bind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve open_bind_wrt_exp_rec_degree_bind_wrt_exp : lngen.
Hint Rewrite open_bind_wrt_exp_rec_degree_bind_wrt_exp using solve [auto] : lngen.

(* end hide *)

Lemma open_exp_wrt_exp_lc_exp :
forall e2 e1,
  lc_exp e2 ->
  open_exp_wrt_exp e2 e1 = e2.
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve open_exp_wrt_exp_lc_exp : lngen.
Hint Rewrite open_exp_wrt_exp_lc_exp using solve [auto] : lngen.

Lemma open_peb_wrt_exp_lc_peb :
forall peb1 e1,
  lc_peb peb1 ->
  open_peb_wrt_exp peb1 e1 = peb1.
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve open_peb_wrt_exp_lc_peb : lngen.
Hint Rewrite open_peb_wrt_exp_lc_peb using solve [auto] : lngen.

Lemma open_pe_wrt_exp_lc_pe :
forall pe1 e1,
  lc_pe pe1 ->
  open_pe_wrt_exp pe1 e1 = pe1.
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve open_pe_wrt_exp_lc_pe : lngen.
Hint Rewrite open_pe_wrt_exp_lc_pe using solve [auto] : lngen.

Lemma open_type_wrt_exp_lc_type :
forall t1 e1,
  lc_type t1 ->
  open_type_wrt_exp t1 e1 = t1.
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve open_type_wrt_exp_lc_type : lngen.
Hint Rewrite open_type_wrt_exp_lc_type using solve [auto] : lngen.

Lemma open_ptt_wrt_exp_lc_ptt :
forall ptt1 e1,
  lc_ptt ptt1 ->
  open_ptt_wrt_exp ptt1 e1 = ptt1.
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve open_ptt_wrt_exp_lc_ptt : lngen.
Hint Rewrite open_ptt_wrt_exp_lc_ptt using solve [auto] : lngen.

Lemma open_pt_wrt_exp_lc_pt :
forall pt1 e1,
  lc_pt pt1 ->
  open_pt_wrt_exp pt1 e1 = pt1.
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve open_pt_wrt_exp_lc_pt : lngen.
Hint Rewrite open_pt_wrt_exp_lc_pt using solve [auto] : lngen.

Lemma open_kind_wrt_exp_lc_kind :
forall k1 e1,
  lc_kind k1 ->
  open_kind_wrt_exp k1 e1 = k1.
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve open_kind_wrt_exp_lc_kind : lngen.
Hint Rewrite open_kind_wrt_exp_lc_kind using solve [auto] : lngen.

Lemma open_bind_wrt_exp_lc_bind :
forall bind1 e1,
  lc_bind bind1 ->
  open_bind_wrt_exp bind1 e1 = bind1.
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve open_bind_wrt_exp_lc_bind : lngen.
Hint Rewrite open_bind_wrt_exp_lc_bind using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [fv] *)

Ltac default_auto ::= auto with set lngen; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma efv_exp_close_exp_wrt_exp_rec_efv_peb_close_peb_wrt_exp_rec_efv_pe_close_pe_wrt_exp_rec_mutual :
(forall e1 x1 n1,
  efv_exp (close_exp_wrt_exp_rec n1 x1 e1) [=] remove x1 (efv_exp e1)) /\
(forall peb1 x1 n1,
  efv_peb (close_peb_wrt_exp_rec n1 x1 peb1) [=] remove x1 (efv_peb peb1)) /\
(forall pe1 x1 n1,
  efv_pe (close_pe_wrt_exp_rec n1 x1 pe1) [=] remove x1 (efv_pe pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_exp_close_exp_wrt_exp_rec :
forall e1 x1 n1,
  efv_exp (close_exp_wrt_exp_rec n1 x1 e1) [=] remove x1 (efv_exp e1).
Proof.
pose proof efv_exp_close_exp_wrt_exp_rec_efv_peb_close_peb_wrt_exp_rec_efv_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_close_exp_wrt_exp_rec : lngen.
Hint Rewrite efv_exp_close_exp_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_peb_close_peb_wrt_exp_rec :
forall peb1 x1 n1,
  efv_peb (close_peb_wrt_exp_rec n1 x1 peb1) [=] remove x1 (efv_peb peb1).
Proof.
pose proof efv_exp_close_exp_wrt_exp_rec_efv_peb_close_peb_wrt_exp_rec_efv_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_close_peb_wrt_exp_rec : lngen.
Hint Rewrite efv_peb_close_peb_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pe_close_pe_wrt_exp_rec :
forall pe1 x1 n1,
  efv_pe (close_pe_wrt_exp_rec n1 x1 pe1) [=] remove x1 (efv_pe pe1).
Proof.
pose proof efv_exp_close_exp_wrt_exp_rec_efv_peb_close_peb_wrt_exp_rec_efv_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_close_pe_wrt_exp_rec : lngen.
Hint Rewrite efv_pe_close_pe_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_type_close_type_wrt_exp_rec_efv_ptt_close_ptt_wrt_exp_rec_efv_pt_close_pt_wrt_exp_rec_mutual :
(forall t1 x1 n1,
  efv_type (close_type_wrt_exp_rec n1 x1 t1) [=] remove x1 (efv_type t1)) /\
(forall ptt1 x1 n1,
  efv_ptt (close_ptt_wrt_exp_rec n1 x1 ptt1) [=] remove x1 (efv_ptt ptt1)) /\
(forall pt1 x1 n1,
  efv_pt (close_pt_wrt_exp_rec n1 x1 pt1) [=] remove x1 (efv_pt pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_type_close_type_wrt_exp_rec :
forall t1 x1 n1,
  efv_type (close_type_wrt_exp_rec n1 x1 t1) [=] remove x1 (efv_type t1).
Proof.
pose proof efv_type_close_type_wrt_exp_rec_efv_ptt_close_ptt_wrt_exp_rec_efv_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_close_type_wrt_exp_rec : lngen.
Hint Rewrite efv_type_close_type_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_ptt_close_ptt_wrt_exp_rec :
forall ptt1 x1 n1,
  efv_ptt (close_ptt_wrt_exp_rec n1 x1 ptt1) [=] remove x1 (efv_ptt ptt1).
Proof.
pose proof efv_type_close_type_wrt_exp_rec_efv_ptt_close_ptt_wrt_exp_rec_efv_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_close_ptt_wrt_exp_rec : lngen.
Hint Rewrite efv_ptt_close_ptt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pt_close_pt_wrt_exp_rec :
forall pt1 x1 n1,
  efv_pt (close_pt_wrt_exp_rec n1 x1 pt1) [=] remove x1 (efv_pt pt1).
Proof.
pose proof efv_type_close_type_wrt_exp_rec_efv_ptt_close_ptt_wrt_exp_rec_efv_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_close_pt_wrt_exp_rec : lngen.
Hint Rewrite efv_pt_close_pt_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_kind_close_kind_wrt_exp_rec_mutual :
(forall k1 x1 n1,
  efv_kind (close_kind_wrt_exp_rec n1 x1 k1) [=] remove x1 (efv_kind k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_kind_close_kind_wrt_exp_rec :
forall k1 x1 n1,
  efv_kind (close_kind_wrt_exp_rec n1 x1 k1) [=] remove x1 (efv_kind k1).
Proof.
pose proof efv_kind_close_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_close_kind_wrt_exp_rec : lngen.
Hint Rewrite efv_kind_close_kind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_bind_close_bind_wrt_exp_rec_mutual :
(forall bind1 x1 n1,
  efv_bind (close_bind_wrt_exp_rec n1 x1 bind1) [=] remove x1 (efv_bind bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_bind_close_bind_wrt_exp_rec :
forall bind1 x1 n1,
  efv_bind (close_bind_wrt_exp_rec n1 x1 bind1) [=] remove x1 (efv_bind bind1).
Proof.
pose proof efv_bind_close_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_close_bind_wrt_exp_rec : lngen.
Hint Rewrite efv_bind_close_bind_wrt_exp_rec using solve [auto] : lngen.

(* end hide *)

Lemma efv_exp_close_exp_wrt_exp :
forall e1 x1,
  efv_exp (close_exp_wrt_exp x1 e1) [=] remove x1 (efv_exp e1).
Proof.
unfold close_exp_wrt_exp; default_simp.
Qed.

Hint Resolve efv_exp_close_exp_wrt_exp : lngen.
Hint Rewrite efv_exp_close_exp_wrt_exp using solve [auto] : lngen.

Lemma efv_peb_close_peb_wrt_exp :
forall peb1 x1,
  efv_peb (close_peb_wrt_exp x1 peb1) [=] remove x1 (efv_peb peb1).
Proof.
unfold close_peb_wrt_exp; default_simp.
Qed.

Hint Resolve efv_peb_close_peb_wrt_exp : lngen.
Hint Rewrite efv_peb_close_peb_wrt_exp using solve [auto] : lngen.

Lemma efv_pe_close_pe_wrt_exp :
forall pe1 x1,
  efv_pe (close_pe_wrt_exp x1 pe1) [=] remove x1 (efv_pe pe1).
Proof.
unfold close_pe_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pe_close_pe_wrt_exp : lngen.
Hint Rewrite efv_pe_close_pe_wrt_exp using solve [auto] : lngen.

Lemma efv_type_close_type_wrt_exp :
forall t1 x1,
  efv_type (close_type_wrt_exp x1 t1) [=] remove x1 (efv_type t1).
Proof.
unfold close_type_wrt_exp; default_simp.
Qed.

Hint Resolve efv_type_close_type_wrt_exp : lngen.
Hint Rewrite efv_type_close_type_wrt_exp using solve [auto] : lngen.

Lemma efv_ptt_close_ptt_wrt_exp :
forall ptt1 x1,
  efv_ptt (close_ptt_wrt_exp x1 ptt1) [=] remove x1 (efv_ptt ptt1).
Proof.
unfold close_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_ptt_close_ptt_wrt_exp : lngen.
Hint Rewrite efv_ptt_close_ptt_wrt_exp using solve [auto] : lngen.

Lemma efv_pt_close_pt_wrt_exp :
forall pt1 x1,
  efv_pt (close_pt_wrt_exp x1 pt1) [=] remove x1 (efv_pt pt1).
Proof.
unfold close_pt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pt_close_pt_wrt_exp : lngen.
Hint Rewrite efv_pt_close_pt_wrt_exp using solve [auto] : lngen.

Lemma efv_kind_close_kind_wrt_exp :
forall k1 x1,
  efv_kind (close_kind_wrt_exp x1 k1) [=] remove x1 (efv_kind k1).
Proof.
unfold close_kind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_kind_close_kind_wrt_exp : lngen.
Hint Rewrite efv_kind_close_kind_wrt_exp using solve [auto] : lngen.

Lemma efv_bind_close_bind_wrt_exp :
forall bind1 x1,
  efv_bind (close_bind_wrt_exp x1 bind1) [=] remove x1 (efv_bind bind1).
Proof.
unfold close_bind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_bind_close_bind_wrt_exp : lngen.
Hint Rewrite efv_bind_close_bind_wrt_exp using solve [auto] : lngen.

(* begin hide *)

Lemma efv_exp_open_exp_wrt_exp_rec_lower_efv_peb_open_peb_wrt_exp_rec_lower_efv_pe_open_pe_wrt_exp_rec_lower_mutual :
(forall e1 e2 n1,
  efv_exp e1 [<=] efv_exp (open_exp_wrt_exp_rec n1 e2 e1)) /\
(forall peb1 e1 n1,
  efv_peb peb1 [<=] efv_peb (open_peb_wrt_exp_rec n1 e1 peb1)) /\
(forall pe1 e1 n1,
  efv_pe pe1 [<=] efv_pe (open_pe_wrt_exp_rec n1 e1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_exp_open_exp_wrt_exp_rec_lower :
forall e1 e2 n1,
  efv_exp e1 [<=] efv_exp (open_exp_wrt_exp_rec n1 e2 e1).
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_lower_efv_peb_open_peb_wrt_exp_rec_lower_efv_pe_open_pe_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_open_exp_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_peb_open_peb_wrt_exp_rec_lower :
forall peb1 e1 n1,
  efv_peb peb1 [<=] efv_peb (open_peb_wrt_exp_rec n1 e1 peb1).
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_lower_efv_peb_open_peb_wrt_exp_rec_lower_efv_pe_open_pe_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_open_peb_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pe_open_pe_wrt_exp_rec_lower :
forall pe1 e1 n1,
  efv_pe pe1 [<=] efv_pe (open_pe_wrt_exp_rec n1 e1 pe1).
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_lower_efv_peb_open_peb_wrt_exp_rec_lower_efv_pe_open_pe_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_open_pe_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_type_open_type_wrt_exp_rec_lower_efv_ptt_open_ptt_wrt_exp_rec_lower_efv_pt_open_pt_wrt_exp_rec_lower_mutual :
(forall t1 e1 n1,
  efv_type t1 [<=] efv_type (open_type_wrt_exp_rec n1 e1 t1)) /\
(forall ptt1 e1 n1,
  efv_ptt ptt1 [<=] efv_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1)) /\
(forall pt1 e1 n1,
  efv_pt pt1 [<=] efv_pt (open_pt_wrt_exp_rec n1 e1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_type_open_type_wrt_exp_rec_lower :
forall t1 e1 n1,
  efv_type t1 [<=] efv_type (open_type_wrt_exp_rec n1 e1 t1).
Proof.
pose proof efv_type_open_type_wrt_exp_rec_lower_efv_ptt_open_ptt_wrt_exp_rec_lower_efv_pt_open_pt_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_open_type_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_ptt_open_ptt_wrt_exp_rec_lower :
forall ptt1 e1 n1,
  efv_ptt ptt1 [<=] efv_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1).
Proof.
pose proof efv_type_open_type_wrt_exp_rec_lower_efv_ptt_open_ptt_wrt_exp_rec_lower_efv_pt_open_pt_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_open_ptt_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pt_open_pt_wrt_exp_rec_lower :
forall pt1 e1 n1,
  efv_pt pt1 [<=] efv_pt (open_pt_wrt_exp_rec n1 e1 pt1).
Proof.
pose proof efv_type_open_type_wrt_exp_rec_lower_efv_ptt_open_ptt_wrt_exp_rec_lower_efv_pt_open_pt_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_open_pt_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_kind_open_kind_wrt_exp_rec_lower_mutual :
(forall k1 e1 n1,
  efv_kind k1 [<=] efv_kind (open_kind_wrt_exp_rec n1 e1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_kind_open_kind_wrt_exp_rec_lower :
forall k1 e1 n1,
  efv_kind k1 [<=] efv_kind (open_kind_wrt_exp_rec n1 e1 k1).
Proof.
pose proof efv_kind_open_kind_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_open_kind_wrt_exp_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_bind_open_bind_wrt_exp_rec_lower_mutual :
(forall bind1 e1 n1,
  efv_bind bind1 [<=] efv_bind (open_bind_wrt_exp_rec n1 e1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_bind_open_bind_wrt_exp_rec_lower :
forall bind1 e1 n1,
  efv_bind bind1 [<=] efv_bind (open_bind_wrt_exp_rec n1 e1 bind1).
Proof.
pose proof efv_bind_open_bind_wrt_exp_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_open_bind_wrt_exp_rec_lower : lngen.

(* end hide *)

Lemma efv_exp_open_exp_wrt_exp_lower :
forall e1 e2,
  efv_exp e1 [<=] efv_exp (open_exp_wrt_exp e1 e2).
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve efv_exp_open_exp_wrt_exp_lower : lngen.

Lemma efv_peb_open_peb_wrt_exp_lower :
forall peb1 e1,
  efv_peb peb1 [<=] efv_peb (open_peb_wrt_exp peb1 e1).
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve efv_peb_open_peb_wrt_exp_lower : lngen.

Lemma efv_pe_open_pe_wrt_exp_lower :
forall pe1 e1,
  efv_pe pe1 [<=] efv_pe (open_pe_wrt_exp pe1 e1).
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pe_open_pe_wrt_exp_lower : lngen.

Lemma efv_type_open_type_wrt_exp_lower :
forall t1 e1,
  efv_type t1 [<=] efv_type (open_type_wrt_exp t1 e1).
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve efv_type_open_type_wrt_exp_lower : lngen.

Lemma efv_ptt_open_ptt_wrt_exp_lower :
forall ptt1 e1,
  efv_ptt ptt1 [<=] efv_ptt (open_ptt_wrt_exp ptt1 e1).
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_ptt_open_ptt_wrt_exp_lower : lngen.

Lemma efv_pt_open_pt_wrt_exp_lower :
forall pt1 e1,
  efv_pt pt1 [<=] efv_pt (open_pt_wrt_exp pt1 e1).
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pt_open_pt_wrt_exp_lower : lngen.

Lemma efv_kind_open_kind_wrt_exp_lower :
forall k1 e1,
  efv_kind k1 [<=] efv_kind (open_kind_wrt_exp k1 e1).
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_kind_open_kind_wrt_exp_lower : lngen.

Lemma efv_bind_open_bind_wrt_exp_lower :
forall bind1 e1,
  efv_bind bind1 [<=] efv_bind (open_bind_wrt_exp bind1 e1).
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_bind_open_bind_wrt_exp_lower : lngen.

(* begin hide *)

Lemma efv_exp_open_exp_wrt_exp_rec_upper_efv_peb_open_peb_wrt_exp_rec_upper_efv_pe_open_pe_wrt_exp_rec_upper_mutual :
(forall e1 e2 n1,
  efv_exp (open_exp_wrt_exp_rec n1 e2 e1) [<=] efv_exp e2 `union` efv_exp e1) /\
(forall peb1 e1 n1,
  efv_peb (open_peb_wrt_exp_rec n1 e1 peb1) [<=] efv_exp e1 `union` efv_peb peb1) /\
(forall pe1 e1 n1,
  efv_pe (open_pe_wrt_exp_rec n1 e1 pe1) [<=] efv_exp e1 `union` efv_pe pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_exp_open_exp_wrt_exp_rec_upper :
forall e1 e2 n1,
  efv_exp (open_exp_wrt_exp_rec n1 e2 e1) [<=] efv_exp e2 `union` efv_exp e1.
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_upper_efv_peb_open_peb_wrt_exp_rec_upper_efv_pe_open_pe_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_open_exp_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_peb_open_peb_wrt_exp_rec_upper :
forall peb1 e1 n1,
  efv_peb (open_peb_wrt_exp_rec n1 e1 peb1) [<=] efv_exp e1 `union` efv_peb peb1.
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_upper_efv_peb_open_peb_wrt_exp_rec_upper_efv_pe_open_pe_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_open_peb_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pe_open_pe_wrt_exp_rec_upper :
forall pe1 e1 n1,
  efv_pe (open_pe_wrt_exp_rec n1 e1 pe1) [<=] efv_exp e1 `union` efv_pe pe1.
Proof.
pose proof efv_exp_open_exp_wrt_exp_rec_upper_efv_peb_open_peb_wrt_exp_rec_upper_efv_pe_open_pe_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_open_pe_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_type_open_type_wrt_exp_rec_upper_efv_ptt_open_ptt_wrt_exp_rec_upper_efv_pt_open_pt_wrt_exp_rec_upper_mutual :
(forall t1 e1 n1,
  efv_type (open_type_wrt_exp_rec n1 e1 t1) [<=] efv_exp e1 `union` efv_type t1) /\
(forall ptt1 e1 n1,
  efv_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1) [<=] efv_exp e1 `union` efv_ptt ptt1) /\
(forall pt1 e1 n1,
  efv_pt (open_pt_wrt_exp_rec n1 e1 pt1) [<=] efv_exp e1 `union` efv_pt pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_type_open_type_wrt_exp_rec_upper :
forall t1 e1 n1,
  efv_type (open_type_wrt_exp_rec n1 e1 t1) [<=] efv_exp e1 `union` efv_type t1.
Proof.
pose proof efv_type_open_type_wrt_exp_rec_upper_efv_ptt_open_ptt_wrt_exp_rec_upper_efv_pt_open_pt_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_open_type_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_ptt_open_ptt_wrt_exp_rec_upper :
forall ptt1 e1 n1,
  efv_ptt (open_ptt_wrt_exp_rec n1 e1 ptt1) [<=] efv_exp e1 `union` efv_ptt ptt1.
Proof.
pose proof efv_type_open_type_wrt_exp_rec_upper_efv_ptt_open_ptt_wrt_exp_rec_upper_efv_pt_open_pt_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_open_ptt_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_pt_open_pt_wrt_exp_rec_upper :
forall pt1 e1 n1,
  efv_pt (open_pt_wrt_exp_rec n1 e1 pt1) [<=] efv_exp e1 `union` efv_pt pt1.
Proof.
pose proof efv_type_open_type_wrt_exp_rec_upper_efv_ptt_open_ptt_wrt_exp_rec_upper_efv_pt_open_pt_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_open_pt_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_kind_open_kind_wrt_exp_rec_upper_mutual :
(forall k1 e1 n1,
  efv_kind (open_kind_wrt_exp_rec n1 e1 k1) [<=] efv_exp e1 `union` efv_kind k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_kind_open_kind_wrt_exp_rec_upper :
forall k1 e1 n1,
  efv_kind (open_kind_wrt_exp_rec n1 e1 k1) [<=] efv_exp e1 `union` efv_kind k1.
Proof.
pose proof efv_kind_open_kind_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_open_kind_wrt_exp_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma efv_bind_open_bind_wrt_exp_rec_upper_mutual :
(forall bind1 e1 n1,
  efv_bind (open_bind_wrt_exp_rec n1 e1 bind1) [<=] efv_exp e1 `union` efv_bind bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma efv_bind_open_bind_wrt_exp_rec_upper :
forall bind1 e1 n1,
  efv_bind (open_bind_wrt_exp_rec n1 e1 bind1) [<=] efv_exp e1 `union` efv_bind bind1.
Proof.
pose proof efv_bind_open_bind_wrt_exp_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_open_bind_wrt_exp_rec_upper : lngen.

(* end hide *)

Lemma efv_exp_open_exp_wrt_exp_upper :
forall e1 e2,
  efv_exp (open_exp_wrt_exp e1 e2) [<=] efv_exp e2 `union` efv_exp e1.
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve efv_exp_open_exp_wrt_exp_upper : lngen.

Lemma efv_peb_open_peb_wrt_exp_upper :
forall peb1 e1,
  efv_peb (open_peb_wrt_exp peb1 e1) [<=] efv_exp e1 `union` efv_peb peb1.
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve efv_peb_open_peb_wrt_exp_upper : lngen.

Lemma efv_pe_open_pe_wrt_exp_upper :
forall pe1 e1,
  efv_pe (open_pe_wrt_exp pe1 e1) [<=] efv_exp e1 `union` efv_pe pe1.
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pe_open_pe_wrt_exp_upper : lngen.

Lemma efv_type_open_type_wrt_exp_upper :
forall t1 e1,
  efv_type (open_type_wrt_exp t1 e1) [<=] efv_exp e1 `union` efv_type t1.
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve efv_type_open_type_wrt_exp_upper : lngen.

Lemma efv_ptt_open_ptt_wrt_exp_upper :
forall ptt1 e1,
  efv_ptt (open_ptt_wrt_exp ptt1 e1) [<=] efv_exp e1 `union` efv_ptt ptt1.
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_ptt_open_ptt_wrt_exp_upper : lngen.

Lemma efv_pt_open_pt_wrt_exp_upper :
forall pt1 e1,
  efv_pt (open_pt_wrt_exp pt1 e1) [<=] efv_exp e1 `union` efv_pt pt1.
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve efv_pt_open_pt_wrt_exp_upper : lngen.

Lemma efv_kind_open_kind_wrt_exp_upper :
forall k1 e1,
  efv_kind (open_kind_wrt_exp k1 e1) [<=] efv_exp e1 `union` efv_kind k1.
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_kind_open_kind_wrt_exp_upper : lngen.

Lemma efv_bind_open_bind_wrt_exp_upper :
forall bind1 e1,
  efv_bind (open_bind_wrt_exp bind1 e1) [<=] efv_exp e1 `union` efv_bind bind1.
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve efv_bind_open_bind_wrt_exp_upper : lngen.

(* begin hide *)

Lemma efv_exp_esubst_exp_fresh_efv_peb_esubst_peb_fresh_efv_pe_esubst_pe_fresh_mutual :
(forall e1 e2 x1,
  x1 `notin` efv_exp e1 ->
  efv_exp (esubst_exp e2 x1 e1) [=] efv_exp e1) /\
(forall peb1 e1 x1,
  x1 `notin` efv_peb peb1 ->
  efv_peb (esubst_peb e1 x1 peb1) [=] efv_peb peb1) /\
(forall pe1 e1 x1,
  x1 `notin` efv_pe pe1 ->
  efv_pe (esubst_pe e1 x1 pe1) [=] efv_pe pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_exp_esubst_exp_fresh :
forall e1 e2 x1,
  x1 `notin` efv_exp e1 ->
  efv_exp (esubst_exp e2 x1 e1) [=] efv_exp e1.
Proof.
pose proof efv_exp_esubst_exp_fresh_efv_peb_esubst_peb_fresh_efv_pe_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_esubst_exp_fresh : lngen.
Hint Rewrite efv_exp_esubst_exp_fresh using solve [auto] : lngen.

Lemma efv_peb_esubst_peb_fresh :
forall peb1 e1 x1,
  x1 `notin` efv_peb peb1 ->
  efv_peb (esubst_peb e1 x1 peb1) [=] efv_peb peb1.
Proof.
pose proof efv_exp_esubst_exp_fresh_efv_peb_esubst_peb_fresh_efv_pe_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_esubst_peb_fresh : lngen.
Hint Rewrite efv_peb_esubst_peb_fresh using solve [auto] : lngen.

Lemma efv_pe_esubst_pe_fresh :
forall pe1 e1 x1,
  x1 `notin` efv_pe pe1 ->
  efv_pe (esubst_pe e1 x1 pe1) [=] efv_pe pe1.
Proof.
pose proof efv_exp_esubst_exp_fresh_efv_peb_esubst_peb_fresh_efv_pe_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_esubst_pe_fresh : lngen.
Hint Rewrite efv_pe_esubst_pe_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma efv_type_esubst_type_fresh_efv_ptt_esubst_ptt_fresh_efv_pt_esubst_pt_fresh_mutual :
(forall t1 e1 x1,
  x1 `notin` efv_type t1 ->
  efv_type (esubst_type e1 x1 t1) [=] efv_type t1) /\
(forall ptt1 e1 x1,
  x1 `notin` efv_ptt ptt1 ->
  efv_ptt (esubst_ptt e1 x1 ptt1) [=] efv_ptt ptt1) /\
(forall pt1 e1 x1,
  x1 `notin` efv_pt pt1 ->
  efv_pt (esubst_pt e1 x1 pt1) [=] efv_pt pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_type_esubst_type_fresh :
forall t1 e1 x1,
  x1 `notin` efv_type t1 ->
  efv_type (esubst_type e1 x1 t1) [=] efv_type t1.
Proof.
pose proof efv_type_esubst_type_fresh_efv_ptt_esubst_ptt_fresh_efv_pt_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_esubst_type_fresh : lngen.
Hint Rewrite efv_type_esubst_type_fresh using solve [auto] : lngen.

Lemma efv_ptt_esubst_ptt_fresh :
forall ptt1 e1 x1,
  x1 `notin` efv_ptt ptt1 ->
  efv_ptt (esubst_ptt e1 x1 ptt1) [=] efv_ptt ptt1.
Proof.
pose proof efv_type_esubst_type_fresh_efv_ptt_esubst_ptt_fresh_efv_pt_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_esubst_ptt_fresh : lngen.
Hint Rewrite efv_ptt_esubst_ptt_fresh using solve [auto] : lngen.

Lemma efv_pt_esubst_pt_fresh :
forall pt1 e1 x1,
  x1 `notin` efv_pt pt1 ->
  efv_pt (esubst_pt e1 x1 pt1) [=] efv_pt pt1.
Proof.
pose proof efv_type_esubst_type_fresh_efv_ptt_esubst_ptt_fresh_efv_pt_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_esubst_pt_fresh : lngen.
Hint Rewrite efv_pt_esubst_pt_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma efv_kind_esubst_kind_fresh_mutual :
(forall k1 e1 x1,
  x1 `notin` efv_kind k1 ->
  efv_kind (esubst_kind e1 x1 k1) [=] efv_kind k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_kind_esubst_kind_fresh :
forall k1 e1 x1,
  x1 `notin` efv_kind k1 ->
  efv_kind (esubst_kind e1 x1 k1) [=] efv_kind k1.
Proof.
pose proof efv_kind_esubst_kind_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_esubst_kind_fresh : lngen.
Hint Rewrite efv_kind_esubst_kind_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma efv_bind_esubst_bind_fresh_mutual :
(forall bind1 e1 x1,
  x1 `notin` efv_bind bind1 ->
  efv_bind (esubst_bind e1 x1 bind1) [=] efv_bind bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_bind_esubst_bind_fresh :
forall bind1 e1 x1,
  x1 `notin` efv_bind bind1 ->
  efv_bind (esubst_bind e1 x1 bind1) [=] efv_bind bind1.
Proof.
pose proof efv_bind_esubst_bind_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_esubst_bind_fresh : lngen.
Hint Rewrite efv_bind_esubst_bind_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma efv_exp_esubst_exp_lower_efv_peb_esubst_peb_lower_efv_pe_esubst_pe_lower_mutual :
(forall e1 e2 x1,
  remove x1 (efv_exp e1) [<=] efv_exp (esubst_exp e2 x1 e1)) /\
(forall peb1 e1 x1,
  remove x1 (efv_peb peb1) [<=] efv_peb (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 x1,
  remove x1 (efv_pe pe1) [<=] efv_pe (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_exp_esubst_exp_lower :
forall e1 e2 x1,
  remove x1 (efv_exp e1) [<=] efv_exp (esubst_exp e2 x1 e1).
Proof.
pose proof efv_exp_esubst_exp_lower_efv_peb_esubst_peb_lower_efv_pe_esubst_pe_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_esubst_exp_lower : lngen.

Lemma efv_peb_esubst_peb_lower :
forall peb1 e1 x1,
  remove x1 (efv_peb peb1) [<=] efv_peb (esubst_peb e1 x1 peb1).
Proof.
pose proof efv_exp_esubst_exp_lower_efv_peb_esubst_peb_lower_efv_pe_esubst_pe_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_esubst_peb_lower : lngen.

Lemma efv_pe_esubst_pe_lower :
forall pe1 e1 x1,
  remove x1 (efv_pe pe1) [<=] efv_pe (esubst_pe e1 x1 pe1).
Proof.
pose proof efv_exp_esubst_exp_lower_efv_peb_esubst_peb_lower_efv_pe_esubst_pe_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_esubst_pe_lower : lngen.

(* begin hide *)

Lemma efv_type_esubst_type_lower_efv_ptt_esubst_ptt_lower_efv_pt_esubst_pt_lower_mutual :
(forall t1 e1 x1,
  remove x1 (efv_type t1) [<=] efv_type (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 x1,
  remove x1 (efv_ptt ptt1) [<=] efv_ptt (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 x1,
  remove x1 (efv_pt pt1) [<=] efv_pt (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_type_esubst_type_lower :
forall t1 e1 x1,
  remove x1 (efv_type t1) [<=] efv_type (esubst_type e1 x1 t1).
Proof.
pose proof efv_type_esubst_type_lower_efv_ptt_esubst_ptt_lower_efv_pt_esubst_pt_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_esubst_type_lower : lngen.

Lemma efv_ptt_esubst_ptt_lower :
forall ptt1 e1 x1,
  remove x1 (efv_ptt ptt1) [<=] efv_ptt (esubst_ptt e1 x1 ptt1).
Proof.
pose proof efv_type_esubst_type_lower_efv_ptt_esubst_ptt_lower_efv_pt_esubst_pt_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_esubst_ptt_lower : lngen.

Lemma efv_pt_esubst_pt_lower :
forall pt1 e1 x1,
  remove x1 (efv_pt pt1) [<=] efv_pt (esubst_pt e1 x1 pt1).
Proof.
pose proof efv_type_esubst_type_lower_efv_ptt_esubst_ptt_lower_efv_pt_esubst_pt_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_esubst_pt_lower : lngen.

(* begin hide *)

Lemma efv_kind_esubst_kind_lower_mutual :
(forall k1 e1 x1,
  remove x1 (efv_kind k1) [<=] efv_kind (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_kind_esubst_kind_lower :
forall k1 e1 x1,
  remove x1 (efv_kind k1) [<=] efv_kind (esubst_kind e1 x1 k1).
Proof.
pose proof efv_kind_esubst_kind_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_esubst_kind_lower : lngen.

(* begin hide *)

Lemma efv_bind_esubst_bind_lower_mutual :
(forall bind1 e1 x1,
  remove x1 (efv_bind bind1) [<=] efv_bind (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_bind_esubst_bind_lower :
forall bind1 e1 x1,
  remove x1 (efv_bind bind1) [<=] efv_bind (esubst_bind e1 x1 bind1).
Proof.
pose proof efv_bind_esubst_bind_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_esubst_bind_lower : lngen.

(* begin hide *)

Lemma efv_exp_esubst_exp_notin_efv_peb_esubst_peb_notin_efv_pe_esubst_pe_notin_mutual :
(forall e1 e2 x1 x2,
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_exp e2 ->
  x2 `notin` efv_exp (esubst_exp e2 x1 e1)) /\
(forall peb1 e1 x1 x2,
  x2 `notin` efv_peb peb1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_peb (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 x1 x2,
  x2 `notin` efv_pe pe1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_pe (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_exp_esubst_exp_notin :
forall e1 e2 x1 x2,
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_exp e2 ->
  x2 `notin` efv_exp (esubst_exp e2 x1 e1).
Proof.
pose proof efv_exp_esubst_exp_notin_efv_peb_esubst_peb_notin_efv_pe_esubst_pe_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_esubst_exp_notin : lngen.

Lemma efv_peb_esubst_peb_notin :
forall peb1 e1 x1 x2,
  x2 `notin` efv_peb peb1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_peb (esubst_peb e1 x1 peb1).
Proof.
pose proof efv_exp_esubst_exp_notin_efv_peb_esubst_peb_notin_efv_pe_esubst_pe_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_esubst_peb_notin : lngen.

Lemma efv_pe_esubst_pe_notin :
forall pe1 e1 x1 x2,
  x2 `notin` efv_pe pe1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_pe (esubst_pe e1 x1 pe1).
Proof.
pose proof efv_exp_esubst_exp_notin_efv_peb_esubst_peb_notin_efv_pe_esubst_pe_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_esubst_pe_notin : lngen.

(* begin hide *)

Lemma efv_type_esubst_type_notin_efv_ptt_esubst_ptt_notin_efv_pt_esubst_pt_notin_mutual :
(forall t1 e1 x1 x2,
  x2 `notin` efv_type t1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_type (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 x1 x2,
  x2 `notin` efv_ptt ptt1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_ptt (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 x1 x2,
  x2 `notin` efv_pt pt1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_pt (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_type_esubst_type_notin :
forall t1 e1 x1 x2,
  x2 `notin` efv_type t1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_type (esubst_type e1 x1 t1).
Proof.
pose proof efv_type_esubst_type_notin_efv_ptt_esubst_ptt_notin_efv_pt_esubst_pt_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_esubst_type_notin : lngen.

Lemma efv_ptt_esubst_ptt_notin :
forall ptt1 e1 x1 x2,
  x2 `notin` efv_ptt ptt1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_ptt (esubst_ptt e1 x1 ptt1).
Proof.
pose proof efv_type_esubst_type_notin_efv_ptt_esubst_ptt_notin_efv_pt_esubst_pt_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_esubst_ptt_notin : lngen.

Lemma efv_pt_esubst_pt_notin :
forall pt1 e1 x1 x2,
  x2 `notin` efv_pt pt1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_pt (esubst_pt e1 x1 pt1).
Proof.
pose proof efv_type_esubst_type_notin_efv_ptt_esubst_ptt_notin_efv_pt_esubst_pt_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_esubst_pt_notin : lngen.

(* begin hide *)

Lemma efv_kind_esubst_kind_notin_mutual :
(forall k1 e1 x1 x2,
  x2 `notin` efv_kind k1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_kind (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_kind_esubst_kind_notin :
forall k1 e1 x1 x2,
  x2 `notin` efv_kind k1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_kind (esubst_kind e1 x1 k1).
Proof.
pose proof efv_kind_esubst_kind_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_esubst_kind_notin : lngen.

(* begin hide *)

Lemma efv_bind_esubst_bind_notin_mutual :
(forall bind1 e1 x1 x2,
  x2 `notin` efv_bind bind1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_bind (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_bind_esubst_bind_notin :
forall bind1 e1 x1 x2,
  x2 `notin` efv_bind bind1 ->
  x2 `notin` efv_exp e1 ->
  x2 `notin` efv_bind (esubst_bind e1 x1 bind1).
Proof.
pose proof efv_bind_esubst_bind_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_esubst_bind_notin : lngen.

(* begin hide *)

Lemma efv_exp_esubst_exp_upper_efv_peb_esubst_peb_upper_efv_pe_esubst_pe_upper_mutual :
(forall e1 e2 x1,
  efv_exp (esubst_exp e2 x1 e1) [<=] efv_exp e2 `union` remove x1 (efv_exp e1)) /\
(forall peb1 e1 x1,
  efv_peb (esubst_peb e1 x1 peb1) [<=] efv_exp e1 `union` remove x1 (efv_peb peb1)) /\
(forall pe1 e1 x1,
  efv_pe (esubst_pe e1 x1 pe1) [<=] efv_exp e1 `union` remove x1 (efv_pe pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_exp_esubst_exp_upper :
forall e1 e2 x1,
  efv_exp (esubst_exp e2 x1 e1) [<=] efv_exp e2 `union` remove x1 (efv_exp e1).
Proof.
pose proof efv_exp_esubst_exp_upper_efv_peb_esubst_peb_upper_efv_pe_esubst_pe_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_exp_esubst_exp_upper : lngen.

Lemma efv_peb_esubst_peb_upper :
forall peb1 e1 x1,
  efv_peb (esubst_peb e1 x1 peb1) [<=] efv_exp e1 `union` remove x1 (efv_peb peb1).
Proof.
pose proof efv_exp_esubst_exp_upper_efv_peb_esubst_peb_upper_efv_pe_esubst_pe_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_peb_esubst_peb_upper : lngen.

Lemma efv_pe_esubst_pe_upper :
forall pe1 e1 x1,
  efv_pe (esubst_pe e1 x1 pe1) [<=] efv_exp e1 `union` remove x1 (efv_pe pe1).
Proof.
pose proof efv_exp_esubst_exp_upper_efv_peb_esubst_peb_upper_efv_pe_esubst_pe_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pe_esubst_pe_upper : lngen.

(* begin hide *)

Lemma efv_type_esubst_type_upper_efv_ptt_esubst_ptt_upper_efv_pt_esubst_pt_upper_mutual :
(forall t1 e1 x1,
  efv_type (esubst_type e1 x1 t1) [<=] efv_exp e1 `union` remove x1 (efv_type t1)) /\
(forall ptt1 e1 x1,
  efv_ptt (esubst_ptt e1 x1 ptt1) [<=] efv_exp e1 `union` remove x1 (efv_ptt ptt1)) /\
(forall pt1 e1 x1,
  efv_pt (esubst_pt e1 x1 pt1) [<=] efv_exp e1 `union` remove x1 (efv_pt pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_type_esubst_type_upper :
forall t1 e1 x1,
  efv_type (esubst_type e1 x1 t1) [<=] efv_exp e1 `union` remove x1 (efv_type t1).
Proof.
pose proof efv_type_esubst_type_upper_efv_ptt_esubst_ptt_upper_efv_pt_esubst_pt_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_type_esubst_type_upper : lngen.

Lemma efv_ptt_esubst_ptt_upper :
forall ptt1 e1 x1,
  efv_ptt (esubst_ptt e1 x1 ptt1) [<=] efv_exp e1 `union` remove x1 (efv_ptt ptt1).
Proof.
pose proof efv_type_esubst_type_upper_efv_ptt_esubst_ptt_upper_efv_pt_esubst_pt_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_ptt_esubst_ptt_upper : lngen.

Lemma efv_pt_esubst_pt_upper :
forall pt1 e1 x1,
  efv_pt (esubst_pt e1 x1 pt1) [<=] efv_exp e1 `union` remove x1 (efv_pt pt1).
Proof.
pose proof efv_type_esubst_type_upper_efv_ptt_esubst_ptt_upper_efv_pt_esubst_pt_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_pt_esubst_pt_upper : lngen.

(* begin hide *)

Lemma efv_kind_esubst_kind_upper_mutual :
(forall k1 e1 x1,
  efv_kind (esubst_kind e1 x1 k1) [<=] efv_exp e1 `union` remove x1 (efv_kind k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_kind_esubst_kind_upper :
forall k1 e1 x1,
  efv_kind (esubst_kind e1 x1 k1) [<=] efv_exp e1 `union` remove x1 (efv_kind k1).
Proof.
pose proof efv_kind_esubst_kind_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_kind_esubst_kind_upper : lngen.

(* begin hide *)

Lemma efv_bind_esubst_bind_upper_mutual :
(forall bind1 e1 x1,
  efv_bind (esubst_bind e1 x1 bind1) [<=] efv_exp e1 `union` remove x1 (efv_bind bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma efv_bind_esubst_bind_upper :
forall bind1 e1 x1,
  efv_bind (esubst_bind e1 x1 bind1) [<=] efv_exp e1 `union` remove x1 (efv_bind bind1).
Proof.
pose proof efv_bind_esubst_bind_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve efv_bind_esubst_bind_upper : lngen.


(* *********************************************************************** *)
(** * Theorems about [subst] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma esubst_exp_close_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_mutual :
(forall e2 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_exp e1 x1 (close_exp_wrt_exp_rec n1 x2 e2) = close_exp_wrt_exp_rec n1 x2 (esubst_exp e1 x1 e2)) /\
(forall peb1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_peb e1 x1 (close_peb_wrt_exp_rec n1 x2 peb1) = close_peb_wrt_exp_rec n1 x2 (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pe e1 x1 (close_pe_wrt_exp_rec n1 x2 pe1) = close_pe_wrt_exp_rec n1 x2 (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_close_exp_wrt_exp_rec :
forall e2 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_exp e1 x1 (close_exp_wrt_exp_rec n1 x2 e2) = close_exp_wrt_exp_rec n1 x2 (esubst_exp e1 x1 e2).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_close_exp_wrt_exp_rec : lngen.

Lemma esubst_peb_close_peb_wrt_exp_rec :
forall peb1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_peb e1 x1 (close_peb_wrt_exp_rec n1 x2 peb1) = close_peb_wrt_exp_rec n1 x2 (esubst_peb e1 x1 peb1).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_close_peb_wrt_exp_rec : lngen.

Lemma esubst_pe_close_pe_wrt_exp_rec :
forall pe1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pe e1 x1 (close_pe_wrt_exp_rec n1 x2 pe1) = close_pe_wrt_exp_rec n1 x2 (esubst_pe e1 x1 pe1).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_close_pe_wrt_exp_rec : lngen.

(* begin hide *)

Lemma esubst_type_close_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_mutual :
(forall t1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_type e1 x1 (close_type_wrt_exp_rec n1 x2 t1) = close_type_wrt_exp_rec n1 x2 (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_ptt e1 x1 (close_ptt_wrt_exp_rec n1 x2 ptt1) = close_ptt_wrt_exp_rec n1 x2 (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pt e1 x1 (close_pt_wrt_exp_rec n1 x2 pt1) = close_pt_wrt_exp_rec n1 x2 (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_close_type_wrt_exp_rec :
forall t1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_type e1 x1 (close_type_wrt_exp_rec n1 x2 t1) = close_type_wrt_exp_rec n1 x2 (esubst_type e1 x1 t1).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_close_type_wrt_exp_rec : lngen.

Lemma esubst_ptt_close_ptt_wrt_exp_rec :
forall ptt1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_ptt e1 x1 (close_ptt_wrt_exp_rec n1 x2 ptt1) = close_ptt_wrt_exp_rec n1 x2 (esubst_ptt e1 x1 ptt1).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_close_ptt_wrt_exp_rec : lngen.

Lemma esubst_pt_close_pt_wrt_exp_rec :
forall pt1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pt e1 x1 (close_pt_wrt_exp_rec n1 x2 pt1) = close_pt_wrt_exp_rec n1 x2 (esubst_pt e1 x1 pt1).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_close_pt_wrt_exp_rec : lngen.

(* begin hide *)

Lemma esubst_kind_close_kind_wrt_exp_rec_mutual :
(forall k1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_kind e1 x1 (close_kind_wrt_exp_rec n1 x2 k1) = close_kind_wrt_exp_rec n1 x2 (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_close_kind_wrt_exp_rec :
forall k1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_kind e1 x1 (close_kind_wrt_exp_rec n1 x2 k1) = close_kind_wrt_exp_rec n1 x2 (esubst_kind e1 x1 k1).
Proof.
pose proof esubst_kind_close_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_close_kind_wrt_exp_rec : lngen.

(* begin hide *)

Lemma esubst_bind_close_bind_wrt_exp_rec_mutual :
(forall bind1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_bind e1 x1 (close_bind_wrt_exp_rec n1 x2 bind1) = close_bind_wrt_exp_rec n1 x2 (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_close_bind_wrt_exp_rec :
forall bind1 e1 x1 x2 n1,
  degree_exp_wrt_exp n1 e1 ->
  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_bind e1 x1 (close_bind_wrt_exp_rec n1 x2 bind1) = close_bind_wrt_exp_rec n1 x2 (esubst_bind e1 x1 bind1).
Proof.
pose proof esubst_bind_close_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_close_bind_wrt_exp_rec : lngen.

Lemma esubst_exp_close_exp_wrt_exp :
forall e2 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_exp e1 x1 (close_exp_wrt_exp x2 e2) = close_exp_wrt_exp x2 (esubst_exp e1 x1 e2).
Proof.
unfold close_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_close_exp_wrt_exp : lngen.

Lemma esubst_peb_close_peb_wrt_exp :
forall peb1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_peb e1 x1 (close_peb_wrt_exp x2 peb1) = close_peb_wrt_exp x2 (esubst_peb e1 x1 peb1).
Proof.
unfold close_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_close_peb_wrt_exp : lngen.

Lemma esubst_pe_close_pe_wrt_exp :
forall pe1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pe e1 x1 (close_pe_wrt_exp x2 pe1) = close_pe_wrt_exp x2 (esubst_pe e1 x1 pe1).
Proof.
unfold close_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_close_pe_wrt_exp : lngen.

Lemma esubst_type_close_type_wrt_exp :
forall t1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_type e1 x1 (close_type_wrt_exp x2 t1) = close_type_wrt_exp x2 (esubst_type e1 x1 t1).
Proof.
unfold close_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_close_type_wrt_exp : lngen.

Lemma esubst_ptt_close_ptt_wrt_exp :
forall ptt1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_ptt e1 x1 (close_ptt_wrt_exp x2 ptt1) = close_ptt_wrt_exp x2 (esubst_ptt e1 x1 ptt1).
Proof.
unfold close_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_close_ptt_wrt_exp : lngen.

Lemma esubst_pt_close_pt_wrt_exp :
forall pt1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_pt e1 x1 (close_pt_wrt_exp x2 pt1) = close_pt_wrt_exp x2 (esubst_pt e1 x1 pt1).
Proof.
unfold close_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_close_pt_wrt_exp : lngen.

Lemma esubst_kind_close_kind_wrt_exp :
forall k1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_kind e1 x1 (close_kind_wrt_exp x2 k1) = close_kind_wrt_exp x2 (esubst_kind e1 x1 k1).
Proof.
unfold close_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_close_kind_wrt_exp : lngen.

Lemma esubst_bind_close_bind_wrt_exp :
forall bind1 e1 x1 x2,
  lc_exp e1 ->  x1 <> x2 ->
  x2 `notin` efv_exp e1 ->
  esubst_bind e1 x1 (close_bind_wrt_exp x2 bind1) = close_bind_wrt_exp x2 (esubst_bind e1 x1 bind1).
Proof.
unfold close_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_close_bind_wrt_exp : lngen.

(* begin hide *)

Lemma esubst_exp_degree_exp_wrt_exp_esubst_peb_degree_peb_wrt_exp_esubst_pe_degree_pe_wrt_exp_mutual :
(forall e1 e2 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp n1 e2 ->
  degree_exp_wrt_exp n1 (esubst_exp e2 x1 e1)) /\
(forall peb1 e1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_peb_wrt_exp n1 (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pe_wrt_exp n1 (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_degree_exp_wrt_exp :
forall e1 e2 x1 n1,
  degree_exp_wrt_exp n1 e1 ->
  degree_exp_wrt_exp n1 e2 ->
  degree_exp_wrt_exp n1 (esubst_exp e2 x1 e1).
Proof.
pose proof esubst_exp_degree_exp_wrt_exp_esubst_peb_degree_peb_wrt_exp_esubst_pe_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_degree_exp_wrt_exp : lngen.

Lemma esubst_peb_degree_peb_wrt_exp :
forall peb1 e1 x1 n1,
  degree_peb_wrt_exp n1 peb1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_peb_wrt_exp n1 (esubst_peb e1 x1 peb1).
Proof.
pose proof esubst_exp_degree_exp_wrt_exp_esubst_peb_degree_peb_wrt_exp_esubst_pe_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_degree_peb_wrt_exp : lngen.

Lemma esubst_pe_degree_pe_wrt_exp :
forall pe1 e1 x1 n1,
  degree_pe_wrt_exp n1 pe1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pe_wrt_exp n1 (esubst_pe e1 x1 pe1).
Proof.
pose proof esubst_exp_degree_exp_wrt_exp_esubst_peb_degree_peb_wrt_exp_esubst_pe_degree_pe_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_degree_pe_wrt_exp : lngen.

(* begin hide *)

Lemma esubst_type_degree_type_wrt_exp_esubst_ptt_degree_ptt_wrt_exp_esubst_pt_degree_pt_wrt_exp_mutual :
(forall t1 e1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_type_wrt_exp n1 (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_ptt_wrt_exp n1 (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pt_wrt_exp n1 (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_degree_type_wrt_exp :
forall t1 e1 x1 n1,
  degree_type_wrt_exp n1 t1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_type_wrt_exp n1 (esubst_type e1 x1 t1).
Proof.
pose proof esubst_type_degree_type_wrt_exp_esubst_ptt_degree_ptt_wrt_exp_esubst_pt_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_degree_type_wrt_exp : lngen.

Lemma esubst_ptt_degree_ptt_wrt_exp :
forall ptt1 e1 x1 n1,
  degree_ptt_wrt_exp n1 ptt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_ptt_wrt_exp n1 (esubst_ptt e1 x1 ptt1).
Proof.
pose proof esubst_type_degree_type_wrt_exp_esubst_ptt_degree_ptt_wrt_exp_esubst_pt_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_degree_ptt_wrt_exp : lngen.

Lemma esubst_pt_degree_pt_wrt_exp :
forall pt1 e1 x1 n1,
  degree_pt_wrt_exp n1 pt1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_pt_wrt_exp n1 (esubst_pt e1 x1 pt1).
Proof.
pose proof esubst_type_degree_type_wrt_exp_esubst_ptt_degree_ptt_wrt_exp_esubst_pt_degree_pt_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_degree_pt_wrt_exp : lngen.

(* begin hide *)

Lemma esubst_kind_degree_kind_wrt_exp_mutual :
(forall k1 e1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_kind_wrt_exp n1 (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_degree_kind_wrt_exp :
forall k1 e1 x1 n1,
  degree_kind_wrt_exp n1 k1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_kind_wrt_exp n1 (esubst_kind e1 x1 k1).
Proof.
pose proof esubst_kind_degree_kind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_degree_kind_wrt_exp : lngen.

(* begin hide *)

Lemma esubst_bind_degree_bind_wrt_exp_mutual :
(forall bind1 e1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_bind_wrt_exp n1 (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_degree_bind_wrt_exp :
forall bind1 e1 x1 n1,
  degree_bind_wrt_exp n1 bind1 ->
  degree_exp_wrt_exp n1 e1 ->
  degree_bind_wrt_exp n1 (esubst_bind e1 x1 bind1).
Proof.
pose proof esubst_bind_degree_bind_wrt_exp_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_degree_bind_wrt_exp : lngen.

(* begin hide *)

Lemma esubst_exp_fresh_eq_esubst_peb_fresh_eq_esubst_pe_fresh_eq_mutual :
(forall e2 e1 x1,
  x1 `notin` efv_exp e2 ->
  esubst_exp e1 x1 e2 = e2) /\
(forall peb1 e1 x1,
  x1 `notin` efv_peb peb1 ->
  esubst_peb e1 x1 peb1 = peb1) /\
(forall pe1 e1 x1,
  x1 `notin` efv_pe pe1 ->
  esubst_pe e1 x1 pe1 = pe1).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_fresh_eq :
forall e2 e1 x1,
  x1 `notin` efv_exp e2 ->
  esubst_exp e1 x1 e2 = e2.
Proof.
pose proof esubst_exp_fresh_eq_esubst_peb_fresh_eq_esubst_pe_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_fresh_eq : lngen.
Hint Rewrite esubst_exp_fresh_eq using solve [auto] : lngen.

Lemma esubst_peb_fresh_eq :
forall peb1 e1 x1,
  x1 `notin` efv_peb peb1 ->
  esubst_peb e1 x1 peb1 = peb1.
Proof.
pose proof esubst_exp_fresh_eq_esubst_peb_fresh_eq_esubst_pe_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_fresh_eq : lngen.
Hint Rewrite esubst_peb_fresh_eq using solve [auto] : lngen.

Lemma esubst_pe_fresh_eq :
forall pe1 e1 x1,
  x1 `notin` efv_pe pe1 ->
  esubst_pe e1 x1 pe1 = pe1.
Proof.
pose proof esubst_exp_fresh_eq_esubst_peb_fresh_eq_esubst_pe_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_fresh_eq : lngen.
Hint Rewrite esubst_pe_fresh_eq using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_type_fresh_eq_esubst_ptt_fresh_eq_esubst_pt_fresh_eq_mutual :
(forall t1 e1 x1,
  x1 `notin` efv_type t1 ->
  esubst_type e1 x1 t1 = t1) /\
(forall ptt1 e1 x1,
  x1 `notin` efv_ptt ptt1 ->
  esubst_ptt e1 x1 ptt1 = ptt1) /\
(forall pt1 e1 x1,
  x1 `notin` efv_pt pt1 ->
  esubst_pt e1 x1 pt1 = pt1).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_fresh_eq :
forall t1 e1 x1,
  x1 `notin` efv_type t1 ->
  esubst_type e1 x1 t1 = t1.
Proof.
pose proof esubst_type_fresh_eq_esubst_ptt_fresh_eq_esubst_pt_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_fresh_eq : lngen.
Hint Rewrite esubst_type_fresh_eq using solve [auto] : lngen.

Lemma esubst_ptt_fresh_eq :
forall ptt1 e1 x1,
  x1 `notin` efv_ptt ptt1 ->
  esubst_ptt e1 x1 ptt1 = ptt1.
Proof.
pose proof esubst_type_fresh_eq_esubst_ptt_fresh_eq_esubst_pt_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_fresh_eq : lngen.
Hint Rewrite esubst_ptt_fresh_eq using solve [auto] : lngen.

Lemma esubst_pt_fresh_eq :
forall pt1 e1 x1,
  x1 `notin` efv_pt pt1 ->
  esubst_pt e1 x1 pt1 = pt1.
Proof.
pose proof esubst_type_fresh_eq_esubst_ptt_fresh_eq_esubst_pt_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_fresh_eq : lngen.
Hint Rewrite esubst_pt_fresh_eq using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_kind_fresh_eq_mutual :
(forall k1 e1 x1,
  x1 `notin` efv_kind k1 ->
  esubst_kind e1 x1 k1 = k1).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_fresh_eq :
forall k1 e1 x1,
  x1 `notin` efv_kind k1 ->
  esubst_kind e1 x1 k1 = k1.
Proof.
pose proof esubst_kind_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_fresh_eq : lngen.
Hint Rewrite esubst_kind_fresh_eq using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_bind_fresh_eq_mutual :
(forall bind1 e1 x1,
  x1 `notin` efv_bind bind1 ->
  esubst_bind e1 x1 bind1 = bind1).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_fresh_eq :
forall bind1 e1 x1,
  x1 `notin` efv_bind bind1 ->
  esubst_bind e1 x1 bind1 = bind1.
Proof.
pose proof esubst_bind_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_fresh_eq : lngen.
Hint Rewrite esubst_bind_fresh_eq using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_exp_fresh_same_esubst_peb_fresh_same_esubst_pe_fresh_same_mutual :
(forall e2 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_exp (esubst_exp e1 x1 e2)) /\
(forall peb1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_peb (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pe (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_fresh_same :
forall e2 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_exp (esubst_exp e1 x1 e2).
Proof.
pose proof esubst_exp_fresh_same_esubst_peb_fresh_same_esubst_pe_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_fresh_same : lngen.
Hint Rewrite esubst_exp_fresh_same using solve [auto] : lngen.

Lemma esubst_peb_fresh_same :
forall peb1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_peb (esubst_peb e1 x1 peb1).
Proof.
pose proof esubst_exp_fresh_same_esubst_peb_fresh_same_esubst_pe_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_fresh_same : lngen.
Hint Rewrite esubst_peb_fresh_same using solve [auto] : lngen.

Lemma esubst_pe_fresh_same :
forall pe1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pe (esubst_pe e1 x1 pe1).
Proof.
pose proof esubst_exp_fresh_same_esubst_peb_fresh_same_esubst_pe_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_fresh_same : lngen.
Hint Rewrite esubst_pe_fresh_same using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_type_fresh_same_esubst_ptt_fresh_same_esubst_pt_fresh_same_mutual :
(forall t1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_type (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_ptt (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pt (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_fresh_same :
forall t1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_type (esubst_type e1 x1 t1).
Proof.
pose proof esubst_type_fresh_same_esubst_ptt_fresh_same_esubst_pt_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_fresh_same : lngen.
Hint Rewrite esubst_type_fresh_same using solve [auto] : lngen.

Lemma esubst_ptt_fresh_same :
forall ptt1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_ptt (esubst_ptt e1 x1 ptt1).
Proof.
pose proof esubst_type_fresh_same_esubst_ptt_fresh_same_esubst_pt_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_fresh_same : lngen.
Hint Rewrite esubst_ptt_fresh_same using solve [auto] : lngen.

Lemma esubst_pt_fresh_same :
forall pt1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pt (esubst_pt e1 x1 pt1).
Proof.
pose proof esubst_type_fresh_same_esubst_ptt_fresh_same_esubst_pt_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_fresh_same : lngen.
Hint Rewrite esubst_pt_fresh_same using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_kind_fresh_same_mutual :
(forall k1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_kind (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_fresh_same :
forall k1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_kind (esubst_kind e1 x1 k1).
Proof.
pose proof esubst_kind_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_fresh_same : lngen.
Hint Rewrite esubst_kind_fresh_same using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_bind_fresh_same_mutual :
(forall bind1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_bind (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_fresh_same :
forall bind1 e1 x1,
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_bind (esubst_bind e1 x1 bind1).
Proof.
pose proof esubst_bind_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_fresh_same : lngen.
Hint Rewrite esubst_bind_fresh_same using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_exp_fresh_esubst_peb_fresh_esubst_pe_fresh_mutual :
(forall e2 e1 x1 x2,
  x1 `notin` efv_exp e2 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_exp (esubst_exp e1 x2 e2)) /\
(forall peb1 e1 x1 x2,
  x1 `notin` efv_peb peb1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_peb (esubst_peb e1 x2 peb1)) /\
(forall pe1 e1 x1 x2,
  x1 `notin` efv_pe pe1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pe (esubst_pe e1 x2 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_fresh :
forall e2 e1 x1 x2,
  x1 `notin` efv_exp e2 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_exp (esubst_exp e1 x2 e2).
Proof.
pose proof esubst_exp_fresh_esubst_peb_fresh_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_fresh : lngen.
Hint Rewrite esubst_exp_fresh using solve [auto] : lngen.

Lemma esubst_peb_fresh :
forall peb1 e1 x1 x2,
  x1 `notin` efv_peb peb1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_peb (esubst_peb e1 x2 peb1).
Proof.
pose proof esubst_exp_fresh_esubst_peb_fresh_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_fresh : lngen.
Hint Rewrite esubst_peb_fresh using solve [auto] : lngen.

Lemma esubst_pe_fresh :
forall pe1 e1 x1 x2,
  x1 `notin` efv_pe pe1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pe (esubst_pe e1 x2 pe1).
Proof.
pose proof esubst_exp_fresh_esubst_peb_fresh_esubst_pe_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_fresh : lngen.
Hint Rewrite esubst_pe_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_type_fresh_esubst_ptt_fresh_esubst_pt_fresh_mutual :
(forall t1 e1 x1 x2,
  x1 `notin` efv_type t1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_type (esubst_type e1 x2 t1)) /\
(forall ptt1 e1 x1 x2,
  x1 `notin` efv_ptt ptt1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_ptt (esubst_ptt e1 x2 ptt1)) /\
(forall pt1 e1 x1 x2,
  x1 `notin` efv_pt pt1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pt (esubst_pt e1 x2 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_fresh :
forall t1 e1 x1 x2,
  x1 `notin` efv_type t1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_type (esubst_type e1 x2 t1).
Proof.
pose proof esubst_type_fresh_esubst_ptt_fresh_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_fresh : lngen.
Hint Rewrite esubst_type_fresh using solve [auto] : lngen.

Lemma esubst_ptt_fresh :
forall ptt1 e1 x1 x2,
  x1 `notin` efv_ptt ptt1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_ptt (esubst_ptt e1 x2 ptt1).
Proof.
pose proof esubst_type_fresh_esubst_ptt_fresh_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_fresh : lngen.
Hint Rewrite esubst_ptt_fresh using solve [auto] : lngen.

Lemma esubst_pt_fresh :
forall pt1 e1 x1 x2,
  x1 `notin` efv_pt pt1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_pt (esubst_pt e1 x2 pt1).
Proof.
pose proof esubst_type_fresh_esubst_ptt_fresh_esubst_pt_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_fresh : lngen.
Hint Rewrite esubst_pt_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_kind_fresh_mutual :
(forall k1 e1 x1 x2,
  x1 `notin` efv_kind k1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_kind (esubst_kind e1 x2 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_fresh :
forall k1 e1 x1 x2,
  x1 `notin` efv_kind k1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_kind (esubst_kind e1 x2 k1).
Proof.
pose proof esubst_kind_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_fresh : lngen.
Hint Rewrite esubst_kind_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_bind_fresh_mutual :
(forall bind1 e1 x1 x2,
  x1 `notin` efv_bind bind1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_bind (esubst_bind e1 x2 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_fresh :
forall bind1 e1 x1 x2,
  x1 `notin` efv_bind bind1 ->
  x1 `notin` efv_exp e1 ->
  x1 `notin` efv_bind (esubst_bind e1 x2 bind1).
Proof.
pose proof esubst_bind_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_fresh : lngen.
Hint Rewrite esubst_bind_fresh using solve [auto] : lngen.

Lemma esubst_exp_lc_exp :
forall e1 e2 x1,
  lc_exp e1 ->
  lc_exp e2 ->
  lc_exp (esubst_exp e2 x1 e1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_exp_lc_exp : lngen.

Lemma esubst_peb_lc_peb :
forall peb1 e1 x1,
  lc_peb peb1 ->
  lc_exp e1 ->
  lc_peb (esubst_peb e1 x1 peb1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_peb_lc_peb : lngen.

Lemma esubst_pe_lc_pe :
forall pe1 e1 x1,
  lc_pe pe1 ->
  lc_exp e1 ->
  lc_pe (esubst_pe e1 x1 pe1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_pe_lc_pe : lngen.

Lemma esubst_type_lc_type :
forall t1 e1 x1,
  lc_type t1 ->
  lc_exp e1 ->
  lc_type (esubst_type e1 x1 t1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_type_lc_type : lngen.

Lemma esubst_ptt_lc_ptt :
forall ptt1 e1 x1,
  lc_ptt ptt1 ->
  lc_exp e1 ->
  lc_ptt (esubst_ptt e1 x1 ptt1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_ptt_lc_ptt : lngen.

Lemma esubst_pt_lc_pt :
forall pt1 e1 x1,
  lc_pt pt1 ->
  lc_exp e1 ->
  lc_pt (esubst_pt e1 x1 pt1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_pt_lc_pt : lngen.

Lemma esubst_kind_lc_kind :
forall k1 e1 x1,
  lc_kind k1 ->
  lc_exp e1 ->
  lc_kind (esubst_kind e1 x1 k1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_kind_lc_kind : lngen.

Lemma esubst_bind_lc_bind :
forall bind1 e1 x1,
  lc_bind bind1 ->
  lc_exp e1 ->
  lc_bind (esubst_bind e1 x1 bind1).
Proof.
default_simp.
Qed.

Hint Resolve esubst_bind_lc_bind : lngen.

(* begin hide *)

Lemma esubst_exp_open_exp_wrt_exp_rec_esubst_peb_open_peb_wrt_exp_rec_esubst_pe_open_pe_wrt_exp_rec_mutual :
(forall e3 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_exp e1 x1 (open_exp_wrt_exp_rec n1 e2 e3) = open_exp_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_exp e1 x1 e3)) /\
(forall peb1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 e2 peb1) = open_peb_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 e2 pe1) = open_pe_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_exp_open_exp_wrt_exp_rec :
forall e3 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_exp e1 x1 (open_exp_wrt_exp_rec n1 e2 e3) = open_exp_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_exp e1 x1 e3).
Proof.
pose proof esubst_exp_open_exp_wrt_exp_rec_esubst_peb_open_peb_wrt_exp_rec_esubst_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_open_exp_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_peb_open_peb_wrt_exp_rec :
forall peb1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 e2 peb1) = open_peb_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_peb e1 x1 peb1).
Proof.
pose proof esubst_exp_open_exp_wrt_exp_rec_esubst_peb_open_peb_wrt_exp_rec_esubst_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_open_peb_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pe_open_pe_wrt_exp_rec :
forall pe1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 e2 pe1) = open_pe_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_pe e1 x1 pe1).
Proof.
pose proof esubst_exp_open_exp_wrt_exp_rec_esubst_peb_open_peb_wrt_exp_rec_esubst_pe_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_open_pe_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_type_open_type_wrt_exp_rec_esubst_ptt_open_ptt_wrt_exp_rec_esubst_pt_open_pt_wrt_exp_rec_mutual :
(forall t1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_type e1 x1 (open_type_wrt_exp_rec n1 e2 t1) = open_type_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 e2 ptt1) = open_ptt_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 e2 pt1) = open_pt_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_type_open_type_wrt_exp_rec :
forall t1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_type e1 x1 (open_type_wrt_exp_rec n1 e2 t1) = open_type_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_type e1 x1 t1).
Proof.
pose proof esubst_type_open_type_wrt_exp_rec_esubst_ptt_open_ptt_wrt_exp_rec_esubst_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_open_type_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_ptt_open_ptt_wrt_exp_rec :
forall ptt1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 e2 ptt1) = open_ptt_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_ptt e1 x1 ptt1).
Proof.
pose proof esubst_type_open_type_wrt_exp_rec_esubst_ptt_open_ptt_wrt_exp_rec_esubst_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_open_ptt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pt_open_pt_wrt_exp_rec :
forall pt1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 e2 pt1) = open_pt_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_pt e1 x1 pt1).
Proof.
pose proof esubst_type_open_type_wrt_exp_rec_esubst_ptt_open_ptt_wrt_exp_rec_esubst_pt_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_open_pt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_open_kind_wrt_exp_rec_mutual :
(forall k1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 e2 k1) = open_kind_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_open_kind_wrt_exp_rec :
forall k1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 e2 k1) = open_kind_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_kind e1 x1 k1).
Proof.
pose proof esubst_kind_open_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_open_kind_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_open_bind_wrt_exp_rec_mutual :
(forall bind1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 e2 bind1) = open_bind_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_open_bind_wrt_exp_rec :
forall bind1 e1 e2 x1 n1,
  lc_exp e1 ->
  esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 e2 bind1) = open_bind_wrt_exp_rec n1 (esubst_exp e1 x1 e2) (esubst_bind e1 x1 bind1).
Proof.
pose proof esubst_bind_open_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_open_bind_wrt_exp_rec : lngen.

(* end hide *)

Lemma esubst_exp_open_exp_wrt_exp :
forall e3 e1 e2 x1,
  lc_exp e1 ->
  esubst_exp e1 x1 (open_exp_wrt_exp e3 e2) = open_exp_wrt_exp (esubst_exp e1 x1 e3) (esubst_exp e1 x1 e2).
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_open_exp_wrt_exp : lngen.

Lemma esubst_peb_open_peb_wrt_exp :
forall peb1 e1 e2 x1,
  lc_exp e1 ->
  esubst_peb e1 x1 (open_peb_wrt_exp peb1 e2) = open_peb_wrt_exp (esubst_peb e1 x1 peb1) (esubst_exp e1 x1 e2).
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_open_peb_wrt_exp : lngen.

Lemma esubst_pe_open_pe_wrt_exp :
forall pe1 e1 e2 x1,
  lc_exp e1 ->
  esubst_pe e1 x1 (open_pe_wrt_exp pe1 e2) = open_pe_wrt_exp (esubst_pe e1 x1 pe1) (esubst_exp e1 x1 e2).
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_open_pe_wrt_exp : lngen.

Lemma esubst_type_open_type_wrt_exp :
forall t1 e1 e2 x1,
  lc_exp e1 ->
  esubst_type e1 x1 (open_type_wrt_exp t1 e2) = open_type_wrt_exp (esubst_type e1 x1 t1) (esubst_exp e1 x1 e2).
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_open_type_wrt_exp : lngen.

Lemma esubst_ptt_open_ptt_wrt_exp :
forall ptt1 e1 e2 x1,
  lc_exp e1 ->
  esubst_ptt e1 x1 (open_ptt_wrt_exp ptt1 e2) = open_ptt_wrt_exp (esubst_ptt e1 x1 ptt1) (esubst_exp e1 x1 e2).
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_open_ptt_wrt_exp : lngen.

Lemma esubst_pt_open_pt_wrt_exp :
forall pt1 e1 e2 x1,
  lc_exp e1 ->
  esubst_pt e1 x1 (open_pt_wrt_exp pt1 e2) = open_pt_wrt_exp (esubst_pt e1 x1 pt1) (esubst_exp e1 x1 e2).
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_open_pt_wrt_exp : lngen.

Lemma esubst_kind_open_kind_wrt_exp :
forall k1 e1 e2 x1,
  lc_exp e1 ->
  esubst_kind e1 x1 (open_kind_wrt_exp k1 e2) = open_kind_wrt_exp (esubst_kind e1 x1 k1) (esubst_exp e1 x1 e2).
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_open_kind_wrt_exp : lngen.

Lemma esubst_bind_open_bind_wrt_exp :
forall bind1 e1 e2 x1,
  lc_exp e1 ->
  esubst_bind e1 x1 (open_bind_wrt_exp bind1 e2) = open_bind_wrt_exp (esubst_bind e1 x1 bind1) (esubst_exp e1 x1 e2).
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_open_bind_wrt_exp : lngen.

Lemma esubst_exp_open_exp_wrt_exp_var :
forall e2 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_exp_wrt_exp (esubst_exp e1 x1 e2) (e_Var_f x2) = esubst_exp e1 x1 (open_exp_wrt_exp e2 (e_Var_f x2)).
Proof.
intros; rewrite esubst_exp_open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_open_exp_wrt_exp_var : lngen.

Lemma esubst_peb_open_peb_wrt_exp_var :
forall peb1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_peb_wrt_exp (esubst_peb e1 x1 peb1) (e_Var_f x2) = esubst_peb e1 x1 (open_peb_wrt_exp peb1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_peb_open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_open_peb_wrt_exp_var : lngen.

Lemma esubst_pe_open_pe_wrt_exp_var :
forall pe1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_pe_wrt_exp (esubst_pe e1 x1 pe1) (e_Var_f x2) = esubst_pe e1 x1 (open_pe_wrt_exp pe1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_pe_open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_open_pe_wrt_exp_var : lngen.

Lemma esubst_type_open_type_wrt_exp_var :
forall t1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_type_wrt_exp (esubst_type e1 x1 t1) (e_Var_f x2) = esubst_type e1 x1 (open_type_wrt_exp t1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_type_open_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_open_type_wrt_exp_var : lngen.

Lemma esubst_ptt_open_ptt_wrt_exp_var :
forall ptt1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_ptt_wrt_exp (esubst_ptt e1 x1 ptt1) (e_Var_f x2) = esubst_ptt e1 x1 (open_ptt_wrt_exp ptt1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_ptt_open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_open_ptt_wrt_exp_var : lngen.

Lemma esubst_pt_open_pt_wrt_exp_var :
forall pt1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_pt_wrt_exp (esubst_pt e1 x1 pt1) (e_Var_f x2) = esubst_pt e1 x1 (open_pt_wrt_exp pt1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_pt_open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_open_pt_wrt_exp_var : lngen.

Lemma esubst_kind_open_kind_wrt_exp_var :
forall k1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_kind_wrt_exp (esubst_kind e1 x1 k1) (e_Var_f x2) = esubst_kind e1 x1 (open_kind_wrt_exp k1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_kind_open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_open_kind_wrt_exp_var : lngen.

Lemma esubst_bind_open_bind_wrt_exp_var :
forall bind1 e1 x1 x2,
  x1 <> x2 ->
  lc_exp e1 ->
  open_bind_wrt_exp (esubst_bind e1 x1 bind1) (e_Var_f x2) = esubst_bind e1 x1 (open_bind_wrt_exp bind1 (e_Var_f x2)).
Proof.
intros; rewrite esubst_bind_open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_open_bind_wrt_exp_var : lngen.

(* begin hide *)

Lemma esubst_exp_spec_rec_esubst_peb_spec_rec_esubst_pe_spec_rec_mutual :
(forall e1 e2 x1 n1,
  esubst_exp e2 x1 e1 = open_exp_wrt_exp_rec n1 e2 (close_exp_wrt_exp_rec n1 x1 e1)) /\
(forall peb1 e1 x1 n1,
  esubst_peb e1 x1 peb1 = open_peb_wrt_exp_rec n1 e1 (close_peb_wrt_exp_rec n1 x1 peb1)) /\
(forall pe1 e1 x1 n1,
  esubst_pe e1 x1 pe1 = open_pe_wrt_exp_rec n1 e1 (close_pe_wrt_exp_rec n1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_exp_spec_rec :
forall e1 e2 x1 n1,
  esubst_exp e2 x1 e1 = open_exp_wrt_exp_rec n1 e2 (close_exp_wrt_exp_rec n1 x1 e1).
Proof.
pose proof esubst_exp_spec_rec_esubst_peb_spec_rec_esubst_pe_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_peb_spec_rec :
forall peb1 e1 x1 n1,
  esubst_peb e1 x1 peb1 = open_peb_wrt_exp_rec n1 e1 (close_peb_wrt_exp_rec n1 x1 peb1).
Proof.
pose proof esubst_exp_spec_rec_esubst_peb_spec_rec_esubst_pe_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pe_spec_rec :
forall pe1 e1 x1 n1,
  esubst_pe e1 x1 pe1 = open_pe_wrt_exp_rec n1 e1 (close_pe_wrt_exp_rec n1 x1 pe1).
Proof.
pose proof esubst_exp_spec_rec_esubst_peb_spec_rec_esubst_pe_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_type_spec_rec_esubst_ptt_spec_rec_esubst_pt_spec_rec_mutual :
(forall t1 e1 x1 n1,
  esubst_type e1 x1 t1 = open_type_wrt_exp_rec n1 e1 (close_type_wrt_exp_rec n1 x1 t1)) /\
(forall ptt1 e1 x1 n1,
  esubst_ptt e1 x1 ptt1 = open_ptt_wrt_exp_rec n1 e1 (close_ptt_wrt_exp_rec n1 x1 ptt1)) /\
(forall pt1 e1 x1 n1,
  esubst_pt e1 x1 pt1 = open_pt_wrt_exp_rec n1 e1 (close_pt_wrt_exp_rec n1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_type_spec_rec :
forall t1 e1 x1 n1,
  esubst_type e1 x1 t1 = open_type_wrt_exp_rec n1 e1 (close_type_wrt_exp_rec n1 x1 t1).
Proof.
pose proof esubst_type_spec_rec_esubst_ptt_spec_rec_esubst_pt_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_ptt_spec_rec :
forall ptt1 e1 x1 n1,
  esubst_ptt e1 x1 ptt1 = open_ptt_wrt_exp_rec n1 e1 (close_ptt_wrt_exp_rec n1 x1 ptt1).
Proof.
pose proof esubst_type_spec_rec_esubst_ptt_spec_rec_esubst_pt_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pt_spec_rec :
forall pt1 e1 x1 n1,
  esubst_pt e1 x1 pt1 = open_pt_wrt_exp_rec n1 e1 (close_pt_wrt_exp_rec n1 x1 pt1).
Proof.
pose proof esubst_type_spec_rec_esubst_ptt_spec_rec_esubst_pt_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_spec_rec_mutual :
(forall k1 e1 x1 n1,
  esubst_kind e1 x1 k1 = open_kind_wrt_exp_rec n1 e1 (close_kind_wrt_exp_rec n1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_spec_rec :
forall k1 e1 x1 n1,
  esubst_kind e1 x1 k1 = open_kind_wrt_exp_rec n1 e1 (close_kind_wrt_exp_rec n1 x1 k1).
Proof.
pose proof esubst_kind_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_spec_rec_mutual :
(forall bind1 e1 x1 n1,
  esubst_bind e1 x1 bind1 = open_bind_wrt_exp_rec n1 e1 (close_bind_wrt_exp_rec n1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_spec_rec :
forall bind1 e1 x1 n1,
  esubst_bind e1 x1 bind1 = open_bind_wrt_exp_rec n1 e1 (close_bind_wrt_exp_rec n1 x1 bind1).
Proof.
pose proof esubst_bind_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_spec_rec : lngen.

(* end hide *)

Lemma esubst_exp_spec :
forall e1 e2 x1,
  esubst_exp e2 x1 e1 = open_exp_wrt_exp (close_exp_wrt_exp x1 e1) e2.
Proof.
unfold close_exp_wrt_exp; unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_spec : lngen.

Lemma esubst_peb_spec :
forall peb1 e1 x1,
  esubst_peb e1 x1 peb1 = open_peb_wrt_exp (close_peb_wrt_exp x1 peb1) e1.
Proof.
unfold close_peb_wrt_exp; unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_spec : lngen.

Lemma esubst_pe_spec :
forall pe1 e1 x1,
  esubst_pe e1 x1 pe1 = open_pe_wrt_exp (close_pe_wrt_exp x1 pe1) e1.
Proof.
unfold close_pe_wrt_exp; unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_spec : lngen.

Lemma esubst_type_spec :
forall t1 e1 x1,
  esubst_type e1 x1 t1 = open_type_wrt_exp (close_type_wrt_exp x1 t1) e1.
Proof.
unfold close_type_wrt_exp; unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_spec : lngen.

Lemma esubst_ptt_spec :
forall ptt1 e1 x1,
  esubst_ptt e1 x1 ptt1 = open_ptt_wrt_exp (close_ptt_wrt_exp x1 ptt1) e1.
Proof.
unfold close_ptt_wrt_exp; unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_spec : lngen.

Lemma esubst_pt_spec :
forall pt1 e1 x1,
  esubst_pt e1 x1 pt1 = open_pt_wrt_exp (close_pt_wrt_exp x1 pt1) e1.
Proof.
unfold close_pt_wrt_exp; unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_spec : lngen.

Lemma esubst_kind_spec :
forall k1 e1 x1,
  esubst_kind e1 x1 k1 = open_kind_wrt_exp (close_kind_wrt_exp x1 k1) e1.
Proof.
unfold close_kind_wrt_exp; unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_spec : lngen.

Lemma esubst_bind_spec :
forall bind1 e1 x1,
  esubst_bind e1 x1 bind1 = open_bind_wrt_exp (close_bind_wrt_exp x1 bind1) e1.
Proof.
unfold close_bind_wrt_exp; unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_spec : lngen.

(* begin hide *)

Lemma esubst_exp_esubst_exp_esubst_peb_esubst_peb_esubst_pe_esubst_pe_mutual :
(forall e1 e2 e3 x2 x1,
  x2 `notin` efv_exp e2 ->
  x2 <> x1 ->
  esubst_exp e2 x1 (esubst_exp e3 x2 e1) = esubst_exp (esubst_exp e2 x1 e3) x2 (esubst_exp e2 x1 e1)) /\
(forall peb1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_peb e1 x1 (esubst_peb e2 x2 peb1) = esubst_peb (esubst_exp e1 x1 e2) x2 (esubst_peb e1 x1 peb1)) /\
(forall pe1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_pe e1 x1 (esubst_pe e2 x2 pe1) = esubst_pe (esubst_exp e1 x1 e2) x2 (esubst_pe e1 x1 pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_esubst_exp :
forall e1 e2 e3 x2 x1,
  x2 `notin` efv_exp e2 ->
  x2 <> x1 ->
  esubst_exp e2 x1 (esubst_exp e3 x2 e1) = esubst_exp (esubst_exp e2 x1 e3) x2 (esubst_exp e2 x1 e1).
Proof.
pose proof esubst_exp_esubst_exp_esubst_peb_esubst_peb_esubst_pe_esubst_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_esubst_exp : lngen.

Lemma esubst_peb_esubst_peb :
forall peb1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_peb e1 x1 (esubst_peb e2 x2 peb1) = esubst_peb (esubst_exp e1 x1 e2) x2 (esubst_peb e1 x1 peb1).
Proof.
pose proof esubst_exp_esubst_exp_esubst_peb_esubst_peb_esubst_pe_esubst_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_esubst_peb : lngen.

Lemma esubst_pe_esubst_pe :
forall pe1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_pe e1 x1 (esubst_pe e2 x2 pe1) = esubst_pe (esubst_exp e1 x1 e2) x2 (esubst_pe e1 x1 pe1).
Proof.
pose proof esubst_exp_esubst_exp_esubst_peb_esubst_peb_esubst_pe_esubst_pe_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_esubst_pe : lngen.

(* begin hide *)

Lemma esubst_type_esubst_type_esubst_ptt_esubst_ptt_esubst_pt_esubst_pt_mutual :
(forall t1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_type e1 x1 (esubst_type e2 x2 t1) = esubst_type (esubst_exp e1 x1 e2) x2 (esubst_type e1 x1 t1)) /\
(forall ptt1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_ptt e1 x1 (esubst_ptt e2 x2 ptt1) = esubst_ptt (esubst_exp e1 x1 e2) x2 (esubst_ptt e1 x1 ptt1)) /\
(forall pt1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_pt e1 x1 (esubst_pt e2 x2 pt1) = esubst_pt (esubst_exp e1 x1 e2) x2 (esubst_pt e1 x1 pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_esubst_type :
forall t1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_type e1 x1 (esubst_type e2 x2 t1) = esubst_type (esubst_exp e1 x1 e2) x2 (esubst_type e1 x1 t1).
Proof.
pose proof esubst_type_esubst_type_esubst_ptt_esubst_ptt_esubst_pt_esubst_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_esubst_type : lngen.

Lemma esubst_ptt_esubst_ptt :
forall ptt1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_ptt e1 x1 (esubst_ptt e2 x2 ptt1) = esubst_ptt (esubst_exp e1 x1 e2) x2 (esubst_ptt e1 x1 ptt1).
Proof.
pose proof esubst_type_esubst_type_esubst_ptt_esubst_ptt_esubst_pt_esubst_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_esubst_ptt : lngen.

Lemma esubst_pt_esubst_pt :
forall pt1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_pt e1 x1 (esubst_pt e2 x2 pt1) = esubst_pt (esubst_exp e1 x1 e2) x2 (esubst_pt e1 x1 pt1).
Proof.
pose proof esubst_type_esubst_type_esubst_ptt_esubst_ptt_esubst_pt_esubst_pt_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_esubst_pt : lngen.

(* begin hide *)

Lemma esubst_kind_esubst_kind_mutual :
(forall k1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_kind e1 x1 (esubst_kind e2 x2 k1) = esubst_kind (esubst_exp e1 x1 e2) x2 (esubst_kind e1 x1 k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_esubst_kind :
forall k1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_kind e1 x1 (esubst_kind e2 x2 k1) = esubst_kind (esubst_exp e1 x1 e2) x2 (esubst_kind e1 x1 k1).
Proof.
pose proof esubst_kind_esubst_kind_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_esubst_kind : lngen.

(* begin hide *)

Lemma esubst_bind_esubst_bind_mutual :
(forall bind1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_bind e1 x1 (esubst_bind e2 x2 bind1) = esubst_bind (esubst_exp e1 x1 e2) x2 (esubst_bind e1 x1 bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_esubst_bind :
forall bind1 e1 e2 x2 x1,
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  esubst_bind e1 x1 (esubst_bind e2 x2 bind1) = esubst_bind (esubst_exp e1 x1 e2) x2 (esubst_bind e1 x1 bind1).
Proof.
pose proof esubst_bind_esubst_bind_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_esubst_bind : lngen.

(* begin hide *)

Lemma esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual :
(forall e2 e1 x1 x2 n1,
  x2 `notin` efv_exp e2 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_exp e1 x1 e2 = close_exp_wrt_exp_rec n1 x2 (esubst_exp e1 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x2) e2))) *
(forall peb1 e1 x1 x2 n1,
  x2 `notin` efv_peb peb1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_peb e1 x1 peb1 = close_peb_wrt_exp_rec n1 x2 (esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x2) peb1))) *
(forall pe1 e1 x1 x2 n1,
  x2 `notin` efv_pe pe1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_pe e1 x1 pe1 = close_pe_wrt_exp_rec n1 x2 (esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x2) pe1))).
Proof.
apply_mutual_ind exp_peb_pe_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec :
forall e2 e1 x1 x2 n1,
  x2 `notin` efv_exp e2 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_exp e1 x1 e2 = close_exp_wrt_exp_rec n1 x2 (esubst_exp e1 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x2) e2)).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec :
forall peb1 e1 x1 x2 n1,
  x2 `notin` efv_peb peb1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_peb e1 x1 peb1 = close_peb_wrt_exp_rec n1 x2 (esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x2) peb1)).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec :
forall pe1 e1 x1 x2 n1,
  x2 `notin` efv_pe pe1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_pe e1 x1 pe1 = close_pe_wrt_exp_rec n1 x2 (esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x2) pe1)).
Proof.
pose proof esubst_exp_close_exp_wrt_exp_rec_open_exp_wrt_exp_rec_esubst_peb_close_peb_wrt_exp_rec_open_peb_wrt_exp_rec_esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_close_pe_wrt_exp_rec_open_pe_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual :
(forall t1 e1 x1 x2 n1,
  x2 `notin` efv_type t1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_type e1 x1 t1 = close_type_wrt_exp_rec n1 x2 (esubst_type e1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x2) t1))) *
(forall ptt1 e1 x1 x2 n1,
  x2 `notin` efv_ptt ptt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_ptt e1 x1 ptt1 = close_ptt_wrt_exp_rec n1 x2 (esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x2) ptt1))) *
(forall pt1 e1 x1 x2 n1,
  x2 `notin` efv_pt pt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_pt e1 x1 pt1 = close_pt_wrt_exp_rec n1 x2 (esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x2) pt1))).
Proof.
apply_mutual_ind type_ptt_pt_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec :
forall t1 e1 x1 x2 n1,
  x2 `notin` efv_type t1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_type e1 x1 t1 = close_type_wrt_exp_rec n1 x2 (esubst_type e1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x2) t1)).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec :
forall ptt1 e1 x1 x2 n1,
  x2 `notin` efv_ptt ptt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_ptt e1 x1 ptt1 = close_ptt_wrt_exp_rec n1 x2 (esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x2) ptt1)).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec :
forall pt1 e1 x1 x2 n1,
  x2 `notin` efv_pt pt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_pt e1 x1 pt1 = close_pt_wrt_exp_rec n1 x2 (esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x2) pt1)).
Proof.
pose proof esubst_type_close_type_wrt_exp_rec_open_type_wrt_exp_rec_esubst_ptt_close_ptt_wrt_exp_rec_open_ptt_wrt_exp_rec_esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_close_pt_wrt_exp_rec_open_pt_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_close_kind_wrt_exp_rec_open_kind_wrt_exp_rec_mutual :
(forall k1 e1 x1 x2 n1,
  x2 `notin` efv_kind k1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_kind e1 x1 k1 = close_kind_wrt_exp_rec n1 x2 (esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x2) k1))).
Proof.
apply_mutual_ind kind_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_kind_close_kind_wrt_exp_rec_open_kind_wrt_exp_rec :
forall k1 e1 x1 x2 n1,
  x2 `notin` efv_kind k1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_kind e1 x1 k1 = close_kind_wrt_exp_rec n1 x2 (esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x2) k1)).
Proof.
pose proof esubst_kind_close_kind_wrt_exp_rec_open_kind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_close_kind_wrt_exp_rec_open_kind_wrt_exp_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_close_bind_wrt_exp_rec_open_bind_wrt_exp_rec_mutual :
(forall bind1 e1 x1 x2 n1,
  x2 `notin` efv_bind bind1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_bind e1 x1 bind1 = close_bind_wrt_exp_rec n1 x2 (esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x2) bind1))).
Proof.
apply_mutual_ind bind_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma esubst_bind_close_bind_wrt_exp_rec_open_bind_wrt_exp_rec :
forall bind1 e1 x1 x2 n1,
  x2 `notin` efv_bind bind1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  degree_exp_wrt_exp n1 e1 ->
  esubst_bind e1 x1 bind1 = close_bind_wrt_exp_rec n1 x2 (esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x2) bind1)).
Proof.
pose proof esubst_bind_close_bind_wrt_exp_rec_open_bind_wrt_exp_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_close_bind_wrt_exp_rec_open_bind_wrt_exp_rec : lngen.

(* end hide *)

Lemma esubst_exp_close_exp_wrt_exp_open_exp_wrt_exp :
forall e2 e1 x1 x2,
  x2 `notin` efv_exp e2 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_exp e1 x1 e2 = close_exp_wrt_exp x2 (esubst_exp e1 x1 (open_exp_wrt_exp e2 (e_Var_f x2))).
Proof.
unfold close_exp_wrt_exp; unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_close_exp_wrt_exp_open_exp_wrt_exp : lngen.

Lemma esubst_peb_close_peb_wrt_exp_open_peb_wrt_exp :
forall peb1 e1 x1 x2,
  x2 `notin` efv_peb peb1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_peb e1 x1 peb1 = close_peb_wrt_exp x2 (esubst_peb e1 x1 (open_peb_wrt_exp peb1 (e_Var_f x2))).
Proof.
unfold close_peb_wrt_exp; unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_close_peb_wrt_exp_open_peb_wrt_exp : lngen.

Lemma esubst_pe_close_pe_wrt_exp_open_pe_wrt_exp :
forall pe1 e1 x1 x2,
  x2 `notin` efv_pe pe1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_pe e1 x1 pe1 = close_pe_wrt_exp x2 (esubst_pe e1 x1 (open_pe_wrt_exp pe1 (e_Var_f x2))).
Proof.
unfold close_pe_wrt_exp; unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_close_pe_wrt_exp_open_pe_wrt_exp : lngen.

Lemma esubst_type_close_type_wrt_exp_open_type_wrt_exp :
forall t1 e1 x1 x2,
  x2 `notin` efv_type t1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_type e1 x1 t1 = close_type_wrt_exp x2 (esubst_type e1 x1 (open_type_wrt_exp t1 (e_Var_f x2))).
Proof.
unfold close_type_wrt_exp; unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_close_type_wrt_exp_open_type_wrt_exp : lngen.

Lemma esubst_ptt_close_ptt_wrt_exp_open_ptt_wrt_exp :
forall ptt1 e1 x1 x2,
  x2 `notin` efv_ptt ptt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_ptt e1 x1 ptt1 = close_ptt_wrt_exp x2 (esubst_ptt e1 x1 (open_ptt_wrt_exp ptt1 (e_Var_f x2))).
Proof.
unfold close_ptt_wrt_exp; unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_close_ptt_wrt_exp_open_ptt_wrt_exp : lngen.

Lemma esubst_pt_close_pt_wrt_exp_open_pt_wrt_exp :
forall pt1 e1 x1 x2,
  x2 `notin` efv_pt pt1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_pt e1 x1 pt1 = close_pt_wrt_exp x2 (esubst_pt e1 x1 (open_pt_wrt_exp pt1 (e_Var_f x2))).
Proof.
unfold close_pt_wrt_exp; unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_close_pt_wrt_exp_open_pt_wrt_exp : lngen.

Lemma esubst_kind_close_kind_wrt_exp_open_kind_wrt_exp :
forall k1 e1 x1 x2,
  x2 `notin` efv_kind k1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_kind e1 x1 k1 = close_kind_wrt_exp x2 (esubst_kind e1 x1 (open_kind_wrt_exp k1 (e_Var_f x2))).
Proof.
unfold close_kind_wrt_exp; unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_close_kind_wrt_exp_open_kind_wrt_exp : lngen.

Lemma esubst_bind_close_bind_wrt_exp_open_bind_wrt_exp :
forall bind1 e1 x1 x2,
  x2 `notin` efv_bind bind1 ->
  x2 `notin` efv_exp e1 ->
  x2 <> x1 ->
  lc_exp e1 ->
  esubst_bind e1 x1 bind1 = close_bind_wrt_exp x2 (esubst_bind e1 x1 (open_bind_wrt_exp bind1 (e_Var_f x2))).
Proof.
unfold close_bind_wrt_exp; unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_close_bind_wrt_exp_open_bind_wrt_exp : lngen.

Lemma esubst_exp_e_Lam :
forall x2 e2 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_exp e2 `union` singleton x1 ->
  esubst_exp e1 x1 (e_Lam e2) = e_Lam (close_exp_wrt_exp x2 (esubst_exp e1 x1 (open_exp_wrt_exp e2 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_exp_e_Lam : lngen.

Lemma esubst_exp_e_Fix :
forall f1 e2 e1 x1,
  lc_exp e1 ->
  f1 `notin` efv_exp e1 `union` efv_exp e2 `union` singleton x1 ->
  esubst_exp e1 x1 (e_Fix e2) = e_Fix (close_exp_wrt_exp f1 (esubst_exp e1 x1 (open_exp_wrt_exp e2 (e_Var_f f1)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_exp_e_Fix : lngen.

Lemma esubst_peb_peb_Branch :
forall x2 C1 e2 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_exp e2 `union` singleton x1 ->
  esubst_peb e1 x1 (peb_Branch C1 e2) = peb_Branch (C1) (close_exp_wrt_exp x2 (esubst_exp e1 x1 (open_exp_wrt_exp e2 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_peb_peb_Branch : lngen.

Lemma esubst_type_t_Pi :
forall x2 t1 t2 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_type t2 `union` singleton x1 ->
  esubst_type e1 x1 (t_Pi t1 t2) = t_Pi (esubst_type e1 x1 t1) (close_type_wrt_exp x2 (esubst_type e1 x1 (open_type_wrt_exp t2 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_type_t_Pi : lngen.

Lemma esubst_type_t_Sigma :
forall x2 t1 t2 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_type t2 `union` singleton x1 ->
  esubst_type e1 x1 (t_Sigma t1 t2) = t_Sigma (esubst_type e1 x1 t1) (close_type_wrt_exp x2 (esubst_type e1 x1 (open_type_wrt_exp t2 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_type_t_Sigma : lngen.

Lemma esubst_ptt_ptt_pattern :
forall x2 C1 t1 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_type t1 `union` singleton x1 ->
  esubst_ptt e1 x1 (ptt_pattern C1 t1) = ptt_pattern (C1) (close_type_wrt_exp x2 (esubst_type e1 x1 (open_type_wrt_exp t1 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_ptt_ptt_pattern : lngen.

Lemma esubst_kind_k_pi :
forall x2 t1 k1 e1 x1,
  lc_exp e1 ->
  x2 `notin` efv_exp e1 `union` efv_kind k1 `union` singleton x1 ->
  esubst_kind e1 x1 (k_pi t1 k1) = k_pi (esubst_type e1 x1 t1) (close_kind_wrt_exp x2 (esubst_kind e1 x1 (open_kind_wrt_exp k1 (e_Var_f x2)))).
Proof.
default_simp.
Qed.

Hint Resolve esubst_kind_k_pi : lngen.

(* begin hide *)

Lemma esubst_exp_intro_rec_esubst_peb_intro_rec_esubst_pe_intro_rec_mutual :
(forall e1 x1 e2 n1,
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp_rec n1 e2 e1 = esubst_exp e2 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1)) /\
(forall peb1 x1 e1 n1,
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp_rec n1 e1 peb1 = esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1)) /\
(forall pe1 x1 e1 n1,
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp_rec n1 e1 pe1 = esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1)).
Proof.
apply_mutual_ind exp_peb_pe_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_exp_intro_rec :
forall e1 x1 e2 n1,
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp_rec n1 e2 e1 = esubst_exp e2 x1 (open_exp_wrt_exp_rec n1 (e_Var_f x1) e1).
Proof.
pose proof esubst_exp_intro_rec_esubst_peb_intro_rec_esubst_pe_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_exp_intro_rec : lngen.
Hint Rewrite esubst_exp_intro_rec using solve [auto] : lngen.

Lemma esubst_peb_intro_rec :
forall peb1 x1 e1 n1,
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp_rec n1 e1 peb1 = esubst_peb e1 x1 (open_peb_wrt_exp_rec n1 (e_Var_f x1) peb1).
Proof.
pose proof esubst_exp_intro_rec_esubst_peb_intro_rec_esubst_pe_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_peb_intro_rec : lngen.
Hint Rewrite esubst_peb_intro_rec using solve [auto] : lngen.

Lemma esubst_pe_intro_rec :
forall pe1 x1 e1 n1,
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp_rec n1 e1 pe1 = esubst_pe e1 x1 (open_pe_wrt_exp_rec n1 (e_Var_f x1) pe1).
Proof.
pose proof esubst_exp_intro_rec_esubst_peb_intro_rec_esubst_pe_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pe_intro_rec : lngen.
Hint Rewrite esubst_pe_intro_rec using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_type_intro_rec_esubst_ptt_intro_rec_esubst_pt_intro_rec_mutual :
(forall t1 x1 e1 n1,
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp_rec n1 e1 t1 = esubst_type e1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x1) t1)) /\
(forall ptt1 x1 e1 n1,
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp_rec n1 e1 ptt1 = esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1)) /\
(forall pt1 x1 e1 n1,
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp_rec n1 e1 pt1 = esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1)).
Proof.
apply_mutual_ind type_ptt_pt_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_type_intro_rec :
forall t1 x1 e1 n1,
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp_rec n1 e1 t1 = esubst_type e1 x1 (open_type_wrt_exp_rec n1 (e_Var_f x1) t1).
Proof.
pose proof esubst_type_intro_rec_esubst_ptt_intro_rec_esubst_pt_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_type_intro_rec : lngen.
Hint Rewrite esubst_type_intro_rec using solve [auto] : lngen.

Lemma esubst_ptt_intro_rec :
forall ptt1 x1 e1 n1,
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp_rec n1 e1 ptt1 = esubst_ptt e1 x1 (open_ptt_wrt_exp_rec n1 (e_Var_f x1) ptt1).
Proof.
pose proof esubst_type_intro_rec_esubst_ptt_intro_rec_esubst_pt_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_ptt_intro_rec : lngen.
Hint Rewrite esubst_ptt_intro_rec using solve [auto] : lngen.

Lemma esubst_pt_intro_rec :
forall pt1 x1 e1 n1,
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp_rec n1 e1 pt1 = esubst_pt e1 x1 (open_pt_wrt_exp_rec n1 (e_Var_f x1) pt1).
Proof.
pose proof esubst_type_intro_rec_esubst_ptt_intro_rec_esubst_pt_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_pt_intro_rec : lngen.
Hint Rewrite esubst_pt_intro_rec using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_kind_intro_rec_mutual :
(forall k1 x1 e1 n1,
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp_rec n1 e1 k1 = esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_kind_intro_rec :
forall k1 x1 e1 n1,
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp_rec n1 e1 k1 = esubst_kind e1 x1 (open_kind_wrt_exp_rec n1 (e_Var_f x1) k1).
Proof.
pose proof esubst_kind_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_kind_intro_rec : lngen.
Hint Rewrite esubst_kind_intro_rec using solve [auto] : lngen.

(* begin hide *)

Lemma esubst_bind_intro_rec_mutual :
(forall bind1 x1 e1 n1,
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp_rec n1 e1 bind1 = esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1)).
Proof.
apply_mutual_ind bind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma esubst_bind_intro_rec :
forall bind1 x1 e1 n1,
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp_rec n1 e1 bind1 = esubst_bind e1 x1 (open_bind_wrt_exp_rec n1 (e_Var_f x1) bind1).
Proof.
pose proof esubst_bind_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve esubst_bind_intro_rec : lngen.
Hint Rewrite esubst_bind_intro_rec using solve [auto] : lngen.

Lemma esubst_exp_intro :
forall x1 e1 e2,
  x1 `notin` efv_exp e1 ->
  open_exp_wrt_exp e1 e2 = esubst_exp e2 x1 (open_exp_wrt_exp e1 (e_Var_f x1)).
Proof.
unfold open_exp_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_exp_intro : lngen.

Lemma esubst_peb_intro :
forall x1 peb1 e1,
  x1 `notin` efv_peb peb1 ->
  open_peb_wrt_exp peb1 e1 = esubst_peb e1 x1 (open_peb_wrt_exp peb1 (e_Var_f x1)).
Proof.
unfold open_peb_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_peb_intro : lngen.

Lemma esubst_pe_intro :
forall x1 pe1 e1,
  x1 `notin` efv_pe pe1 ->
  open_pe_wrt_exp pe1 e1 = esubst_pe e1 x1 (open_pe_wrt_exp pe1 (e_Var_f x1)).
Proof.
unfold open_pe_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pe_intro : lngen.

Lemma esubst_type_intro :
forall x1 t1 e1,
  x1 `notin` efv_type t1 ->
  open_type_wrt_exp t1 e1 = esubst_type e1 x1 (open_type_wrt_exp t1 (e_Var_f x1)).
Proof.
unfold open_type_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_type_intro : lngen.

Lemma esubst_ptt_intro :
forall x1 ptt1 e1,
  x1 `notin` efv_ptt ptt1 ->
  open_ptt_wrt_exp ptt1 e1 = esubst_ptt e1 x1 (open_ptt_wrt_exp ptt1 (e_Var_f x1)).
Proof.
unfold open_ptt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_ptt_intro : lngen.

Lemma esubst_pt_intro :
forall x1 pt1 e1,
  x1 `notin` efv_pt pt1 ->
  open_pt_wrt_exp pt1 e1 = esubst_pt e1 x1 (open_pt_wrt_exp pt1 (e_Var_f x1)).
Proof.
unfold open_pt_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_pt_intro : lngen.

Lemma esubst_kind_intro :
forall x1 k1 e1,
  x1 `notin` efv_kind k1 ->
  open_kind_wrt_exp k1 e1 = esubst_kind e1 x1 (open_kind_wrt_exp k1 (e_Var_f x1)).
Proof.
unfold open_kind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_kind_intro : lngen.

Lemma esubst_bind_intro :
forall x1 bind1 e1,
  x1 `notin` efv_bind bind1 ->
  open_bind_wrt_exp bind1 e1 = esubst_bind e1 x1 (open_bind_wrt_exp bind1 (e_Var_f x1)).
Proof.
unfold open_bind_wrt_exp; default_simp.
Qed.

Hint Resolve esubst_bind_intro : lngen.


(* *********************************************************************** *)
(** * "Restore" tactics *)

Ltac default_auto := auto; tauto.
Ltac default_autorewrite := fail.
