Add LoadPath "./metatheory".

Set Implicit Arguments.

Require Import Metatheory.
Require Import lang.
Require Import langExtra.
Require Import isEq.
Require Import infrastructure. 

Axiom wfs_sig : wfs Signature.
Axiom datacon_bind : forall (C:contmvar), 
  exists t, exists s, exists T, exists u, 
    binds C (bind_t t) Signature /\ t= (t_Pi s (t_App (t_Con T) u)).
Axiom datatype_bind : forall T, 
  exists k, exists s,
    binds T (bind_k k) Signature /\ k = (k_pi s k_type).
Axiom datacon_bind_uniq : forall (C:contmvar) s s' T T' u u',
  binds C (bind_t (t_Pi s (t_App (t_Con T) u))) Signature ->
  binds C (bind_t (t_Pi s' (t_App (t_Con T') u'))) Signature ->
  s = s' /\ T = T' /\ u = u'.
Axiom datatype_bind_uniq : forall (T:contmvar) s s',
  binds T (bind_k (k_pi s k_type)) Signature ->
  binds T (bind_k (k_pi s' k_type)) Signature ->
  s = s'.
Axiom datacon_bind_efv_empty : forall (C:contmvar) s T u,
  binds C (bind_t (t_Pi s (t_App (t_Con T) u))) Signature ->
  efv_type s = {} /\ efv_exp u = {}. 
Axiom datatype_bind_efv_empty : forall (T:contmvar) s,
  binds T (bind_k (k_pi s k_type)) Signature ->
  efv_type s = {}.
Axiom datacon_bind_lc : forall (C:contmvar) t,
  binds C (bind_t t) Signature -> lc_type t. 
Axiom datatype_bind_lc : forall (T:contmvar) k,
  binds T (bind_k k) Signature -> lc_kind k.

Lemma lc_open_ui : forall Ci si T ui xi,
  binds Ci (bind_t (t_Pi si (t_App (t_Con T) ui))) Signature ->
  lc_exp (open_exp_wrt_exp ui (e_Var_f xi)).
Proof.
  intros.
  apply datacon_bind_lc in H.
  inversion H; subst.
  assert (J:=@H3 xi). inversion J; subst. auto.
Qed.

Lemma pure_is_v_of_exp: forall v,
    lc_exp v ->
    is_v_of_exp v ->
    pure v.
Proof.
  intros v H1 H2.
  induction H1; simpl in H2; try solve [inversion H2|intuition|auto].
Qed.

Lemma notmatching_ptt_comm : forall ptt1 ptt2,
  notmatching_ptt ptt1 ptt2 ->
  notmatching_ptt ptt2 ptt1.
Proof.
   inversion 1; constructor; auto.
Qed.
Hint Resolve notmatching_ptt_comm.

Lemma find_pt_deterministic_helper : forall ptt1 ptt2 pt,
  find_pt ptt1 pt ->
  find_pt ptt2 pt ->
  (ptt1 = ptt2) \/  (notmatching_ptt ptt1 ptt2).
Proof.
  induction 1; inversion 1; subst; auto.
Qed.

Lemma find_pt_deterministic : forall C ti tj pt1,
  find_pt (ptt_pattern C ti) pt1 ->
  find_pt (ptt_pattern C tj) pt1 ->
  ti = tj.
Proof. 
  intros C ti tj pt1 F1 F2.
  destruct (find_pt_deterministic_helper F1 F2).
    injection H; tauto.
    inversion H; absurd (C = C); auto.
Qed.

(*************************************************************************)


(************* Local closure *******************************)

Definition lc_wft_prop G t k (WF : wft G t k) :=
  lc_ctx G /\ lc_type t /\ lc_kind k.
Definition lc_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  lc_ctx G /\ lc_exp e /\ lc_exp u /\ lc_pt pt /\ lc_kind k.
Definition lc_wfk_prop G k (WF : wfk G k) :=
  lc_ctx G /\ lc_kind k.
Definition lc_wfe_prop G e t (WF : wfe G e t) :=
  lc_ctx G /\ lc_exp e /\ lc_type t.
Definition lc_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  lc_ctx G /\ lc_exp e /\ lc_exp u /\ lc_pe pe /\ lc_type t.
Definition lc_wfc_prop G (WF : wfc G) :=
  lc_ctx G /\ (forall x t, binds x t G -> lc_type t).
Lemma lc_wf:
  (forall G k WF, @lc_wfk_prop G k WF) /\
  (forall G t k WF, @lc_wft_prop G t k WF) /\
  (forall G e u pt k WF, @lc_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @lc_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @lc_wfpe_prop G e u pe t WF) /\
  (forall G WF, @lc_wfc_prop G WF).
(wf_mutind_cases 
  apply wf_mutind with 
    (P  := lc_wfk_prop) 
    (P0 := lc_wft_prop) 
    (P1 := lc_wfpt_prop)
    (P2 := lc_wfe_prop)  
    (P3 := lc_wfpe_prop)
    (P4 := lc_wfc_prop) Case); 
  unfold lc_wfk_prop, lc_wft_prop, lc_wfpt_prop, 
  lc_wfe_prop, lc_wfpe_prop, lc_wfc_prop; intros; subst; simpl; intuition;
  eauto using lc_isEq.
  Case "WFK_Pi".
    SCase "lc_ctx".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
    SCase "lc_k_pi".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      apply lc_k_pi_exists with (x1:=x) ; auto.
        inversion H1; auto.
  Case "K_Const".
    apply datatype_bind_lc in b; auto.
  Case "K_Pi".
    SCase "lc_ctx".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
    SCase "lc_t_pi".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 [H2 H3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
        inversion H1; auto.
  Case "K_Sigma".
    SCase "lc_ctx".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
    SCase "lc_t_Sigma".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 [H2 H3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
        inversion H1; auto.
  Case "PK_Cons".
    SCase "lc_pt".
      apply lc_pt_cons; auto.
        pick fresh x.
        assert (x `notin` L) as xnotinL. auto.
        apply H0 in xnotinL.
        destruct xnotinL as [J1 [J2 J3]].
        apply lc_ptt_pattern_exists with (x1:=x); auto.
  Case "T_Fun".
    SCase "lc_ctx".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
    SCase "lc_e_Lam".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 [H2 H3]].
      apply lc_e_Lam_exists with (x1:=x); auto.
    SCase "lc_t_Pi".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 [H2 H3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
        inversion H1; auto.
  Case "T_Fix".
    SCase "lc_ctx".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
    SCase "lc_e_Fix".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 [H2 H3]].
      apply lc_e_Fix_exists with (f1:=x); auto.
    SCase "lc_t".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [H1 H2].
      simpl in H1. destruct H1; auto.
  Case "T_Sigma".
    SCase "lc_t_Sigma".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H1 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
  Case "T_Proj1".
    SCase "lc_t_Proj1".
      inversion H2; auto.
  Case "AP_Cons".
    SCase "lc_pe".
      apply lc_pe_Cons; auto.
        pick fresh x.
        assert (x `notin` L) as xnotinL. auto.
        apply H in xnotinL.
        destruct xnotinL as [J1 [J2 J3]].
        apply lc_peb_Branch_exists with (x1:=x); auto.
  Case "C_Empty".
    inversion H.
  Case "C_Term".
    analyze_binds H3; eauto.
      inversion H6; inversion H5; subst; auto.
  Case "C_Eq".
    analyze_binds H5; eauto.
      inversion H9; inversion H8.
Qed.

Lemma lc_wft: forall G t k,
  wft G t k ->
  lc_ctx G /\ lc_type t /\ lc_kind k.
intros. 
pose (J := lc_wf).
unfold lc_wft_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_wfpt: forall G e u pt k,
  wfpt G e u pt k ->
  lc_ctx G /\ lc_exp e /\ lc_exp u /\ lc_pt pt /\ lc_kind k.
intros. 
pose (J := lc_wf).
unfold lc_wfpt_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_wfk: forall G k,
  wfk G k ->
  lc_ctx G /\ lc_kind k.
intros. 
pose (J := lc_wf).
unfold lc_wfk_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_wfe: forall G e t,
  wfe G e t ->
  lc_ctx G /\ lc_exp e /\ lc_type t.
intros. 
pose (J := lc_wf).
unfold lc_wfe_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_wfpe: forall G e u pe t,
  wfpe G e u pe t ->
  lc_ctx G /\ lc_exp e /\ lc_exp u /\ lc_pe pe /\ lc_type t.
intros. 
pose (J := lc_wf).
unfold lc_wfpe_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_wfc: forall G,
  wfc G -> 
  lc_ctx G /\ (forall x t, binds x t G -> lc_type t).
intros. 
pose (J := lc_wf).
unfold lc_wfc_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_notmatching_ptt : forall ptt1 ptt2,
  notmatching_ptt ptt1 ptt2 ->
  lc_ptt ptt1 /\ lc_ptt ptt2.
Proof.
  intros ptt1 ptt2 H.
  induction H; intuition.
Qed.

Lemma lc_find_pt : forall ptt' pt',
  find_pt ptt' pt' ->
  lc_ptt ptt' /\ lc_pt pt'.
Proof.
  intros ptt' pt' H.
  induction H; intuition.
    apply lc_notmatching_ptt in H0; intuition.
Qed.

Definition lc_typeeq_prop D t t' (WF : typeeq D t t') :=
  lc_type t /\ lc_type t' /\ lc_ectx D.
Definition lc_pateq_prop D e u e' u' pt pt' (WF : pateq D e u e' u' pt pt') :=
  lc_pt pt /\ lc_pt pt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
Definition lc_patteq_prop D e u e' u' ptt ptt' (WF : patteq D e u e' u' ptt ptt') :=
  lc_ptt ptt /\ lc_ptt ptt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
Lemma lc_eq:
  (forall D t t' WF, @lc_typeeq_prop D t t' WF) /\
  (forall D e u e' u' pt pt' WF, @lc_pateq_prop D e u e' u' pt pt' WF) /\
  (forall D e u e' u' ptt ptt' WF, @lc_patteq_prop D e u e' u' ptt ptt' WF).
(eq_mutind_cases 
  apply eq_mutind with 
    (P  := lc_typeeq_prop) 
    (P0 := lc_pateq_prop) 
    (P1 := lc_patteq_prop) Case); 
  unfold lc_typeeq_prop, lc_pateq_prop, lc_patteq_prop; 
  intros; subst; simpl; intuition;
  eauto using lc_isEq.
Case "TQ_InCon".
  inversion i; subst.
  apply lc_isEq in H1. decompose [and] H1; auto.
Case "TQ_Pi".
    SCase "lc_t_Pi".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
    SCase "lc_t_Pi'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
Case "TQ_Sigma".
    SCase "lc_t_Sigma".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
    SCase "lc_t_Sigma'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
Case "TQ_App".
    SCase "lc_t_App".
      apply lc_t_App; auto.
        apply lc_isEq in i. decompose [and] i; auto.
    SCase "lc_t_App".
      apply lc_t_App; auto.
        apply lc_isEq in i. decompose [and] i; auto.
Case "TQ_Red1".
    SCase "lc_t_case".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Case; auto.
        apply lc_isEq in i. decompose [and] i; auto.

        assert (x `notin` L) as xnotinL. auto.
        apply i0 in xnotinL.
        apply lc_isEq in xnotinL. decompose [and] xnotinL; auto. 

        apply lc_find_pt in f. decompose [and] f; auto.
    SCase "lc_t".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]]; auto.
    SCase "lc_D".
      apply lc_isEq in i. decompose [and] i; auto.
Case "TQ_Red2".
    SCase "lc_t".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]]; auto.
    SCase "lc_t_case".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Case; auto.
        apply lc_isEq in i. decompose [and] i; auto.

        assert (x `notin` L) as xnotinL. auto.
        apply i0 in xnotinL.
        apply lc_isEq in xnotinL. decompose [and] xnotinL; auto. 

        apply lc_find_pt in f. decompose [and] f; auto.
    SCase "lc_D".
      apply lc_isEq in i. decompose [and] i; auto.
Case "TTPQ_Intro".
    SCase "lc_ptt_pattern".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_ptt_pattern_exists with (x1:=x); auto.
    SCase "lc_ptt_pattern'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_ptt_pattern_exists with (x1:=x); auto.
    SCase "lc_e".
      apply lc_isEq in i. decompose [and] i; auto.
    SCase "lc_u".
      apply lc_isEq in i0. decompose [and] i0; auto.
    SCase "lc_e'".
      apply lc_isEq in i. decompose [and] i; auto.
    SCase "lc_u'".
      apply lc_isEq in i0. decompose [and] i0; auto.
    SCase "lc_D".
      apply lc_isEq in i0. decompose [and] i0; auto.
Qed.

Lemma lc_typeeq: forall D t t',
  typeeq D t t' -> 
  lc_type t /\ lc_type t' /\ lc_ectx D.
intros. 
pose (J := lc_eq).
unfold lc_typeeq_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_pateq: forall D e u e' u' pt pt',
  pateq D e u e' u' pt pt' ->
  lc_pt pt /\ lc_pt pt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
intros. 
pose (J := lc_eq).
unfold lc_pateq_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_patteq: forall D e u e' u' ptt ptt',
  patteq D e u e' u' ptt ptt' ->
  lc_ptt ptt /\ lc_ptt ptt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
intros. 
pose (J := lc_eq).
unfold lc_patteq_prop in J. 
decompose [and] J; auto.
Qed.

Lemma lc_kindeq: forall D k k',
  kindeq D k k' -> 
  lc_kind k /\ lc_kind k' /\ lc_ectx D.
intros.
  induction H; intuition.
    pick fresh x.
    assert (x `notin` L) as xnotinL. auto.
    apply H1 in xnotinL.
    destruct xnotinL as [J1 [J2 J3]].
    apply lc_k_pi_exists with (x1:=x) ; auto.
      apply lc_typeeq in H.
      decompose [and] H; auto.

    pick fresh x.
    assert (x `notin` L) as xnotinL. auto.
    apply H1 in xnotinL.
    destruct xnotinL as [J1 [J2 J3]].
    apply lc_k_pi_exists with (x1:=x) ; auto.
      apply lc_typeeq in H.
      decompose [and] H; auto.

    apply lc_typeeq in H.
    decompose [and] H; auto.
Qed.

Lemma lc_ctxereq: forall D D',
  ctxereq D D' ->
  lc_ectx D /\ lc_ectx D'.
intros.
  induction H; simpl; intuition.
    apply lc_isEq in H0. decompose [and] H0; auto.
    apply lc_isEq in H1. decompose [and] H1; auto.
    apply lc_isEq in H0. decompose [and] H0; auto.
    apply lc_isEq in H1. decompose [and] H1; auto.
Qed.  

Lemma lc_ctxeq: forall G G',
  ctxeq G G' ->
  lc_ctx G /\ lc_ctx G'.
intros.
  induction H; simpl; intuition.
    apply lc_typeeq in H0. decompose [and] H0; auto.
    apply lc_typeeq in H0. decompose [and] H0; auto.
    apply lc_isEq in H0. decompose [and] H0; auto.
    apply lc_isEq in H1. decompose [and] H1; auto.
    apply lc_isEq in H0. decompose [and] H0; auto.
    apply lc_isEq in H1. decompose [and] H1; auto.
Qed.  

Lemma app_lc_ctx1: forall G1 G2,
  lc_ctx (G1 ++ G2) ->
  lc_ctx G1.
intros.
  induction G1; simpl; auto.
    destruct a.
      simpl in H. destruct H; auto.
      destruct p.
        simpl in H. decompose [and] H; auto.
Qed.

Lemma app_lc_ctx2: forall G1 G2,
  lc_ctx (G1 ++ G2) ->
  lc_ctx G2.
intros.
  induction G1; simpl; auto.
    destruct a.
      simpl in H. destruct H; auto.
      destruct p.
        simpl in H. decompose [and] H; auto.
Qed.

Lemma app_lc_ctx3: forall G1 G2,
 lc_ctx G1 -> lc_ctx G2 ->
 lc_ctx (G1++G2).
intros.
 induction G1; simpl; auto.
   destruct a.
     simpl in H. destruct H; auto.
     destruct p.
       simpl in H. decompose [and] H; auto.
Qed.

Lemma app_lc_ectx1: forall D1 D2,
  lc_ectx (D1 ++ D2) ->
  lc_ectx D1.
intros.
  induction D1; simpl; auto.
    destruct a.
      simpl in H. decompose [and] H; auto.
Qed.

Lemma app_lc_ectx2: forall D1 D2,
  lc_ectx (D1 ++ D2) ->
  lc_ectx D2.
intros.
  induction D1; simpl; auto.
    destruct a.
      simpl in H. decompose [and] H; auto.
Qed.

Lemma app_lc_ectx3: forall D1 D2,
 lc_ectx D1 -> lc_ectx D2 ->
 lc_ectx (D1++D2).
intros.
 induction D1; simpl; auto.
   destruct a.
     simpl in H. decompose [and] H; auto.
Qed.

Lemma lc_notmatching_peb : forall peb1 peb2,
  notmatching_peb peb1 peb2 ->
  lc_peb peb1 /\ lc_peb peb2.
Proof.
  intros peb1 peb2 H.
  induction H; intuition.
Qed.

Lemma lc_find_pe : forall peb' pe',
  find_pe peb' pe' ->
  lc_peb peb' /\ lc_pe pe'.
Proof.
  intros peb' pe' H.
  induction H; intuition.
    apply lc_notmatching_peb in H0; intuition.
Qed.

Lemma lc_reduce : forall e e',
 reduce e e' ->
 lc_exp e /\ lc_exp e'.
Proof.
 intros e e' H.
 induction H; intuition.
   apply lc_find_pe in H2; intuition.
   apply lc_find_pe in H2; intuition.
     inversion H3; subst.
     auto with lngen.
Qed.

Lemma cut_lc_ectx : forall D1 D2 e1 e2,
  lc_ectx (D1 ++ (e1,e2) :: D2) ->
  lc_ectx (D1 ++ D2).
Proof.
  intros D1 D2 e1 e2 H.
  generalize dependent D2.
  generalize dependent e1.
  generalize dependent e2.
  induction D1; intros; simpl in *; intuition auto.
    destruct a; intuition eauto.
Qed.

Lemma weakening_lc_ectx : forall D1 D2 D3,
  lc_ectx (D1 ++ D3) ->
  lc_ectx D2 ->
  lc_ectx (D1 ++ D2 ++ D3).
Proof.
  intros D1 D2 D3 H1 H2.
  generalize dependent D2.
  generalize dependent D3.
  induction D1; intros; simpl in *; intuition auto.
    apply app_lc_ectx3; auto.
    destruct a; intuition eauto.
Qed.

Lemma lc_ctx_erase : forall G,
 lc_ctx G ->
 lc_ectx (erase G).
Proof.
  intros. induction G; eauto.
  destruct a.
  inversion H; subst;  simpl; eauto.
  destruct p. simpl in *. 
  destruct H as (J1, (J2, J3)). repeat split; eauto.
Qed. 

Lemma subst_lc_ectx : forall D w x,
    lc_ectx D ->
    lc_exp w ->
    lc_ectx (esubst_ectx w x D).
Proof.
  intros. induction D; eauto.
  destruct a; simpl in *; intuition eauto using esubst_exp_lc_exp.
Qed. 

Hint Extern 1 (lc_ctx ?G) =>
  match goal with
  | H: wfc G |- _ => apply (proj1 (lc_wfc H))
  | H: wfpe G _ _ _ _ |- _ => apply (proj1 (lc_wfpe H))
  | H: wfe G _ _ |- _ => apply (proj1 (lc_wfe H))
  | H: wfk G _ |- _ => apply (proj1 (lc_wfk H))
  | H: wfpt G _ _ _ _ |- _ => apply (proj1 (lc_wfpt H))
  | H: wft G _ _ |- _ => apply (proj1 (lc_wft H))
  | H: ctxeq G _ |- _ => apply (proj1 (lc_ctxeq H))
  | H: ctxeq _ G |- _ => apply (proj2 (lc_ctxeq H))
  end.

Hint Extern 1 (lc_exp ?e) =>
  match goal with
  | H: wfpe _ e _ _ _ |- _ => apply (proj1 (proj2 (lc_wfpe H)))
  | H: wfpe _ _ e _ _ |- _ => apply (proj1 (proj2 (proj2 (lc_wfpe H))))
  | H: wfe _ e _ |- _ => apply (proj1 (proj2 (lc_wfe H)))
  | H: isEq _ e _ |- _ => apply (proj1 (proj2 (lc_isEq H)))
  | H: isEq _ _ e |- _ => apply (proj2 (proj2 (lc_isEq H)))
  end.

Hint Extern 1 (lc_exp ?w) =>
  match goal with
  | H: isEq _ _ (e_Con ?C w) |- _ =>
       let H3 := fresh in
       destruct (lc_isEq H) as [_ [_ H3]]; inversion H3; subst; auto
  end.

Hint Extern 1 (lc_pe ?pe) =>
  match goal with
  | H: wfpe _ _ _ pe _ |- _ => apply (proj1 (proj2 (proj2 (proj2 (lc_wfpe H)))))
  end.

Hint Extern 1 (lc_type ?t) =>
  match goal with
  | H: wfpe _ _ _ _ t |- _ => apply (proj2 (proj2 (proj2 (proj2 (lc_wfpe H)))))
  | H: wfe _ _ t |- _ => apply (proj2 (proj2 (lc_wfe H)))
  | H: wfpt _ _ _ _ t |- _ => apply (proj1 (proj2 (lc_wfpt H)))
  | H: wft _ _ t |- _ => apply (proj1 (proj2 (lc_wft H)))
  | H: typeeq _ t _ |- _ => apply (proj1 (lc_typeeq H))
  | H: typeeq _ _ t |- _ => apply (proj1 (proj2 (lc_typeeq H)))
  end.

Hint Extern 1 (lc_pt ?pt) =>
  match goal with
  | H: wfpt _ _ _ pt _ |- _ => apply (proj1 (proj2 (proj2 (lc_wfpt H))))
  end.

Hint Extern 1 (lc_kind ?k) =>
  match goal with
  | H: wfk _ k |- _ => apply (proj2 (lc_wfk H))
  | H: wfpt _ _ _ _ k |- _ => apply (proj2 (proj2 (lc_wfpt H)))
  | H: wft _ _ k |- _ => apply (proj2 (proj2 (lc_wft H)))
  | H: kindeq _ k _ |- _ => apply (proj1 (lc_kindeq H))
  | H: kindeq _ _ k |- _ => apply (proj1 (proj2 (lc_kindeq H)))
  end.

Hint Extern 1 (lc_ectx ?D) =>
  match goal with
  | H: kindeq D _ _ |- _ => apply (proj2 (proj2 (lc_kindeq H)))
  | H: typeeq D _ _ |- _ => apply (proj2 (proj2 (lc_typeeq H)))
  | H: ctxereq D _ |- _ => apply (proj1 (lc_ctxereq H))
  | H: ctxereq _ D |- _ => apply (proj2 (lc_ctxereq H))
  | H: isEq D _ _ |- _ => apply (proj1 (lc_isEq H))
  end.

Hint Extern 1 (lc_ectx (erase ?G)) =>
  match goal with
  | H: ctxeq G _ |- _ => destruct (lc_ctxeq H); apply lc_ctx_erase; auto
  | H: ctxeq _ G |- _ => destruct (lc_ctxeq H); apply lc_ctx_erase; auto
  | H: lc_ctx G |- _ => apply lc_ctx_erase; auto
  end.

(*************Properties of ctxeq**************)
(* more later after trans*)

Lemma ctxeq_erase : forall G G', ctxeq G G' -> 
  ctxereq (erase G) (erase G').
intros.
induction H. auto. auto.
simpl. auto.
Qed.

(*********** CUT for EQCTX **********************)
    
Definition cut_typeeq_prop D t1 t2 (EQ:typeeq D t1 t2) :=
  forall D1 e1 e2 D2,
    D = (D1 ++ (e1,e2) :: D2) -> 
    isEq D2 e1 e2 ->
    typeeq (D1 ++ D2) t1 t2.
Definition cut_pateq_prop D e u e' u' pt1 pt2 (EQ:pateq D e u e' u' pt1 pt2) :=
  forall D1 e1 e2 D2,
    D = (D1 ++ (e1,e2) :: D2) -> 
    isEq D2 e1 e2 ->
    pateq (D1 ++ D2) e u e' u' pt1 pt2.
Definition cut_patteq_prop D e u e' u' ptt1 ptt2 (EQ:patteq D e u e' u' ptt1 ptt2) :=
  forall D1 e1 e2 D2,
    D = (D1 ++ (e1,e2) :: D2) -> 
    isEq D2 e1 e2 ->
    patteq (D1 ++ D2) e u e' u' ptt1 ptt2.

Lemma cut_eq : (forall D t1 t2 (EQ1:typeeq D t1 t2), cut_typeeq_prop EQ1)
   /\ (forall D e u e' u' pt1 pt2 (EQ2:pateq D e u e' u' pt1 pt2), cut_pateq_prop EQ2)
   /\ (forall D e u e' u' ptt1 ptt2 (EQ3:patteq D e u e' u' ptt1 ptt2), cut_patteq_prop EQ3).
Proof.
(eq_mutind_cases 
  apply eq_mutind with 
   (P := cut_typeeq_prop)
   (P0:= cut_pateq_prop)
   (P1:= cut_patteq_prop) Case); 
    unfold cut_typeeq_prop; unfold cut_pateq_prop; unfold cut_patteq_prop; 
    intros; subst; eauto using  cut_isEq, cut_incon.
  Case "TQ_TRefl".
    apply TQ_TRefl; auto.
      eapply cut_lc_ectx; eauto.
  Case "TQ_URefl".
    apply TQ_URefl; auto.
      apply cut_lc_ectx in l; auto.
  Case "TQ_Red1".
  eapply TQ_Red1; eauto. 
    eapply cut_isEq; eauto. 
    intros x xnotinL. 
      apply i0 in xnotinL.
      rewrite_env (((w, e_Var_f x) :: D1) ++ D2).
      eapply cut_isEq; eauto.
    intros xj xjnotinL. 
      rewrite_env (((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)++D2).
      eapply H; eauto.
      rewrite_env (((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)++(e1,e2)::D2); auto.
  Case "TQ_Red2".
  eapply TQ_Red2; eauto. 
    eapply cut_isEq; eauto. 
    intros x xnotinL. 
      apply i0 in xnotinL.
      rewrite_env (((w, e_Var_f x) :: D1) ++ D2).
      eapply cut_isEq; eauto.
    intros xj xjnotinL. 
      rewrite_env (((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)++D2).
      eapply H; eauto.
      rewrite_env (((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)++(e1,e2)::D2); auto.
  Case "TPQ_Empty".
    apply TPQ_Empty; auto.
      apply cut_lc_ectx in l3; auto.
  Case "TTPQ_Intro".
  eapply TTPQ_Intro; eauto. 
    eapply cut_isEq; eauto. 
    eapply cut_isEq; eauto. 
    intros x xnotinL. 
      rewrite_env (((e, e_Con Ci (e_Var_f x))::(u, open_exp_wrt_exp ui (e_Var_f x))::D1)++D2).
      eapply H; eauto.
      rewrite_env (((e, e_Con Ci  (e_Var_f x))::(u, open_exp_wrt_exp ui (e_Var_f x))::D1)++(e1,e2)::D2); auto.
Qed.

Lemma cut_typeeq : forall D1 D2 e1 e2 t1 t2,
  typeeq (D1 ++ (e1,e2) :: D2) t1 t2 ->
  isEq D2 e1 e2 ->
  typeeq (D1 ++ D2) t1 t2.
intros. 
pose (J := cut_eq).
unfold cut_typeeq_prop in J. 
destruct J. eauto.
Qed.

Lemma cut_kindeq : forall D1 D2 e1 e2 k1 k2,
  kindeq (D1 ++ (e1,e2) :: D2) k1 k2 ->
  isEq D2 e1 e2 ->
  kindeq (D1 ++ D2) k1 k2.
intros. 
remember (D1 ++ (e1,e2) :: D2) as D'. generalize dependent D'.
induction 2; auto; subst.
  apply KQ_Refl; auto.
    apply cut_lc_ectx in H; auto.
  eapply KQ_Pi with (L:=L). 
    eapply cut_typeeq;  eauto.
    intros. eapply H2; eauto.
Qed.

(************** WEAKENING for equality Judgements ****************)

Definition weakening_typeeq_prop D t t' (WF : typeeq D t t') :=
  forall D1 D2 D3,
  D = D1 ++ D3 ->
  lc_ectx D2 ->
  typeeq (D1 ++ D2 ++ D3) t t'.

Definition weakening_pateq_prop D e u e' u' pt pt' (WF : pateq D e u e' u' pt pt') :=
  forall D1 D2 D3,
  D = D1 ++ D3 ->
  lc_ectx D2 ->
  pateq (D1 ++ D2 ++ D3) e u e' u' pt pt'. 

Definition weakening_patteq_prop D e u e' u' ptt ptt' (WF : patteq D e u e' u' ptt ptt') :=
  forall D1 D2 D3,
  D = D1 ++ D3 ->
  lc_ectx D2 ->
  patteq (D1 ++ D2 ++ D3) e u e' u' ptt ptt'. 

Lemma eq_weakening:
  (forall D t t' WF, @weakening_typeeq_prop D t t'  WF) /\
  (forall D e u e' u' pt pt' WF, @weakening_pateq_prop D e u e' u' pt pt' WF) /\
  (forall D e u e' u' ptt ptt' WF, @weakening_patteq_prop D e u e' u' ptt ptt' WF).
(eq_mutind_cases 
  apply eq_mutind with
  (P  := weakening_typeeq_prop) 
  (P0 := weakening_pateq_prop)
  (P1 := weakening_patteq_prop)Case); 
  unfold weakening_typeeq_prop, weakening_pateq_prop, weakening_patteq_prop; intros; subst; 
  eauto using weakening_isEq, weakening_incon.
Case "TQ_TRefl".
  apply TQ_TRefl; auto.
    apply weakening_lc_ectx; auto.
Case "TQ_URefl".
  apply TQ_URefl; auto.
    apply weakening_lc_ectx; auto.
Case "TQ_Red1".
eapply TQ_Red1 with (L := L); eauto. eapply weakening_isEq; eauto.
  intros. rewrite_env (((w, e_Var_f xj) :: D1) ++ D2 ++ D3).
          eapply weakening_isEq; auto. eapply i0; eauto.
  intros. rewrite_env (((e, e_Con Cj (e_Var_f xj)) :: (w, e_Var_f xj) :: D1) ++ D2 ++ D3).
          eapply H; eauto.
Case "TQ_Red2".
eapply TQ_Red2 with (L := L); eauto. eapply weakening_isEq; eauto.
  intros. rewrite_env (((w, e_Var_f xj) :: D1) ++ D2 ++ D3).
          eapply weakening_isEq; auto. eapply i0; eauto.
  intros. rewrite_env (((e, e_Con Cj (e_Var_f xj)) :: (w, e_Var_f xj) :: D1) ++ D2 ++ D3).
          eapply H; eauto.
Case "TPQ_Empty".
  apply TPQ_Empty; auto.
    apply weakening_lc_ectx; auto.
Case "TTPQ_Intro".
eapply TTPQ_Intro with (L := L); eauto. 
  eapply weakening_isEq; auto.
  eapply weakening_isEq; auto.
  intros. rewrite_env (((e, e_Con Ci (e_Var_f xi)) :: (u, open_exp_wrt_exp ui (e_Var_f xi)) :: D1) ++ D2 ++ D3).
  eapply H; eauto.
Qed.

Lemma weakening_typeeq : forall D1 D2 D3 t1 t2, 
  typeeq (D1 ++ D3) t1 t2 -> 
  lc_ectx D2 ->
  typeeq (D1 ++ D2 ++ D3) t1 t2.
intros. 
pose (J := eq_weakening).
unfold weakening_typeeq_prop in J. 
destruct J. eauto.
Qed.

Lemma weakening_kindeq : forall D1 D2 D3 k1 k2, 
  kindeq (D1 ++ D3) k1 k2 -> 
  lc_ectx D2 ->
  kindeq (D1 ++ D2 ++ D3) k1 k2.
intros. 
remember (D1 ++ D3) as D'. generalize dependent D'.
induction 2; subst.
eapply KQ_Refl.
  apply weakening_lc_ectx; auto.
eapply KQ_Pi with (L:=L). 
  eapply weakening_typeeq; eauto.
  intros. eapply H2; eauto.
Qed.

(* Problem with this lemma statement is where do 
   we get ctxeq G0 G0'? Reflexivity requires a 
   wfc, and G0 may only be well-formed when appended 
   to another context. *)
Lemma weakening_ctxeq : forall G1 G2 G' G0 G0',
  ctxeq (G1 ++ G2) G' 
  -> ctxeq G0 G0'
  -> exists G1', exists G2', 
    G' = G1' ++ G2' /\ ctxeq (G1 ++ G0 ++ G2) (G1' ++ G0' ++ G2').
intro G1. induction G1; intros; simpl in *; auto. 
  exists nil. exists G'. simpl. split. auto.
  generalize dependent G0'.
  induction G0; intros. simpl. simpl in H0. inversion H0.  subst. simpl. auto. 
  inversion H0. subst. simpl. apply CQ_Term. apply IHG0. auto.
  rewrite_env (erase G0 ++ erase G2 ++ nil).
  eapply weakening_typeeq; simpl_env; auto.

  Case "Alt". subst. inversion H0. subst. simpl.
  apply CQ_Eq. eapply IHG0. auto. 
  rewrite_env (erase G0 ++ erase G2 ++ nil). 
    apply weakening_isEq; simpl_env; auto.
  rewrite_env (erase G0 ++ erase G2 ++ nil). 
    apply weakening_isEq; simpl_env; auto.
Case "cons".
  inversion H.
  SCase "VarAsn".
    subst. destruct (IHG1 G2 G'0 G0 G0') as [ G1' [ G2' [Eq CEq]]]; eauto. subst. clear IHG1.
    exists (VarAsn (exp * exp) x t' :: G1').
    exists G2'. split. auto. simpl. apply CQ_Term. auto.
    simpl_env. 
    apply weakening_typeeq; simpl_env in H5; auto.
  SCase "AltAsn".
    subst. destruct (IHG1 G2 G'0 G0 G0') as [ G1' [ G2' [Eq CEq]]]; eauto. subst. clear IHG1.
    exists (AltAsn type (e1', e2') :: G1').
    exists G2'. split. auto. simpl. apply CQ_Eq. auto.
    simpl_env. apply weakening_isEq; simpl_env in *; auto.
    simpl_env. apply weakening_isEq; simpl_env in *; auto.
Qed.

(************* Refl Symm Trans for Eq context equiv *****************)

Lemma D_ref: forall D, lc_ectx D -> ctxereq D D.
intros D H. 
induction D; auto. 
  destruct a.
  simpl in H. decompose [and] H. auto using ref_isEq. 
Qed.
Lemma D_sym: forall D1 D2, ctxereq D1 D2 -> ctxereq D2 D1.
induction 1; eauto.
eapply CEQ_Eq; eauto; eapply ctx_conv_isEq; eauto; eapply sym_isEq;eauto.
Qed.
Lemma D_trans: forall D2 D1 D3, ctxereq D1 D2 -> ctxereq D2 D3 -> ctxereq D1 D3.
intro D2. induction D2; intros; inversion H; inversion H0; auto.
subst. inversion H7. subst.
eapply CEQ_Eq; eauto.  eapply trans_isEq. eauto. eapply ctx_conv_isEq. eapply D_sym. eauto. eauto.
eauto using trans_isEq, ctx_conv_isEq, D_sym.
Qed.

(************* CTX CONVERSION for EQUALITY *****************)

Definition ctx_conv_typeeq_prop D t t' (WF : typeeq D t t') :=
  forall D', ctxereq D D' -> typeeq D' t t'.

Definition ctx_conv_pateq_prop D e u e' u' pt pt' (WF : pateq D e u e' u' pt pt') :=
  forall D', ctxereq D D' -> pateq D' e u e' u' pt pt'.

Definition ctx_conv_patteq_prop D e u e' u' ptt ptt' (WF : patteq D e u e' u' ptt ptt') :=
  forall D', ctxereq D D' -> patteq D' e u e' u' ptt ptt'.

Lemma eq_ctx_conv:
  (forall G t t' WF, @ctx_conv_typeeq_prop G t t'  WF) /\
  (forall G e u e' u' pt pt' WF, @ctx_conv_pateq_prop G e u e' u' pt pt' WF) /\
  (forall G e u e' u' ptt ptt' WF, @ctx_conv_patteq_prop G e u e' u' ptt ptt' WF).
(eq_mutind_cases 
  apply eq_mutind with
  (P  := ctx_conv_typeeq_prop) 
  (P0 := ctx_conv_pateq_prop)
  (P1 := ctx_conv_patteq_prop)Case); 
  unfold ctx_conv_typeeq_prop, ctx_conv_pateq_prop, ctx_conv_patteq_prop; intros; subst; 
  eauto using ctx_conv_isEq, ctx_conv_incon, ref_isEq.
Case "TQ_TRefl".
apply TQ_TRefl; auto.
  apply lc_ctxereq in H. decompose [and] H; auto.
Case "TQ_URefl".
apply TQ_URefl; auto.
  apply lc_ctxereq in H. decompose [and] H; auto.
Case "TQ_Red1".
eapply TQ_Red1 with (L := L); eauto.
  eapply ctx_conv_isEq; eauto.
  intros. eapply ctx_conv_isEq with (D1 := (w, e_Var_f xj)::D); auto.
          eapply CEQ_Eq; eauto using ref_isEq.
  intros. eapply H; eauto. 
          eapply CEQ_Eq; eauto using ref_isEq.
            apply ref_isEq; simpl; auto.
            apply ref_isEq; simpl; auto.
Case "TQ_Red2".
eapply TQ_Red2 with (L := L); eauto.
  eapply ctx_conv_isEq; eauto.
  intros. eapply ctx_conv_isEq with (D1 := (w, e_Var_f xj)::D); auto.
          eapply CEQ_Eq; eauto using ref_isEq.
  intros. eapply H; eauto. 
          eapply CEQ_Eq; eauto using ref_isEq.
            apply ref_isEq; simpl; auto.
            apply ref_isEq; simpl; auto.
Case "TPQ_Empty".
apply TPQ_Empty; auto.
  apply lc_ctxereq in H. decompose [and] H; auto.
Case "TTPQ_Intro".
eapply TTPQ_Intro with (L := L); eauto.
  eapply ctx_conv_isEq; eauto.
  eapply ctx_conv_isEq; eauto.
  intros. eapply H; eauto. 
          eapply CEQ_Eq; eauto using ref_isEq, lc_open_ui.
            apply ref_isEq; simpl; eauto using lc_open_ui.
            apply ref_isEq; simpl; eauto using lc_open_ui.
Qed.

Lemma ctx_conv_typeeq : forall D1 D2 t1 t2, 
  typeeq D1 t1 t2 -> ctxereq D1 D2 ->
  typeeq D2 t1 t2.
intros. 
pose (J := eq_ctx_conv).
unfold ctx_conv_typeeq_prop in J. 
destruct J. eauto.
Qed.

Lemma ctx_conv_patteq : forall D D' e u e' u' ptt ptt',
  patteq D e u e' u' ptt ptt' -> 
  ctxereq D D' -> 
  patteq D' e u e' u' ptt ptt'.
intros. 
pose (J := eq_ctx_conv).
unfold ctx_conv_patteq_prop in J. 
decompose [and] J. eauto.
Qed.

Lemma ctx_conv_kindeq : forall D1 D2 k1 k2,
  kindeq D1 k1 k2 -> ctxereq D1 D2 -> 
  kindeq D2 k1 k2.
induction 1; intros.
apply KQ_Refl; auto.
  apply lc_ctxereq in H0. decompose [and] H0; auto.
eapply KQ_Pi with (L := L); eauto using ctx_conv_typeeq.
Qed.


(******************* REFLEXIVITY *****************)

Definition ref_typeeq_prop t (LJ: lc_type t) :=
  forall D, 
    lc_ectx D -> lc_type t -> 
    typeeq D t t.

Definition ref_pateq_prop pt (LJ : lc_pt pt) :=
  forall D e u, 
    lc_ectx D -> lc_exp e -> lc_exp u -> 
    pateq D e u e u pt pt.

Definition ref_patteq_prop ptt (LJ : lc_ptt ptt) :=
  forall D e u, 
    lc_ectx D -> lc_exp e -> lc_exp u -> 
    patteq D e u e u ptt ptt.

Lemma ref_eq:
  (forall t LJ, @ref_typeeq_prop t LJ) /\
  (forall pt LJ, @ref_pateq_prop pt LJ) /\
  (forall ptt LJ, @ref_patteq_prop ptt LJ).
(lc_tp_mutind_cases 
  apply lc_tp_mutind with 
  (P  := ref_typeeq_prop) 
  (P0 := ref_pateq_prop)
  (P1 := ref_patteq_prop) Case);
  unfold ref_typeeq_prop, ref_pateq_prop, ref_patteq_prop; 
  intros; subst; auto.
Case "lc_t_Pi".
  apply TQ_Pi with (L:={});eauto.
Case "lc_t_Sigma".
  apply TQ_Sigma with (L:={});eauto.
Case "lc_t_App".
  eauto using ref_isEq.
Case "lc_t_Case".
  eauto using ref_isEq.
Case "lc_ptt_pattern".
  assert (J:=@datacon_bind C).
  destruct J as [ti [si [T [ui [Binds EQ]]]]]; subst.
  apply TTPQ_Intro with (L:={}) (si:=si) (T:=T) (ui:=ui); auto.
    eauto using ref_isEq.
    eauto using ref_isEq.
    intros xi xiempty.
      apply H; auto.
      simpl. repeat (split;auto).
      apply datacon_bind_lc in Binds.
      inversion Binds; subst.
      assert (J:=@H6 xi).
      inversion J; subst. auto.
Qed.

Lemma ref_typeeq : forall D t, 
  lc_ectx D -> 
  lc_type t -> 
  typeeq D t t.
intros. 
pose (J := ref_eq).
unfold ref_typeeq_prop in J. 
destruct J; eauto.
Qed.

Lemma ref_kindeq : forall D k,
  lc_ectx D -> lc_kind k -> kindeq D k k.
intros. 
induction H0; auto.
  apply KQ_Pi with (L:={}); auto.
   apply ref_typeeq; auto.
Qed.

Lemma ref_ctxereq : forall D1, 
  lc_ectx D1 ->
  ctxereq D1 D1.
  induction D1; intros; auto. 
    destruct a. 
    simpl in H. decompose [and] H.
    eapply CEQ_Eq. auto. eapply ref_isEq; auto.
    eapply ref_isEq; auto.
Qed.

Lemma ref_ctxeq : forall G, 
  lc_ctx G -> ctxeq G G.
intros. induction G; eauto.
 destruct a. simpl in H.  destruct H.
  apply CQ_Term; auto using ref_typeeq.

 destruct p. simpl in H. destruct H as (J1, (J2, J3)).
 eapply CQ_Eq; eauto using ref_isEq.
Qed.

(**************Weakening for ctxeq *********************)
(*moving*)
Lemma ctxeq_app: forall G1 G2 G3,
  lc_ctx G3 ->
  ctxeq G1 G2 ->
 ctxeq (G3 ++ G1) (G3++G2).
Proof.
  intros.
  induction G3; eauto.
  destruct a. 
  simpl in H. inversion H; subst.
  rewrite_env (VarAsn (exp * exp) a t :: (G3 ++ G1)).
  rewrite_env (VarAsn (exp * exp) a t :: (G3 ++ G2)).
  eapply CQ_Term; eauto.
    apply ref_typeeq; auto.
      simpl_env. apply app_lc_ectx3.
        apply lc_ctx_erase; auto.
        apply lc_ctx_erase.
          apply lc_ctxeq in H0. decompose [and] H0; auto.

  rewrite_env (AltAsn type p :: (G3 ++ G1)).
  rewrite_env (AltAsn type p :: (G3 ++ G2)).
  destruct p. inversion H; subst. destruct H2.
  simpl in H. decompose [and] H.
  apply IHG3 in H7. 
  eapply CQ_Eq; eauto; eapply ref_isEq; auto.
Qed.
 

Lemma ctxeq_length : forall G1 G2, ctxeq G1 G2 -> length G1 = length G2.
intros. induction H. auto.
simpl. auto. simpl. auto.
Qed.

Lemma nil_ctxeq_append : forall G2 G1, 
  ctxeq G1 (G2 ++ G1) -> G2 = nil.
intros. destruct G2. auto.
pose (J := ctxeq_length H). simpl in J. 
rewrite app_length in J.
pose (K := succ_plus_discr (length G1) (length G2) J).
contradiction.
Qed.

Lemma ctxeq_eq_app : forall G1 G2 G3 G4, 
  ctxeq G1 G2 -> ctxeq G3 G4 -> 
  ctxeq (G1 ++ G3) (G2 ++ G4).
Proof.
induction 1; intros; auto. 
 
 rewrite_env (VarAsn (exp * exp) x t :: (G ++ G3)).
 rewrite_env (VarAsn (exp * exp) x t' :: (G' ++ G4)).
 eapply CQ_Term; eauto.
 rewrite_env (erase G ++ erase G3 ++nil).
 replace (erase G) with (erase G++nil) in H0; auto.
 eapply weakening_typeeq; eauto. 
   simpl_env; auto.

 rewrite_env (AltAsn type (e1, e2) :: (G ++ G3)).
 rewrite_env (AltAsn type (e1', e2') :: (G' ++ G4)).
 eapply CQ_Eq; eauto.
 replace (erase G) with (erase G ++ nil) in H0; auto.
 rewrite_env (erase G ++ erase G3++nil). apply weakening_isEq; auto.
 simpl_env; auto.
 replace (erase G) with (erase G ++ nil) in H1; auto.
 rewrite_env (erase G ++ erase G3++nil). apply weakening_isEq; auto.
 simpl_env; auto.
Qed.


Lemma ctxeq_concat: forall G G1 G2,
  ctxeq G (G1 ++ G2) -> (* G' = (G1++G2) -> *)
  exists G1', exists G2', 
    G = G1' ++ G2' /\ ctxeq G2' G2.
Proof.
  intros. generalize dependent G. (*generalize dependent G2.*)
  induction G1; intros; eauto.
   exists nil; exists G; split; auto.

   destruct a.
  Case " VarAsn".
   inversion H; subst.
   destruct  (IHG1 G0 H3) as [G1' [G2' (Jeq, Jctxeq)]].
   exists (VarAsn (exp * exp) a t0 :: G1').
    exists G2'; subst; split; eauto.
  
  Case "AltAsn".
    inversion H; subst.
    destruct  (IHG1 G0 H2) as [G1' [G2' (Jeq, Jctxeq)]].
    exists ( AltAsn type (e1, e2) :: G1').
    exists G2'; subst; split; eauto.
Qed.

Lemma ctxeq_impossible : forall G G1 a, 
  ~(ctxeq G (G1++ one a ++ G)).
Proof.
intros G. induction G; intros; simpl; auto.  
 unfold not; intro J; inversion J.
 eapply app_cons_not_nil; eauto.

 intro J. 
 destruct G1. replace (nil ++ a0::a::G) with (a0::a::G) in J; auto.
 inversion J; subst. 
 replace  (VarAsn (exp * exp) x t :: G)
   with  (nil ++ one (VarAsn (exp * exp) x t) ++ G) in H2; auto.
 eapply IHG; eauto.
 replace (AltAsn type (e1, e2) :: G) 
   with (nil++ one (AltAsn type (e1, e2)) ++ G) in H3; eauto.
 eapply IHG; eauto.

 replace ((a1 :: G1) ++ a0 :: a :: G)
   with (a1 :: (G1 ++ a0 :: a :: G)) in J; auto.
 inversion J; subst.
 replace (G1 ++ a0 :: VarAsn (exp * exp) x t :: G)
   with ((G1 ++ one(a0) )++ one (VarAsn (exp * exp) x t)++ G) in H2; eauto.
 eapply IHG; eauto. simpl_env; auto.

 replace (G1 ++ a0 :: AltAsn type (e1, e2) :: G)
   with ((G1 ++ one a0) ++ one (AltAsn type (e1, e2))++ G) in H3; eauto.
 eapply IHG; eauto. simpl_env; auto.
Qed.


Lemma ctxeq_dom_eq: forall G1 G2,
   ctxeq G1 G2 -> dom G1 [=] dom G2.
Proof.
 intros.
induction H; simpl; eauto. solve_uniq.
solve_uniq.
Qed.  

(******** SYMMETRY for EQUALITY ***************)

Definition sym_typeeq_prop D t t' (WF : typeeq D t t') :=
  typeeq D t' t.

Definition sym_pateq_prop D e u e' u' pt pt' (WF : pateq D e u e' u' pt pt') :=
  pateq D e' u' e u pt' pt.

Definition sym_patteq_prop D e u e' u' ptt ptt' (WF : patteq D e u e' u' ptt ptt') :=
  patteq D e' u' e u ptt' ptt.

Lemma eq_sym:
  (forall D t t' WF, @sym_typeeq_prop D t t'  WF) /\
  (forall D e u e' u' pt pt' WF, @sym_pateq_prop D e u e' u' pt pt' WF) /\
  (forall D e u e' u' ptt ptt' WF, @sym_patteq_prop D e u e' u' ptt ptt' WF).
(eq_mutind_cases 
  apply eq_mutind with 
  (P  := sym_typeeq_prop) 
  (P0 := sym_pateq_prop)
  (P1 := sym_patteq_prop) Case);
  unfold sym_typeeq_prop, sym_pateq_prop, sym_patteq_prop; intros; subst; 
  eauto using sym_isEq.
Case "TTPQ_Intro".
  apply TTPQ_Intro with (L:=L) (si:=si) (T:=T) (ui:=ui); auto using sym_isEq.
    intros xi xinotinL.
    apply H in xinotinL.
      apply ctx_conv_typeeq with (D1:=(e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D); auto.
        apply CEQ_Eq; eauto using ref_isEq, lc_open_ui.
          apply CEQ_Eq; eauto using ref_isEq, lc_open_ui.
            apply ref_ctxereq; auto.
          rewrite_env (nil ++ (one (u, open_exp_wrt_exp ui (e_Var_f xi))) ++ D).
          apply weakening_isEq; auto.
          simpl. repeat(split; auto).
            eauto using lc_open_ui.
          apply ref_isEq; eauto using lc_open_ui.
            simpl. repeat(split; auto).
            eauto using lc_open_ui.
Qed.

Lemma sym_typeeq : forall D1 t1 t2, 
  typeeq D1 t1 t2 ->
  typeeq D1 t2 t1.
intros. 
pose (J := eq_sym).
unfold sym_typeeq_prop in J. 
destruct J.  eauto. 
Qed.

Lemma sym_kindeq : forall D1 k1 k2,
  kindeq D1 k1 k2 -> 
  kindeq D1 k2 k1.
induction 1; eauto using sym_typeeq.
Qed.

Lemma sym_ctxereq : forall D1 D2, 
  ctxereq D1 D2 -> ctxereq D2 D1.
induction 1; eauto.
eapply CEQ_Eq; eauto.
eapply  ctx_conv_isEq. eauto. eapply sym_isEq. eauto.
eapply  ctx_conv_isEq. eauto. eapply sym_isEq. eauto.
Qed.


Lemma sym_ctxeq : forall G1 G2, 
  ctxeq G1 G2 -> ctxeq G2 G1.
induction 1; eauto.
eapply CQ_Term; eauto.
eapply sym_typeeq; eauto. eapply ctx_conv_typeeq; eauto. eapply ctxeq_erase. eauto.
eapply CQ_Eq; eauto.
  eapply ctx_conv_isEq; eauto. eapply ctxeq_erase; eauto. eapply sym_isEq; eauto.
  eapply ctx_conv_isEq; eauto. eapply ctxeq_erase; eauto. eapply sym_isEq; eauto.
Qed.

(******** TRANS for EQUALITY ***************)

(******** Induction on Pair of Nats ********)

Inductive PairLt: (nat * nat) -> (nat *nat) -> Prop :=
 | PairOrdL: forall a1 a2 b1 b2, 
    a1 < a2 ->
    PairLt (a1, b1) (a2, b2)
 | PairOrdR: forall a b1 b2, 
    b1 < b2 ->
    PairLt (a, b1) (a, b2)
.

Hint Constructors PairLt. 

Theorem pairlt_Acc: forall a:(nat*nat), Acc PairLt a.
Proof.
 intros a. 
  destruct a. generalize dependent n0.
  induction n. 

  Case "n=0". 
   intro m. induction m.
   SCase "m = 0".
   apply Acc_intro. 
   intros. inversion H. inversion H2. inversion H2.
   SCase "m= S(m')".
   apply Acc_intro. 
   intros. inversion H; subst.  inversion H2.
   destruct (le_lt_or_eq b1 m (lt_n_Sm_le b1 m H2)); subst; eauto.
   eapply Acc_inv; eauto.

  Case "n=S n'".
   intro m. induction m.
   SCase "m=0". 
   apply Acc_intro. 
   intros. inversion H; subst. 
   destruct (le_lt_or_eq a1 n (lt_n_Sm_le a1 n H2)); subst; eauto.
   eapply Acc_inv with (x:= (n, 0)); eauto.
   inversion H2.
   SCase "m=S m'".
   apply Acc_intro. intros. 
   inversion H; subst.  
   destruct (le_lt_or_eq a1 n (lt_n_Sm_le a1 n H2)); subst; eauto.
   eapply Acc_inv; eauto.
   destruct (le_lt_or_eq b1 m (lt_n_Sm_le b1 m H2)); subst; eauto.
   eapply Acc_inv; eauto.  
Defined.   

Theorem well_founded_pairlt: well_founded PairLt.
Proof.
  unfold well_founded. intros.
  apply (pairlt_Acc a).
Defined.

Lemma pair_lt_ind: 
 forall a (P:nat*nat -> Prop), 
   (forall a, (forall b, PairLt b a -> P b) -> P a) -> P a.
Proof.
  intros.
  elim (well_founded_pairlt a); auto.
Qed.

Require Import Coq.Arith.Max.
Require Import Coq.Arith.Compare.
Require Import Omega.

(******** type eq indexed by steps ********)

Inductive dtypeeq : nat -> ectx -> type -> type -> Prop :=    (* defn typeeq *)
 | DTQ_InCon : forall (D:ectx) (t t':type),
     lc_type t ->
     lc_type t' ->
     incon  D  ->
     dtypeeq 0 D t t'
 | DTQ_TRefl : forall (D:ectx) (T:contpvar),
     (lc_ectx D)  ->
     dtypeeq 0 D (t_Con T) (t_Con T)
 | DTQ_URefl : forall (D:ectx),
      (lc_ectx D)  ->
     dtypeeq 0 D t_Unit t_Unit
 | DTQ_Pi : forall n m (L:vars) (D:ectx) (t1 t2 t1' t2':type),
     dtypeeq n D t1 t1' ->
      ( forall x , x \notin  L  -> dtypeeq m D  ( open_type_wrt_exp t2 (e_Var_f x) )   ( open_type_wrt_exp t2' (e_Var_f x) )  )  ->
     dtypeeq (S (max n m)) D (t_Pi t1 t2) (t_Pi t1' t2')
 | DTQ_Sigma : forall n m (L:vars) (D:ectx) (t1 t2 t1' t2':type),
     dtypeeq n D t1 t1' ->
      ( forall x , x \notin  L  -> dtypeeq m D  ( open_type_wrt_exp t2 (e_Var_f x) )   ( open_type_wrt_exp t2' (e_Var_f x) )  )  ->
     dtypeeq (S (max n m)) D (t_Sigma t1 t2) (t_Sigma t1' t2')
 | DTQ_App : forall n (D:ectx) (t:type) (e:exp) (t':type) (e':exp),
     dtypeeq n D t t' ->
      isEq  D   e   e'  ->
     dtypeeq (S n) D (t_App t e) (t_App t' e')
 | DTQ_Red1 : forall n (L:vars) (D:ectx) (e:exp) (T:contpvar) (u:exp) (pt5:pt) (t:type) (Cj:contmvar) (w:exp) (sj:type) (uj:exp) (tj:type),
      isEq  D   e   (e_Con Cj w)  ->
     pure w ->
      (binds  Cj  (bind_t   (t_Pi sj (t_App (t_Con T) uj))  ) Signature)  ->
      ( forall xj , xj \notin  L  ->  isEq    ( (  w  ,  (e_Var_f xj)  ) ::  D  )     u    ( open_exp_wrt_exp uj (e_Var_f xj) )   )  ->
     find_pt (ptt_pattern Cj tj) pt5 ->
      ( forall xj , xj \notin  L  -> dtypeeq  n ( (  e  ,   (e_Con Cj (e_Var_f xj))   ) ::   ( (  w  ,  (e_Var_f xj)  ) ::  D  )   )   ( open_type_wrt_exp tj (e_Var_f xj) )  t )  ->
     dtypeeq (S n) D (t_Case e T u pt5) t
 | DTQ_Red2 : forall n (L:vars) (D:ectx) (t:type) (e:exp) (T:contpvar) (u:exp) (pt5:pt) (Cj:contmvar) (w:exp) (sj:type) (uj:exp) (tj:type),
      isEq  D   e   (e_Con Cj w)  ->
     pure w ->
      (binds  Cj  (bind_t   (t_Pi sj (t_App (t_Con T) uj))  ) Signature)  ->
      ( forall xj , xj \notin  L  ->  isEq    ( (  w  ,  (e_Var_f xj)  ) ::  D  )     u    ( open_exp_wrt_exp uj (e_Var_f xj) )   )  ->
     find_pt (ptt_pattern Cj tj) pt5 ->
      ( forall xj , xj \notin  L  -> dtypeeq  n ( (  e  ,   (e_Con Cj (e_Var_f xj))   ) ::   ( (  w  ,  (e_Var_f xj)  ) ::  D  )   )  t  ( open_type_wrt_exp tj (e_Var_f xj) )  )  ->
     dtypeeq (S n) D t (t_Case e T u pt5)
 | DTQ_Case : forall n (D:ectx) (e:exp) (T:contpvar) (u:exp) (pt5:pt) (e' u':exp) (pt':pt),
      isEq  D   e   e'  ->
      isEq  D   u   u'  ->
     dpateq n D e u e' u' pt5 pt' ->
     dtypeeq (S n) D (t_Case e T u pt5) (t_Case e' T u' pt')
with dpateq : nat -> ectx -> exp -> exp -> exp -> exp -> pt -> pt -> Prop :=    (* defn pateq *)
 | DTPQ_Empty : forall (D:ectx) (e u e' u':exp),
     lc_exp e ->
     lc_exp u ->
     lc_exp e' ->
     lc_exp u' ->
     (lc_ectx D)  ->
     dpateq 0 D e u e' u' pt_empty pt_empty
 | DTPQ_Cons : forall n m (D:ectx) (e u e' u':exp) (ptt5:ptt) (pt5:pt) (ptt':ptt) (pt':pt),
     dpatteq n D e u e' u' ptt5 ptt' ->
     dpateq m D e u e' u' pt5 pt' ->
     dpateq (S (max n m)) D e u e' u' (pt_cons ptt5 pt5) (pt_cons ptt' pt')
with dpatteq : nat -> ectx -> exp -> exp -> exp -> exp -> ptt -> ptt -> Prop :=    (* defn patteq *)
 | DTTPQ_Intro : forall n (L:vars) (D:ectx) (e u e' u':exp) (Ci:contmvar) (ti ti' si:type) (T:contpvar) (ui:exp),
      (binds  Ci  (bind_t   (t_Pi si (t_App (t_Con T) ui))  ) Signature)  ->
      isEq  D   e   e'  ->
      isEq  D   u   u'  ->
      ( forall xi , xi \notin  L  -> dtypeeq  n ( (  e  ,  (e_Con Ci (e_Var_f xi))  ) ::   ( (  u  ,   ( open_exp_wrt_exp ui (e_Var_f xi) )   ) ::  D  )   )   ( open_type_wrt_exp ti (e_Var_f xi) )   ( open_type_wrt_exp ti' (e_Var_f xi) )  )  ->
     dpatteq (S n) D e u e' u' (ptt_pattern Ci ti) (ptt_pattern Ci ti').

Tactic Notation "dtypeeq_ind_cases" tactic(first) tactic(c) :=
  first;
  [ c "DTQ_InCon" | c "DTQ_TRefl" | c "DTQ_URefl" | c "DTQ_Pi" |
    c "DTQ_Sigma" | c "DTQ_App" | c "DTQ_Red1" |
    c "DTQ_Red2" | c "DTQ_Case" ].

Hint Constructors dtypeeq dpateq dpatteq.

Scheme dtypeeq_ind' := Induction for dtypeeq Sort Prop
  with dpateq_ind' := Induction for dpateq Sort Prop
  with dpatteq_ind' := Induction for dpatteq Sort Prop.

Combined Scheme deq_mutind from dtypeeq_ind', dpateq_ind', dpatteq_ind'.

Tactic Notation "deq_mutind_cases" tactic(first) tactic(c) :=
  first;
  [ c "DTQ_InCon" | c "DTQ_TRefl" | c "DTQ_URefl" | c "DTQ_Pi" |
    c "DTQ_Sigma" | c "DTQ_App" | c "DTQ_Red1" |
    c "DTQ_Red2" | c "DTQ_Case" | c "DTPQ_Empty" |
    c "DTPQ_Cons" | c "DTTPQ_Intro" ].


Definition lc_dtypeeq_prop n D t t' (WF : dtypeeq n D t t') :=
  lc_type t /\ lc_type t' /\ lc_ectx D.
Definition lc_dpateq_prop n D e u e' u' pt pt' (WF : dpateq n D e u e' u' pt pt') :=
  lc_pt pt /\ lc_pt pt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
Definition lc_dpatteq_prop n D e u e' u' ptt ptt' (WF : dpatteq n D e u e' u' ptt ptt') :=
  lc_ptt ptt /\ lc_ptt ptt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
Lemma lc_deq:
  (forall n D t t' WF, @lc_dtypeeq_prop n D t t' WF) /\
  (forall n D e u e' u' pt pt' WF, @lc_dpateq_prop n D e u e' u' pt pt' WF) /\
  (forall n D e u e' u' ptt ptt' WF, @lc_dpatteq_prop n D e u e' u' ptt ptt' WF).
(deq_mutind_cases 
  apply deq_mutind with 
    (P  := lc_dtypeeq_prop) 
    (P0 := lc_dpateq_prop) 
    (P1 := lc_dpatteq_prop) Case); 
  unfold lc_dtypeeq_prop, lc_dpateq_prop, lc_dpatteq_prop; 
  intros; subst; simpl; intuition;
  eauto using lc_isEq.
Case "DTQ_InCon".
  inversion i; subst.
  apply lc_isEq in H1. decompose [and] H1; auto.
Case "DTQ_Pi".
    SCase "lc_t_Pi".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
    SCase "lc_t_Pi'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Pi_exists with (x1:=x); auto.
Case "DTQ_Sigma".
    SCase "lc_t_Sigma".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
    SCase "lc_t_Sigma'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H0 in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Sigma_exists with (x1:=x); auto.
Case "DTQ_Red1".
    SCase "lc_t_case".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Case; auto.
        apply lc_isEq in i. decompose [and] i; auto.

        assert (x `notin` L) as xnotinL. auto.
        apply i0 in xnotinL.
        apply lc_isEq in xnotinL. decompose [and] xnotinL; auto. 

        apply lc_find_pt in f. decompose [and] f; auto.
    SCase "lc_t".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]]; auto.
Case "DTQ_Red2".
    SCase "lc_t".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]]; auto.
    SCase "lc_t_case".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_t_Case; auto.
        apply lc_isEq in i. decompose [and] i; auto.

        assert (x `notin` L) as xnotinL. auto.
        apply i0 in xnotinL.
        apply lc_isEq in xnotinL. decompose [and] xnotinL; auto. 

        apply lc_find_pt in f. decompose [and] f; auto.
Case "DTTPQ_Intro".
    SCase "lc_ptt_pattern".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_ptt_pattern_exists with (x1:=x); auto.
    SCase "lc_ptt_pattern'".
      pick fresh x.
      assert (x `notin` L) as xnotinL. auto.
      apply H in xnotinL.
      destruct xnotinL as [J1 [J2 J3]].
      apply lc_ptt_pattern_exists with (x1:=x); auto.
Qed.

Lemma lc_dtypeeq: forall n D t t',
  dtypeeq n D t t' -> 
  lc_type t /\ lc_type t' /\ lc_ectx D.
intros. 
pose (J := lc_deq).
unfold lc_dtypeeq_prop in J. 
decompose [and] J; eauto.
Qed.

Lemma lc_dpateq: forall n D e u e' u' pt pt',
  dpateq n D e u e' u' pt pt' ->
  lc_pt pt /\ lc_pt pt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
intros. 
pose (J := lc_deq).
unfold lc_dpateq_prop in J. 
decompose [and] J; eauto.
Qed.

Lemma lc_dpatteq: forall n D e u e' u' ptt ptt',
  dpatteq n D e u e' u' ptt ptt' ->
  lc_ptt ptt /\ lc_ptt ptt' /\ lc_exp e /\ lc_exp u /\ lc_exp e' /\ lc_exp u' /\ lc_ectx D.
intros. 
pose (J := lc_deq).
unfold lc_dpatteq_prop in J. 
decompose [and] J; eauto.
Qed.

Hint Extern 1 (lc_type ?t) =>
  match goal with
  | H: dtypeeq _ _ t _ |- _ => apply (proj1 (lc_dtypeeq H))
  | H: dtypeeq _ _ _ t |- _ => apply (proj1 (proj2 (lc_dtypeeq H)))
  end.

Hint Extern 1 (lc_ectx ?D) =>
  match goal with
  | H: dtypeeq _ D _ _ |- _ => apply (proj2 (proj2 (lc_dtypeeq H)))
  end.

Lemma dpateq_inversion_left : forall n D e u e' u' pt pt' Ci ti si T ui,
  dpateq n D e u e' u' pt pt' ->
  find_pt (ptt_pattern Ci ti) pt ->
  (binds  Ci  (bind_t   (t_Pi si (t_App (t_Con T) ui))  ) Signature)  ->
  exists ti', exists m,
    S m <= n /\
    find_pt (ptt_pattern Ci ti') pt' /\
    dpatteq (S m) D e u e' u' (ptt_pattern Ci ti) (ptt_pattern Ci ti').
Proof.
  intros n D e u e' u' pt pt' Ci ti si T ui H1 H2 H3.
  generalize dependent Ci.
  generalize dependent ti.
  generalize dependent si.
  generalize dependent T.
  generalize dependent ui.
  induction H1; intros; subst.
    inversion H4.
    inversion H2; subst.
    Case "pt_here".
      inversion H; subst.    
      exists ti'. exists n0.
      split.
        assert (J:=@le_max_l (S n0) m). omega.
      split.
        apply pt_here.
          apply lc_dpatteq in H. decompose [and] H; auto.
          apply lc_dpateq in H1. decompose [and] H1; auto.
        eapply DTTPQ_Intro; eauto.
    Case "pt_there".
      apply IHdpateq with (ui:=ui)(T:=T)(si:=si) in H6; auto.
        destruct H6 as [ti' [m0 [EQ [Hfind Hpateq]]]].
        exists ti'. exists m0.
        split; auto.
          assert (J:=@le_max_r n m). omega.
        split; auto.
        apply pt_there; auto.
          inversion H; subst.
          inversion H7; subst.
          apply notmatching_ptt_intro; auto.
            apply lc_dpatteq in H. decompose [and] H; auto.
            apply lc_dpatteq in Hpateq. decompose [and] Hpateq; auto.
Qed.

Lemma dpateq_inversion_right : forall n D e u e' u' pt pt' Ci ti' si T ui,
  dpateq n D e u e' u' pt pt' ->
  find_pt (ptt_pattern Ci ti') pt' ->
  (binds  Ci  (bind_t   (t_Pi si (t_App (t_Con T) ui))  ) Signature)  ->
  exists ti, exists m,
    S m <= n /\
    find_pt (ptt_pattern Ci ti) pt /\
    dpatteq (S m) D e u e' u' (ptt_pattern Ci ti) (ptt_pattern Ci ti').
Proof.
  intros n D e u e' u' pt pt' Ci ti' si T ui H1 H2 H3.
  generalize dependent Ci.
  generalize dependent ti'.
  generalize dependent si.
  generalize dependent T.
  generalize dependent ui.
  induction H1; intros; subst.
    inversion H4.
    inversion H2; subst.
    Case "pt_here".
      inversion H; subst.    
      exists ti. exists n0.
      split.
        assert (J:=@le_max_l (S n0) m). omega.
      split.
        apply pt_here.
          apply lc_dpatteq in H. decompose [and] H; auto.
          apply lc_dpateq in H1. decompose [and] H1; auto.
        eapply DTTPQ_Intro; eauto.
    Case "pt_there".
      apply IHdpateq with (ui:=ui)(T:=T)(si:=si) in H6; auto.
        destruct H6 as [ti [m0 [EQ [Hfind Hpateq]]]].
        exists ti. exists m0.
        split; auto.
          assert (J:=@le_max_r n m). omega.
        split; auto.
        apply pt_there; auto.
          inversion H; subst.
          inversion H7; subst.
          apply notmatching_ptt_intro; auto.
            apply lc_dpatteq in H. decompose [and] H; auto.
            apply lc_dpatteq in Hpateq. decompose [and] Hpateq; auto.
Qed.

Definition cut_dtypeeq_prop n := 
  forall D1 D2 t1 t2 e1 e2,
    dtypeeq n (D1 ++ (e1,e2) :: D2) t1 t2 ->
    isEq D2 e1 e2 ->
    dtypeeq n (D1 ++ D2) t1 t2.
Definition cut_dpateq_prop n := 
  forall D1 D2 e u e' u' pt1 pt2 e1 e2,
    dpateq n (D1 ++ (e1,e2) :: D2) e u e' u' pt1 pt2 ->
    isEq D2 e1 e2 ->
    dpateq n (D1 ++ D2) e u e' u' pt1 pt2.
Definition cut_dpatteq_prop n := 
  forall D1 D2 e u e' u' ptt1 ptt2 e1 e2,
    dpatteq n (D1 ++ (e1,e2) :: D2) e u e' u' ptt1 ptt2 ->  
    isEq D2 e1 e2 ->
    dpatteq n (D1 ++ D2) e u e' u' ptt1 ptt2.

Definition cut_deq_P (n :nat) :=
  cut_dtypeeq_prop n /\
  cut_dpateq_prop n /\
  cut_dpatteq_prop n.
 
Lemma cut_deq: forall n,
  cut_deq_P n. 
Proof.
 intro n.
 apply lt_wf_ind. clear n.
 unfold cut_deq_P, cut_dtypeeq_prop, cut_dpateq_prop, cut_dpatteq_prop.
 intros n IH.
 repeat(split).
 Case "cut_dtypeeq_prop".
   intros D1 D2 t1 t2 e1 e2 Hcut HisEq.
   (dtypeeq_ind_cases (inversion Hcut) SCase); subst; eauto using cut_incon.
  SCase "DTQ_TRefl". 
    apply DTQ_TRefl.
      apply cut_lc_ectx in H; auto.
  SCase "DTQ_URefl".
    apply DTQ_URefl.
      apply cut_lc_ectx in H; auto.
  SCase "DTQ_Pi".
    apply DTQ_Pi with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 t0 t1' e1 e2 H HisEq); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D1 D2 
                          (open_type_wrt_exp t3 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          e1 e2 xnotinL HisEq); auto.
  SCase "DTQ_Sigma".
    apply DTQ_Sigma with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 t0 t1' e1 e2 H HisEq); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D1 D2 
                          (open_type_wrt_exp t3 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          e1 e2 xnotinL HisEq); auto.
  SCase "DTQ_App".
    apply DTQ_App; auto.
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 t t' e1 e2 H HisEq); auto.

      eauto using cut_isEq.
  SCase "DTQ_Red1".
    eapply DTQ_Red1 with (L:=L); eauto. 
      eapply cut_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        rewrite_env (((w, e_Var_f xj) :: D1) ++ D2).
        eapply cut_isEq; eauto.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) 
                          D2 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          t2 
                          e1 
                          e2 
                          xjnotinL 
                          HisEq); auto.
  SCase "DTQ_Red2".
    eapply DTQ_Red2 with (L:=L); eauto. 
      eapply cut_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        rewrite_env (((w, e_Var_f xj) :: D1) ++ D2).
        eapply cut_isEq; eauto.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) 
                          D2 
                          t1 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          e1 
                          e2 
                          xjnotinL 
                          HisEq); auto.
  SCase "DTQ_Case".
    eapply DTQ_Case; eauto. 
      eapply cut_isEq; eauto. 
      eapply cut_isEq; eauto. 

      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D1
                        D2
                        e u e' u' 
                        pt5 
                        pt'
                        e1 
                        e2 
                        H1
                        HisEq); auto.
 Case "cut_dpateq_prop".
   intros D1 D2 e u e' u' pt1 pt2 e1 e2 Hcut HisEq.
   inversion Hcut; subst; eauto using cut_incon.
  SCase "DTPQ_Empty".
    apply DTPQ_Empty; auto.
      apply cut_lc_ectx in H3; auto.
  SCase "DTPQ_Cons".
    apply DTPQ_Cons; eauto.
      assert (n0 < S (max n0 m)) as IH1.
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J2.
      assert (Hresult3:=J3 
                        D1
                        D2
                        e u e' u' 
                        ptt5 
                        ptt'
                        e1 
                        e2 
                        H
                        HisEq); auto.

      assert (m < S (max n0 m)) as IH2.
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D1
                        D2
                        e u e' u' 
                        pt5 
                        pt'
                        e1 
                        e2 
                        H0
                        HisEq); auto.
 Case "cut_dpatteq_prop".
   intros D1 D2 e u e' u' ptt1 ptt2 e1 e2 Hcut HisEq.
   inversion Hcut; subst; eauto using cut_incon.
  SCase "DTTPQ_Intro".
    eapply DTTPQ_Intro with (L:=L); eauto. 
      eapply cut_isEq; eauto. 
      eapply cut_isEq; eauto. 
      intros xi xinotinL. 
      apply H2 in xinotinL. 
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 
                        ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D1) 
                        D2 
                        (open_type_wrt_exp ti (e_Var_f xi)) 
                        (open_type_wrt_exp ti' (e_Var_f xi)) 
                        e1 
                        e2 
                        xinotinL
                        HisEq); auto.
Qed.

Lemma cut_dtypeeq : forall n D1 D2 e1 e2 t1 t2,
  dtypeeq n (D1 ++ (e1,e2) :: D2) t1 t2 ->
  isEq D2 e1 e2 ->
  dtypeeq n (D1 ++ D2) t1 t2.
intros. 
pose (J := cut_deq n).
unfold cut_dtypeeq_prop in J. 
destruct J. eauto.
Qed.

Definition weakening_dtypeeq_prop n :=
  forall D1 D2 D3 t t',
    dtypeeq n (D1 ++ D3) t t' ->
    lc_ectx D2 ->
    dtypeeq n (D1 ++ D2 ++ D3) t t'.

Definition weakening_dpateq_prop n :=
  forall D1 D2 D3 e u e' u' pt pt',
    dpateq n (D1 ++ D3) e u e' u' pt pt' ->
    lc_ectx D2 ->
    dpateq n (D1 ++ D2 ++ D3) e u e' u' pt pt'. 

Definition weakening_dpatteq_prop n :=
  forall D1 D2 D3 e u e' u' ptt ptt',
    dpatteq n (D1 ++ D3) e u e' u' ptt ptt' ->
    lc_ectx D2 ->
    dpatteq n (D1 ++ D2 ++ D3) e u e' u' ptt ptt'. 

Definition weakening_deq_P (n :nat) :=
  weakening_dtypeeq_prop n /\
  weakening_dpateq_prop n /\
  weakening_dpatteq_prop n.
 
Lemma weakening_deq: forall n,
  weakening_deq_P n. 
Proof.
 intro n.
 apply lt_wf_ind. clear n.
 unfold weakening_deq_P, weakening_dtypeeq_prop, weakening_dpateq_prop, weakening_dpatteq_prop.
 intros n IH.
 repeat(split).
 Case "weakening_dtypeeq_prop".
   intros D1 D2 D3 t1 t2 Hweakening LD.
   (dtypeeq_ind_cases (inversion Hweakening) SCase); subst; eauto using weakening_incon.
  SCase "DTQ_TRefl".
    apply DTQ_TRefl.  
      apply weakening_lc_ectx; auto.
  SCase "DTQ_URefl".
    apply DTQ_URefl.
      apply weakening_lc_ectx; auto.
  SCase "DTQ_Pi".
    apply DTQ_Pi with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 D3 t0 t1' H); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D1 D2 D3
                          (open_type_wrt_exp t3 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          xnotinL); auto.
  SCase "DTQ_Sigma".
    apply DTQ_Sigma with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 D3 t0 t1' H); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D1 D2 D3
                          (open_type_wrt_exp t3 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          xnotinL); auto.
  SCase "DTQ_App".
    apply DTQ_App; auto.
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D1 D2 D3 t t' H); auto.

      eauto using weakening_isEq.
  SCase "DTQ_Red1".
    eapply DTQ_Red1 with (L:=L); eauto. 
      eapply weakening_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        rewrite_env (((w, e_Var_f xj) :: D1) ++ D2 ++ D3).
        eapply weakening_isEq; eauto.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) 
                          D2 
                          D3
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          t2 
                          xjnotinL); auto.
  SCase "DTQ_Red2".
    eapply DTQ_Red2 with (L:=L); eauto. 
      eapply weakening_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        rewrite_env (((w, e_Var_f xj) :: D1) ++ D2 ++ D3).
        eapply weakening_isEq; eauto.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) 
                          D2 
                          D3
                          t1 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          xjnotinL); auto.
  SCase "DTQ_Case".
    eapply DTQ_Case; eauto. 
      eapply weakening_isEq; eauto. 
      eapply weakening_isEq; eauto. 

      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D1
                        D2
                        D3
                        e u e' u' 
                        pt5 
                        pt'
                        H1); auto.
 Case "cut_dpateq_prop".
   intros D1 D2 D3 e u e' u' pt1 pt2 Hweakening LD.
   inversion Hweakening; subst; eauto using weakening_incon.
  SCase "DTPQ_Empty".
    apply DTPQ_Empty; auto. 
      apply weakening_lc_ectx; auto.
  SCase "DTPQ_Cons".
    apply DTPQ_Cons; eauto.
      assert (n0 < S (max n0 m)) as IH1.
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J2.
      assert (Hresult3:=J3 
                        D1
                        D2
                        D3
                        e u e' u' 
                        ptt5 
                        ptt'
                        H); auto.

      assert (m < S (max n0 m)) as IH2.
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D1
                        D2
                        D3
                        e u e' u' 
                        pt5 
                        pt'
                        H0); auto.
 Case "cut_dpatteq_prop".
   intros D1 D2 D3 e u e' u' ptt1 ptt2 Hweakening LD.
   inversion Hweakening; subst; eauto using weakening_incon.
  SCase "DTTPQ_Intro".
    eapply DTTPQ_Intro with (L:=L); eauto. 
      eapply weakening_isEq; eauto. 
      eapply weakening_isEq; eauto. 
      intros xi xinotinL. 
      apply H2 in xinotinL. 
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 
                        ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D1) 
                        D2 
                        D3
                        (open_type_wrt_exp ti (e_Var_f xi)) 
                        (open_type_wrt_exp ti' (e_Var_f xi)) 
                        xinotinL); auto.
Qed.

Lemma weakening_dtypeeq : forall n D1 D2 D3 t t',
    dtypeeq n (D1 ++ D3) t t' ->
    lc_ectx D2 ->
    dtypeeq n (D1 ++ D2 ++ D3) t t'.
intros. 
pose (J := weakening_deq n).
unfold weakening_dtypeeq_prop in J. 
destruct J. eauto.
Qed.

Definition ctx_conv_dtypeeq_prop n :=
  forall D D' t t',
    dtypeeq n D t t' ->
    ctxereq D D' -> 
    dtypeeq n D' t t'.

Definition ctx_conv_dpateq_prop n :=
  forall D D' e u e' u' pt pt',
    dpateq n D e u e' u' pt pt' ->
    ctxereq D D' -> 
    dpateq n D' e u e' u' pt pt'.

Definition ctx_conv_dpatteq_prop n :=
  forall D D' e u e' u' ptt ptt',
    dpatteq n D e u e' u' ptt ptt' ->
    ctxereq D D' -> 
    dpatteq n D' e u e' u' ptt ptt'.

Definition ctx_conv_deq_P (n :nat) :=
  ctx_conv_dtypeeq_prop n /\
  ctx_conv_dpateq_prop n /\
  ctx_conv_dpatteq_prop n.
 
Lemma ctx_conv_deq: forall n,
  ctx_conv_deq_P n. 
Proof.
 intro n.
 apply lt_wf_ind. clear n.
 unfold ctx_conv_deq_P, ctx_conv_dtypeeq_prop, ctx_conv_dpateq_prop, ctx_conv_dpatteq_prop.
 intros n IH.
 repeat(split).
 Case "ctx_conv_dtypeeq_prop".
   intros D D' t t' Hctx_conv Hctxeq.
   (dtypeeq_ind_cases (inversion Hctx_conv) SCase); subst; eauto using ctx_conv_incon.
  SCase "DTQ_Pi".
    apply DTQ_Pi with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D D' t1 t1' H Hctxeq); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D D'
                          (open_type_wrt_exp t2 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          xnotinL Hctxeq); auto.
  SCase "DTQ_Sigma".
    apply DTQ_Sigma with (L:=L); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D D' t1 t1' H Hctxeq); auto.

      intros x xnotinL.
      apply H0 in xnotinL.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D D'
                          (open_type_wrt_exp t2 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          xnotinL); auto.
  SCase "DTQ_App".
    apply DTQ_App; auto.
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D D' t0 t'0 H Hctxeq); auto.

      eauto using ctx_conv_isEq.
  SCase "DTQ_Red1".
    eapply DTQ_Red1 with (L:=L); eauto. 
      eapply ctx_conv_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        eapply ctx_conv_isEq; eauto.
          apply CEQ_Eq; auto using ref_isEq.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D)
                        ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D')) as Hctxeq'.
          apply CEQ_Eq; auto.
            apply CEQ_Eq; auto using ref_isEq.
            apply ref_isEq; simpl; auto.
            apply ref_isEq; simpl; auto.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D) 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D') 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          t' 
                          xjnotinL
                          Hctxeq'); auto.
  SCase "DTQ_Red2".
    eapply DTQ_Red2 with (L:=L); eauto. 
      eapply ctx_conv_isEq; eauto. 
      intros xj xjnotinL. 
        apply H2 in xjnotinL.
        eapply ctx_conv_isEq; eauto.
          apply CEQ_Eq; auto using ref_isEq.
      intros xj xjnotinL. 
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D)
                        ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D')) as Hctxeq'.
          apply CEQ_Eq; auto.
            apply CEQ_Eq; auto using ref_isEq.
            apply ref_isEq; simpl; auto.
            apply ref_isEq; simpl; auto.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D) 
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D') 
                          t 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          xjnotinL
                          Hctxeq'); auto.
  SCase "DTQ_Case".
    eapply DTQ_Case; eauto. 
      eapply ctx_conv_isEq; eauto. 
      eapply ctx_conv_isEq; eauto. 

      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D
                        D'
                        e u e' u' 
                        pt5
                        pt'
                        H1 Hctxeq); auto.
 Case "ctx_conv_dpateq_prop".
   intros D D' e u e' u' pt1 pt2 Hctx_conv Hctxeq.
   inversion Hctx_conv; subst; eauto using ctx_conv_incon.
  SCase "DTPQ_Cons".
    apply DTPQ_Cons; eauto.
      assert (n0 < S (max n0 m)) as IH1.
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J2.
      assert (Hresult3:=J3 
                        D
                        D'
                        e u e' u' 
                        ptt5 
                        ptt'
                        H Hctxeq); auto.

      assert (m < S (max n0 m)) as IH2.
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D
                        D'
                        e u e' u' 
                        pt5 
                        pt'
                        H0 Hctxeq); auto.
 Case "ctx_conv_dpatteq_prop".
   intros D D' e u e' u' ptt1 ptt2 Hctx_conv Hctxeq.
   inversion Hctx_conv; subst; eauto using ctx_conv_incon.
  SCase "DTTPQ_Intro".
    eapply DTTPQ_Intro with (L:=L); eauto. 
      eapply ctx_conv_isEq; eauto. 
      eapply ctx_conv_isEq; eauto. 
      intros xi xinotinL. 
      apply H2 in xinotinL. 
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (ctxereq ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D)
                      ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D')) as Hctxeq'.
        apply CEQ_Eq; auto using ref_isEq.
          apply CEQ_Eq; auto using ref_isEq.
            apply ref_isEq; simpl; eauto using lc_open_ui.
            apply ref_isEq; simpl; eauto using lc_open_ui.
            apply ref_isEq; simpl; eauto using lc_open_ui.
      assert (Hresult1:=J1 
                        ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D) 
                        ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D') 
                        (open_type_wrt_exp ti (e_Var_f xi)) 
                        (open_type_wrt_exp ti' (e_Var_f xi)) 
                        xinotinL Hctxeq'); auto.
Qed.

Lemma ctx_conv_dtypeeq: forall n D D' t t',
    dtypeeq n D t t' ->
    ctxereq D D' -> 
    dtypeeq n D' t t'.
intros. 
pose (J := ctx_conv_deq n).
unfold ctx_conv_dtypeeq_prop in J. 
destruct J. eauto.
Qed.

Lemma notin_efv_exp_open_exp_wrt_exp_rec_mut : 
(forall e x y kk,
  x `notin` efv_exp e ->
  x <> y ->
  x `notin` efv_exp (open_exp_wrt_exp_rec kk (e_Var_f y) e)) /\
(forall peb x y kk,
  x `notin` efv_peb peb ->
  x <> y ->
  x `notin` efv_peb (open_peb_wrt_exp_rec kk (e_Var_f y) peb)) /\
(forall pe x y kk,
  x `notin` efv_pe pe ->
  x <> y ->
  x `notin` efv_pe (open_pe_wrt_exp_rec kk (e_Var_f y) pe)).
Proof.
  apply_mutual_ind exp_peb_pe_mutind; intros u x y kk H1 H2; simpl in *; auto.
  destruct (lt_eq_lt_dec u kk); simpl; auto.
    destruct s; simpl; auto.
Qed.

Lemma notin_efv_exp_open_exp_wrt_exp_rec : forall e x y kk,
  x `notin` efv_exp e ->
  x <> y ->
  x `notin` efv_exp (open_exp_wrt_exp_rec kk (e_Var_f y) e).
Proof.
  pose proof notin_efv_exp_open_exp_wrt_exp_rec_mut as H; intuition eauto.
Qed.

Lemma subst_notmatching_ptt : forall p p' e x,
  notmatching_ptt p p' ->
  lc_exp e ->
  notmatching_ptt (esubst_ptt e x p) (esubst_ptt e x p').
Proof.
  intros.
  inversion H; subst.
  simpl. eapply notmatching_ptt_intro; eauto.
    apply esubst_ptt_lc_ptt with (x1:=x) (e1:=e) in H1; auto.
    apply esubst_ptt_lc_ptt with (x1:=x) (e1:=e) in H2; auto.
Qed.

Lemma subst_find_pt : forall C t pt e x,
  find_pt (ptt_pattern C t) pt ->
  lc_exp e ->
  find_pt (ptt_pattern C (esubst_type e x t)) (esubst_pt e x pt).
Proof.
  intros.
  induction pt; simpl.
    inversion H.
    inversion H; subst; auto.
      apply pt_here; auto.
        apply esubst_ptt_lc_ptt with (x1:=x) (e1:=e) in H4; auto.
        apply esubst_pt_lc_pt with (x1:=x) (e1:=e) in H5; auto.
      apply pt_there; auto.
        apply subst_notmatching_ptt with (x:=x) (e:=e) in H5; auto.
Qed.


Definition subst_dtypeeq_prop n :=
  forall D t1 t2 w x,
    dtypeeq n D t1 t2 ->
    pure w ->
    dtypeeq n (esubst_ectx w x D) (esubst_type w x t1) (esubst_type w x t2).

Definition subst_dpateq_prop n :=
  forall D e u e' u' pt pt' w x,
    dpateq n D e u e' u' pt pt' ->
    pure w ->
    dpateq n (esubst_ectx w x D) 
             (esubst_exp w x e)
             (esubst_exp w x u)
             (esubst_exp w x e')
             (esubst_exp w x u')
             (esubst_pt w x pt) 
             (esubst_pt w x pt').

Definition subst_dpatteq_prop n :=
  forall D e u e' u' ptt ptt' w x,
    dpatteq n D e u e' u' ptt ptt' ->
    pure w ->
    dpatteq n (esubst_ectx w x D) 
              (esubst_exp w x e)
              (esubst_exp w x u)
              (esubst_exp w x e')
              (esubst_exp w x u')
              (esubst_ptt w x ptt) 
              (esubst_ptt w x ptt').

Definition subst_deq_P (n :nat) :=
  subst_dtypeeq_prop n /\
  subst_dpateq_prop n /\
  subst_dpatteq_prop n.
 
Lemma subst_deq: forall n,
  subst_deq_P n. 
Proof.
 intro n.
 apply lt_wf_ind. clear n.
 unfold subst_deq_P, subst_dtypeeq_prop, subst_dpateq_prop, subst_dpatteq_prop.
 intros n IH.
 repeat(split).
 Case "subst_dtypeeq_prop".
   intros D t1 t2 w x HJ Pw.
   assert (lc_exp w) as Lw. apply lc_pure; auto.
   (dtypeeq_ind_cases (inversion HJ) SCase); subst; simpl; eauto using subst_incon.
  SCase "DTQ_InCon".
   eapply DTQ_InCon; eauto using subst_incon; auto with lngen.
  SCase "DTQ_TRefl".
    apply DTQ_TRefl.
      apply subst_lc_ectx; auto. 
  SCase "DTQ_URefl".
    apply DTQ_URefl. 
      apply subst_lc_ectx; auto. 
  SCase "DTQ_Pi".
    apply DTQ_Pi with (L:=L `union` singleton x); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D t0 t1' w x H Pw); auto.

      intros x0 x0notinL.
      assert (x0 `notin` L) as H2. auto.
      apply H0 in H2.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D
                          (open_type_wrt_exp t3 (e_Var_f x0)) 
                          (open_type_wrt_exp t2' (e_Var_f x0)) 
                          w x
                          H2 Pw); auto.

      rewrite esubst_type_open_type_wrt_exp_var; auto.
      rewrite esubst_type_open_type_wrt_exp_var; auto.
  SCase "DTQ_Sigma".
    apply DTQ_Sigma with (L:=L `union` singleton x); auto.
      assert (n0 < S (max n0 m)) as IH1. 
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D t0 t1' w x H Pw); auto.

      intros x0 x0notinL.
      assert (x0 `notin` L) as H2. auto.
      apply H0 in H2.
      assert (m < S (max n0 m)) as IH2. 
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult2:=J1 D
                          (open_type_wrt_exp t3 (e_Var_f x0)) 
                          (open_type_wrt_exp t2' (e_Var_f x0)) 
                          w x
                          H2); auto.

      rewrite esubst_type_open_type_wrt_exp_var; auto.
      rewrite esubst_type_open_type_wrt_exp_var; auto.
  SCase "DTQ_App".
    apply DTQ_App; auto.
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 D t t' w x H Pw); auto.

      eauto using subst_isEq.
  SCase "DTQ_Red1".
    eapply DTQ_Red1 with (tj:=esubst_type w x tj) (L:=L `union` singleton x) (w:=esubst_exp w x w0); eauto. 
      apply subst_isEq with (x:=x) (w:=w) in H; eauto. 
      apply subst_pure; auto.
      intros xj xjnotin.
        assert (xj `notin` L) as xjnotinL. auto.  
        apply H2 in xjnotinL.
        apply subst_isEq with (x:=x) (w:=w) in xjnotinL; auto.
        rewrite esubst_exp_open_exp_wrt_exp in xjnotinL; auto.       
        simpl in xjnotinL.
        assert ((if xj == x then w else e_Var_f xj) = e_Var_f xj) as EQ.
          destruct (xj == x); subst; auto.
            destruct_notin. contradict NotInTac; auto.
        rewrite <- EQ. 
        rewrite esubst_exp_fresh_eq with (e1:=w)(e2:=uj) in xjnotinL; auto.
        apply datacon_bind_efv_empty in H1.
        decompose [and] H1. rewrite H6. auto.

      apply subst_find_pt with (x:=x) (e:=w) in H3; auto using lc_pure.

      intros xj xjnotin. 
        assert (xj `notin` L) as xjnotinL. auto.  
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w0, e_Var_f xj)::D) 
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          t2
                          w x 
                          xjnotinL Pw); auto.
        rewrite esubst_type_open_type_wrt_exp in Hresult1; auto.       
        simpl in Hresult1.
        assert ((if xj == x then w else e_Var_f xj) = e_Var_f xj) as EQ.
          destruct (xj == x); subst; auto.
            destruct_notin. contradict NotInTac; auto.
        rewrite <- EQ; auto.
  SCase "DTQ_Red2".
    eapply DTQ_Red2 with (tj:=esubst_type w x tj) (L:=L `union` singleton x) (w:=esubst_exp w x w0); eauto. 
      apply subst_isEq with (x:=x) (w:=w) in H; eauto. 
      apply subst_pure; auto.
      intros xj xjnotin.
        assert (xj `notin` L) as xjnotinL. auto.  
        apply H2 in xjnotinL.
        apply subst_isEq with (x:=x) (w:=w) in xjnotinL; auto.
        rewrite esubst_exp_open_exp_wrt_exp in xjnotinL; auto.       
        simpl in xjnotinL.
        assert ((if xj == x then w else e_Var_f xj) = e_Var_f xj) as EQ.
          destruct (xj == x); subst; auto.
            destruct_notin. contradict NotInTac; auto.
        rewrite <- EQ. 
        rewrite esubst_exp_fresh_eq with (e1:=w)(e2:=uj) in xjnotinL; auto.
        apply datacon_bind_efv_empty in H1.
        decompose [and] H1. rewrite H6. auto.

      apply subst_find_pt with (x:=x) (e:=w) in H3; auto using lc_pure.

      intros xj xjnotin. 
        assert (xj `notin` L) as xjnotinL. auto.  
        apply H4 in xjnotinL. 
        assert (n0 < S n0) as IH1. omega.
        apply IH with (m:=n0) in IH1; auto. clear IH.
        destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
        assert (Hresult1:=J1 
                          ((e, e_Con Cj (e_Var_f xj))::(w0, e_Var_f xj)::D) 
                          t1
                          (open_type_wrt_exp tj (e_Var_f xj)) 
                          w x 
                          xjnotinL Pw); auto.
        rewrite esubst_type_open_type_wrt_exp in Hresult1; auto.       
        simpl in Hresult1.
        assert ((if xj == x then w else e_Var_f xj) = e_Var_f xj) as EQ.
          destruct (xj == x); subst; auto.
            destruct_notin. contradict NotInTac; auto.
        rewrite <- EQ; auto.
  SCase "DTQ_Case".
    eapply DTQ_Case; eauto. 
      eapply subst_isEq; eauto. 
      eapply subst_isEq; eauto. 

      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D
                        e u e' u' 
                        pt5
                        pt'
                        w x
                        H1 Pw); auto.
 Case "subst_dpateq_prop".
   intros D e u e' u' pt1 pt2 w x HJ Pw.
   assert (lc_exp w) as Lw. apply lc_pure; auto.
   inversion HJ; subst; simpl; eauto using subst_incon.
  SCase "DTPQ_Empty".
   eapply DTPQ_Empty; eauto using subst_incon, esubst_exp_lc_exp.
      apply subst_lc_ectx; auto. 
  SCase "DTPQ_Cons".
    apply DTPQ_Cons; eauto.
      assert (n0 < S (max n0 m)) as IH1.
        assert (J:=@le_max_l n0 m). omega.
      apply IH with (m0:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J1 J2.
      assert (Hresult3:=J3 
                        D
                        e u e' u' 
                        ptt5 
                        ptt'
                        w x
                        H Pw); auto.

      assert (m < S (max n0 m)) as IH2.
        assert (J:=@le_max_r n0 m). omega.
      apply IH with (m0:=m) in IH2; auto. clear IH.
      destruct IH2 as [J1 [J2 J3]]. clear J1 J3.
      assert (Hresult2:=J2 
                        D
                        e u e' u' 
                        pt5 
                        pt'
                        w x
                        H0 Pw); auto.
 Case "ctx_conv_dpatteq_prop".
   intros D e u e' u' ptt1 ptt2 w x HJ Pw.
   assert (lc_exp w) as Lw. apply lc_pure; auto.
   inversion HJ; subst; simpl; eauto using subst_incon.
  SCase "DTTPQ_Intro".
    eapply DTTPQ_Intro with (L:=L `union` singleton x); eauto. 
      eapply subst_isEq; eauto. 
      eapply subst_isEq; eauto. 
      intros xi xinotin. 
      assert (xi `notin` L) as xinotinL. auto.  
      apply H2 in xinotinL. 
      assert (n0 < S n0) as IH1. omega.
      apply IH with (m:=n0) in IH1; auto. clear IH.
      destruct IH1 as [J1 [J2 J3]]. clear J2 J3.
      assert (Hresult1:=J1 
                        ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D) 
                        (open_type_wrt_exp ti (e_Var_f xi)) 
                        (open_type_wrt_exp ti' (e_Var_f xi)) 
                        w x
                        xinotinL Pw); auto.
      rewrite esubst_type_open_type_wrt_exp in Hresult1; auto.       
      rewrite esubst_type_open_type_wrt_exp in Hresult1; auto.
      assert ((if xi == x then w else e_Var_f xi) = e_Var_f xi) as EQ0.
        destruct (xi == x); subst; auto.
          destruct_notin. contradict NotInTac; auto.
      assert (esubst_exp w x (e_Var_f xi) = (e_Var_f xi)) as EQ1.
        simpl. auto.
      assert (e_Con Ci (if xi == x then w else e_Var_f xi) = e_Con Ci (e_Var_f xi)) as EQ'.
        destruct (xi == x); subst; auto.
      assert (esubst_ectx w x ((e, e_Con Ci (e_Var_f xi))::(u, open_exp_wrt_exp ui (e_Var_f xi))::D) =       
              ((esubst_exp w x e, e_Con Ci (e_Var_f xi))::(esubst_exp w x u, open_exp_wrt_exp ui (e_Var_f xi))::esubst_ectx w x D)) as EQ.
        simpl.
        rewrite esubst_exp_fresh_eq with (e1:=w)(e2:=open_exp_wrt_exp ui (e_Var_f xi)); auto.
          rewrite <- EQ'; auto.
          apply datacon_bind_efv_empty in H.
          decompose [and] H. 
          apply notin_efv_exp_open_exp_wrt_exp_rec; auto.
             rewrite H4; auto.
      rewrite EQ in Hresult1; auto.
      rewrite EQ1 in Hresult1; auto.
Qed.

Lemma subst_dtypeeq: forall n D t1 t2 w x,
    dtypeeq n D t1 t2 ->
    pure w ->
    dtypeeq n (esubst_ectx w x D) (esubst_type w x t1) (esubst_type w x t2).
intros. 
pose (J := subst_deq n).
unfold subst_dtypeeq_prop in J. 
destruct J. eauto.
Qed.

Lemma esubst_ectx_fresh_eq : forall D e1 x1,
  x1 `notin` efv_ectx D ->
  esubst_ectx e1 x1 D = D.
Proof.
  intros D.
  induction D; intros; simpl; auto.
    destruct a.
    simpl in *.
    rewrite (@esubst_exp_fresh_eq e e1 x1); auto.
    rewrite (@esubst_exp_fresh_eq e0 e1 x1); auto.
    rewrite (@IHD e1 x1); auto.
Qed.

Lemma renaming_dtypeeq0 : forall n x y D t1 t2,
  dtypeeq n D (open_type_wrt_exp t1 (e_Var_f x)) (open_type_wrt_exp t2 (e_Var_f x)) ->
  x `notin` (efv_ectx D) `union` (efv_type t1) `union` (efv_type t2) ->
  y `notin` (efv_ectx D) `union` (efv_type t1) `union` (efv_type t2) ->
  dtypeeq n D (open_type_wrt_exp t1 (e_Var_f y)) (open_type_wrt_exp t2 (e_Var_f y)).
Proof.
  intros n x y D t1 t2 Htypeeq xnotin ynotin.
  destruct (x == y); subst; auto.
  rewrite (@esubst_type_intro x t1 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t2 (e_Var_f y)); auto.
  apply subst_dtypeeq with (x:=x) (w:=e_Var_f y) in Htypeeq; auto.
  assert (esubst_ectx (e_Var_f y) x D = D) as EQ.
    apply esubst_ectx_fresh_eq; auto.
  rewrite <- EQ. auto.
Qed.

Lemma renaming_dtypeeq1 : forall n e1 x y e2 D t1 t2,
  dtypeeq n ((open_exp_wrt_exp e1 (e_Var_f x), open_exp_wrt_exp e2 (e_Var_f x))::D) (open_type_wrt_exp t1 (e_Var_f x)) (open_type_wrt_exp t2 (e_Var_f x)) ->
  x `notin` (efv_ectx D) `union` (efv_exp e1) `union` (efv_exp e2) `union` (efv_type t1) `union` (efv_type t2) ->
  y `notin` (efv_ectx D) `union` (efv_exp e1) `union` (efv_exp e2) `union` (efv_type t1) `union` (efv_type t2) ->
  dtypeeq n ((open_exp_wrt_exp e1 (e_Var_f y), open_exp_wrt_exp e2 (e_Var_f y))::D) (open_type_wrt_exp t1 (e_Var_f y)) (open_type_wrt_exp t2 (e_Var_f y)).
Proof.
  intros n e1 x y e2 D t1 t2 Htypeeq xnotin ynotin.
  destruct (x == y); subst; auto.
  rewrite (@esubst_exp_intro x e1 (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t1 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t2 (e_Var_f y)); auto.
  apply subst_dtypeeq with (x:=x) (w:=e_Var_f y) in Htypeeq; auto.
  assert (esubst_ectx (e_Var_f y) x D = D) as EQ.
    apply esubst_ectx_fresh_eq; auto.
  assert (esubst_ectx (e_Var_f y) x ((open_exp_wrt_exp e1 (e_Var_f x), (open_exp_wrt_exp e2 (e_Var_f x)))::D) =
          (esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1 (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2 (e_Var_f x)))::D) as EQ'.
    simpl. rewrite EQ. auto. 
  rewrite <- EQ'. auto.
Qed.

Lemma renaming_dtypeeq2 : forall n e1 x y e2 e1' e2' D t1 t2,
  dtypeeq n ((open_exp_wrt_exp e1' (e_Var_f x), open_exp_wrt_exp e2' (e_Var_f x))::(open_exp_wrt_exp e1 (e_Var_f x), open_exp_wrt_exp e2 (e_Var_f x))::D) (open_type_wrt_exp t1 (e_Var_f x)) (open_type_wrt_exp t2 (e_Var_f x)) ->
  x `notin` (efv_ectx D) `union` (efv_exp e1) `union` (efv_exp e2) `union` (efv_exp e1') `union` (efv_exp e2') `union` (efv_type t1) `union` (efv_type t2) ->
  y `notin` (efv_ectx D) `union` (efv_exp e1) `union` (efv_exp e2) `union` (efv_exp e1') `union` (efv_exp e2') `union` (efv_type t1) `union` (efv_type t2) ->
  dtypeeq n ((open_exp_wrt_exp e1' (e_Var_f y), open_exp_wrt_exp e2' (e_Var_f y))::(open_exp_wrt_exp e1 (e_Var_f y), open_exp_wrt_exp e2 (e_Var_f y))::D) (open_type_wrt_exp t1 (e_Var_f y)) (open_type_wrt_exp t2 (e_Var_f y)).
Proof.
  intros n e1 x y e2 e1' e2' D t1 t2 Htypeeq xnotin ynotin.
  destruct (x == y); subst; auto.
  rewrite (@esubst_exp_intro x e1 (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t1 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t2 (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e1' (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2' (e_Var_f y)); auto.
  apply subst_dtypeeq with (x:=x) (w:=e_Var_f y) in Htypeeq; auto.
  assert (esubst_ectx (e_Var_f y) x D = D) as EQ.
    apply esubst_ectx_fresh_eq; auto.
  assert (esubst_ectx (e_Var_f y) x ((open_exp_wrt_exp e1' (e_Var_f x), (open_exp_wrt_exp e2' (e_Var_f x)))::(open_exp_wrt_exp e1 (e_Var_f x), (open_exp_wrt_exp e2 (e_Var_f x)))::D) =
          (esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1' (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2' (e_Var_f x)))::(esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1 (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2 (e_Var_f x)))::D) as EQ'.
    simpl. rewrite EQ. auto. 
  rewrite <- EQ'. auto.
Qed.

Lemma renaming_dtypeeq3 : forall n e1 x y e2 e1' e2' e1'' e2'' D t1 t2,
  dtypeeq n ((open_exp_wrt_exp e1'' (e_Var_f x), open_exp_wrt_exp e2'' (e_Var_f x))::
             (open_exp_wrt_exp e1' (e_Var_f x), open_exp_wrt_exp e2' (e_Var_f x))::
             (open_exp_wrt_exp e1 (e_Var_f x), open_exp_wrt_exp e2 (e_Var_f x))::
             D) (open_type_wrt_exp t1 (e_Var_f x)) (open_type_wrt_exp t2 (e_Var_f x)) ->
  x `notin` (efv_ectx D) `union` 
            (efv_exp e1) `union` (efv_exp e2) `union` 
            (efv_exp e1') `union` (efv_exp e2') `union` 
            (efv_exp e1'') `union` (efv_exp e2'') `union` 
            (efv_type t1) `union` (efv_type t2) ->
  y `notin` (efv_ectx D) `union` 
            (efv_exp e1) `union` (efv_exp e2) `union` 
            (efv_exp e1') `union` (efv_exp e2') `union` 
            (efv_exp e1'') `union` (efv_exp e2'') `union` 
            (efv_type t1) `union` (efv_type t2) ->
  dtypeeq n ((open_exp_wrt_exp e1'' (e_Var_f y), open_exp_wrt_exp e2'' (e_Var_f y))::
             (open_exp_wrt_exp e1' (e_Var_f y), open_exp_wrt_exp e2' (e_Var_f y))::
             (open_exp_wrt_exp e1 (e_Var_f y), open_exp_wrt_exp e2 (e_Var_f y))::
             D) (open_type_wrt_exp t1 (e_Var_f y)) (open_type_wrt_exp t2 (e_Var_f y)).
Proof.
  intros n e1 x y e2 e1' e2' e1'' e2'' D t1 t2 Htypeeq xnotin ynotin.
  destruct (x == y); subst; auto.
  rewrite (@esubst_exp_intro x e1 (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t1 (e_Var_f y)); auto.
  rewrite (@esubst_type_intro x t2 (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e1' (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2' (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e1'' (e_Var_f y)); auto.
  rewrite (@esubst_exp_intro x e2'' (e_Var_f y)); auto.
  apply subst_dtypeeq with (x:=x) (w:=e_Var_f y) in Htypeeq; auto.
  assert (esubst_ectx (e_Var_f y) x D = D) as EQ.
    apply esubst_ectx_fresh_eq; auto.
  assert (esubst_ectx (e_Var_f y) x ((open_exp_wrt_exp e1'' (e_Var_f x), (open_exp_wrt_exp e2'' (e_Var_f x)))::
                                     (open_exp_wrt_exp e1' (e_Var_f x), (open_exp_wrt_exp e2' (e_Var_f x)))::
                                     (open_exp_wrt_exp e1 (e_Var_f x), (open_exp_wrt_exp e2 (e_Var_f x)))::D) =
          (esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1'' (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2'' (e_Var_f x)))::
          (esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1' (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2' (e_Var_f x)))::
          (esubst_exp (e_Var_f y) x (open_exp_wrt_exp e1 (e_Var_f x)), esubst_exp (e_Var_f y) x (open_exp_wrt_exp e2 (e_Var_f x)))::
          D) as EQ'.
    simpl. rewrite EQ. auto. 
  rewrite <- EQ'. auto.
Qed.

Definition trans_dtypeeq_prop n m :=
  forall D1 D2 t1 t2 t3, 
    dtypeeq n D1 t1 t2 ->
    dtypeeq m D2 t2 t3 ->  
    ctxereq D1 D2 -> 
    exists p, dtypeeq p D1 t1 t3.

Definition trans_dpateq_prop n m :=
  forall D1 D2 e1 e2 e3 u1 u2 u3 pt1 pt2 pt3,
    dpateq n D1 e1 u1 e2 u2 pt1 pt2 ->
    dpateq m D2 e2 u2 e3 u3 pt2 pt3 ->
    ctxereq D1 D2 ->
    exists p, dpateq p D1 e1 u1 e3 u3 pt1 pt3.

Definition trans_dpatteq_prop n m :=
  forall D1 D2 e1 e2 e3 u1 u2 u3 ptt1 ptt2 ptt3,
    dpatteq n D1 e1 u1 e2 u2 ptt1 ptt2 ->
    dpatteq m D2 e2 u2 e3 u3 ptt2 ptt3 ->
    ctxereq D1 D2 ->
    exists p, dpatteq p D1 e1 u1 e3 u3 ptt1 ptt3.

Definition trans_deq_P (nm:nat*nat) :=  forall n m,
  nm = (n, m) ->
  trans_dtypeeq_prop n m /\
  trans_dpateq_prop n m /\
  trans_dpatteq_prop n m.
 
Lemma trans_deq: forall nm,
  trans_deq_P nm. 
Proof.
 intro nm.
 apply pair_lt_ind.
 unfold trans_deq_P, trans_dtypeeq_prop, trans_dpateq_prop, trans_dpatteq_prop.
 intros [n m] IH.
 intros n0 m0 EQ. inversion EQ; subst. clear EQ.
 repeat(split).
 Case "trans_dtypeeq_prop".
   intros D1 D2 t1 t2 t3 Htrans1 Htrans2 Hctxeq.
   (dtypeeq_ind_cases (inversion Htrans1) SCase); subst.
   SCase "DTQ_InCon".
     exists 0.
     apply DTQ_InCon; auto.
   SCase "DTQ_TRefl".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_TRefl".
       exists 0.
       apply DTQ_TRefl; auto.
     SSCase "DTQ_Red2".
       assert (PairLt (0, n) (0, S n)) as J. auto.
       apply IH with (n:=0) (m:=n) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L) as Hright. auto.
       apply H5 in Hright.
       assert (dtypeeq 0 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) (t_Con T) (t_Con T)) as Hleft.
         apply DTQ_TRefl; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
       assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            (t_Con T)
                            (t_Con T)
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H3 in xj0notinL.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
               apply lc_isEq in H0. decompose [and] H0; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp (t_Con T) (e_Var_f xj0) =
                   (t_Con T)) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=(t_Con T)); eauto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp (t_Con T) (e_Var_f xj) =
                     (t_Con T)) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=(t_Con T)); eauto.
             rewrite EQ; auto.
   SCase "DTQ_URefl".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_URefl".
       exists 0.
       apply DTQ_URefl; auto.
     SSCase "DTQ_Red2".
       assert (PairLt (0, n) (0, S n)) as J. auto.
       apply IH with (n:=0) (m:=n) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L) as Hright. auto.
       apply H5 in Hright.
       assert (dtypeeq 0 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) t_Unit t_Unit) as Hleft.
         apply DTQ_URefl; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
       assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            t_Unit
                            t_Unit
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H3 in xj0notinL.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
               apply lc_isEq in H0. decompose [and] H0; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp t_Unit (e_Var_f xj0) =
                   t_Unit) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=t_Unit); eauto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp t_Unit (e_Var_f xj) =
                     t_Unit) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=t_Unit); eauto.
             rewrite EQ; auto.
   SCase "DTQ_Pi".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
            destruct Htrans1; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_Pi".
       assert (PairLt (n, n0) (S (max n m), S (max n0 m1))) as J1.
         apply PairOrdL. 
           assert (J:=@le_max_l n m). omega.
       apply IH with (n:=n) (m:=n0) in J1; auto.
       destruct J1 as [J1 [J2 J3]]. clear J2 J3.

       assert (PairLt (m, m1) (S (max n m), S (max n0 m1))) as J2.
         apply PairOrdL. 
           assert (J:=@le_max_r n m). omega.
       apply IH with (n:=m) (m:=m1) in J2; auto.
       destruct J2 as [J2 [J3 J4]]. clear J3 J4.
       clear IH.

       assert (Hresult1:=J1 D1 D2 t0 t1' t1'0 H H5 Hctxeq).
       clear H H5.

       pick fresh x.
       assert (x `notin` L) as Hleft2. auto.
       apply H0 in Hleft2.
       assert (x `notin` L0) as Hright2. auto.
       apply H7 in Hright2.
       assert (Hresult2:=J2 D1 D2 
                          (open_type_wrt_exp t4 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          (open_type_wrt_exp t2'0 (e_Var_f x)) 
                          Hleft2 
                          Hright2 
                          Hctxeq).
       clear H0 H7 J1 J2.

       destruct Hresult1 as [p1 Hresult1].
       destruct Hresult2 as [p2 Hresult2].

       exists (S (max p1 p2)).
       apply DTQ_Pi with (L:=L0 `union` L `union` efv_ectx D1 `union` efv_type t4 `union` efv_type t2'0); eauto.
         intros x0 x0notin.
         apply renaming_dtypeeq0 with (x:=x); auto.
     SSCase "DTQ_Red2".
       assert (PairLt (S (max n m), n0) (S (max n m), S n0)) as J. auto.
       apply IH with (n:=S (max n m)) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L0) as Hright. auto.
       apply H6 in Hright.
       assert (dtypeeq (S (max n m)) ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) (t_Pi t0 t4) (t_Pi t1' t2')) as Hleft.
         rewrite_env (nil ++ ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::nil) ++ D1).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
       assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            (t_Pi t0 t4) 
                            (t_Pi t1' t2')
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` L0 `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj `union` efv_type t0 `union` efv_type t4)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L0) as xj0notinL0. auto.
           apply H4 in xj0notinL0.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
               apply lc_isEq in H1. decompose [and] H1; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp (t_Pi t0 t4) (e_Var_f xj0) =
                   (t_Pi t0 t4)) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=(t_Pi t0 t4)); eauto.
               apply lc_dtypeeq in Htrans1.
                 destruct Htrans1; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp (t_Pi t0 t4) (e_Var_f xj) =
                     (t_Pi t0 t4)) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=(t_Pi t0 t4)); eauto.
             rewrite EQ; auto.
   SCase "DTQ_Sigma".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
            destruct Htrans1; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_Sigma".
       assert (PairLt (n, n0) (S (max n m), S (max n0 m1))) as J1.
         apply PairOrdL. 
           assert (J:=@le_max_l n m). omega.
       apply IH with (n:=n) (m:=n0) in J1; auto.
       destruct J1 as [J1 [J2 J3]]. clear J2 J3.

       assert (PairLt (m, m1) (S (max n m), S (max n0 m1))) as J2.
         apply PairOrdL. 
           assert (J:=@le_max_r n m). omega.
       apply IH with (n:=m) (m:=m1) in J2; auto.
       destruct J2 as [J2 [J3 J4]]. clear J3 J4.
       clear IH.

       assert (Hresult1:=J1 D1 D2 t0 t1' t1'0 H H5 Hctxeq).
       clear H H5.

       pick fresh x.
       assert (x `notin` L) as Hleft2. auto.
       apply H0 in Hleft2.
       assert (x `notin` L0) as Hright2. auto.
       apply H7 in Hright2.
       assert (Hresult2:=J2 D1 D2 
                          (open_type_wrt_exp t4 (e_Var_f x)) 
                          (open_type_wrt_exp t2' (e_Var_f x)) 
                          (open_type_wrt_exp t2'0 (e_Var_f x)) 
                          Hleft2 
                          Hright2 
                          Hctxeq).
       clear H0 H7 J1 J2.

       destruct Hresult1 as [p1 Hresult1].
       destruct Hresult2 as [p2 Hresult2].

       exists (S (max p1 p2)).
       apply DTQ_Sigma with (L:=L0 `union` L `union` efv_ectx D1 `union` efv_type t4 `union` efv_type t2'0); eauto.
         intros x0 x0notin.
         apply renaming_dtypeeq0 with (x:=x); auto.
     SSCase "DTQ_Red2".
       assert (PairLt (S (max n m), n0) (S (max n m), S n0)) as J. auto.
       apply IH with (n:=S (max n m)) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L0) as Hright. auto.
       apply H6 in Hright.
       assert (dtypeeq (S (max n m)) ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) (t_Sigma t0 t4) (t_Sigma t1' t2')) as Hleft.
         rewrite_env (nil ++ ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::nil) ++ D1).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
       assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            (t_Sigma t0 t4) 
                            (t_Sigma t1' t2')
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` L0 `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj `union` efv_type t0 `union` efv_type t4)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L0) as xj0notinL0. auto.
           apply H4 in xj0notinL0.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
               apply lc_isEq in H1. decompose [and] H1; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp (t_Sigma t0 t4) (e_Var_f xj0) =
                   (t_Sigma t0 t4)) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=(t_Sigma t0 t4)); eauto.
               apply lc_dtypeeq in Htrans1.
                 destruct Htrans1; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp (t_Sigma t0 t4) (e_Var_f xj) =
                     (t_Sigma t0 t4)) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=(t_Sigma t0 t4)); eauto.
             rewrite EQ; auto.
   SCase "DTQ_App".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
           destruct Htrans1; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_App".
       assert (PairLt (n, n0) (S n, S n0)) as J.
         apply PairOrdL. auto.
       apply IH with (n:=n) (m:=n0) in J; auto.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       clear IH.

       assert (Hresult1:=J1 D1 D2 t t' t'0 H H5 Hctxeq).
       clear H H5 J1.

       destruct Hresult1 as [p1 Hresult1].

       exists (S p1).
       apply DTQ_App; eauto.
         apply trans_isEq with (e2:=e'); auto.
           apply ctx_conv_isEq with (D1:=D2); auto.
             apply sym_ctxereq; auto.
     SSCase "DTQ_Red2".
       assert (PairLt (S n, n0) (S n, S n0)) as J. auto.
       apply IH with (n:=S n) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L) as Hright. auto.
       apply H6 in Hright.
       assert (dtypeeq (S n) ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) (t_App t e) (t_App t' e')) as Hleft.
         rewrite_env (nil ++ ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::nil) ++ D1).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
      assert (ctxereq ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
      assert (Hresult:=@J1 ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            (t_App t e) 
                            (t_App t' e')
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_exp e0  `union` efv_type tj `union` efv_type t)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H4 in xj0notinL.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e0 (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e0, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
               apply lc_isEq in H1. decompose [and] H1; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp (t_App t e) (e_Var_f xj0) =
                   (t_App t e)) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=(t_App t e)); eauto.
               apply lc_dtypeeq in Htrans1.
                 destruct Htrans1; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e0 (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e0, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp (t_App t e) (e_Var_f xj) =
                     (t_App t e)) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=(t_App t e)); eauto.
             rewrite EQ; auto.
   SCase "DTQ_Red1".
     assert (PairLt (n, m0) (S n, m0)) as J. auto.
     apply IH with (n:=n) (m:=m0) in J; auto. clear IH.
     destruct J as [J1 [J2 J3]]. clear J2 J3.
     pick fresh xj.
     assert (xj `notin` L) as Hleft. auto.
     apply H4 in Hleft.
     assert (dtypeeq m0 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2) t2 t3) as Hright.
       rewrite_env (nil ++ ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::nil) ++ D2).
       eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
     assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                     ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
       apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
     assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                          ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                          (open_type_wrt_exp tj (e_Var_f xj))
                          t2 
                          t3
                          Hleft
                          Hright
                          Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red1 with (L:=L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj `union` efv_type t3)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
               apply lc_isEq in H. decompose [and] H; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp t3 (e_Var_f xj0) =
                   t3) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=t3); eauto.
               apply lc_dtypeeq in Htrans2.
                 decompose [and] Htrans2; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp t3 (e_Var_f xj) =
                     t3) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=t3); eauto.
             rewrite EQ; auto.
   SCase "DTQ_Red2".
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
           destruct Htrans1; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_Red1".
     destruct (incon_decidable D1) as [inconD1 | conD1].
       SSSCase "inconD1".
       exists 0.       
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
           destruct Htrans1; auto.
         apply lc_dtypeeq in Htrans2.
           decompose [and] Htrans2; auto.
       SSSCase "conD1".
       assert (PairLt (n, n0) (S n, S n0)) as J.
         apply PairOrdL. auto.
       apply IH with (n:=n) (m:=n0) in J; auto.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       clear IH.

       pick fresh xj.
       assert (xj `notin` L0) as Hright. auto.
       apply H17 in Hright.
       assert (xj `notin` L) as Hleft. auto.
       apply H4 in Hleft.
       assert (isEq D1 (e_Con Cj w) (e_Con Cj0 w0)) as CwEQCw0.
         apply trans_isEq with (e2:=e); auto.
           apply sym_isEq; auto.
             apply ctx_conv_isEq with (D1:=D2); auto.
               apply sym_ctxereq; auto.
       assert (Cj = Cj0) as EQC. 
         apply injection1_isEq with (w1:=w) (w2:=w0) (D:=D1); auto.
       subst Cj0.
       assert (isEq D1 w w0) as wEQw0.
         apply injection2_isEq in CwEQCw0; auto.
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(w0, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; auto.
             simpl; split; auto.
           apply ref_isEq; auto.
             simpl; split; auto.
       assert (K:=@datacon_bind_uniq Cj sj sj0 T T uj uj0 H1 H13).
       decompose [and] K. subst sj0 uj0. clear H7 K.
       assert (K:=@find_pt_deterministic Cj tj tj0 pt5 H3 H16).
       subst tj0.

       assert (Hresult:=@J1 ((e, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e, e_Con Cj (e_Var_f xj))::(w0, e_Var_f xj)::D2)
                            t1 
                            (open_type_wrt_exp tj (e_Var_f xj))
                            t3
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists p.
       apply subst_dtypeeq with (x:=xj) (w:=w) in Hresult; auto.
       rewrite esubst_type_fresh_eq in Hresult; auto.
       rewrite esubst_type_fresh_eq in Hresult; auto.
       simpl in Hresult.
       rewrite esubst_exp_fresh_eq in Hresult; auto.
       rewrite esubst_exp_fresh_eq in Hresult; auto.
       rewrite_env (((e,e_Con Cj (if xj == xj then w else e_Var_f xj))::nil)++(w, if xj==xj then w else e_Var_f xj)::esubst_ectx w xj D1) in Hresult.
       apply cut_dtypeeq in Hresult; auto.
         rewrite_env (nil ++ (e,e_Con Cj (if xj == xj then w else e_Var_f xj))::esubst_ectx w xj D1) in Hresult.
         apply cut_dtypeeq in Hresult; auto.
           simpl in Hresult.
           rewrite esubst_ectx_fresh_eq in Hresult; auto.
           destruct (xj == xj); subst; auto.
             rewrite esubst_ectx_fresh_eq; auto.
             contradict n1; auto.
         destruct (xj == xj); subst; auto using ref_isEq.
           apply ref_isEq; auto. apply subst_lc_ectx; auto.
           contradict n1; auto.
     SSCase "DTQ_Red2".
       assert (PairLt (S n, n0) (S n, S n0)) as J. auto.
       apply IH with (n:=S n) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L0) as Hright. auto.
       apply H10 in Hright.
       assert (dtypeeq (S n) ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::D1) t1 (t_Case e T u pt5)) as Hleft.
         rewrite_env (nil ++ ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::nil) ++ D1).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::D1)
                       ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; simpl; auto.
           apply ref_isEq; simpl; auto.
       assert (Hresult:=@J1 ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::D1)
                            ((e0, e_Con Cj0 (e_Var_f xj))::(w0, e_Var_f xj)::D2)
                            t1 
                            (t_Case e T u pt5)
                            (open_type_wrt_exp tj0 (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L0 `union` efv_ectx D1 `union` efv_exp w0 `union` efv_exp e0 `union` efv_type tj0 `union` efv_type t1)(Cj:=Cj0)(w:=w0)(sj:=sj0)(uj:=uj0)(tj:=tj0); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L0) as xj0notinL0. auto.
           apply H8 in xj0notinL0.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e0 (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj0 (e_Var_b 0))(e_Var_f xj0)) =
                   (e0, (e_Con Cj0 (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
               apply lc_isEq in H5. decompose [and] H5; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w0 (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w0, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w0); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp t1 (e_Var_f xj0) =
                   t1) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=t1); eauto.
               apply lc_dtypeeq in Htrans1.
                 destruct Htrans1; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e0 (e_Var_f xj), open_exp_wrt_exp (e_Con Cj0 (e_Var_b 0))(e_Var_f xj)) =
                     (e0, (e_Con Cj0 (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w0 (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w0, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w0); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp t1 (e_Var_f xj) =
                     t1) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=t1); eauto.
             rewrite EQ; auto.
     SSCase "DTQ_Case".
       apply dpateq_inversion_left with (ui:=uj)(T:=T)(si:=sj)(Ci:=Cj)(ti:=tj) in H14; auto.
       destruct H14 as [tj' [m [EQ [Hfind Hpatteq]]]].
       inversion Hpatteq; subst.
       assert (PairLt (n, m) (S n, S n0)) as J. auto.
       apply IH with (n:=n) (m:=m) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.

       pick fresh xj.
       assert (xj `notin` L) as Hleft'. auto.
       apply H4 in Hleft'.
       assert (xj `notin` L0) as Hright'. auto.
       apply H19 in Hright'.

       assert (K:=@datacon_bind_uniq Cj sj si T T0 uj ui H1 H9).
       decompose [and] K. subst si ui T0. clear K.

       assert (dtypeeq n ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1) t1 (open_type_wrt_exp tj (e_Var_f xj))) as Hleft.
         rewrite_env (((e, e_Con Cj (e_Var_f xj))::nil) ++ ((u,open_exp_wrt_exp uj (e_Var_f xj))::nil) ++ ((w, e_Var_f xj):: D1)).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
             eapply lc_open_ui; eauto.
       assert (dtypeeq m ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2) (open_type_wrt_exp tj (e_Var_f xj)) (open_type_wrt_exp tj' (e_Var_f xj))) as Hright.
         rewrite_env (((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::nil)++((w, e_Var_f xj)::nil) ++ D2).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; simpl; auto.
           apply ref_isEq; simpl; eauto using lc_open_ui.
           apply ref_isEq; simpl; auto. split; eauto using lc_open_ui.
           apply ref_isEq; simpl; auto. split; eauto using lc_open_ui.
       assert (Hresult1:=J1 
                         ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                         ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                         t1
                         (open_type_wrt_exp tj (e_Var_f xj)) 
                         (open_type_wrt_exp tj' (e_Var_f xj))
                         Hleft Hright Hctxereq).
       clear J1.

       destruct Hresult1 as [p1 Hresult1].
       exists (S p1).
       apply DTQ_Red2 with (L:=L0 `union` L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e' `union` efv_type tj' `union` efv_type t1 `union` efv_exp u`union` efv_exp uj)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj'); auto.
         apply trans_isEq with (e2:=e); auto.
           apply sym_isEq; auto.
             eapply ctx_conv_isEq; eauto.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H2 in xj0notinL.
           apply trans_isEq with (e2:=u); auto.
             apply sym_isEq; auto.
             rewrite_env (nil ++ (one (w, e_Var_f xj0)) ++ D1).
             apply weakening_isEq; auto.
               simpl. split; auto.
               eapply ctx_conv_isEq; eauto.
                 apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           rewrite_env ((one (e', e_Con Cj (e_Var_f xj0)))++((w, e_Var_f xj0)::D1)).
           apply cut_dtypeeq with (e1:=u)(e2:=open_exp_wrt_exp uj (e_Var_f xj0)); auto.
             apply ctx_conv_dtypeeq with (D':=((e',e_Con Cj (e_Var_f xj))::(u, open_exp_wrt_exp uj (e_Var_f xj))::(w,e_Var_f xj)::D1)) in Hresult1; auto.
               simpl.
               assert ((open_exp_wrt_exp e' (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                       (e', (e_Con Cj (e_Var_f xj0)))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=e'); eauto.
                   apply lc_isEq in H12. decompose [and] H12; auto.
               rewrite <- EQ'. clear EQ'.
               assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                       (w, (e_Var_f xj0))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               rewrite <- EQ'. clear EQ'.
               assert ((open_exp_wrt_exp u (e_Var_f xj0), open_exp_wrt_exp uj (e_Var_f xj0)) =
                       (u, open_exp_wrt_exp uj (e_Var_f xj0))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
                   apply lc_isEq in H13. decompose [and] H13; auto.
               rewrite <- EQ'. clear EQ'.
               assert (open_type_wrt_exp t1 (e_Var_f xj0) =
                       t1) as EQ'.
                 rewrite open_type_wrt_exp_lc_type with (t1:=t1); eauto.
                   apply lc_dtypeeq in Htrans1.
                     destruct Htrans1; auto.
               rewrite <- EQ'. clear EQ'.
               apply renaming_dtypeeq3 with (x:=xj); auto.
                 assert ((open_exp_wrt_exp e' (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                         (e', (e_Con Cj (e_Var_f xj)))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=e'); eauto.
                 rewrite EQ'. clear EQ'.
                 assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                         (w, (e_Var_f xj))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 rewrite EQ'. clear EQ'.
                 assert ((open_exp_wrt_exp u (e_Var_f xj), open_exp_wrt_exp uj (e_Var_f xj)) =
                         (u, open_exp_wrt_exp uj (e_Var_f xj))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
                 rewrite EQ'. clear EQ'.
                 assert (open_type_wrt_exp t1 (e_Var_f xj) =
                         t1) as EQ'.
                   rewrite open_type_wrt_exp_lc_type with (t1:=t1); eauto.
                 rewrite EQ'; auto.

               apply CEQ_Eq; auto using ref_isEq.
                 apply CEQ_Eq; auto using ref_isEq.
                   apply CEQ_Eq; auto using ref_isEq.
                     apply ref_ctxereq; auto.
                     apply ref_isEq; simpl; auto.
                     apply ref_isEq; simpl; eauto using lc_open_ui.
                 rewrite_env (nil ++ ((u, open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::nil) ++ D1).
                 apply weakening_isEq; auto.
                   simpl. split; eauto using lc_open_ui.
                   eapply ctx_conv_isEq; eauto.
                     apply sym_ctxereq; auto.
                     apply ref_isEq; simpl; auto. 
                       split; eauto using lc_open_ui.
   SCase "DTQ_Case".  
     inversion Htrans2; subst.
     SSCase "DTQ_InCon".
       exists 0.
       apply DTQ_InCon; auto.
         apply lc_dtypeeq in Htrans1.
           destruct Htrans1; auto.
         eapply ctx_conv_incon; eauto.
           apply sym_ctxereq; auto.
     SSCase "DTQ_Red1".
       apply dpateq_inversion_right with (ui:=uj)(T:=T)(si:=sj)(Ci:=Cj)(ti':=tj) in H1; auto.
       destruct H1 as [tj' [m [EQ [Hfind Hpatteq]]]].
       inversion Hpatteq; subst.
       assert (PairLt (m, n0) (S n, S n0)) as J. auto.
       apply IH with (n:=m) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.

       assert (K:=@datacon_bind_uniq Cj sj si T T0 uj ui H10 H5).
       decompose [and] K. subst si ui T0. clear K.

       pick fresh xj.
       assert (xj `notin` L) as Hright'. auto.
       apply H14 in Hright'.
       assert (xj `notin` L0) as Hleft'. auto.
       apply H19 in Hleft'.
       assert (dtypeeq m ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1) (open_type_wrt_exp tj' (e_Var_f xj)) (open_type_wrt_exp tj (e_Var_f xj))) as Hleft.
         rewrite_env (((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::nil) ++ ((w, e_Var_f xj)::nil) ++ D1).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
       assert (dtypeeq n0 ((e', e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2) (open_type_wrt_exp tj (e_Var_f xj)) t3) as Hright.
         rewrite_env (((e', e_Con Cj (e_Var_f xj))::nil) ++ ((u,open_exp_wrt_exp uj (e_Var_f xj))::nil)++((w, e_Var_f xj)::D2)).
         eapply weakening_dtypeeq; eauto.
           simpl. repeat(split; auto).
             apply lc_isEq in H18. decompose [and] H18; auto.
             eapply lc_open_ui; eauto.
       assert (ctxereq ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e', e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; simpl; auto.
           apply ref_isEq; simpl; eauto using lc_open_ui.
           rewrite_env (nil ++ ((u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::nil)++D1).
           apply weakening_isEq; auto.
             simpl; split; eauto using lc_open_ui.
             apply ref_isEq; simpl; eauto using lc_open_ui.
               simpl; split; eauto using lc_open_ui.
       assert (Hresult1:=J1 
                         ((e, e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                         ((e', e_Con Cj (e_Var_f xj))::(u,open_exp_wrt_exp uj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                         (open_type_wrt_exp tj' (e_Var_f xj)) 
                         (open_type_wrt_exp tj (e_Var_f xj))
                         t3
                         Hleft Hright Hctxereq).
       clear J1.

       destruct Hresult1 as [p1 Hresult1].
       exists (S p1).
       apply DTQ_Red1 with (L:=L0 `union` L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e `union` efv_type tj' `union` efv_type t3 `union` efv_exp u `union` efv_exp uj)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj'); auto.
         apply trans_isEq with (e2:=e'); auto.
           eapply ctx_conv_isEq; eauto.
             apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H12 in xj0notinL.
           apply trans_isEq with (e2:=u'); auto.
             rewrite_env (nil ++ (one (w, e_Var_f xj0)) ++ D1).
             apply weakening_isEq; auto.
               simpl; split; auto.

             eapply ctx_conv_isEq; eauto.
               apply CEQ_Eq; auto using ref_isEq.
                 apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           rewrite_env ((one (e, e_Con Cj (e_Var_f xj0)))++((w, e_Var_f xj0)::D1)).
           apply cut_dtypeeq with (e1:=u)(e2:=open_exp_wrt_exp uj (e_Var_f xj0)); auto.
               simpl.
               assert ((open_exp_wrt_exp e (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                       (e, (e_Con Cj (e_Var_f xj0)))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
                   apply lc_isEq in H17. decompose [and] H17; auto.
               rewrite <- EQ'. clear EQ'.
               assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                       (w, (e_Var_f xj0))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               rewrite <- EQ'. clear EQ'.
               assert ((open_exp_wrt_exp u (e_Var_f xj0), open_exp_wrt_exp uj (e_Var_f xj0)) =
                       (u, open_exp_wrt_exp uj (e_Var_f xj0))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
                   apply lc_isEq in H18. decompose [and] H18; auto.
               rewrite <- EQ'. clear EQ'.
               assert (open_type_wrt_exp t3 (e_Var_f xj0) =
                       t3) as EQ'.
                 rewrite open_type_wrt_exp_lc_type with (t1:=t3); eauto.
                   apply lc_dtypeeq in Htrans2.
                     decompose [and] Htrans2; auto.
               rewrite <- EQ'. clear EQ'.
               apply renaming_dtypeeq3 with (x:=xj); auto.
                 assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                         (e, (e_Con Cj (e_Var_f xj)))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
                 rewrite EQ'. clear EQ'.
                 assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                         (w, (e_Var_f xj))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 rewrite EQ'. clear EQ'.
                 assert ((open_exp_wrt_exp u (e_Var_f xj), open_exp_wrt_exp uj (e_Var_f xj)) =
                         (u, open_exp_wrt_exp uj (e_Var_f xj))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
                 rewrite EQ'. clear EQ'.
                 assert (open_type_wrt_exp t3 (e_Var_f xj) =
                         t3) as EQ'.
                   rewrite open_type_wrt_exp_lc_type with (t1:=t3); eauto.
                 rewrite EQ'; auto.

             rewrite_env (nil ++ ((w, e_Var_f xj0)::nil) ++ D1).
             apply trans_isEq with (e2:=u'); auto.
               apply weakening_isEq; auto.
               simpl; split; auto.
               eapply ctx_conv_isEq; eauto.
                 simpl. 
                 apply CEQ_Eq; auto using ref_isEq.
                   apply sym_ctxereq; auto.
     SSCase "DTQ_Red2".
       assert (PairLt (S n, n0) (S n, S n0)) as J. auto.
       apply IH with (n:=S n) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J2 J3.
       pick fresh xj.
       assert (xj `notin` L) as Hright. auto.
       apply H7 in Hright.
       assert (dtypeeq (S n) ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1) (t_Case e T u pt5) (t_Case e' T u' pt')) as Hleft.
         rewrite_env (nil ++ ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::nil)++D1).
         apply weakening_dtypeeq; auto.
           simpl. repeat(split; auto).
       assert (ctxereq ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                       ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.   
             apply ref_isEq; simpl; auto.
             apply ref_isEq; simpl; auto.
       assert (Hresult:=@J1 ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D1)
                            ((e0, e_Con Cj (e_Var_f xj))::(w, e_Var_f xj)::D2)
                            (t_Case e T u pt5)
                            (t_Case e' T u' pt')
                            (open_type_wrt_exp tj (e_Var_f xj))
                            Hleft
                            Hright
                            Hctxereq).
       destruct Hresult as [p Hresult].
       exists (S p) .
       apply DTQ_Red2 with (L:=L `union` efv_ectx D1 `union` efv_exp w `union` efv_exp e0 `union` efv_exp e `union` efv_type tj `union` efv_pt pt5 `union` efv_exp u `union` efv_exp uj)(Cj:=Cj)(w:=w)(sj:=sj)(uj:=uj)(tj:=tj); auto.
         eapply ctx_conv_isEq; eauto.
           apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert (xj0 `notin` L) as xj0notinL. auto.
           apply H5 in xj0notinL.
           eapply ctx_conv_isEq; eauto.
             apply CEQ_Eq; auto using ref_isEq.
               apply sym_ctxereq; auto.
         intros xj0 xj0notin.
           assert ((open_exp_wrt_exp e0 (e_Var_f xj0), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj0)) =
                   (e0, (e_Con Cj (e_Var_f xj0)))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
               apply lc_isEq in H2. decompose [and] H2; auto.
           rewrite <- EQ. clear EQ.
           assert ((open_exp_wrt_exp w (e_Var_f xj0), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj0)) =
                   (w, (e_Var_f xj0))) as EQ.
             rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
               apply lc_pure; auto.
           rewrite <- EQ. clear EQ.
           assert (open_type_wrt_exp (t_Case e T u pt5) (e_Var_f xj0) =
                   t_Case e T u pt5) as EQ.
             rewrite open_type_wrt_exp_lc_type with (t1:=t_Case e T u pt5); eauto.
               apply lc_dtypeeq in Htrans1.
                 destruct Htrans1; auto.
           rewrite <- EQ. clear EQ.
           apply renaming_dtypeeq2 with (x:=xj); auto.
             assert ((open_exp_wrt_exp e0 (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
                     (e0, (e_Con Cj (e_Var_f xj)))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=e0); eauto.
             rewrite EQ. clear EQ.
             assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0)(e_Var_f xj)) =
                     (w, (e_Var_f xj))) as EQ.
               rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
                 apply lc_pure; auto.
             rewrite EQ. clear EQ.
             assert (open_type_wrt_exp (t_Case e T u pt5) (e_Var_f xj) =
                     (t_Case e T u pt5)) as EQ.
               rewrite open_type_wrt_exp_lc_type with (t1:=(t_Case e T u pt5)); eauto.
             rewrite EQ; auto.
     SSCase "DTQ_Case".
       assert (PairLt (n, n0) (S n, S n0)) as J. auto.
       apply IH with (n:=n) (m:=n0) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J1 J3.
       assert (Hresult2:=J2 D1 D2 e e' e'0 u u' u'0 pt5 pt' pt'0 H1 H11 Hctxeq).
       clear H1 H11 J2.

       destruct Hresult2 as [p2 Hresult2].

       exists (S p2).
       apply DTQ_Case; eauto.
         apply trans_isEq with (e2:=e'); auto.
           apply ctx_conv_isEq with (D1:=D2); auto.
             apply sym_ctxereq; auto.
         apply trans_isEq with (e2:=u'); auto.
           apply ctx_conv_isEq with (D1:=D2); auto.
             apply sym_ctxereq; auto.
 Case "trans_dpateq_prop".
   intros D1 D2 e1 e2 e3 u1 u2 u3 pt1 pt2 pt3 Htrans1 Htrans2 Hctxeq.
   inversion Htrans1; subst.
   SCase "DTPQ_Empty".
     inversion Htrans2; subst.
     SCase "DTPQ_Empty".
       exists 0. apply DTPQ_Empty; auto.
   SCase "DTPQ_Cons".
     inversion Htrans2; subst.
     SCase "DTPQ_Cons".
       assert (PairLt (n, n0) (S (max n m), S (max n0 m1))) as J.
         apply PairOrdL. 
           assert (J:=@le_max_l n m). omega.
       apply IH with (n:=n) (m:=n0) in J; auto.
       destruct J as [J1 [J2 J3]]. clear J1 J2.
       assert (Hresult3:=J3 D1 D2 e1 e2 e3 u1 u2 u3 ptt5 ptt' ptt'0 H H9 Hctxeq).
       clear H H9 J3.
       destruct Hresult3 as [p3 Hresult3].

       assert (PairLt (m, m1) (S (max n m), S (max n0 m1))) as J.
         apply PairOrdL. 
           assert (J:=@le_max_r n m). omega.
       apply IH with (n:=m) (m:=m1) in J; auto. clear IH.
       destruct J as [J1 [J2 J3]]. clear J1 J3.
       assert (Hresult2:=J2 D1 D2 e1 e2 e3 u1 u2 u3 pt5 pt' pt'0 H0 H11 Hctxeq).
       clear H11 H0 J2.
       destruct Hresult2 as [p2 Hresult2].

       exists (S (max p3 p2)).
       apply DTPQ_Cons; eauto.
 Case "trans_dpatteq_prop".
   intros D1 D2 e1 e2 e3 u1 u2 u3 ptt1 ptt2 ptt3 Htrans1 Htrans2 Hctxeq.
   inversion Htrans1; subst.
   SCase "DTTPQ_Intro".
     inversion Htrans2; subst.
     SCase "DTTPQ_Intro".
       assert (PairLt (n, n0) (S n, S n0)) as J.
         apply PairOrdL. auto.
       apply IH with (n:=n) (m:=n0) in J; auto.
       destruct J as [J1 [J2 J3]]. clear J3 J2 IH.

       pick fresh x.
       assert (x `notin` L) as Hleft. auto.
       apply H2 in Hleft.
       assert (x `notin` L0) as Hright. auto.
       apply H15 in Hright.
       assert (K:=@datacon_bind_uniq Ci si si0 T T0 ui ui0 H H5).
       decompose [and] K. subst si0 ui0 T0. clear K.
       assert (ctxereq ((e1, e_Con Ci (e_Var_f x))::(u1, open_exp_wrt_exp ui (e_Var_f x))::D1)
                       ((e2, e_Con Ci (e_Var_f x))::(u2, open_exp_wrt_exp ui (e_Var_f x))::D2)) as Hctxereq.
         apply CEQ_Eq; auto using ref_isEq.
         apply CEQ_Eq; auto using ref_isEq.
           apply ref_isEq; simpl; eauto using lc_open_ui.
           rewrite_env (nil ++ (one (u1, open_exp_wrt_exp ui (e_Var_f x))) ++ D1).
           apply weakening_isEq; auto.
             simpl; eauto using lc_open_ui.
             apply ref_isEq; simpl; eauto using lc_open_ui.
       assert (Hresult1:=J1 
                         ((e1, e_Con Ci (e_Var_f x))::(u1, open_exp_wrt_exp ui (e_Var_f x))::D1)
                         ((e2, e_Con Ci (e_Var_f x))::(u2, open_exp_wrt_exp ui (e_Var_f x))::D2)
                         (open_type_wrt_exp ti (e_Var_f x))
                         (open_type_wrt_exp ti' (e_Var_f x))
                         (open_type_wrt_exp ti'0 (e_Var_f x))
                         Hleft 
                         Hright 
                         Hctxereq).
       destruct Hresult1 as [p1 Hresult1].

       exists (S p1).
       eapply DTTPQ_Intro with (L:=L `union` efv_ectx D1 `union` efv_exp e1 `union` efv_exp u1 `union` efv_exp ui `union` efv_type ti `union` efv_type ti'0); eauto.
         eapply trans_isEq; eauto.
           eapply ctx_conv_isEq; eauto.
             apply sym_ctxereq; auto.
         eapply trans_isEq; eauto.
           eapply ctx_conv_isEq; eauto.
             apply sym_ctxereq; auto.
         intros xi xinotinL.
               assert ((open_exp_wrt_exp e1 (e_Var_f xi), open_exp_wrt_exp (e_Con Ci (e_Var_b 0))(e_Var_f xi)) =
                       (e1, (e_Con Ci (e_Var_f xi)))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=e1); eauto.
                   apply lc_isEq in H0. decompose [and] H0; auto.
               rewrite <- EQ'. clear EQ'.
               assert ((open_exp_wrt_exp u1 (e_Var_f xi), open_exp_wrt_exp ui (e_Var_f xi)) =
                       (u1, open_exp_wrt_exp ui (e_Var_f xi))) as EQ'.
                 rewrite open_exp_wrt_exp_lc_exp with (e2:=u1); eauto.
                   apply lc_isEq in H1. decompose [and] H1; auto.
               rewrite <- EQ'. clear EQ'.
               apply renaming_dtypeeq2 with (x:=x); auto.
                 assert ((open_exp_wrt_exp e1 (e_Var_f x), open_exp_wrt_exp (e_Con Ci (e_Var_b 0))(e_Var_f x)) =
                         (e1, (e_Con Ci (e_Var_f x)))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=e1); eauto.
                 rewrite EQ'. clear EQ'.
                 assert ((open_exp_wrt_exp u1 (e_Var_f x), open_exp_wrt_exp ui (e_Var_f x)) =
                         (u1, open_exp_wrt_exp ui (e_Var_f x))) as EQ'.
                   rewrite open_exp_wrt_exp_lc_exp with (e2:=u1); eauto.
                 rewrite EQ'; auto.
Qed.

Lemma trans_dtypeeq : forall n m D1 D2 t1 t2 t3, 
  dtypeeq n D1 t1 t2 ->
  dtypeeq m D2 t2 t3 ->  
  ctxereq D1 D2 -> 
  exists p, dtypeeq p D1 t1 t3.
intros.
assert ((n,m) = (n, m)) as K. auto.
pose (J := trans_deq K).
unfold trans_dtypeeq_prop in J. 
destruct J. eauto.
Qed.

Definition sound_dtypeeq_prop n D t t' (WF : dtypeeq n D t t') :=
  typeeq D t t'.

Definition sound_dpateq_prop n D e u e' u' pt pt' (WF : dpateq n D e u e' u' pt pt') :=
  pateq D e u e' u' pt pt'.

Definition sound_dpatteq_prop n D e u e' u' ptt ptt' (WF : dpatteq n D e u e' u' ptt ptt') :=
  patteq D e u e' u' ptt ptt'.

Lemma sound_deq:
  (forall n D t t' WF, @sound_dtypeeq_prop n D t t'  WF) /\
  (forall n D e u e' u' pt pt' WF, @sound_dpateq_prop n D e u e' u' pt pt' WF) /\
  (forall n D e u e' u' ptt ptt' WF, @sound_dpatteq_prop n D e u e' u' ptt ptt' WF).
(deq_mutind_cases 
  apply deq_mutind with 
  (P  := sound_dtypeeq_prop) 
  (P0 := sound_dpateq_prop)
  (P1 := sound_dpatteq_prop) Case);
  unfold sound_dtypeeq_prop, sound_dpateq_prop, sound_dpatteq_prop; intros; subst; eauto.
Qed.

Lemma sound_dtypeeq: forall n D t t',
  dtypeeq n D t t' ->
  typeeq D t t'.
intros. 
pose (J := sound_deq).
unfold sound_dtypeeq_prop in J. 
destruct J.  eauto. 
Qed.

Definition complete_dtypeeq_prop D t t' (WF : typeeq D t t') :=
  exists n, dtypeeq n D t t'.

Definition complete_dpateq_prop D e u e' u' pt pt' (WF : pateq D e u e' u' pt pt') :=
  exists n, dpateq n D e u e' u' pt pt'.

Definition complete_dpatteq_prop D e u e' u' ptt ptt' (WF : patteq D e u e' u' ptt ptt') :=
  exists n, dpatteq n D e u e' u' ptt ptt'.

Lemma complete_deq:
  (forall D t t' WF, @complete_dtypeeq_prop D t t'  WF) /\
  (forall D e u e' u' pt pt' WF, @complete_dpateq_prop D e u e' u' pt pt' WF) /\
  (forall D e u e' u' ptt ptt' WF, @complete_dpatteq_prop D e u e' u' ptt ptt' WF).
(eq_mutind_cases 
  apply eq_mutind with 
  (P  := complete_dtypeeq_prop) 
  (P0 := complete_dpateq_prop)
  (P1 := complete_dpatteq_prop) Case);
  unfold complete_dtypeeq_prop, complete_dpateq_prop, complete_dpatteq_prop; intros; subst; eauto.
Case "TQ_Pi".
  pick fresh x.
  assert (x `notin` L) as xnotinL. auto.
  apply H0 in xnotinL.
  destruct H as [n1 H1].
  destruct xnotinL as [n2 H2].
  exists (S (max n1 n2)).
  apply DTQ_Pi with (L:=L `union` efv_ectx D `union` efv_type t2 `union` efv_type t2'); auto.
    intros x0 x0notinL.
    apply renaming_dtypeeq0 with (x:=x); auto.
Case "TQ_Sigma".
  pick fresh x.
  assert (x `notin` L) as xnotinL. auto.
  apply H0 in xnotinL.
  destruct H as [n1 H1].
  destruct xnotinL as [n2 H2].
  exists (S (max n1 n2)).
  apply DTQ_Sigma with (L:=L `union` efv_ectx D `union` efv_type t2 `union` efv_type t2'); auto.
    intros x0 x0notinL.
    apply renaming_dtypeeq0 with (x:=x); auto.
Case "TQ_App".
  destruct H as [n H].
  exists (S n).
  apply DTQ_App; auto.
Case "TQ_Red1".
  pick fresh xj.
  assert (xj `notin` L) as xjnotinL. auto.
  apply H in xjnotinL.
  destruct xjnotinL as [n J].
  exists (S n).
  eapply DTQ_Red1 with (L:=L `union` efv_ectx D `union` efv_type t `union` efv_type tj `union` efv_exp e `union` efv_exp u  `union` efv_exp w `union` efv_pt pt5); eauto.
    intros xj' xj'notin.
    assert ((open_exp_wrt_exp e (e_Var_f xj'), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj')) =
            (e, (e_Con Cj (e_Var_f xj')))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      apply lc_isEq in i. decompose [and] i; auto.
    rewrite <- EQ'. clear EQ'.
    assert ((open_exp_wrt_exp w (e_Var_f xj'), open_exp_wrt_exp (e_Var_b 0) (e_Var_f xj')) =
            (w, (e_Var_f xj'))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
        apply lc_pure; auto.
    rewrite <- EQ'. clear EQ'.
    assert (open_type_wrt_exp t (e_Var_f xj') =
      t) as EQ'.
      rewrite open_type_wrt_exp_lc_type with (t1:=t); eauto.
    rewrite <- EQ'; auto.
    apply renaming_dtypeeq2 with (x:=xj); auto.
      assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
              (e, (e_Con Cj (e_Var_f xj)))) as EQ.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      rewrite EQ. clear EQ.
      assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0) (e_Var_f xj)) =
              (w, (e_Var_f xj))) as EQ.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
          apply lc_pure; auto.
      rewrite EQ. clear EQ.
      assert (open_type_wrt_exp t (e_Var_f xj) =
        t) as EQ.
        rewrite open_type_wrt_exp_lc_type with (t1:=t); eauto.
      rewrite EQ; auto.
Case "TQ_Red2".
  pick fresh xj.
  assert (xj `notin` L) as xjnotinL. auto.
  apply H in xjnotinL.
  destruct xjnotinL as [n J].
  exists (S n).
  eapply DTQ_Red2 with (L:=L `union` efv_ectx D `union` efv_type t `union` efv_type tj `union` efv_exp e `union` efv_exp u  `union` efv_exp w `union` efv_pt pt5); eauto.
    intros xj' xj'notin.
    assert ((open_exp_wrt_exp e (e_Var_f xj'), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj')) =
            (e, (e_Con Cj (e_Var_f xj')))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      apply lc_isEq in i. decompose [and] i; auto.
    rewrite <- EQ'. clear EQ'.
    assert ((open_exp_wrt_exp w (e_Var_f xj'), open_exp_wrt_exp (e_Var_b 0) (e_Var_f xj')) =
            (w, (e_Var_f xj'))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
        apply lc_pure; auto.
    rewrite <- EQ'. clear EQ'.
    assert (open_type_wrt_exp t (e_Var_f xj') =
      t) as EQ'.
      rewrite open_type_wrt_exp_lc_type with (t1:=t); eauto.
    rewrite <- EQ'; auto.
    apply renaming_dtypeeq2 with (x:=xj); auto.
      assert ((open_exp_wrt_exp e (e_Var_f xj), open_exp_wrt_exp (e_Con Cj (e_Var_b 0))(e_Var_f xj)) =
              (e, (e_Con Cj (e_Var_f xj)))) as EQ.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      rewrite EQ. clear EQ.
      assert ((open_exp_wrt_exp w (e_Var_f xj), open_exp_wrt_exp (e_Var_b 0) (e_Var_f xj)) =
              (w, (e_Var_f xj))) as EQ.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=w); eauto.
          apply lc_pure; auto.
      rewrite EQ. clear EQ.
      assert (open_type_wrt_exp t (e_Var_f xj) =
        t) as EQ.
        rewrite open_type_wrt_exp_lc_type with (t1:=t); eauto.
      rewrite EQ; auto.
Case "TQ_Case".
  destruct H as [n H].
  exists (S n). auto.
Case "TPQ_Cons".
  destruct H as [n H].
  destruct H0 as [n0 H0].
  exists (S (max n n0)). auto.
Case "TTPQ_Intro".
  pick fresh xi.
  assert (xi `notin` L) as xinotinL. auto.
  apply H in xinotinL.
  destruct xinotinL as [n J].
  exists (S n).
  eapply DTTPQ_Intro with (L:=L `union` efv_exp e `union` efv_ectx D `union` efv_exp u `union` efv_exp ui `union` efv_type ti `union` efv_type ti'); eauto.
    intros xi0 xi0notin.
    assert ((open_exp_wrt_exp e (e_Var_f xi0), open_exp_wrt_exp (e_Con Ci (e_Var_b 0))(e_Var_f xi0)) =
            (e, (e_Con Ci (e_Var_f xi0)))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      apply lc_isEq in i. decompose [and] i; auto.
    rewrite <- EQ'. clear EQ'.
    assert ((open_exp_wrt_exp u (e_Var_f xi0), open_exp_wrt_exp ui (e_Var_f xi0)) =
            (u, open_exp_wrt_exp ui (e_Var_f xi0))) as EQ'.
      rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
      apply lc_isEq in i0. decompose [and] i0; auto.
    rewrite <- EQ'. clear EQ'.
    apply renaming_dtypeeq2 with (x:=xi); auto.
      assert ((open_exp_wrt_exp e (e_Var_f xi), open_exp_wrt_exp (e_Con Ci (e_Var_b 0))(e_Var_f xi)) =
              (e, (e_Con Ci (e_Var_f xi)))) as EQ'.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=e); eauto.
      rewrite EQ'. clear EQ'.
      assert ((open_exp_wrt_exp u (e_Var_f xi), open_exp_wrt_exp ui (e_Var_f xi)) =
              (u, open_exp_wrt_exp ui (e_Var_f xi))) as EQ'.
        rewrite open_exp_wrt_exp_lc_exp with (e2:=u); eauto.
      rewrite EQ'; auto.
Qed.

Lemma complete_dtypeeq: forall D t t',
  typeeq D t t' ->
  exists n, dtypeeq n D t t'.
Proof.
intros. 
pose (J := complete_deq).
unfold complete_dtypeeq_prop in J. 
destruct J.  eauto. 
Qed.

Lemma trans_typeeq : forall D t1 t2 t3, 
  typeeq D t1 t2 ->
  typeeq D t2 t3 ->
  typeeq D t1 t3.
Proof.
  intros D t1 t2 t3 Htrans1 Htrans2.
  apply complete_dtypeeq in Htrans1.
  apply complete_dtypeeq in Htrans2.
  destruct Htrans1 as [n1 Htrans1].
  destruct Htrans2 as [n2 Htrans2].
  assert (lc_ectx D) as LD. auto.
  assert(K:=@ref_ctxereq D LD).
  assert(J:=@trans_dtypeeq n1 n2 D D t1 t2 t3 Htrans1 Htrans2 K); auto.
  destruct J as [p J].
  apply sound_dtypeeq in J; auto.
Qed.

Definition trans_kindeq_prop (n:nat) :=
  forall D k1 k2 k3,
    n = size_kind k2 ->
    kindeq D k1 k2 -> 
    kindeq D k2 k3 -> 
    kindeq D k1 k3.

Lemma _trans_kindeq : forall n,
  trans_kindeq_prop n.
Proof.
  intro n.
  apply lt_wf_ind. clear n.
  unfold trans_kindeq_prop.
  intros n IH.
  intros D k1 k2 k3 EQ Htrans1 Htrans2; subst.
  destruct k2.
    inversion Htrans1; subst.
    inversion Htrans2; subst; auto.

    inversion Htrans1; subst.
    inversion Htrans2; subst.
    apply KQ_Pi with (L:=L0 `union` L); auto.
      eapply trans_typeeq; eauto.
      intros x xnotin.
      assert (x `notin` L0) as xnotinL0. auto.
      apply H6 in xnotinL0; auto.
      assert (x `notin` L) as xnotinL. auto.
      apply H4 in xnotinL; auto.
      assert (size_kind (open_kind_wrt_exp k2 (e_Var_f x)) < size_kind (k_pi t k2)) as J.
        simpl. rewrite size_kind_open_kind_wrt_exp_var. omega.
      eauto.
Qed.

Lemma trans_kindeq : forall D k1 k2 k3,
  kindeq D k1 k2 -> 
  kindeq D k2 k3 -> 
  kindeq D k1 k3.
intros. 
assert (size_kind k2 = size_kind k2) as K. auto.
pose (J := @_trans_kindeq (size_kind k2) D k1 k2 k3 K); auto.
Qed.

Lemma trans_ctxeq : forall G1 G2 G3, 
  ctxeq G1 G2 -> 
  ctxeq G2 G3 -> 
  ctxeq G1 G3.
Proof.
  intros G1 G2 G3 Htrans1 Htrans2.
  generalize dependent G1.
  generalize dependent G3.
  induction G2; intros.
    inversion Htrans1; subst.
    inversion Htrans2; subst; auto.

    inversion Htrans1; subst.
      inversion Htrans2; subst.
      apply ctx_conv_typeeq with (D2:=erase G) in H6; auto.
        apply CQ_Term; eauto using trans_typeeq.
        apply sym_ctxereq; auto.    
          apply ctxeq_erase in H2; auto.

      inversion Htrans2; subst.
      apply CQ_Eq; auto.
        apply trans_isEq with (e2:=e1'); auto.
          apply ctx_conv_isEq with (D1:=erase G2); auto.
            apply sym_ctxereq; auto.    
              apply ctxeq_erase in H1; auto.
        apply trans_isEq with (e2:=e2'); auto.
          apply ctx_conv_isEq with (D1:=erase G2); auto.
            apply sym_ctxereq; auto.    
              apply ctxeq_erase in H1; auto.
Qed.

(********Substitution for Equality Judgments **********************)

Lemma subst_typeeq: forall D t1 t2 w x,
 typeeq D t1 t2 ->
 pure w ->
 typeeq (esubst_ectx w x D) (esubst_type w x t1) (esubst_type w x t2).
Proof.
  intros D t1 t2 w x H Pw.
  apply complete_dtypeeq in H.
  destruct H as [n H].
  assert(J:=@subst_dtypeeq n D t1 t2 w x H Pw); auto.
  apply sound_dtypeeq in J; auto.
Qed.

Lemma subst_kindeq: forall D k1 k2 w x,
 kindeq D k1 k2 ->
 pure w ->
 kindeq (esubst_ectx w x D) (esubst_kind w x k1) (esubst_kind w x k2).
Proof.
 intros D k1 k2 w x H Pw.
 induction H; simpl; auto.
   apply KQ_Refl.
     apply subst_lc_ectx; auto.
       apply lc_pure; auto.
   apply KQ_Pi with (L:=L `union` {{x}}); auto.
     apply subst_typeeq; auto.

     intros. 
     assert (lc_exp w) as Lw. apply lc_pure; auto.
     assert (x0 `notin` L) as J. auto.
     apply H1 in J; auto.
       rewrite esubst_kind_open_kind_wrt_exp in J; auto.
       rewrite esubst_kind_open_kind_wrt_exp in J; auto.
       simpl in J.
       destruct (@eq_dec expvar _ x0 x); subst; auto. 
         destruct_notin. contradict NotInTac; auto.
Qed.

Lemma subst_ctxereq : forall D1 D2 w x,
 ctxereq D1 D2 ->
 pure w ->
 ctxereq (esubst_ectx w x D1) (esubst_ectx w x D2). 
Proof.
 intros D1 D2 w x H Pw.
 induction H; simpl; auto.
   apply CEQ_Eq; eauto using subst_isEq.
Qed.   

Lemma subst_erase : forall G w x,
  esubst_ectx w x (erase G) = erase (esubst_G w x G).
Proof.
  induction G; intros; simpl; auto.
    destruct a; auto.
      destruct p; simpl; auto.
        rewrite <- IHG; auto.
Qed.   

Lemma _subst_ctxeq: forall G1 G2 w x,
 ctxeq G1 G2 ->
 pure w ->
 ctxeq (esubst_G w x G1) (esubst_G w x G2). 
Proof.
 intros G1 G2 w x H Pw.
 induction H; simpl; auto.
   apply CQ_Term; auto.
     apply subst_typeeq with (x:=x) (w:=w) in H0; auto.
       rewrite <- subst_erase; auto.

   apply CQ_Eq; auto. 
     rewrite <- subst_erase.
       apply subst_isEq; auto.
     rewrite <- subst_erase.
       apply subst_isEq; auto.
Qed.

(************** CUT for Wellformedness Judgements ****************)

Lemma ctxeq_mid_inv: forall G G1 G2 e1 e2,
  ctxeq G (G1 ++ AltAsn type (e1, e2) :: G2) ->
  isEq (erase G2) e1 e2 ->
  exists G1', exists G2', exists e1', exists e2',
    G = G1' ++ AltAsn type (e1', e2') :: G2' /\ ctxeq G2' G2 /\
    isEq(erase G2') e1' e1 /\ isEq(erase G2') e2' e2 /\
    ctxeq (G1' ++ G2') (G1 ++ G2). 
Proof.
  intros. generalize dependent G1. generalize dependent G2.
  induction G; intros; eauto.
   inversion H; subst.
   contradict H1. simpl_env. apply nil_neq_one_mid; auto.

   destruct a.
  Case " VarAsn".
   inversion H; subst.
   simpl_env in H5. apply one_eq_app in H5.
   destruct H5 as [[G1' [EQ1 EQ2]] | [EQ1 EQ2]]; subst.
     simpl in H3. apply IHG in H3; auto.
     destruct H3 as [G3 [G4 [e1' [e2' [J1 [J2 [J3 [J4 J5]]]]]]]]; subst.
     exists (VarAsn (exp*exp) a t :: G3). exists G4.
     exists e1'. exists e2'.
     repeat(split; auto).
       simpl_env. simpl.
       apply CQ_Term; auto.
         simpl_env in H6. simpl in H6.
         simpl_env. apply cut_typeeq in H6; auto.
           apply trans_isEq with (e2:=e1); auto.
             apply trans_isEq with (e2:=e2); auto.
               apply ctx_conv_isEq with (D1:=erase G2); auto.
                 apply sym_ctxereq; auto. 
                   apply ctxeq_erase; auto.
               apply sym_isEq; auto.

     simpl in EQ2. inversion EQ2; subst.

  Case "AltAsn".
   inversion H; subst.
   simpl_env in H3. apply one_eq_app in H3.
   destruct H3 as [[G1' [EQ1 EQ2]] | [EQ1 EQ2]]; subst.
     simpl in H4. apply IHG in H4; auto.
     destruct H4 as [G3 [G4 [e1'' [e2'' [J1 [J2 [J3 [J4 J5]]]]]]]]; subst.
     exists (AltAsn type (e0, e3) :: G3). exists G4.
     exists e1''. exists e2''.
     repeat(split; auto).
       simpl_env. simpl.
       apply CQ_Eq; simpl_env in *; simpl in*; auto.
         apply cut_isEq with (u1:=e1'') (u2:=e2''); auto.
           apply trans_isEq with (e2:=e1); auto.
             apply trans_isEq with (e2:=e2); auto.
               apply ctx_conv_isEq with (D1:=erase G2); auto.
                 apply sym_ctxereq; auto. 
                   apply ctxeq_erase; auto.
               apply sym_isEq; auto.
         apply cut_isEq with (u1:=e1'') (u2:=e2''); auto.
           apply trans_isEq with (e2:=e1); auto.
             apply trans_isEq with (e2:=e2); auto.
               apply ctx_conv_isEq with (D1:=erase G2); auto.
                 apply sym_ctxereq; auto. 
                   apply ctxeq_erase; auto.
               apply sym_isEq; auto.

     simpl in EQ2. inversion EQ2; subst.
       exists (nil). exists G.
       exists e0. exists e3.
       repeat(split; eauto).
Qed.

Definition cut_wfk_prop G k (WF : wfk G k) :=
  forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wfk (G2 ++ G1) k.
Definition cut_wft_prop G t k (WF : wft G t k) :=
  forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wft (G2 ++ G1) t k.
Definition cut_wfe_prop G e t (WF : wfe G e t) :=
  forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wfe (G2 ++ G1) e t.
Definition cut_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wfpt (G2 ++ G1) e u pt k. 
Definition cut_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wfpe (G2 ++ G1) e u pe t.
Definition cut_wfc_prop G (WF : wfc G) :=
 forall G1 G2 e1 e2, 
  G = G2++ AltAsn type (e1, e2):: G1 ->
  isEq (erase G1) e1 e2 ->
  wfc (G2 ++ G1).

Lemma cut_wf:
  (forall G k WF, @cut_wfk_prop G k WF) /\
  (forall G t k WF, @cut_wft_prop G t k WF) /\
  (forall G e u pt k WF, @cut_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @cut_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @cut_wfpe_prop G e u pe t WF) /\
  (forall G WF, @cut_wfc_prop G WF).
Proof.
 apply wf_mutind with 
  (P  := cut_wfk_prop) 
  (P0 := cut_wft_prop) 
  (P1 := cut_wfpt_prop)
  (P2 := cut_wfe_prop)  
  (P3 := cut_wfpe_prop)
  (P4 := cut_wfc_prop); 
  unfold cut_wfk_prop, cut_wft_prop, cut_wfpt_prop, 
  cut_wfe_prop, cut_wfpe_prop, cut_wfc_prop; intros; subst; 
  eauto using cut_isEq, cut_incon, cut_kindeq.

Case "WFK_Pi".
eapply WFK_Pi with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto. 
  intros. rewrite_env (((VarAsn (exp * exp) x t) :: G2) ++ G1).
  eapply H; eauto. solve_uniq. 

Case "K_InCon".
 apply K_InCon; auto.  eapply H; eauto. simpl_env in *.
 simpl in i. eauto using cut_incon. 
 solve_uniq.  eapply H0; eauto.

Case "K_Pi".
eapply K_Pi with (L := L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
  intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1).
  eapply H; eauto. 
  solve_uniq.
  
Case "K_Sigma". 
eapply K_Sigma with (L := L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
  intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1).
  eapply H; eauto. solve_uniq.
  
Case "K_App".
eapply K_App with (L := L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
  intros. simpl_env. rewrite_env ((one (e_Var_f x, e) ++ erase G2) ++ erase G1).
  eapply cut_kindeq with (e1:=e1)(e2:=e2); auto. simpl. simpl_env in k0. eapply k0; eauto.

Case "K_KConv".
eapply K_KConv; eauto. simpl_env in *. simpl in k0. eapply cut_kindeq; eauto.
(*
Case "K_CConv".
 destruct (ctxeq_mid_inv G2 G1 c) as [G2' [G1' [e1' [e2' [Jeq [Jctxeq [J1 [J2 J3]]]]]]]]; subst; auto.
 eapply K_CConv with (G:= G2' ++ G1'); eauto.
  eapply H; eauto.
   apply trans_isEq with (e2:=e1); auto.
    apply trans_isEq with (e2:=e2); auto.
      apply ctx_conv_isEq with (D1:=erase G1); auto.
        apply sym_ctxereq; auto. 
          apply ctxeq_erase; auto.
    apply sym_isEq; auto.
*)
Case "PK_Cons".
 eapply PK_Cons with (L := L\u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto. 
 intros.
 rewrite_env (
   (AltAsn type (e, e_Con C (e_Var_f x))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f x))
         :: VarAsn (exp * exp) x s :: G2)++ G1).
 eapply H0; eauto. solve_uniq.

Case "T_incon".
 apply T_InCon; auto. eapply H; eauto.
   simpl_env in *. simpl in i.
   apply cut_incon in i; auto.
   solve_uniq.  eapply H0; eauto.
Case "T_Var".
 eapply T_Var; eauto.
  analyze_binds b; eauto.
    eapply binds_app_1; eauto.
    inversion H3; subst.
      inversion H0.
      inversion H0.
    eapply binds_app_2; eauto.
Case "T_Ctr".
 eapply T_Ctr with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
 intros. rewrite_env (((e_Var_f x, e) :: (erase G2)) ++ (erase G1)).
 eapply cut_typeeq; eauto.
 rewrite_env ((e_Var_f x, e) :: erase (G2 ++ AltAsn type (e1, e2):: G1)); eauto.

Case "T_Fun".
  eapply T_Fun with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
  intros.  
  rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1).
  eapply H; eauto. 
  solve_uniq.

Case "T_App".
 eapply T_App with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
 intros. 
 rewrite_env (((e_Var_f x, e2) :: (erase G2)) ++ erase G1).
 eapply cut_typeeq; eauto.
 rewrite_env ((e_Var_f x, e2) :: erase (G2 ++ AltAsn type (e0, e3):: G1)); eauto.

Case "T_Fix".
 eapply T_Fix with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
 intros.  
 rewrite_env (((VarAsn (exp * exp) f (t_Pi t1 t2)) :: G2) ++ G1).
 eapply H; eauto.
 solve_uniq.

Case "T_Sigma".
 eapply T_Sigma with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
 intros.  
 rewrite_env (((e_Var_f x, e1) :: erase G2) ++ erase G1).
 eapply cut_typeeq; eauto.
 rewrite_env ((e_Var_f x, e1) :: erase (G2 ++ AltAsn type (e0, e3):: G1)); eauto.

 intros.
 rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1).
 eapply H1; eauto. solve_uniq.

Case "T_Proj2".
  eapply T_Proj2 with (L:=L \u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto.
  intros.
 rewrite_env (((e_Var_f x, e_Proj1 e) :: erase G2) ++ erase G1).
 eapply cut_typeeq; eauto.
 rewrite_env ((e_Var_f x, e_Proj1 e) :: erase (G2 ++ AltAsn type (e1, e2):: G1)); eauto.

Case "T_TConv".
 eapply T_TConv with (t:=t); eauto.
 simpl_env in *; simpl in *; eauto using cut_typeeq.

(*
Case "T_CConv".
 destruct (ctxeq_mid_inv G2 G1 c) as [G2' [G1' [e1' [e2' [Jeq [Jctxeq [J1 [J2 J3]]]]]]]]; subst; auto.
 eapply T_CConv with (G:= G2' ++ G1'); eauto.
  eapply H; eauto.
   apply trans_isEq with (e2:=e1); auto.
    apply trans_isEq with (e2:=e2); auto.
      apply ctx_conv_isEq with (D1:=erase G1); auto.
        apply sym_ctxereq; auto. 
          apply ctxeq_erase; auto.
    apply sym_isEq; auto.
*)
Case "AP_Cons". 
 eapply AP_Cons with (L := L\u dom (G1 ++ AltAsn type (e1, e2):: G2)); eauto. 
 intros.
 rewrite_env (
   (AltAsn type (e, e_Con Ci (e_Var_f xi))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi))
         :: VarAsn (exp * exp) xi si :: G2) ++ G1).
 eapply H; eauto.
 solve_uniq.

Case "wf_nil".  
 contradict H. simpl_env. apply nil_neq_one_mid.

Case "C_Term".
 destruct G2; subst; eauto.
   inversion H1; subst; eauto.
   destruct a.
     inversion H1; subst; eauto.
     rewrite_env ((VarAsn (exp * exp) a t0) :: (G2 ++ G1)).
     eapply C_Term; eauto.

     inversion H1; subst; eauto.
Case "C_Eq".
 destruct G2; subst; eauto.
   inversion H2; subst; eauto.
   inversion H2; subst; eauto.
     rewrite_env (AltAsn type (e1, e2):: (G2 ++ G1)).
     eapply C_Eq; eauto.
Qed.

Lemma cut_wfe : forall G1 G2 e1 e2 e t,
  wfe (G1++ AltAsn type (e1, e2):: G2)  e t ->
  isEq (erase G2) e1 e2 ->
  wfe (G1++ G2)  e t.
Proof.
intros. 
pose (J := cut_wf).
unfold cut_wfe_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]]. eauto.
Qed.

Lemma cut_wfk : forall G1 G2 e1 e2 k,
  wfk (G2++ AltAsn type (e1, e2):: G1) k ->
  isEq (erase G1) e1 e2 ->
  wfk (G2 ++ G1) k.
Proof.
intros. 
pose (J := cut_wf).
unfold cut_wfk_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]]. eauto.
Qed.

Lemma cut_wft : forall G1 G2 e1 e2 t k, 
  wft (G2++ AltAsn type (e1, e2):: G1) t k ->
  isEq (erase G1) e1 e2 ->
  wft (G2 ++ G1) t k.
Proof.
intros. 
pose (J := cut_wf).
unfold cut_wft_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]]. eauto.
Qed.

Lemma cut_wfc : forall G1 G2 e1 e2, 
  wfc (G2++ AltAsn type (e1, e2):: G1) ->
  isEq (erase G1) e1 e2 ->
  wfc (G2 ++ G1).
Proof.
intros. 
pose (J := cut_wf).
unfold cut_wfc_prop in J. 
destruct J as [JK [JT [JPT [JE [JPE JC]]]]]. eauto.
Qed.

(*****************some properties of ctxeq ******************)

Lemma weakening_ctxeq_simple: forall G G' G1 G2 G3 G1' G2' G3',
  ctxeq G G' ->
  lc_ctx G2 ->
  G = G1 ++ G3 ->
  G' = G1' ++ G3' ->
  ctxeq G3 G3' ->
  ctxeq (G2 ++ G3) (G2' ++ G3') ->
  ctxeq (G1 ++ G2 ++ G3) (G1'++G2'++G3').
Proof.
  intros. generalize dependent G1;  generalize dependent G1';
   generalize dependent G3; generalize dependent G3'; 
     generalize dependent G2;   generalize dependent G2'.
  induction H; intros; eauto.
Case "CQ_Empty".
  destruct (app_eq_nil G1 G3) as (J1, J3); eauto.
  destruct (app_eq_nil G1' G3') as (J2, J4); eauto.
  subst; simpl_env in *; auto. 

Case "CQ_Term".
  simpl_env in *.
  assert (J: (exists qs, G1' =  VarAsn (exp * exp) x t' :: qs /\ G'= qs ++ G3') \/ 
             (G1' = nil /\ G3' = VarAsn (exp * exp) x t' :: G')).
  eapply app_eq_one; eauto. 
  destruct J as [[G0' [J1' J2']] | [J1' J2']].
  assert (J: (exists qs, G1 =  VarAsn (exp * exp) x t :: qs /\ G= qs ++ G3) \/ 
             (G1 = nil /\ G3 = VarAsn (exp * exp) x t :: G)).
  eapply app_eq_one; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].

  subst. rewrite_env (VarAsn (exp * exp) x t :: (G0 ++ G2 ++ G3)).
  rewrite_env (VarAsn (exp * exp) x t' :: (G0' ++ G2' ++ G3')).
  eapply CQ_Term; eauto.
  simpl_env in *. eapply weakening_typeeq; eauto.

  subst. assert (J: ctxeq G (G0' ++ one (VarAsn (exp * exp) x t) ++ G)).
  eapply trans_ctxeq with (G2:= G0' ++G3'); eauto.
  simpl_env; eapply ctxeq_app; eauto using sym_ctxeq.
  eapply app_lc_ctx1 with (G2:=G3').
  destruct (lc_ctxeq H); auto.
  assert False. eapply ctxeq_impossible; eauto. destruct H6.

  assert (J: (exists qs, G1 =  VarAsn (exp * exp) x t :: qs /\ G= qs ++ G3) \/ 
             (G1 = nil /\ G3 = VarAsn (exp * exp) x t :: G)).
  eapply app_eq_one; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].

  subst. assert (J: ctxeq G' (G0 ++ one (VarAsn (exp * exp) x t') ++ G')).
  apply sym_ctxeq.
  eapply trans_ctxeq with (G2:= G0 ++G3); eauto.
  simpl_env; eapply ctxeq_app; eauto using sym_ctxeq.
  eapply app_lc_ctx1 with (G2:=G3).
  destruct (lc_ctxeq H); auto.
  assert False. eapply ctxeq_impossible; eauto. destruct H6.

  subst. 
  rewrite_env (nil ++ (G2 ++ one (VarAsn (exp * exp) x t)) ++ G);
  rewrite_env (nil ++ (G2' ++ one (VarAsn (exp * exp) x t')) ++ G').
  eapply IHctxeq; eauto.
  eapply app_lc_ctx3; eauto.
  eapply app_lc_ctx1 with (G2:=G). destruct (lc_ctxeq H3); eauto.
  simpl_env in *; auto.
  
Case "CQ_Eq".
  simpl_env in *.
  assert (J: (exists qs, G1' =  AltAsn type (e1', e2') :: qs /\ G'= qs ++ G3') \/ 
             (G1' = nil /\ G3' = AltAsn type (e1', e2') :: G')).
  eapply app_eq_one; eauto. 
  destruct J as [[G0' [J1' J2']] | [J1' J2']].
  assert (J: (exists qs, G1 =  AltAsn type (e1, e2) :: qs /\ G= qs ++ G3) \/ 
             (G1 = nil /\ G3 = AltAsn type (e1, e2) :: G)).
   eapply app_eq_one; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].

  subst. rewrite_env (AltAsn type (e1, e2):: (G0 ++ G2 ++ G3)).
  rewrite_env (AltAsn type (e1', e2')  :: (G0' ++ G2' ++ G3')).
  eapply CQ_Eq; eauto.
  simpl_env in *. eapply weakening_isEq; eauto.
  simpl_env in *. eapply weakening_isEq; eauto.

  subst. assert (J: ctxeq G (G0' ++ one (AltAsn type (e1, e2)) ++ G)).
  eapply trans_ctxeq with (G2:= G0' ++G3'); eauto.
  simpl_env; eapply ctxeq_app; eauto using sym_ctxeq.
  eapply app_lc_ctx1 with (G2:=G3').
  destruct (lc_ctxeq H); auto.
  assert False. eapply ctxeq_impossible; eauto. destruct H7.
  
  assert (J: (exists qs, G1 =  AltAsn type (e1, e2) :: qs /\ G= qs ++ G3) \/ 
             (G1 = nil /\ G3 = AltAsn type (e1, e2) :: G)).
   eapply app_eq_one; eauto. 
  destruct J as [[G0 [J1 J2]] | [J1 J2]].

  subst. assert (J: ctxeq G' (G0 ++ one (AltAsn type (e1', e2') ) ++ G')).
  apply sym_ctxeq.
  eapply trans_ctxeq with (G2:= G0 ++G3); eauto.
  simpl_env; eapply ctxeq_app; eauto using sym_ctxeq.
  eapply app_lc_ctx1 with (G2:=G3).
  destruct (lc_ctxeq H); auto.
  assert False. eapply ctxeq_impossible; eauto. destruct H7.

  subst. 
  rewrite_env (nil ++ (G2 ++ one (AltAsn type (e1, e2) )) ++ G);
  rewrite_env (nil ++ (G2' ++ one (AltAsn type (e1', e2') )) ++ G').
  eapply IHctxeq; eauto.
  eapply app_lc_ctx3; eauto.
  eapply app_lc_ctx1 with (G2:=G). destruct (lc_ctxeq H3); eauto.
  simpl_env in *; auto.
Qed.  

Lemma ctxeq_dom_eq2: forall G1 G2 G1' G2' G3 G3',
  ctxeq G1 G1' ->
  ctxeq G3 G3' ->
  G3 = (G2 ++ G1) -> 
  G3' =(G2'++ G1') -> 
  uniq G3 ->
  dom G2 [=] dom G2'.
Proof.
 intros. generalize dependent G1; generalize dependent G1'; generalize dependent G2'; generalize dependent G2.  
 induction H0; intros; eauto.

Case "CQ_Empty".
 symmetry in H1. symmetry in H2. 
 destruct (app_eq_nil G2' G1' H2) as (J1, J2).
 destruct (app_eq_nil G2 G1 H1) as (J3, J4).
 subst; solve_uniq.

Case "CQ_Term".   
 destruct G2; subst.  
 (* G2 = nil *)
 destruct G2'; subst. 
 (* G2' = nil *) solve_uniq.
 (*G2' <> nil*)
 replace ((a :: G2') ++ G1') with (a :: G2' ++ G1') in H2; auto.
 inversion H2; subst. 
 simpl_asnlist in *. 
 assert (J: disjoint (one (VarAsn (exp * exp) x t))  G). 
 apply uniq_app_3; eauto.
 assert (J1:= ctxeq_dom_eq H1).  assert (J2:= ctxeq_dom_eq H0). 
 rewrite <- H4 in J1. simpl_asnlist in *. solve_uniq.

 (*G2 <> nil*)
 destruct G2'; subst. 
 (* G2' = nil *) 
 replace ((a :: G2) ++ G1) with (a :: G2 ++ G1) in H4; auto.
 inversion H4; subst. 
 simpl_asnlist in *. 
 assert (J: disjoint (one (VarAsn (exp * exp) x t))  (G2++G1)). 
 apply uniq_app_3; eauto.
 assert (J1:= ctxeq_dom_eq H1).  assert (J2:= ctxeq_dom_eq H0). 
 rewrite <- H2 in J1. simpl_asnlist in *. solve_uniq.

 (*G2' <> nil*) 
 replace ((a0 :: G2') ++ G1') with (a0 :: G2' ++ G1') in H2; auto.
 inversion H2; subst. 
 replace ((a :: G2') ++ G1') with (a :: G2' ++ G1') in H4; auto.
 inversion H4; subst. clear H2 H4 H.
 assert (J: dom G2 [=] dom G2'). 
 eapply IHctxeq; eauto. inversion H3; auto.
 solve_uniq.

Case "CQ_Eq".
destruct G2; subst.  
 (* G2 = nil *)
 destruct G2'; subst. 
 (* G2' = nil *) solve_uniq.
 (*G2' <> nil*)
 replace ((a :: G2') ++ G1') with (a :: G2' ++ G1') in H2; auto.
 inversion H2; subst. 
 simpl_asnlist in *.
 rewrite <- H5 in H4.
 assert (J: ctxeq G (G2' ++ one (AltAsn type (e1, e2)) ++ G)).
 eapply trans_ctxeq with (G2:= G2' ++ G1'); eauto.
 eapply ctxeq_app; eauto using sym_ctxeq.
 eapply app_lc_ctx1 with (G2:=G1').
 destruct (lc_ctxeq H0); auto.
 assert False. eapply ctxeq_impossible; eauto. destruct H6.

 (*G2 <> nil*)
 destruct G2'; subst. 
 (* G2' = nil *) 
 replace ((a :: G2) ++ G1) with (a :: G2 ++ G1) in H5; auto.
 inversion H5; subst. 
 simpl_env in *. rewrite <- H2 in H4.
 assert (J: ctxeq G' (G2++ one (AltAsn type (e1', e2')) ++ G')).
 apply sym_ctxeq. eapply trans_ctxeq with (G2:=G2++G1); auto.
 eapply ctxeq_app; eauto. 
 eapply app_lc_ctx1 with (G2:=G1).
 destruct (lc_ctxeq H0); auto.
 apply sym_ctxeq; auto.
 assert False. eapply ctxeq_impossible; eauto. destruct H6.

 (*G2' <> nil*) 
 replace ((a0 :: G2') ++ G1') with (a0 :: G2' ++ G1') in H2; auto.
 inversion H2; subst. 
 replace ((a :: G2) ++ G1) with (a :: G2 ++ G1) in H5; auto.
 inversion H5; subst.
 assert (J: dom G2 [=] dom G2'). 
 eapply IHctxeq; eauto. inversion H3; auto. 
 solve_uniq.
Qed.


(************* Regularity of context well-formed ness ***********)

Lemma reg_wfk_c: forall G k,
 wfk G k -> wfc G.
Proof.
  intros.
  induction H; eauto.
  pick fresh z. 
  assert (J: z \notin L); auto.
  assert (J1:= H0 z J).  inversion J1; auto. 
Qed.


Definition reg_wft_c_prop G t k (WF : wft G t k) := wfc G.
Definition reg_wfpt_c_prop G e u pt k (WF : wfpt G e u pt k) := wfc G.
 

Lemma reg_wftpt_c:
 (forall G t k WF, @reg_wft_c_prop G t k WF)
 /\ (forall G e u pt k WF, @reg_wfpt_c_prop G e u pt k WF).
Proof.
 eapply wf_tpt_mutind
 with (P  := reg_wft_c_prop ) 
  (P0 := reg_wfpt_c_prop );
  unfold reg_wft_c_prop, reg_wfpt_c_prop; intros;
  eauto using reg_wfk_c; try solve [
 pick fresh z;
 assert (J:  wfc (VarAsn (exp * exp) z t1 :: G)); eauto;
 inversion J;auto].
Qed.


Lemma reg_wft_c: forall G t k,
 wft G t k -> wfc G.
Proof.
 intros.
 apply ((proj1 reg_wftpt_c) G t k); eauto.
Qed.

Definition reg_wfe_c_prop G e t (WF : wfe G e t) := wfc G.
Definition reg_wfpe_c_prop G e u pe t (WF : wfpe G e u pe t) := wfc G.
 
Lemma reg_wfepe_c:
 (forall G e t WF, @reg_wfe_c_prop G e t WF)
 /\ (forall G e u pe t WF, @reg_wfpe_c_prop G e u pe t WF).
Proof.
 eapply wf_epe_mutind
 with (P  := reg_wfe_c_prop ) 
  (P0 := reg_wfpe_c_prop );
  unfold reg_wfe_c_prop, reg_wfpe_c_prop; intros; 
  eauto using reg_wft_c; try solve [
 pick fresh z;
 assert (J:  wfc (VarAsn (exp * exp) z t1 :: G)); eauto;
 inversion J;auto].

 pick fresh z;
 assert (J:  wfc (VarAsn (exp * exp) z (t_Pi t1 t2) :: G)); eauto;
 inversion J;auto.
Qed.


Lemma reg_wfe_c: forall G e t,
 wfe G e t -> wfc G.
Proof.
 intros.
 apply ((proj1 reg_wfepe_c) G e t); eauto.
Qed.

Lemma wf_ctr_type: forall C S t,
  wfs S ->
  binds C (bind_t t) S ->
  wft nil t k_type.
Proof.
 intros C S t HwfS Hbind.
 induction S; eauto.
 inversion Hbind.
 destruct a.

 Case "VarAsn".
  assert (J:= binds_cons_1 bind nat C a (bind_t t) b S Hbind).
  destruct J as [(Jeq1, Jea2) |Jbind ].
  inversion HwfS; subst; eauto.
   inversion H0; subst; auto.
   inversion H0; eauto.
  
  inversion HwfS; subst; eapply IHS; auto.

 inversion Hbind. inversion H.
  inversion HwfS; subst; eapply IHS; eauto.
Qed.


Lemma wf_tctr_kind: forall T S k,
  wfs S ->
  binds T (bind_k k) S ->
  wfk nil k .
Proof.
 intros T S k HwfS Hbind.
 induction S; eauto.
 inversion Hbind.
 destruct a.

 Case "VarAsn".
  assert (J:= binds_cons_1 bind nat T a (bind_k k) b S Hbind).
  destruct J as [(Jeq1, Jea2) |Jbind ].
  inversion HwfS; subst; eauto.
   inversion H0; subst; auto.
   inversion H0; eauto.
  
  inversion HwfS; subst; eapply IHS; auto.
Case "AltAsn".
 inversion Hbind. inversion H.
  inversion HwfS; subst; eapply IHS; eauto.
Qed.

Lemma binds_ctxeq: forall G G' x t,
binds x t G ->
ctxeq G G' ->
exists t', binds x t' G' /\ typeeq (erase G') t' t.
Proof.
  intros. generalize dependent G'.
  induction G; intros; auto. inversion H.
  destruct G'.
  assert (J:= sym_ctxeq H0).
  replace (a::G) with ((a::G)++nil) in J; auto.
  assert (J1: (a::G) = nil). eapply nil_ctxeq_append ; eauto.
  inversion J1. simpl_env; auto.

  inversion H0; subst.   
  inversion H; subst. inversion H1; subst. 
  exists t'; split; eauto.
  eapply binds_cons_2; eauto. simpl.
  eapply sym_typeeq; eapply ctx_conv_typeeq; eauto.
  eapply ctxeq_erase; eauto.

  destruct IHG with (G':=G'); eauto. destruct H2.
  exists x1; split; eauto.
  eapply binds_cons_3; eauto.

  inversion H; inversion H0; subst.  inversion H1.
  destruct IHG with (G':=G'); eauto. destruct H2.
  exists x0; split; eauto.
  simpl_env.  eapply binds_app_2; eauto.
  simpl_env.
  rewrite_env (nil++ erase (one(AltAsn type (e1', e2'))) ++ erase G').
  replace (erase G') with (nil++ erase G') in H3; auto.
  eapply weakening_typeeq; eauto.
  simpl. destruct ( lc_isEq H12).  destruct H8.
  destruct ( lc_isEq H13).  destruct H14; auto.
Qed.
 
(************* Weakening for Well-formed Judgments ********************)


Lemma wfc_head: forall G1 G2,
 wfc (G1 ++ G2) ->
 wfc G2.
Proof.
  intros. 
  induction G1; intros; eauto. 
  eapply IHG1.
  inversion H; subst; eauto.
Qed.

Lemma wfc_disjoint: forall G1 G2,
 wfc (G1++G2) ->
 disjoint G1 G2.
Proof.
  intros. induction G1; subst; eauto. 
  unfold disjoint; simpl; auto. solve_uniq.

  inversion H; subst; eauto.
  solve_uniq.
Qed.
  

Lemma uniq_wfc: forall G,
 wfc G -> uniq G.
Proof.
  intros. induction H; eauto; try solve [solve_uniq]. 
  simpl_asnlist in *. apply uniq_alt; auto. 
Qed.


Definition weakening_wfk_prop G k (WF : wfk G k) :=
  forall G1 G2 G3, 
  G = G3 ++ G1 ->
  wfc (G2 ++ G1) -> 
  disjoint G2 G3 ->
  wfk (G3 ++ G2 ++ G1) k.
Definition weakening_wft_prop G t k (WF : wft G t k) :=
  forall G1 G2 G3,
  G = G3 ++ G1 ->
  wfc (G2 ++ G1) -> 
  disjoint G2 G3 ->
  wft (G3 ++ G2 ++ G1) t k.
Definition weakening_wfe_prop G e t (WF : wfe G e t) :=
  forall G1 G2 G3,
  G = G3 ++ G1 ->
  wfc (G2 ++ G1) -> 
  disjoint G2 G3 ->
  wfe (G3 ++ G2 ++ G1) e t.
Definition weakening_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  forall G1 G2 G3,
  G = G3 ++ G1 ->
  wfc (G2 ++ G1) -> 
  disjoint G2 G3 ->
  wfpt (G3 ++ G2 ++ G1) e u pt k. 
Definition weakening_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  forall G1 G2 G3,
  G = G3 ++ G1 ->
  wfc (G2 ++ G1) -> 
  disjoint G2 G3 ->
  wfpe (G3 ++ G2 ++ G1) e u pe t.
Definition weakening_wfc_prop G (WF : wfc G) :=
 forall G1 G2 G3,
  G = G3 ++ G1 -> 
  wfc (G2 ++ G1) ->
  disjoint G2 G3 ->
  wfc (G3 ++ G2 ++ G1).



Lemma wf_weakening:
(*  wfs Signature ->*)
  (forall G k WF, @weakening_wfk_prop G k WF) /\
  (forall G t k WF, @weakening_wft_prop G t k WF) /\
  (forall G e u pt k WF, @weakening_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @weakening_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @weakening_wfpe_prop G e u pe t WF) /\
  (forall G WF, @weakening_wfc_prop G WF).
Proof.
(*  intros HwfS. *)
 apply wf_mutind with 
  (P  := weakening_wfk_prop) 
  (P0 := weakening_wft_prop) 
  (P1 := weakening_wfpt_prop)
  (P2 := weakening_wfe_prop)  
  (P3 := weakening_wfpe_prop)
  (P4 := weakening_wfc_prop); 
  unfold weakening_wfk_prop, weakening_wft_prop, weakening_wfpt_prop, 
  weakening_wfe_prop, weakening_wfpe_prop, weakening_wfc_prop; intros; subst; 
  eauto using weakening_isEq, weakening_incon, weakening_kindeq.

Case "WFK_Pi".
eapply WFK_Pi with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto. 
  intros. rewrite_env (((VarAsn (exp * exp) x t) :: G3) ++ G2 ++ G1).
  eapply H; eauto. solve_uniq. 

Case "K_InCon".
apply K_InCon; auto. simpl_env in *. 
 apply weakening_incon; auto.
   apply app_lc_ectx1 with (D2:=erase G1).
   apply lc_wfc in H2. decompose [and] H2.
   apply lc_ctx_erase in H1; simpl_env in H1; auto.
 solve_uniq.  

Case "K_Pi".
eapply K_Pi with (L := L \u dom (G1 ++ G2 ++ G3)); eauto.
  intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G3) ++ G2 ++ G1).
  eapply H; eauto. 
  solve_uniq.
  
Case "K_Sigma". 
eapply K_Sigma with (L := L \u dom (G1 ++ G2 ++ G3)); eauto.
  intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G3) ++ G2 ++ G1).
  eapply H; eauto. solve_uniq.
  
Case "K_App".
eapply K_App with (L := L \u dom (G1 ++ G2 ++ G3)); eauto.
  intros. simpl_env. rewrite_env ((one (e_Var_f x, e) ++ erase G3) ++ erase G2 ++ erase G1).
  eapply weakening_kindeq. simpl. simpl_env in k0. eapply k0; eauto.
    apply lc_wfc in H3.
    destruct H3.
    apply app_lc_ctx1 in H3.
    apply lc_ctx_erase; auto.

Case "K_KConv".
eapply K_KConv; eauto. simpl_env in *. eapply weakening_kindeq; eauto.
    apply lc_wfc in H2.
    destruct H2.
    apply app_lc_ctx1 in H1.
    apply lc_ctx_erase; auto.

(*
Case "K_CConv".
 destruct (ctxeq_concat G3 G1 c) as [G3' [G1' (Jeq, Jctxeq)]].
 eapply K_CConv with (G:= G3' ++ G2 ++ G1'); eauto.
 eapply H; eauto.
 eapply  wfc_head_conv; eauto.
 eapply wfc_head; eauto. subst; eapply reg_wft_c ; eauto.
  assert (J3: dom G3 [=] dom G3'). 
  eapply ctxeq_dom_eq2 with (G3':=G) (G3:= G3++G1) (G1:=G1) (G1':=G1'); eauto.
  apply sym_ctxeq; auto. apply sym_ctxeq; auto.
  apply uniq_wfc; auto.
  unfold disjoint. solve_uniq. 

  eapply weakening_ctxeq_simple with (G:= G3'++G1') (G':= G3++G1); eauto.
  subst; auto.
  eapply app_lc_ctx1 with (G2:=G1).
  eapply lc_wfc; auto. eapply ctxeq_app; eauto.
  eapply app_lc_ctx1 with (G2:=G1).
  eapply lc_wfc; auto. 
*)
Case "PK_Cons".
 eapply PK_Cons with (L := L\u dom (G1 ++ G2 ++ G3) ); auto. eauto. intros.
 rewrite_env (
   (AltAsn type (e, e_Con C (e_Var_f x))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f x))
         :: VarAsn (exp * exp) x s :: G3) ++ G2 ++ G1).
 eapply H0; eauto. solve_uniq.
 unfold disjoint; simpl. solve_uniq.
 unfold disjoint; simpl.  solve_uniq.

Case "T_incon".
 apply T_InCon; auto. simpl_env in *.
 apply weakening_incon; auto.
   apply app_lc_ectx1 with (D2:=erase G1).
   apply lc_wfc in H2. decompose [and] H2.
   apply lc_ctx_erase in H1; simpl_env in H1; auto.
 solve_uniq. 
Case "T_Var".
 eapply T_Var; eauto.
 assert (J: binds x t G3 \/ binds x t G1).
  eapply binds_app_3; eauto.
 destruct J as [Jb| Jb].
  eapply binds_app_1; eauto.
  rewrite_env ((G3++G2) ++G1).
  eapply binds_app_2; eauto.
Case "T_Ctr".
 eapply T_Ctr with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
 intros. rewrite_env (((e_Var_f x, e) :: (erase G3)) ++ (erase G2) ++ (erase G1)).
 eapply weakening_typeeq; eauto.
    rewrite_env ((e_Var_f x, e) :: erase (G3 ++ G1)); eauto.
    apply lc_wfc in H2.
    destruct H2.
    apply app_lc_ctx1 in H2.
    apply lc_ctx_erase; auto.

Case "T_Fun".
  eapply T_Fun with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
  intros.  
  rewrite_env (((VarAsn (exp * exp) x t1) :: G3) ++ G2 ++ G1).
  eapply H; eauto. 
  solve_uniq.

Case "T_App".
 eapply T_App with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
 intros. 
 rewrite_env (((e_Var_f x, e2) :: (erase G3)) ++ erase G2 ++ erase G1).
 eapply weakening_typeeq; eauto.
    rewrite_env ((e_Var_f x, e2) :: erase (G3 ++ G1)); eauto.
    apply lc_wfc in H3.
    destruct H3.
    apply app_lc_ctx1 in H3.
    apply lc_ctx_erase; auto.

Case "T_Fix".
 eapply T_Fix with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
 intros.  
 rewrite_env (((VarAsn (exp * exp) f (t_Pi t1 t2)) :: G3) ++ G2 ++ G1).
 eapply H; eauto.
 solve_uniq.

Case "T_Sigma".
 eapply T_Sigma with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
 intros.  
 rewrite_env (((e_Var_f x, e1) :: erase G3) ++ erase G2 ++ erase G1).
 eapply weakening_typeeq; eauto.
    rewrite_env ((e_Var_f x, e1) :: erase (G3 ++ G1)); eauto.
    apply lc_wfc in H3.
    destruct H3.
    apply app_lc_ctx1 in H3.
    apply lc_ctx_erase; auto.

 intros.
 rewrite_env (((VarAsn (exp * exp) x t1) :: G3) ++ G2 ++ G1).
 eapply H1; eauto. solve_uniq.

Case "T_Proj2".
  eapply T_Proj2 with (L:=L \u dom (G1 ++ G2 ++ G3)); eauto.
  intros.
 rewrite_env (((e_Var_f x, e_Proj1 e) :: erase G3) ++ erase G2 ++ erase G1).
 eapply weakening_typeeq; eauto.
    rewrite_env ((e_Var_f x, e_Proj1 e) :: erase (G3 ++ G1)); eauto.
    apply lc_wfc in H2.
    destruct H2.
    apply app_lc_ctx1 in H2.
    apply lc_ctx_erase; auto.

Case "T_TConv".
 eapply T_TConv with (t:=t); eauto.
 simpl_env in *. apply weakening_typeeq; auto.
    apply lc_wfc in H2.
    destruct H2.
    apply app_lc_ctx1 in H1.
    apply lc_ctx_erase; auto.

(*
Case "T_CConv".
 destruct (ctxeq_concat G3 G1 c) as [G3' [G1' (Jeq, Jctxeq)]].
 eapply T_CConv with (G:= G3' ++ G2 ++ G1'); eauto.
 eapply H; eauto.
 eapply  wfc_head_conv; eauto.
 eapply wfc_head; eauto. subst; eapply reg_wfe_c ; eauto.
  assert (J3: dom G3 [=] dom G3'). 
  eapply ctxeq_dom_eq2 with (G3':=G) (G3:= G3++G1) (G1:=G1) (G1':=G1'); eauto.
  apply sym_ctxeq; auto. apply sym_ctxeq; auto.
  apply uniq_wfc; auto.
  unfold disjoint. solve_uniq. 

  eapply weakening_ctxeq_simple with (G:= G3'++G1') (G':= G3++G1); eauto.
  subst; auto.
  eapply app_lc_ctx1 with (G2:=G1).
  eapply lc_wfc; auto. eapply ctxeq_app; eauto.
  eapply app_lc_ctx1 with (G2:=G1).
  eapply lc_wfc; auto. 
*)
Case "AP_Cons". 
 eapply AP_Cons with (L := L\u dom (G1 ++ G2 ++ G3) ); auto. eauto. intros.
 rewrite_env (
   (AltAsn type (e, e_Con Ci (e_Var_f xi))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi))
         :: VarAsn (exp * exp) xi si :: G3) ++ G2 ++ G1).
 eapply H; eauto.
 solve_uniq.
 unfold disjoint; simpl. solve_uniq.
 unfold disjoint; simpl.  solve_uniq.

Case "wf_nil".  
 symmetry in H.
 destruct (app_eq_nil G3 G1 H) as (J1, J2).
 subst; simpl_env in *; eauto.

Case "C_Term".
 destruct G3; subst; eauto.
 inversion H1; subst; eauto.
 rewrite_env ((VarAsn (exp * exp) x t) :: (G3 ++ G2 ++ G1)).
 eapply C_Term; eauto.
 eapply H; eauto.  solve_uniq. 
 replace  (VarAsn (exp * exp) x t :: G3) 
   with  (one(VarAsn (exp * exp) x t) ++ G3) in H3; auto.

 rewrite disjoint_sym in H3.
 assert (J: disjoint (one (VarAsn (exp * exp) x t)) G2).
 solve_uniq. simpl_asnlist in *. 
 assert (J1: x \notin (dom G2)). solve_uniq.
 solve_notin.
 
 eapply H0; eauto.  solve_uniq.

Case "C_Eq".
 destruct G3; subst; eauto.
 inversion H2; subst; eauto.
 rewrite_env (AltAsn type (e1, e2):: (G3 ++ G2 ++ G1)).
 eapply C_Eq; eauto.
Qed.

Lemma weakening_wfk : forall G1 G2 G3 k ,
  wfk (G1 ++ G3) k ->
  wfc (G1 ++ G2 ++ G3) -> 
  wfk (G1 ++ G2 ++ G3) k.
intros. 
pose (J := wf_weakening).
unfold weakening_wfk_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]].
eapply JK; eauto using wfc_head.
assert (J: disjoint G1 (G2++G3)).
eapply  wfc_disjoint; eauto.
solve_uniq.
Qed.

Lemma weakening_wft : 
  forall G1 G2 G3 t k,
  wft (G1 ++ G3) t k ->
  wfc (G1 ++ G2 ++ G3) -> 
  wft (G1 ++ G2 ++ G3) t k.
intros. 
pose (J := wf_weakening).
unfold weakening_wft_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]]. eauto.
eapply JT; eauto using wfc_head.
assert (J: disjoint G1 (G2++G3)).
eapply  wfc_disjoint; eauto.
solve_uniq.
Qed.

Lemma weakening_wfe:
  forall G1 G2 G3 e t, 
  wfe (G1 ++ G3) e t ->
  wfc (G1 ++ G2 ++ G3) -> 
  wfe (G1 ++ G2 ++ G3) e t.
intros. 
pose (J := wf_weakening).
unfold weakening_wfe_prop in J. 
destruct J as [JK [JT [JPT [JE JPE]]]]. eauto.
eapply JE; eauto using wfc_head.
assert (J: disjoint G1 (G2++G3)).
eapply  wfc_disjoint; eauto.
solve_uniq.
Qed.

  
(************ Context conversion *************************)

Definition ctx_conv_wfk_prop G k (WF : wfk G k) :=
  forall G1 G2 G1', 
  G = G2 ++ G1 ->
  wfc G1' -> 
  ctxeq G1 G1' ->
  wfk (G2 ++ G1') k.
Definition ctx_conv_wft_prop G t k (WF : wft G t k) :=
  forall G1 G2 G1',
  G = G2 ++ G1 ->
  wfc G1' -> 
  ctxeq G1 G1' ->
  wft (G2 ++ G1') t k.
Definition ctx_conv_wfe_prop G e t (WF : wfe G e t) :=
  forall G1 G2 G1',
  G = G2 ++ G1 ->
  wfc G1' -> 
  ctxeq G1 G1' ->
  wfe (G2 ++ G1') e t.
Definition ctx_conv_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  forall G1 G2 G1',
  G = G2 ++ G1 ->
  wfc G1' -> 
  ctxeq G1 G1' ->
  wfpt (G2 ++ G1') e u pt k. 
Definition ctx_conv_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  forall G1 G2 G1',
  G = G2 ++ G1 ->
  wfc G1' -> 
  ctxeq G1 G1' ->
  wfpe (G2 ++ G1') e u pe t.
Definition ctx_conv_wfc_prop G (WF : wfc G) :=
 forall G1 G2 G1',
  G = G2 ++ G1 -> 
  wfc G1' ->
  ctxeq G1 G1' ->
  (wfc (G2 ++ G1') 
    /\ forall x t, binds x t G -> wft (G2++G1') t k_type).

Lemma wf_ctx_conv:
  (forall G k WF, @ctx_conv_wfk_prop G k WF) /\
  (forall G t k WF, @ctx_conv_wft_prop G t k WF) /\
  (forall G e u pt k WF, @ctx_conv_wfpt_prop G e u pt k WF) /\
  (forall G e t WF, @ctx_conv_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @ctx_conv_wfpe_prop G e u pe t WF) /\
  (forall G WF, @ctx_conv_wfc_prop G WF).
Proof.
 apply wf_mutind with 
  (P  := ctx_conv_wfk_prop) 
  (P0 := ctx_conv_wft_prop) 
  (P1 := ctx_conv_wfpt_prop)
  (P2 := ctx_conv_wfe_prop)  
  (P3 := ctx_conv_wfpe_prop)
  (P4 := ctx_conv_wfc_prop); 
  unfold ctx_conv_wfk_prop, ctx_conv_wft_prop, ctx_conv_wfpt_prop, 
  ctx_conv_wfe_prop, ctx_conv_wfpe_prop, ctx_conv_wfc_prop; intros; subst; 
  eauto using ctx_conv_typeeq, ctx_conv_kindeq.

Case "*".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 

Case "WFK_Pi".
 eapply WFK_Pi with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto. 
  intros. rewrite_env ((VarAsn (exp * exp) x t :: G2) ++ G1').
  eapply H; eauto. simpl_env; auto. 

Case "K_InCon".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 apply K_InCon; auto. 
 eapply ctx_conv_incon with (D:= erase (G2 ++ G1)); auto. 
 eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 assert (J:= ctxeq_dom_eq H3). solve_uniq.
 eapply H0; eauto.

Case "Unit".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 

Case "K_Pi".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 eapply K_Pi with (L := L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1').
 eapply H; eauto. simpl_env; auto.
  
Case "K_Sigma". 
 eapply K_Sigma with (L := L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros. rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1').
 eapply H; eauto.  simpl_env; auto.
  
Case "K_App".
 eapply K_App with (L := L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros. simpl_env. 
 rewrite_env ((one (e_Var_f x, e) ++ erase G2) ++ erase G1').
 eapply ctx_conv_kindeq; eauto.
 rewrite_env ((e_Var_f x, e) :: erase (G2 ++ G1')); eauto.
 eapply CEQ_Eq; eauto using ref_isEq.  eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto. 
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto. 

Case "K_KConv".
 eapply K_KConv; eauto.
 eapply ctx_conv_kindeq; eauto. eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.

Case "PK_Cons".
 eapply PK_Cons with (L := L\u dom (G1 ++ G2 ++ G1') ); eauto. intros.
 rewrite_env (
   (AltAsn type (e, e_Con C (e_Var_f x))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f x))
         :: VarAsn (exp * exp) x s :: G2) ++ G1').
 eapply H0; eauto.  simpl_env; auto.

Case "T_incon".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 apply T_InCon; auto. 
 eapply ctx_conv_incon with (D:= erase (G2 ++ G1)); auto. 
 eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 assert (J:= ctxeq_dom_eq H3). solve_uniq.
 eapply H0; eauto.

Case "unit".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
Case "T_Var".
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 assert (J:exists t', binds x t' (G2++G1') /\ typeeq (erase (G2++G1')) t' t).
 eapply binds_ctxeq; eauto.
 eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 destruct J as [t' (Jb', Jeq)].
 eapply T_TConv; eauto.

Case "T_Ctr".
 eapply T_Ctr with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros. rewrite_env (((e_Var_f x, e) :: (erase G2)) ++ (erase G1')).
 eapply ctx_conv_typeeq; eauto using ref_isEq.
 rewrite_env ((e_Var_f x, e) :: erase (G2 ++ G1')); eauto.
 eapply CEQ_Eq; eauto using ref_isEq.  eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
Case "T_Fun".
  eapply T_Fun with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
  intros.  
  rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++  G1').
  eapply H; eauto.  simpl_env; auto.

Case "T_App".
 eapply T_App with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros. 
 rewrite_env (((e_Var_f x, e2) :: (erase G2)) ++  erase G1').
 eapply ctx_conv_typeeq; eauto.
 rewrite_env ((e_Var_f x, e2) :: erase (G2 ++ G1')); eauto.
 eapply CEQ_Eq; eauto using ref_isEq.  eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.

Case "T_Fix".
 eapply T_Fix with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros.  
 rewrite_env (((VarAsn (exp * exp) f (t_Pi t1 t2)) :: G2) ++  G1').
 eapply H; eauto. simpl_env; auto.

Case "T_Sigma".
 eapply T_Sigma with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros.  
 rewrite_env (((e_Var_f x, e1) :: erase G2) ++ erase G1').
 eapply ctx_conv_typeeq; eauto.
 rewrite_env ((e_Var_f x, e1) :: erase (G2 ++ G1')); eauto.
 eapply CEQ_Eq; eauto using ref_isEq.  eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 
 intros.
 rewrite_env (((VarAsn (exp * exp) x t1) :: G2) ++ G1').
 eapply H1; eauto.  simpl_env; auto. 

Case "T_Proj2".
 eapply T_Proj2 with (L:=L \u dom (G1 ++ G2 ++ G1')); eauto.
 intros.
 rewrite_env (((e_Var_f x, e_Proj1 e) :: erase G2) ++ erase G1').
 eapply ctx_conv_typeeq; eauto.
 rewrite_env ((e_Var_f x, e_Proj1 e) :: erase (G2 ++ G1')); eauto.
 eapply CEQ_Eq; eauto using ref_isEq.  eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 eapply ref_isEq; eauto.
 eapply lc_ctx_erase; eauto.
 
Case "T_TConv".
 eapply T_TConv with (t:=t); eauto.
 eapply ctx_conv_typeeq; eauto.
 eapply ctxeq_erase; eauto.  eapply ctxeq_app; eauto.
 apply app_lc_ctx1 with (G2:=G1); auto.
 
Case "AP_Cons". 
 eapply AP_Cons with (L := L\u dom (G1 ++ G2 ++ G1') ); eauto. intros.
 rewrite_env (
   (AltAsn type (e, e_Con Ci (e_Var_f xi))
      :: AltAsn type (u, open_exp_wrt_exp ui (e_Var_f xi))
         :: VarAsn (exp * exp) xi si :: G2) ++ G1').
 eapply H; eauto. simpl_env; auto.

Case "wf_nil".  
 symmetry in H.
 destruct (app_eq_nil G2 G1 H) as (J1, J2).
 subst; simpl_env in *; split; auto.  intros. inversion H2.

Case "C_Term". 
 destruct G2; subst.  split; auto. 
 intros. simpl_env in *.  subst. 
 inversion H3; subst. inversion H2; subst.
 destruct H with (G1':=G') (G1:= G) (G2:=nil); eauto.
 assert (J: x0 = x /\ t0 = t \/ binds x0 t0 G).
 eapply binds_cons_1; eauto. destruct J.
 destruct H6; subst; auto.
 rewrite_env (nil ++ one (VarAsn (exp * exp) x t') ++ G').
 eapply weakening_wft; eauto. eapply H0; eauto. simpl_env; auto.
 
 rewrite_env (nil ++ one (VarAsn (exp * exp) x t') ++ G').
 eapply weakening_wft; eauto. 

 replace ((a::G2)++G1) with (a::(G2++G1)) in H1; auto.
 inversion H1; subst.
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 split. rewrite_env (VarAsn (exp * exp) x t :: G2 ++ G1').
 eapply C_Term; eauto. assert (J:=ctxeq_dom_eq H3). solve_uniq.

 intros. 
 assert (J: x0 = x /\ t0 = t \/ binds x0 t0 (G2++G1)).
 eapply binds_cons_1; eauto. destruct J.
 destruct H7; subst; auto.
 assert (J: wft (G2 ++ G1') t k_type).
 eapply H0; eauto.
 rewrite_env ( nil ++ one (VarAsn (exp * exp) x t) ++ (G2 ++G1')).
 eapply weakening_wft; eauto. simpl_env. 
 rewrite_env (VarAsn (exp * exp) x t:: G2 ++ G1').
 eapply C_Term; eauto. assert (J1:= ctxeq_dom_eq H3). solve_uniq.
 
  rewrite_env ( nil ++ one (VarAsn (exp * exp) x t) ++ (G2 ++G1')).
 eapply weakening_wft; eauto.
  rewrite_env (VarAsn (exp * exp) x t:: G2 ++ G1').
  eapply C_Term; eauto. assert (J:= ctxeq_dom_eq H3). solve_uniq.

Case "C_Eq".
 destruct G2; subst.  split; auto. 
 intros. simpl_env in *.  subst. 
 inversion H4; subst. inversion H3; subst. 
 destruct H with (G1':=G') (G1:=G) (G2:=nil); eauto.
 rewrite_env (nil ++ one (AltAsn type (e1', e2') ) ++ G').
 eapply weakening_wft; eauto. simpl_env; eapply (H6 x); eauto. 
 replace (one (AltAsn type (e1, e2)) ++ G) with (AltAsn type (e1, e2) :: G) in H5; auto.
 inversion H5; subst; auto.  inversion H7. 

 replace ((a::G2)++G1) with (a::(G2++G1)) in H2; auto.
 inversion H2; subst.
 destruct H with (G1':=G1') (G3:=G1) (G4:=G2); eauto. 
 split. rewrite_env (AltAsn type (e1, e2) :: G2 ++ G1').
 eapply C_Eq; eauto. 

 intros. 
 inversion H7; subst.  inversion H8.
 rewrite_env ( nil ++ one (AltAsn type (e1, e2)) ++ (G2 ++G1')).
 eapply weakening_wft; eauto. 
  rewrite_env (AltAsn type (e1, e2):: G2 ++ G1').
 eapply C_Eq; eauto.
Qed.


Lemma ctx_conv_wfk: forall G k G',
  wfk G k ->
  wfc G' ->
  ctxeq G G' ->
  wfk G' k.
Proof.
 pose (J := wf_ctx_conv).
unfold ctx_conv_wfk_prop in J. 
destruct J as [JK [JT [JPT [JE [JPE JC]]]]].
intros. rewrite_env (nil++G'). eapply (JK G); eauto. simpl_env; auto.
Qed.


Lemma ctx_conv_wft: forall G k t G',
  wft G t k ->
  wfc G' ->
  ctxeq G G' ->
  wft G' t k.
Proof.
 pose (J := wf_ctx_conv).
unfold ctx_conv_wft_prop in J. 
destruct J as [JK [JT [JPT [JE [JPE JC]]]]].
intros. rewrite_env (nil++G'). eapply (JT G); eauto. simpl_env; auto.
Qed.


Lemma ctx_conv_wfe: forall G e t G',
  wfe G e t ->
  wfc G' ->
  ctxeq G G' ->
  wfe G' e t.
Proof.
 pose (J := wf_ctx_conv).
unfold ctx_conv_wfe_prop in J. 
destruct J as [JK [JT [JPT [JE [JPE JC]]]]].
intros. rewrite_env (nil++G'). eapply (JE G); eauto. simpl_env; auto.
Qed.

(* Lemma wfc_head_conv: forall G1 G1' G2,
 wfc G1 -> ctxeq G1 G1'-> wfc (G2++G1') -> wfc (G2++G1).
*)

Lemma ctx_conv_head_wfc: forall G1 G1' G2,
 wfc G1 -> 
 wfc G1' ->
 ctxeq G1 G1'-> 
 wfc (G2++G1) -> 
 wfc (G2++G1').
Proof. 
  pose (J := wf_ctx_conv).
  unfold ctx_conv_wfc_prop in J. 
  destruct J as [JK [JT [JPT [JE [JPE JC]]]]]. 
  intros. 
  assert (J: wfc (G2 ++ G1') /\
       (forall (x : atom) (t : type), binds x t (G2++G1) -> wft (G2 ++ G1') t k_type)).
 eapply JC; eauto.
 destruct J; auto.
Qed.

(************** Inversion Kinding WFK typeeq*********************)

Lemma inversion_wfk_pi1: forall G t k k1,
 wfk G k ->
 k = k_pi t k1  ->
 wft G t k_type.
Proof.
 intros. 
 induction H; subst; try solve [inversion H0]; eauto.
 
 inversion H0; subst.
 pick fresh z. 
 assert (J: wfc (VarAsn (exp * exp) z t :: G) ).
 eapply reg_wfk_c; eauto.
 inversion J; subst; auto.
Qed.

Lemma inversion_wfk_pi2: forall G t k k1 L,
 wfk G k ->
 k = k_pi t k1  ->
 ( exists x , x \notin  L  /\ 
   wfk  (AssumeList.VarAsn (exp * exp)  x   t  ::  G  ) 
   ( open_kind_wrt_exp k1 (e_Var_f x) ) ) .
Proof.
 intros. 
 assert (Jt1: wft G t k_type). eapply inversion_wfk_pi1; eauto.
 induction H; subst; try solve [inversion H0]; eauto. intros.
 inversion H0; subst. 
 pick fresh z. exists z; intros;split; eauto.
Qed.

Lemma inversion_wft_sigma1: forall G t k t1 t2,
 wft G t k ->
 t = t_Sigma t1 t2  ->
 k = k_type ->
 wft G t1 k_type.
Proof.
 intros. 
 induction H; subst; try solve [inversion H0]; eauto.
 apply K_InCon; auto.
 subst.  simpl_env in *. simpl in H3. solve_uniq.
 inversion H4; auto.

 inversion H0; subst. 
 pick fresh z. 
 assert (J: wfc (VarAsn (exp * exp) z t1 :: G) ).
 eapply reg_wft_c; eauto.
 inversion J; subst; auto.

 inversion H2; subst; eauto.
Qed.

Lemma inversion_wft_sigma2: forall G t k t1 t2 L,
 wft G t k ->
 t = t_Sigma t1 t2  ->
 k = k_type ->
 ( exists x , x \notin  L  /\ 
   wft  (AssumeList.VarAsn (exp * exp)  x   t1  ::  G  ) 
   ( open_type_wrt_exp t2 (e_Var_f x) )  k_type ) .
Proof.
 intros. 
 assert (Jt1: wft G t1 k_type). eapply inversion_wft_sigma1; eauto.
 induction H; subst; try solve [inversion H0]; eauto. intros.
 pick fresh z. exists z; intros. split; auto.
 apply K_InCon; auto.
  simpl in H3. subst; simpl_env in *. 
 assert (J: efv_type (open_type_wrt_exp t2 (e_Var_f z)) [<=]
     efv_exp (e_Var_f z) \u efv_type t2).
 eapply efv_type_open_type_wrt_exp_rec_upper; eauto.
 simpl in J. solve_uniq. inversion H4; auto.

 inversion H0; subst. 
 pick fresh z. exists z; intros;split; eauto.
 
 inversion H2; subst; eauto. 
(*
 destruct IHwft; eauto. 
  apply K_CConv with (G:=G'); eauto using sym_ctxeq. eapply reg_wft_c; eauto.
  destruct H0;
  exists x; intros; split;eauto.
 eapply K_CConv; eauto.
 eapply CQ_Term; eauto. 
   eapply ref_typeeq; auto. 
     apply lc_ctxeq in H2. destruct H2.
   eapply lc_wft; eauto.
 eapply C_Term; eauto. 
 assert (J: wfc (VarAsn (exp * exp) x t1 :: G)).
 eapply reg_wft_c; eauto.
 assert (J1: dom G [=] dom G'). eapply ctxeq_dom_eq; eauto.
 inversion J; subst.
 rewrite <- J1;  solve_notin.*)
Qed.


Lemma inversion_wft_pi1: forall G t k t1 t2,
 wft G t k ->
 t = t_Pi t1 t2  ->
 k = k_type ->
 wft G t1 k_type.
Proof.
 intros. 
 induction H; subst; try solve [inversion H0]; eauto.
 apply K_InCon; auto.
 subst.  simpl_env in *. simpl in H3. solve_uniq.
 inversion H4; auto.

 inversion H0; subst. 
 pick fresh z. 
 assert (J: wfc (VarAsn (exp * exp) z t1 :: G) ).
 eapply reg_wft_c; eauto.
 inversion J; subst; auto.

 inversion H2; subst; eauto.
Qed.

Lemma inversion_wft_pi2: forall G t k t1 t2 L,
 wft G t k ->
 t = t_Pi t1 t2  ->
 k = k_type ->
 ( exists x , x \notin  L  /\ 
   wft  (AssumeList.VarAsn (exp * exp)  x   t1  ::  G  ) 
   ( open_type_wrt_exp t2 (e_Var_f x) )  k_type ) .
Proof.
 intros. 
 assert (Jt1: wft G t1 k_type). eapply inversion_wft_pi1; eauto.
 induction H; subst; try solve [inversion H0]; eauto. intros.
 pick fresh z. exists z; intros. split; auto.
 apply K_InCon; auto.
  simpl in H3. subst; simpl_env in *. 
 assert (J: efv_type (open_type_wrt_exp t2 (e_Var_f z)) [<=]
     efv_exp (e_Var_f z) \u efv_type t2).
 eapply efv_type_open_type_wrt_exp_rec_upper; eauto.
 simpl in J. solve_uniq. inversion H4; auto.

 inversion H0; subst. 
 pick fresh z. exists z; intros;split; eauto.
 
 inversion H2; subst; eauto. 
(*
 destruct IHwft; eauto. 
  apply K_CConv with (G:=G'); eauto using sym_ctxeq. eapply reg_wft_c; eauto.
  destruct H0;
  exists x; intros; split;eauto.
 eapply K_CConv; eauto.
 eapply CQ_Term; eauto. 
   eapply ref_typeeq; auto. 
     apply lc_ctxeq in H2. destruct H2.
   eapply lc_wft; eauto.
 eapply C_Term; eauto. 
 assert (J: wfc (VarAsn (exp * exp) x t1 :: G)).
 eapply reg_wft_c; eauto.
 assert (J1: dom G [=] dom G'). eapply ctxeq_dom_eq; eauto.
 inversion J; subst.
 rewrite <- J1;  solve_notin.*)
Qed.


Lemma inversion_typeeq_pi1: forall D t s t1 t2 s1 s2,
(* ~ (incon D) ->*)
 typeeq D t s ->
 t = t_Pi t1 t2 ->
 s = t_Pi s1 s2 ->
  typeeq D t1 s1 .
Proof.
  intros.
  induction H; subst; try solve [inversion H1 | inversion H0]; eauto.
   
  eapply TQ_InCon; eauto.
  inversion H; auto. inversion H2; auto.
 inversion H1; inversion H0; subst ;auto.
Qed.


 
Lemma inversion_typeeq_pi2: forall L D t s t1 t2 s1 s2,
(* ~ (incon D) ->*)
 typeeq D t s ->
 t = t_Pi t1 t2 ->
 s = t_Pi s1 s2 ->
 exists x,
   x \notin  L  /\
   typeeq D  ( open_type_wrt_exp t2 (e_Var_f x) )  
   ( open_type_wrt_exp s2 (e_Var_f x) ).
Proof.
  intros.
  induction H; subst; try solve [inversion H1 | inversion H0]; eauto.
   
 pick fresh z. exists z; split; eauto.
   eapply TQ_InCon; eauto.
   inversion H; auto. inversion H2; auto.

 inversion H1; inversion H0; subst ;auto.
 pick fresh z. exists z; intros; eauto.
Qed.


Lemma inversion_typeeq_sigma1: forall D t s t1 t2 s1 s2,
(* ~ (incon D) ->*)
 typeeq D t s ->
 t = t_Sigma t1 t2 ->
 s = t_Sigma s1 s2 ->
  typeeq D t1 s1 .
Proof.
  intros.
  induction H; subst; try solve [inversion H1 | inversion H0]; eauto.
   
  eapply TQ_InCon; eauto.
  inversion H; auto. inversion H2; auto.
 inversion H1; inversion H0; subst ;auto.
Qed.

Lemma inversion_typeeq_sigma2: forall L D t s t1 t2 s1 s2,
(* ~ (incon D) ->*)
 typeeq D t s ->
 t = t_Sigma t1 t2 ->
 s = t_Sigma s1 s2 ->
 exists x,
   x \notin  L /\
   typeeq D  ( open_type_wrt_exp t2 (e_Var_f x) )  
   ( open_type_wrt_exp s2 (e_Var_f x) ).
Proof. 
  intros.
  induction H; subst; try solve [inversion H1 | inversion H0]; eauto.
   
 pick fresh z. exists z; split; eauto.
   eapply TQ_InCon; eauto.
   inversion H; auto. inversion H2; auto.

 inversion H1; inversion H0; subst ;auto.
 pick fresh z. exists z; intros; eauto.
Qed.


Lemma inversion_typeeq_app1: forall D t s t1 t2 w1 w2,
(* ~ (incon D) ->*)
 typeeq D t s ->
 t = t_App t1 w1 ->
 s = t_App t2 w2 ->
  typeeq D t1 t2 .
Proof.
 intros.
  induction H; subst; try solve [inversion H1 | inversion H0]; eauto.
   
  eapply TQ_InCon; eauto.
  inversion H; auto. inversion H2; auto.
 inversion H1; inversion H0; subst ;auto.
Qed.

Lemma inversion_typeeq_app2: forall D t s t1 t2 w1 w2,
 ~ (incon D) ->
 typeeq D t s ->
 t = t_App t1 w1 ->
 s = t_App t2 w2 ->
  isEq D w1 w2.
Proof.
  intros.
  induction H0; subst; try solve [inversion H1 | inversion H2]; eauto.
   contradiction.
  inversion H1; inversion H2; subst; auto.
Qed.

Lemma inversion_typeeq_T: forall D t s T1 T2,
  ~ (incon D) ->
  typeeq D t s ->
  t =  t_Con T1->
  s =  t_Con T2 ->
  T1 = T2.
Proof.
  intros.
  induction H0; subst; try solve [inversion H1 | inversion H2]; eauto.
   contradiction.
  inversion H1; inversion H2; subst; auto.
Qed.

(************* REGULARITY ********************)

Lemma incon_wfk : forall G k, 
  lc_kind k -> 
  wfc G ->  
  incon (erase G) -> 
  efv_kind k [<=] dom G ->
  wfk G k. 
intros G k LC.
 generalize dependent G.
 induction LC; intros; eauto. 
 simpl in *. 
 eapply WFK_Pi with (L:= dom G \u efv_type t \u efv_kind k1); eauto.
 intros. eapply H1; eauto.
 eapply C_Term; eauto. 
 eapply K_InCon; eauto. solve_uniq. 
 assert (J:efv_kind (open_kind_wrt_exp k1 (e_Var_f x)) [<=]
     efv_exp (e_Var_f x) \u efv_kind k1).
 eapply efv_kind_open_kind_wrt_exp_rec_upper; eauto.
 simpl.  simpl in J. solve_uniq.
Qed.


Lemma regularity_wfc :
  forall G x t, wfc G -> binds x t G -> wft G t k_type.
Proof.
 intros.  
induction G; auto.
inversion H0.

simpl_env in *.
assert (J: binds x t ( one a) \/ binds x t G). 
solve_binds.
destruct J.
destruct a.
assert (J: (x = a /\ t = t0)).
 split.  eapply binds_one_1; eauto.  eapply binds_one_2; eauto.
 destruct J; inversion H; subst.
 rewrite_env (nil++one (VarAsn (exp * exp) a t0) ++ G).
 replace G with (nil++G) in H9; auto.
 eapply weakening_wft; eauto.
 inversion H1.  inversion H2. destruct H2.

rewrite_env (nil++one a++G).
replace G with (nil++G) in IHG; auto.
eapply weakening_wft; eauto.
eapply IHG; eauto. simpl_env. eapply wfc_head; eauto.
Qed.


(*
Lemma cconv_wfk: forall G G' k,
 wfk G k ->
 ctxeq G G' ->
 wfc G' ->
 wfk G' k.
Proof.
 intros. generalize dependent G'.
 induction H; intros; eauto. 
 eapply WFK_Pi with (L:=L\u dom(G++G')); eauto.
 intros. 
 eapply H0; eauto.
 eapply CQ_Term; eauto. 
   eapply ref_typeeq; eauto.
     apply lc_ctxeq in H1. destruct H1.

     assert (J:lc_ctx ((VarAsn (exp * exp) x t :: G))).
       eapply lc_wfk; eauto. 
     simpl in J; destruct J; eauto.
 eapply C_Term; eauto. 
 eapply K_CConv with (G :=G); eauto.
 pick fresh z. 
 assert (J: wfc (VarAsn (exp * exp) z t :: G)).
 eapply reg_wfk_c; eauto.
 inversion J; eauto.
Qed.*)
 
 
Definition regularity_wft_prop G t k (WF : wft G t k) :=
   wfk G k.
Definition regularity_wfpt_prop G e u pt k (WF : wfpt G e u pt k) :=
  wfk G k.

Lemma regularity_wft_mutual:
  (forall G t k WF, @regularity_wft_prop G t k WF) /\
  (forall G e u pt k WF, @regularity_wfpt_prop G e u pt k WF).
Proof.
apply wf_tpt_mutind with 
  (P := regularity_wft_prop) 
  (P0 := regularity_wfpt_prop);
  unfold  regularity_wft_prop, regularity_wfpt_prop; intros; auto.
(* Case "incon".
 apply incon_wfk; auto.
 assert (J: wfk nil k). eapply wf_tctr_kind with (S:=Signature); eauto.
 apply wfs_sig. eapply lc_wfk; eauto.
*)

Case "K_Const".
 rewrite_env (nil++G++nil).
 eapply weakening_wfk; simpl_env; eauto.
 eapply wf_tctr_kind with (S:=Signature); eauto.
 apply wfs_sig. 

Case "K_Pi".
 apply WFK_Type. 
 pick fresh z.
 apply wfc_head with (G1:= one(VarAsn (exp * exp) z t1)).
 eapply reg_wfk_c; eauto.
 rewrite_env  ((VarAsn (exp * exp) z t1) ::G).
 eapply H; eauto.

Case "K_Sigma".
 apply WFK_Type. 
 pick fresh z.
 apply wfc_head with (G1:= one(VarAsn (exp * exp) z t1)).
 eapply reg_wfk_c; eauto.
 rewrite_env  ((VarAsn (exp * exp) z t1) ::G).
 eapply H; eauto.
(*
Case "K_CConv".
 apply cconv_wfk with (G:=G); auto. *)
Qed. 

Definition regularity_wfe_prop G e t (WF : wfe G e t) :=
  wft G t k_type.
Definition regularity_wfpe_prop G e u pe t (WF : wfpe G e u pe t) :=
  wft G t k_type.

Lemma regularity_wfe_mutual:
  (forall G e t WF, @regularity_wfe_prop G e t WF) /\
  (forall G e u pe t WF, @regularity_wfpe_prop G e u pe t WF).
Proof.
apply wf_epe_mutind with 
  (P := regularity_wfe_prop) 
  (P0 := regularity_wfpe_prop);
  unfold  regularity_wfe_prop, regularity_wfpe_prop; intros; auto.

Case "T_Var".
 eapply regularity_wfc; eauto.

Case "T_Fun".
 eapply K_Pi with (L:= L); eauto.
Case "T_Fix".
 pick fresh z.
 assert (J:  wft (VarAsn (exp * exp) z (t_Pi t1 t2) :: G) (t_Pi t1 t2) k_type); eauto.
 assert (J1:= reg_wft_c J). inversion J1; auto.
Case "T_Pair".
 eapply K_Sigma; eauto.
Case "T_Proj1".
 eapply inversion_wft_sigma1; eauto.
(*Case "T_CCOnv".
 eapply K_CConv with (G:=G); eauto. *)
Qed.


Lemma wfk_wft: forall G t k,
 wft G t k -> wfk G k.
Proof.
 intros. 
 destruct (regularity_wft_mutual) as [JT JPT].
 unfold regularity_wft_prop in JT.
 eapply JT with (t:=t); eauto.
Qed.

Lemma wft_wfe: forall G e t,
 wfe G e t -> wft G t k_type.
Proof.
 intros.
 destruct (regularity_wfe_mutual) as [JE JPE].
 unfold regularity_wfe_prop in JE.
 eapply JE with (t:=t); eauto.
Qed.

